#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>

#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/delay.h>

#include <linux/init.h>
#include <asm/mpc5xxx.h>
#include <bestcomm_api.h>
#include <mgt5200/mgt5200.h>
#include <mgt5200/sdma.h>
#include <asm/mpc5xxx_i2s.h>

#define I2S_MINOR 242

#define MAX_DMA_BUFFERS 16

#define RX_TASK_START
#define TX_TASK_START

/* #define RDEBUG */
/* #define WDEBUG */
/* #define LITEDEBUG */
/* #define DEBUG_LOST_INT */

#ifdef RDEBUG
#define RDPRINTK(fmt, args...)	printk(fmt, ## args)
#else
#define RDPRINTK(fmt, args...)
#endif

#ifdef WDEBUG
#define WDPRINTK(fmt, args...)	printk(fmt, ## args)
#else
#define WDPRINTK(fmt, args...)
#endif

/* Some bit assignments for SICR */
#define DELAY_TIME_SLOT 0x20000000
#define GEN_CLK_INT 0x00800000
#define MULTIWD_ENABLE 0x00400000
#define CLK_POL_RISING 0x00200000
#define LSB_FIRST 0x10000000

static volatile ulong rx_released; /* Counter of released BDs */
static volatile ulong rx_assigned; /* Counter of BDs to be assigned */
static volatile ulong rx_data;     /* Counter of BDs read */
static volatile ulong tx_released; /* Counter of released BDs */
static volatile ulong tx_assigned; /* Counter of BDs to be assigned */

#ifdef DEBUG_LOST_INT
static volatile int tx_int;
static volatile ulong tx_acount; /* real number of assigned BDs */
#endif

static int txtask;
static int rxtask;
static int initiator_tx;
static int initiator_rx;
static int psc_num;
static unsigned long i2s_is_open;
static int configured;
static int once;

static void i2s_start_rx_tx (int task);
static void psc_enable(void);

DECLARE_WAIT_QUEUE_HEAD (i2s_wait_tx);
DECLARE_WAIT_QUEUE_HEAD (i2s_wait_rx);

extern int mgt_sdma_load_tasks_image(void);

struct i2s_buffer {
	char *va;
	dma_addr_t pa;
	size_t len;
};

static struct i2s_buffer i2s_tx_bufs[MAX_DMA_BUFFERS];
static struct i2s_buffer i2s_rx_bufs[MAX_DMA_BUFFERS];

static int i2s_buffers_init(void)
{
	int i;
	for (i = 0; i < MAX_DMA_BUFFERS; i++) {
		if ((i2s_tx_bufs[i].va = 
			 pci_alloc_consistent(NULL, BUFSIZE, &i2s_tx_bufs[i].pa)) == NULL) {
			return -1;
		}
		if ((i2s_rx_bufs[i].va = 
			 pci_alloc_consistent(NULL, BUFSIZE, &i2s_rx_bufs[i].pa)) == NULL) {
			return -1;
		}
	}
	return 0;
}

static int i2s_buffers_release(void)
{
	int i;
	for (i = 0; i < MAX_DMA_BUFFERS; i++) {
		pci_free_consistent(NULL, BUFSIZE, i2s_tx_bufs[i].va, i2s_tx_bufs[i].pa);
		TaskBDRelease(txtask);
		pci_free_consistent(NULL, BUFSIZE, i2s_tx_bufs[i].va, i2s_tx_bufs[i].pa);
		TaskBDRelease(rxtask);
	}

	return 0;
}


static void i2s_tx_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	int bdnum;

#ifdef DEBUG_LOST_INT
	tx_int = 1;
#endif

	TaskIntClear(txtask);
	
	if (tx_assigned == tx_released) {
		printk("%s: %ld:%ld - false int??\n", __FUNCTION__, tx_assigned, tx_released);
		return;
	}

	bdnum = TaskBDRelease(txtask);
	tx_released++;
	if (bdnum != tx_released % MAX_DMA_BUFFERS)
		printk("%s: %d:%ld\n", __FUNCTION__, bdnum, tx_released % MAX_DMA_BUFFERS);
#ifdef LITEDEBUG
	printk("-");
#endif
	WDPRINTK("%s: released bd %d\n", __FUNCTION__, 
			 (bdnum + MAX_DMA_BUFFERS - 1) % MAX_DMA_BUFFERS);
	if (tx_assigned > tx_released) {
		bdnum = TaskBDAssign(txtask, (void *)i2s_tx_bufs[bdnum].pa, NULL, 
							 i2s_tx_bufs[bdnum].len, 0);
		WDPRINTK("%s: assigned bd %d\n", __FUNCTION__, bdnum);
#ifdef LITEDEBUG
		printk("+");
#endif
#ifdef DEBUG_LOST_INT
		tx_acount++;
#endif
	}
#ifdef DEBUG_LOST_INT
	else
		printk("%s: %ld:%ld:%ld\n", __FUNCTION__, tx_assigned, tx_acount, tx_released);
#endif
	wake_up(&i2s_wait_tx);
}

static void i2s_rx_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	int bdnum;

	if (rx_assigned == rx_released) {
		printk("%s: %ld:%ld - false int??\n", __FUNCTION__, rx_assigned, rx_released);
		return;
	}

	TaskIntClear(rxtask);

	bdnum = TaskBDRelease(rxtask);
	rx_released++;
	if (bdnum != rx_released % MAX_DMA_BUFFERS)
		printk("%s: %d:%ld\n", __FUNCTION__, bdnum, rx_released % MAX_DMA_BUFFERS);
	RDPRINTK("%s: released bd %d\n", __FUNCTION__, 
			 (bdnum + MAX_DMA_BUFFERS - 1) % MAX_DMA_BUFFERS);
#ifdef LITEDEBUG
	printk("-");
#endif
	if (rx_assigned > rx_released) {
		bdnum = TaskBDAssign(rxtask, (void *)i2s_rx_bufs[bdnum].pa, NULL, BUFSIZE, 0);
		RDPRINTK("%s: assigned bd %d\n", __FUNCTION__, bdnum);
#ifdef LITEDEBUG
		printk("+");
#endif
	}
	wake_up(&i2s_wait_rx);
}

/**
 *	i2s_write:
 *	@file: file handle
 *	@buf: buffer to write (unused as data does not matter here 
 *	@count: count of bytes
 *	@ppos: pointer to the position to write. No seeks allowed
 *
 */
static ssize_t i2s_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	size_t cnt = count;
	size_t tc;
	char *ptr = (char *)buf;
	int bdnum, err;
	int i, n, flags;

	if (!configured)
		return -EINVAL;

	/*  Can't seek (pwrite) on this device  */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (once) {
		i2s_start_rx_tx(txtask);
		once = 0;
	}

	WDPRINTK("%s: count %d\n", __FUNCTION__, count);

	while(cnt) {
		/* Wait until all buffers will be released */
		err = wait_event_interruptible(i2s_wait_tx, tx_assigned == tx_released);
		if (err) {
#ifdef DEBUG_LOST_INT
			printk("%s: interrupted, %ld:%ld:%ld, tx_int %d, err %d\n", __FUNCTION__, 
				   tx_assigned, tx_acount, tx_released, tx_int, err);
#endif
			break;
		}

		/* prefill the buffers */
		for (i = 0; i < MAX_DMA_BUFFERS && cnt; i++) {
			tc = (cnt > BUFSIZE) ? BUFSIZE : cnt;
			if (copy_from_user(i2s_tx_bufs[tx_assigned % MAX_DMA_BUFFERS].va, 
							   ptr, tc)) {
				err =  -EFAULT;
				break;
			}

			i2s_tx_bufs[tx_assigned % MAX_DMA_BUFFERS].len = tc;
			tx_assigned++;
			ptr += tc;
			cnt -= tc;
		}

		/* Now assign the first buffer to the Bestcomm task */
#ifdef LITEDEBUG
		printk("W");
#endif
		n = tx_released % MAX_DMA_BUFFERS;

		save_flags(flags);cli(); /* Bestcomm can loose interrupts otherwise ??? */

#ifdef DEBUG_LOST_INT
		tx_int = 0;
		tx_acount++;
#endif
		bdnum = TaskBDAssign(txtask, (void *)i2s_tx_bufs[n].pa, NULL, 
							 i2s_tx_bufs[n].len, 0);
		restore_flags(flags);
		if (bdnum < 0) {
			printk("%s:TaskBDAssign error %d(tx stuck??)\n",
				   __FUNCTION__, bdnum);
			return -EIO;
		}
		WDPRINTK("%s: assigned bd %d (buf %d), tc %d\n", __FUNCTION__,
				 bdnum, n, tc);
	}

	WDPRINTK("%s: sent %d, err %d\n", __FUNCTION__, count - cnt, err);
	return (count - cnt) ? (count - cnt) : err;
} 


/**
 *	i2s_read:
 *	@file: file handle
 *	@buf: buffer to read
 *	@count: length of buffer
 *	@ptr: offset (no seek allowed)
 *
 */
 
static ssize_t i2s_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	size_t cnt = count;
	char *ptr = buf;
	size_t bd_count;
	int bdnum;
	int err;

	if (!configured)
		return -EINVAL;

	/*  Can't seek (pread) on this device  */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (once) {
		i2s_start_rx_tx(rxtask);
		once = 0;
	}

	while (cnt) {
		/* How much BDs are needed (up to MAX_DMA_BUFFERS) */
		bd_count = (cnt + BUFSIZE - 1) / BUFSIZE;
		if (bd_count > MAX_DMA_BUFFERS)
			bd_count = MAX_DMA_BUFFERS;
		rx_assigned += bd_count;
		RDPRINTK("%s: cnt %d, rx_data %ld, ass %ld, rel %ld\n", __FUNCTION__, cnt,
				 rx_data, rx_assigned, rx_released);

		/* Start reading the data */
#ifdef LITEDEBUG
		printk("R");
#endif
		bdnum = TaskBDAssign(rxtask, 
							 (void *)i2s_rx_bufs[rx_released % MAX_DMA_BUFFERS].pa,
							 NULL, BUFSIZE, 0);
		if (bdnum < 0) {
			printk("%s:TaskBDAssign error %d\n",
				   __FUNCTION__, bdnum);
			return -EIO;
		}
		RDPRINTK("%s:assigned bd %d (buf %ld)\n", __FUNCTION__,
				 bdnum, rx_assigned % MAX_DMA_BUFFERS);

		/* Copy the data to user as it arrives */
		while (bd_count--) {
			int len;
			len = (cnt > BUFSIZE) ? BUFSIZE : cnt;

			RDPRINTK("%s:to user:rx_data %ld, ass %ld, rel %ld\n", 
__FUNCTION__, rx_data,
				 rx_assigned, rx_released);
			/* Wait at least one BD released */
			err = wait_event_interruptible(i2s_wait_rx, rx_data < rx_released);
			if(err) {
				goto out;
			}
#ifdef LITEDEBUG
			printk("U");
#endif
			if (copy_to_user(ptr, i2s_rx_bufs[rx_data % MAX_DMA_BUFFERS].va, len)) {
				err = -EFAULT;
				goto out;
			}
			ptr += len;
			cnt -= len;
			rx_data ++;
		}
	}
 out:
	RDPRINTK("%s: read %d, err %d\n", __FUNCTION__, count - cnt, err);
	return (count - cnt) ? (count - cnt) : err;
}


static void i2s_start_rx_tx (int task)
{
	if (task == rxtask) {
#ifdef RX_TASK_START
		TaskStart(rxtask, TASK_AUTOSTART_ENABLE, rxtask, TASK_INTERRUPT_ENABLE);
		/*		printk("%s: Rx task %d started\n", __FUNCTION__, rxtask);*/
#endif
	} else {
#ifdef TX_TASK_START
		TaskStart(txtask, TASK_AUTOSTART_ENABLE, txtask, TASK_INTERRUPT_ENABLE);
		/*		printk("%s: Tx task %d started\n", __FUNCTION__, txtask);*/
#endif
	}
	psc_enable();
}

static void psc_enable(void)
{
	struct mpc5xxx_psc *psc = (struct mpc5xxx_psc *)(MPC5xxx_PSC1 + ((psc_num - 1) * 0x200));
	psc->command = MPC5xxx_PSC_RX_ENABLE | MPC5xxx_PSC_TX_ENABLE;
}

static void psc_reset(void)
{
	struct mpc5xxx_psc *psc = (struct mpc5xxx_psc *)(MPC5xxx_PSC1 + ((psc_num - 1) * 0x200));

	psc->command = MPC5xxx_PSC_RST_RX;
	psc->command = MPC5xxx_PSC_RST_TX;
	psc->command = MPC5xxx_PSC_SEL_MODE_REG_1;
	psc->command = MPC5xxx_PSC_RST_ERR_STAT;
}

static int i2s_channel_setup(struct i2s_ioctl *arg)
{
	struct mpc5xxx_psc *psc;
	struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	struct mpc5xxx_cdm *cdm = (struct mpc5xxx_cdm *)MPC5xxx_CDM;
	TaskSetupParamSet_t i2s_setup_tx;
	TaskSetupParamSet_t i2s_setup_rx;

#if 0
	/* Valid PSCs are 1,2,3 and 6 */
	if (arg->psc_num < 1 ||
		arg->psc_num == 4 || arg->psc_num == 5 ||
		arg->psc_num > 6)
		return -EINVAL;
#else
	if (arg->psc_num < 1 || arg->psc_num > 3)
		return -EINVAL;
#endif

	psc_num = arg->psc_num;

	switch (psc_num) {
	case 1:
		initiator_tx = INITIATOR_PSC1_TX;
		initiator_rx = INITIATOR_PSC1_RX;
		cdm->mclken_div_psc1 |= (0x8000 | arg->mclk_div);
		break;
	case 2:
		initiator_tx = INITIATOR_PSC2_TX;
		initiator_rx = INITIATOR_PSC2_RX;
		cdm->mclken_div_psc2 |= (0x8000 | arg->mclk_div);
		break;
	case 3:
		initiator_tx = INITIATOR_PSC3_TX;
		initiator_rx = INITIATOR_PSC3_RX;
		cdm->mclken_div_psc3 |= (0x8000 | arg->mclk_div);
		break;
	}

	psc = (struct mpc5xxx_psc *)(MPC5xxx_PSC1 + ((psc_num - 1) * 0x200));
	psc->command = (MPC5xxx_PSC_TX_DISABLE | MPC5xxx_PSC_RX_DISABLE); /* disable Tx & Rx */
	gpio->port_config &= ~(0x7 << ((psc_num - 1) * 4));

	if (arg->mclk) {
		gpio->port_config |= (0x7 << ((arg->psc_num - 1) * 4)); /* CODEC with MCLK */
	} else {
		gpio->port_config |= (0x6 << ((arg->psc_num - 1) * 4)); /* CODEC without MCLK */
	}

	cdm->clk_enables |= (1 << (arg->psc_num + 4)); /* PSC clock enable */
	/* reset PSC */
	psc_reset();
	
	psc->mode = 0;
	psc->rfalarm = 0x120; /* alarm threshold level */
	psc->tfalarm = 0x120;
	psc->rfcntl = 4; /* granularity */
	psc->tfcntl = 4;
	psc->mpc5xxx_psc_imr = 0; /* disable interrupts */

	/* Configure codec parameters  (MSB first) */
	switch (arg->data_width) {
	case 8:
		psc->sicr = 0x01000000;
		psc->ctur = 0x7; /* 8 bits per frame */
		psc->ccr = arg->frsync_div  | arg->bclk_div;
		break;
	case 16:
		psc->sicr = 0x02000000;
		psc->ctur = 0xf; /* 16 bits per frame */
		psc->ccr = arg->frsync_div | arg->bclk_div;
		break;
	case 24:
		psc->sicr = 0x07000000;
		psc->ctur = 0x17; /* 24 bits per frame */
		psc->ccr = arg->frsync_div | arg->bclk_div;
		break;
	case 32:
		psc->sicr = 0x0f000000;
		psc->ctur = 0x1f; /* 32 bits per frame */
		psc->ccr = arg->frsync_div | arg->bclk_div;
		break;
	default:
		return -EINVAL;
	}

	/* DTS = 1, Data MSB first, use 5200 clk and frame sync, multiword.
	 * Frame sync active low, data sampled on CLK high.
	 */
	psc->sicr |= DELAY_TIME_SLOT | MULTIWD_ENABLE | CLK_POL_RISING;

	if (arg->master) /* Master mode - PSC will drive SCLK and LRCK */
		psc->sicr |= GEN_CLK_INT;

#if 0
	if (arg->le) /* Shift data LSB fist */
		psc->sicr |= LSB_FIRST;
#endif

#if 1
	/* reset PSC again */
	psc_reset();
#endif
#if 0
	/* Local loopback mode enable */
	{
		volatile int val;
		val = psc->mode; /* to switch from mode1 to mode2 */
		val++;
		psc->mode = 0x80;
	}
#endif

	/* setup the tasks */
	memset(&i2s_setup_tx, 0, sizeof(i2s_setup_tx));
	i2s_setup_tx.Initiator = (MPC5200Initiator_t)(initiator_tx);
	i2s_setup_tx.NumBD = MAX_DMA_BUFFERS;
	i2s_setup_tx.Size.MaxBuf = BUFSIZE;
	i2s_setup_tx.StartAddrDst = (volatile u32)&(psc->tfdata);
	i2s_setup_tx.IncrDst = 0;
	i2s_setup_tx.SzDst = 4;
	txtask = TaskSetup(TASK_GEN_TX_BD, &i2s_setup_tx);

	if (request_irq(MPC5xxx_SDMA_IRQ_BASE + txtask, i2s_tx_irq,
					SA_INTERRUPT, "i2s tx dma", NULL)) {
		printk(KERN_ERR "i2s: SDMA tx irq allocation failed\n");
		return -EINVAL;
	}

	memset(&i2s_setup_rx, 0, sizeof(i2s_setup_rx));
	i2s_setup_rx.Initiator = (MPC5200Initiator_t)(initiator_rx);
	i2s_setup_rx.NumBD = MAX_DMA_BUFFERS;
	i2s_setup_rx.Size.MaxBuf = BUFSIZE;
	i2s_setup_rx.StartAddrSrc = (volatile u32)&(psc->rfdata);
	i2s_setup_rx.IncrDst = 4;
	i2s_setup_rx.IncrSrc = 0;
	i2s_setup_rx.SzDst = 4;
	rxtask = TaskSetup(TASK_GEN_RX_BD, &i2s_setup_rx);

	if (request_irq(MPC5xxx_SDMA_IRQ_BASE + rxtask, i2s_rx_irq,
					SA_INTERRUPT, "i2s rx dma", NULL)) {
		printk(KERN_ERR "i2s: SDMA rx irq allocation failed\n");
		return -EINVAL;
	}

	configured = 1;
	return 0;
}

/**
 *	i2s_ioctl:
 *	@inode: inode of the device
 *	@file: file handle to the device
 *	@arg: argument pointer
 *
 */
 
static int i2s_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg)
{
	int ret;
	struct i2s_ioctl i2s_arg;
	switch (cmd) {
	case I2S_SET_CHANNEL:
		if (copy_from_user (&i2s_arg, (struct i2s_ioctl *)arg, sizeof(i2s_arg)))
			return -EFAULT;
		ret = i2s_channel_setup(&i2s_arg);
		break;
	default:
		ret = -EINVAL;
		break;
	}
	return ret;
}

/**
 *	i2s_open:
 *	@inode: inode of device
 *	@file: file handle to device
 */
 
static int i2s_open(struct inode *inode, struct file *file)
{
	if(test_and_set_bit(0, &i2s_is_open))
		return -EBUSY;

	i2s_is_open = 1;
	tx_assigned = tx_released = 0;
	rx_assigned = rx_released = 0;
#ifdef DEBUG_LOST_INT
	tx_acount = 0;
#endif
	rx_data = 0;
	once = 1;
	return 0;

}

/**
 *	i2s_release
 *	@inode: inode to board
 *	@file: file handle to board
 */
 
static int i2s_release(struct inode *inode, struct file *file)
{
	wait_event_interruptible(i2s_wait_tx, tx_assigned == tx_released);
	psc_reset();

	if (rxtask) {
		TaskStop(rxtask);
		free_irq(MPC5xxx_SDMA_IRQ_BASE + rxtask, NULL);
		rxtask = 0;
	}
	if (txtask) {
		TaskStop(txtask);
		free_irq(MPC5xxx_SDMA_IRQ_BASE + txtask, NULL);
		txtask = 0;
	}

	configured = 0;
	clear_bit(0, &i2s_is_open);
	return 0;
}

static unsigned int i2s_poll(struct file *file, poll_table *wait)
{
	int ret = 0;
	if (configured) {
#if 0
		poll_wait (file, &i2s_wait_rx, wait);
		if (rx_int)
			ret = (POLLIN | POLLRDNORM);
#else
		ret = (POLLIN | POLLRDNORM);
#endif
	} 

	return ret;
}
 

static struct file_operations i2s_fops = {
	owner:		THIS_MODULE,
	llseek:		no_llseek,
	read:		i2s_read,
	write:		i2s_write,
	ioctl:		i2s_ioctl,
	open:		i2s_open,
	release:	i2s_release,
	poll:       i2s_poll,
};

static struct miscdevice i2s_miscdev=
{
	I2S_MINOR,
	"I2S",
	&i2s_fops
};


/**
 *	cleanup_module:
 */
 
static void __exit i2s_exit(void)
{
	misc_deregister(&i2s_miscdev);
	i2s_buffers_release();
	TaskStop(txtask);
	free_irq(MPC5xxx_SDMA_IRQ_BASE + txtask, NULL);
	TaskStop(rxtask);
	free_irq(MPC5xxx_SDMA_IRQ_BASE + rxtask, NULL);
}

/**
 * 	i2s_init:
 */
 
static int __init i2s_init(void)
{
	int ret;
	struct mpc5xxx_sdma *sdma = (struct mpc5xxx_sdma *)MPC5xxx_SDMA;

	ret = misc_register(&i2s_miscdev);
	if (ret) {
		printk(KERN_ERR "i2s: can't misc_register on minor=%d\n", I2S_MINOR);
		goto out;
	}
	i2s_buffers_init();

	TasksInitAPI((uint8 *)MPC5xxx_MBAR);
	sdma->taskBar = MPC5xxx_SRAM;
	TasksLoadImage((sdma_regs *)MPC5xxx_SDMA);

	ret = 0;
	printk(KERN_INFO "MPC5200 I2S driver (%d BDs x %d bytes).\n", 
		   MAX_DMA_BUFFERS, BUFSIZE);

out:
	return ret;

}

module_init(i2s_init);
module_exit(i2s_exit);

MODULE_DESCRIPTION("Simple driver for the MPC5200 PSC in CODEC I2S mode.\n");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
