/*
 * arch/ppc/8220_io/fec.c
 *
 * Driver for the mpc8220 Fast Ethernet Controller
 *
 * Author: MontaVista Software, Inc.
 *         source@mvista.com
 *
 * Based on fec.c version with modifications for Alaska by Dale Farnsworth
 *  <dale.farnsworth@mvista.com>.
 *
 * 2003-2004 MontaVista, Software, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

static const char *drv_name = "MPC8220 FEC";
static const char *version = "0.2.1";

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/crc32.h>
#include <linux/pci.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/skbuff.h>
#include <asm/delay.h>
#include <asm/uaccess.h>

#include <linux/vmalloc.h>
#include <linux/mm.h>
#include "bestdma/mcdapi/MCD_dma.h"
#include "bestdma/8220DmaApi.h"
#include "fec.h"

#undef USE_MIB			/* get stats from the MIB counters */

static int rcv_ring_size = RCV_RING_SIZE;

static int __init
rcvsize(char *str)
{
	get_option(&str, &rcv_ring_size);
	return 1;
}

__setup("rcvsize=", rcvsize);

static int xmt_ring_size = XMT_RING_SIZE;

static int __init
xmtsize(char *str)
{
	get_option(&str, &xmt_ring_size);
	return 1;
}

__setup("xmtsize=", xmtsize);

/* How big is the static pool */
#define R_BUFFER_SIZE 8192

/* How big is each block */
#define BUF_SIZE L1_CACHE_BYTES

/* How many buffers */
#define NUM_BUFS ((R_BUFFER_SIZE + BUF_SIZE -1)/BUF_SIZE)

#define MTU_SIZE 1600

static void
phy_processCB (struct net_device *dev)
{
	struct mpc8220_fec_priv* priv = (struct mpc8220_fec_priv*)dev->priv;
	wake_up_interruptible (&priv->phy_queue);
}

/*
 * This routine is used for two purposes.  The first is to keep the
 * FEC's duplex setting in sync with the PHY's.  The second is to keep
 * the system apprised of the state of the link.  Note that this driver
 * does not configure the PHY.  Either the PHY should be configured for
 * auto-negotiation or it should be handled by something like mii-tool.
 */
#define MIIPOLL_EMPTY		0
#define MIIPOLL_ANAR_GOT	1
#define MIIPOLL_ANPAR_GOT	2
#define MIIPOLL_LINKUP		3

static void cont_poll_mii(struct net_device *dev);
static int phy_action (struct mpc8220_fec_priv *fec_env, unsigned int action,
		unsigned char reg);
static void
SM_poll_mii(struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *)dev->priv;
	u32 x_reg;
	u32 r_reg;
	struct mpc8220_fec *fec = priv->fec;
	int ret;

	/* First, find out what's going on with the PHY. */
	switch (priv->miipoll_SMstate) {
	case MIIPOLL_EMPTY:
		priv->phy_readCB = cont_poll_mii;
		if (test_and_set_bit(0, &priv->phy_lock))
			break;
		priv->miipoll_SMstate = MIIPOLL_ANAR_GOT;
		if ((ret = phy_action (priv, PHY_OP_READ, MII_ADVERTISE)) < 0) {
			printk(KERN_INFO "%s: PHY duplex detection anar ERROR %d\n",
				dev->name, ret);
			BUG();
		}
		break;
	case MIIPOLL_ANAR_GOT:
		priv->negotiated = priv->phy_data.data;
		priv->miipoll_SMstate = MIIPOLL_ANPAR_GOT;
		if ((ret = phy_action (priv, PHY_OP_READ, MII_LPA)) < 0) {
			printk(KERN_INFO "%s: PHY duplex detection anpar ERROR %d\n",
				dev->name, ret);
			BUG();
		}
		break;
	case MIIPOLL_ANPAR_GOT:
		priv->negotiated &= priv->phy_data.data;
		priv->phy_duplex = mii_duplex (0, priv->negotiated);
		priv->miipoll_SMstate = MIIPOLL_LINKUP;
		if ((ret = phy_action (priv, PHY_OP_READ, MII_BMSR)) < 0) {
			printk(KERN_INFO "%s: PHY carrier detection ERROR %d\n",
				dev->name, ret);
			BUG();
		}
		break;
	case MIIPOLL_LINKUP:
		priv->phy_carrier = ((priv->phy_data.data & BMSR_LSTATUS) != 0);

		r_reg = in_be32(&fec->r_cntrl);
		x_reg = in_be32(&fec->x_cntrl);
		if (((r_reg & 0x00000002) == 0x00000000) &&
			((x_reg & 0x00000006) == 0x00000006))
				priv->fec_duplex = 1;
		else if (((r_reg & 0x00000002) == 0x00000002) &&
			((x_reg & 0x00000006) == 0x00000000))
				priv->fec_duplex = 0;
		else {
			printk (KERN_INFO "duplex ERROR: x_reg=%08x,r_reg=%08x\n",
				x_reg, r_reg);
			BUG ();
		}
		if (priv->fec_duplex != priv->phy_duplex) {
			/* Stop the FEC */
			out_be32(&fec->ecntrl, (in_be32(&fec->ecntrl) & ~MPC8220_FEC_ECNTRL_ENABLE));
			if (priv->phy_duplex) {
				printk(KERN_INFO "%s: Duplex has been changed: now %s\n",
					dev->name, "FULL DUPLEX");
				out_be32 (&fec->r_cntrl, (in_be32(&fec->r_cntrl) & (~0x00000002)));
				out_be32 (&fec->x_cntrl, (in_be32(&fec->x_cntrl) | 0x00000006));
			} else {
				printk(KERN_INFO "%s: Duplex has been changed: now %s\n",
					dev->name, "HALF DUPLEX");
				out_be32 (&fec->r_cntrl, (in_be32(&fec->r_cntrl) | 0x00000002));
				out_be32 (&fec->x_cntrl, (in_be32(&fec->x_cntrl) & (~0x00000006)));

			}
			/* Start up the FEC */
			out_be32(&fec->ecntrl, (in_be32(&fec->ecntrl) | MPC8220_FEC_ECNTRL_ENABLE));
			MCD_continDma(priv->r_tasknum);
			MCD_continDma(priv->t_tasknum);
		}

		priv->netif_carrier = netif_carrier_ok(dev) != 0;
		if (priv->phy_carrier != priv->netif_carrier) {
			if (priv->phy_carrier) {
				printk(KERN_INFO "%s: Link carrier restored.\n",
					dev->name);
				netif_carrier_on(dev);
			} else {
				printk(KERN_INFO "%s: Link carrier lost.\n",
					dev->name);
				netif_carrier_off(dev);
			}
		}
		priv->miipoll_SMstate = MIIPOLL_EMPTY;
		/* Set up the timer so we'll get called again in 2 seconds. */
		priv->phy_timer.expires = jiffies + 2 * HZ;
		add_timer(&priv->phy_timer);
		clear_bit (0, &priv->phy_lock);
		wake_up_interruptible (&priv->phy_queue);
		break;
	default:
		printk (KERN_INFO "SM_poll_mii state %d ERROR\n",
			priv->miipoll_SMstate);
		BUG ();
	}

}

static void
poll_mii(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	unsigned long flags;

	if (priv->miipoll_SMstate != MIIPOLL_EMPTY) {
		printk (KERN_INFO "poll_mii state (%d) ERROR\n",
				priv->miipoll_SMstate);
		BUG();
	} else {
		spin_lock_irqsave(&priv->mii_irq_lock, flags);
		SM_poll_mii(dev);
		spin_unlock_irqrestore(&priv->mii_irq_lock, flags);
	}
}

static void
cont_poll_mii(struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	unsigned long flags;

	if (priv->miipoll_SMstate == MIIPOLL_EMPTY) {
		printk (KERN_INFO "mii interrupt state (%d) ERROR\n",
				priv->miipoll_SMstate);
		BUG();
	} else {
		spin_lock_irqsave(&priv->mii_irq_lock, flags);
		SM_poll_mii(dev);
		spin_unlock_irqrestore(&priv->mii_irq_lock, flags);
	}
}

/* PHY state machine states */
#define PHY_SM_IDLE	0
#define PHY_SM_READING	1
#define PHY_SM_WRITING	2

static int
phy_action (struct mpc8220_fec_priv *fec_env, unsigned int action, unsigned char reg)
{
	struct mpc8220_fec *fec = fec_env->fec;
	int ret;
	u32 rval;

	if (action != PHY_OP_CONTINUE) {
		if (fec_env->phy_data.state != PHY_SM_IDLE) {
			printk (KERN_DEBUG "PHY operation state (%x) ERROR\n",
				fec_env->phy_data.state);
			return -EALREADY;
		}
		fec_env->phy_data.state = PHY_SM_IDLE;
	}
	out_be32 (&fec->mii_speed, (0xa << 1));	/* 0x0a - value, recommended by
						   MPC8220i manual */
	ret = PHY_RET_WAIT;
	switch (fec_env->phy_data.state) {
	case PHY_SM_IDLE:
		if ((action != PHY_OP_READ) && (action != PHY_OP_WRITE)) {
			printk (KERN_DEBUG "PHY operation code (%x) ERROR\n",
				action);
			ret = -EPERM;
			break;
		}
		if (reg > 30) {
			ret = -EINVAL;
			break;
		}
		rval = ((0x1<<30) |
			(action << 28) |
			(fec_env->phy_data.address << 23) |
			(reg << 18) | (0x2 << 16) | fec_env->phy_data.data);

		out_be32(&fec->mii_data, rval);
		fec_env->phy_data.state = (action == PHY_OP_READ) ?
			PHY_SM_READING : PHY_SM_WRITING;
		ret = PHY_RET_WAIT;
		break;
	case PHY_SM_READING:
		rval = in_be32(&fec->mii_data);
		fec_env->phy_data.data = rval & 0xffff;
	case PHY_SM_WRITING:
		fec_env->phy_data.state = PHY_SM_IDLE;
		ret = PHY_RET_COMPLETE;
		break;
	default:
		printk (KERN_DEBUG "PHY SM state (%d) wrong ERROR\n",
			fec_env->phy_data.state);
		ret = -EPERM;
	}
	return ret;
}

static inline int
phy_process_wait (struct net_device *dev, unsigned int action, unsigned char reg, char* err_text)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *)dev->priv;
	wait_queue_t __wait;
	init_waitqueue_entry(&__wait, current);
	int ret;
	unsigned long flags;

	add_wait_queue(&priv->phy_queue, &__wait);
	set_current_state(TASK_INTERRUPTIBLE);
	spin_lock_irqsave(&priv->mii_irq_lock, flags);
	if ((ret = phy_action (priv, action, reg)) < 0)
		printk(KERN_DEBUG "%s: %s ERROR %d\n",dev->name, err_text, ret);
	spin_unlock_irqrestore(&priv->mii_irq_lock, flags);
	if (ret >= 0) {
		ret = schedule_timeout(2);
		if (signal_pending(current))
			ret = -ERESTARTSYS;
		else if (!ret) {
			printk(KERN_DEBUG "%s: %s ERROR - timeout\n",
				dev->name,err_text);
			ret = -ETIME;
		}
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&priv->phy_queue, &__wait);
	return ret;
}

extern u32 scEthernetRecv_TDT;
extern u32 scEthernetXmit_TDT;

static struct net_device *mpc8220_fec_dev[2];
static int mpc8220_fec_dev_cnt = 0;

static void mpc8220_fec_interrupt(int, void *, struct pt_regs *);

static void my_sdma_receive_interrupt(int, void *, struct pt_regs *);
static void my_sdma_transmit_interrupt(int, void *, struct pt_regs *);

static struct net_device_stats *mpc8220_fec_get_stats(struct net_device *);

static void
mpc8220_fec_tx_timeout(struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;

	printk(KERN_INFO "FEC %p: xmit timeout: TF_STS %08x TF_CTRL %08x\n",
	       priv->fec, priv->fec->tfifo_status, priv->fec->tfifo_cntrl);
	if (!priv->link) {
		printk(KERN_INFO
		       "%s: transmit timed out, network cable problem?\n",
		       dev->name);
	}

	priv->stats.tx_errors++;

	if (!priv->tx_full)
		netif_wake_queue(dev);
}

static void
mpc8220_fec_set_paddr(struct net_device *dev, u8 * mac)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	struct mpc8220_fec *fec = priv->fec;

	out_be32(&fec->paddr1, (mac[0] << 24) | (mac[1] << 16)
		 | (mac[2] << 8) | (mac[3] << 0));
	out_be32(&fec->paddr2, (mac[4] << 24) | (mac[5] << 16) | 0x8808);
}

static int
mpc8220_fec_set_mac_address(struct net_device *dev, void *addr)
{
	struct sockaddr *sock = (struct sockaddr *) addr;

	/* Save the address in the dev_addr so SIOCGIFHWADDR can get it back */
	memcpy(dev->dev_addr, sock->sa_data, dev->addr_len);

	mpc8220_fec_set_paddr(dev, sock->sa_data);
	return 0;
}

void
startXmtDma(struct mpc8220_fec_priv *priv, MCD_bufDescFec * bufs)
{
	struct mpc8220_fec *fec = priv->fec;
	int ret;

	ret = MCD_startDma(priv->t_tasknum,	/* channel */
			   (s8 *) bufs,	/* source */
			   sizeof (int),	/* src incr */
			   (s8 *) & fec->tfifo_data.longs,	/* dst */
			   0,	/* dst incr */
			   0,	/* length in bytes */
			   sizeof (int),	/* width of transfer */
			   priv->xmit_initiator,	/* initiator */
			   4,	/* priority */
			   MCD_FECTX_DMA | MCD_TT_FLAGS_RL,
			   MCD_NO_CSUM | MCD_NO_BYTE_SWAP | MCD_NO_BIT_REV);
}

void
startRcvDma(struct mpc8220_fec_priv *priv, MCD_bufDescFec * bufs)
{
	struct mpc8220_fec *fec = priv->fec;
	int ret;

	ret = MCD_startDma(priv->r_tasknum,	/* Channel Number */
			   (char *) bufs,	/* Buffer descriptor chain */
			   0,	/* No Incrementing Source Addr */
			   (s8 *) & fec->rfifo_data,	/* Ignored */
			   sizeof (int),	/* Increment Dest Addr 1 word */
			   0,	/* Ignored */
			   sizeof (int),	/* 32-bit / word transfers */
			   priv->recv_initiator,	/* Initiator */
			   4,	/* Priority */
			   MCD_FECRX_DMA | MCD_TT_FLAGS_CW,
			   MCD_NO_BYTE_SWAP | MCD_NO_BIT_REV | MCD_NO_CSUM);
}

static void
markRcvPairReady(struct mpc8220_packet_set *b, int last)
{
	/* Make sure to set the 2nd buffer first since if not, then we
	 * could stop *in between* the packet data and the frame which will
	 * cause untold havoc */
	if (last)
		b->data.statCtrl = (MCD_FEC_BUF_READY | MCD_FEC_END_FRAME |
			MCD_FEC_INTERRUPT | MCD_FEC_WRAP);
	else
		b->data.statCtrl = (MCD_FEC_BUF_READY | MCD_FEC_END_FRAME |
			MCD_FEC_INTERRUPT);
}

static void
markXmtPacketReady(MCD_bufDescFec * p, int last)
{
	if (last)
		p->statCtrl = (MCD_FEC_BUF_READY | MCD_FEC_INTERRUPT | MCD_FEC_END_FRAME | MCD_FEC_WRAP);
	else
		p->statCtrl = MCD_FEC_BUF_READY | MCD_FEC_INTERRUPT | MCD_FEC_END_FRAME;
}

static void
initXmtPacket(struct mpc8220_xmt_packet *p, struct mpc8220_fec *fec,
	      MCD_bufDescFec * b)
{
	p->b = b;
	p->fec = fec;
	p->skb = 0;
	p->len = 0;
	p->rdy = 0;

	b->length = 0;		/* filled in by hard_xmit */
	b->dataPointer = 0;	/* filled in by hard_xmit */

	/* Clear the flags so the buffer is not ready to transmit! */
	b->statCtrl = 0;
}

static void
initRcvPacketPair(struct mpc8220_fec *fec, struct mpc8220_packet_pair *p,
		  int last)
{
	struct mpc8220_packet_set *b = p->pair;

	b->data.length = p->skb_size;
	b->data.dataPointer =
	    (u32) pci_map_single(NULL, skb_put(p->skb, 0), p->skb_size,
				 PCI_DMA_FROMDEVICE);

	markRcvPairReady(b, last);
}

void
initRcvRingHeader(struct mpc8220_rcv_ring *p, int n)
{
	int i;

	for (i = 0; n && (n & -n) != n; ++i) {
		n ^= (n & -n);
	}

	if (i) {
		printk(KERN_DEBUG
		       "%s: failure - n(%x) must be a power of two\n", __func__,
		       n);
		BUG();
	}

	if (!n) {
		printk(KERN_DEBUG "%s: failure - n can't be zero\n", __func__);
		BUG();
	}
	p->idx = 0;
	p->size = n;
	p->mask = n - 1;

	p->ring =
	    (struct mpc8220_packet_pair *) vmalloc(n *
						   sizeof (struct
							   mpc8220_packet_pair
							   *));
	if (!p->ring) {
		printk(KERN_DEBUG "%s: failure - vmalloc returns zero\n",
		       __func__);
		BUG();
	}
}

void
initXmtRingHeader(struct mpc8220_xmt_ring *p, int n)
{
	int i;

	for (i = 0; n && (n & -n) != n; ++i) {
		n ^= (n & -n);
	}

	if (i) {
		printk(KERN_DEBUG
		       "%s: failure - n(%x) must be a power of two\n", __func__,
		       n);
		BUG();
	}

	if (!n) {
		printk(KERN_DEBUG "%s: failure - n can't be zero\n", __func__);
		BUG();
	}

	p->head = p->tail = 0;
	p->size = n;
	p->mask = n - 1;
	p->stopped = 0;
	p->ring = 0;
}

static void
initRcvRing(struct net_device *dev, int n)
{
	MCD_bufDescFec *bufs;
	struct mpc8220_fec_priv *priv = dev->priv;
	struct mpc8220_fec *fec = priv->fec;
	struct mpc8220_rcv_ring *ring;
	struct mpc8220_packet_pair *r;
	int i;
	char *p;

	initRcvRingHeader(ring = (&priv->rcvRing), n);
	r = ring->ring;

	/* first allocate a list of buffer descriptors, two per packet since
	   one of them has to hold the frame word */
	bufs = MSD_allocBufDescriptorsFec(n);
	if (!bufs) {
		printk(KERN_DEBUG "%s: failure - bufs can not be allocated\n",
		       __func__);
		BUG();
	}
	memset((void *) bufs, 0, n * sizeof (*bufs));

	/* Fill in the ring */
	for (i = 0; i < n; ++i, ++r) {
		r->pair = (struct mpc8220_packet_set *) &bufs[i];
		r->skb = dev_alloc_skb(r->skb_size = MTU_SIZE);

		if (!r->skb) {
			printk(KERN_DEBUG "%s: failure - wrong skb pointer\n",
			       __func__);
			BUG();
		}

		if ((i + 1) == n)
			initRcvPacketPair(fec, r, 1);
		else
			initRcvPacketPair(fec, r, 0);

		p = skb_push(r->skb, 0);
		flush_dcache_range((u32) p, (u32) p + r->skb_size);
		invalidate_dcache_range((u32) p, (u32) p + r->skb_size);
	}

	/* Link last buffer to the first */

}

static void
initXmtRing(struct net_device *dev, int n)
{
	MCD_bufDescFec *bufs;
	struct mpc8220_fec_priv *priv = dev->priv;
	struct mpc8220_fec *fec = priv->fec;
	struct mpc8220_xmt_ring *ring;
	int i;

	initXmtRingHeader(ring = (&priv->xmtRing), n);

	/* first allocate a list of buffer descriptors, one per packet */
	bufs = MSD_allocBufDescriptorsFec(n);
	if (!bufs) {
		printk(KERN_DEBUG "%s: failure - bufs can not be allocated\n",
		       __func__);
		BUG();
	}
	memset((void *) bufs, 0, n * sizeof (*bufs));

	ring->ring =
	    (struct mpc8220_xmt_packet *) vmalloc(n *
						  sizeof (struct
							  mpc8220_xmt_packet));

	/* Fill in the ring */
	for (i = 0; i < n; ++i) {
		initXmtPacket(&ring->ring[i], fec, &bufs[i]);
	}
}

void
freeRcvRing(struct mpc8220_rcv_ring *ring)
{
	struct mpc8220_packet_pair *r = ring->ring;
	int i;

	for (i = 0; i < ring->size; ++i, ++r) {
		dev_kfree_skb(r->skb);
		MSD_freeBufDescriptorsFec(&r->pair->data, 1);
	}

	vfree(ring->ring);
}

void
freeXmtRing(struct mpc8220_xmt_ring *ring)
{
	struct mpc8220_xmt_packet *r = ring->ring;
	int i;

	for (i = 0; i < ring->size; ++i, ++r) {
		MSD_freeBufDescriptorsFec(r->b, 1);
	}

	vfree(ring->ring);
}

static int
mpc8220_fec_open(struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	struct mpc8220_fec *fec = priv->fec;
	u32 u32_value;
	int irq;
	char *irq_name;
	char *xmit_channel_name, *recv_channel_name;
	struct mpc8220_rcv_ring *ring;
	struct mpc8220_xmt_ring *xmtRing;
	unsigned int negotiated;
	unsigned int duplex;
	int ret;
	int i;

	priv->tx_skb = 0;

	if (dev->base_addr == MPC8220_FEC1) {
		irq = MPC8220_FEC1_IRQ;
		irq_name = "FEC1 Fifo";
	} else {
		irq = MPC8220_FEC2_IRQ;
		irq_name = "FEC2 Fifo";
	}

	if (request_irq(irq, &mpc8220_fec_interrupt,
			SA_INTERRUPT | SA_SAMPLE_RANDOM, irq_name, dev)) {
		goto eagain;
	}
	dev->irq = irq;

	out_be32(&fec->rfifo_status, in_be32(&fec->rfifo_status) & 0x80f00000);
	out_be32(&fec->tfifo_status, in_be32(&fec->tfifo_status) & 0x80f00000);

	out_be32(&fec->ecntrl, MPC8220_FEC_ECNTRL_RESET);
	udelay(10);		/* delay inserted by hardware manufacturer */

	out_be32(&fec->ievent, 0xffffffff);	/* clear intr events */
	out_be32(&fec->imask, 0xffff0000);	/* enable all intr *including* tfint */

	/* MII address of the PHY determination */
	priv->phy_readCB = phy_processCB;
	init_waitqueue_head (&priv->phy_queue);
	for (i=0; i < 33; i++) {
		if (i > 31) {
			printk (KERN_INFO "%s: No PHY detected\n", dev->name);
			return -ERESTARTSYS;
		}
		priv->phy_data.address = i;
		priv->phy_data.data = 0xffff;
		if ((ret = phy_process_wait (dev, PHY_OP_READ, MII_PHYSID1,
			"PHY detection")) < 0) {
				if (ret == -ERESTARTSYS)
					return ret;
				else
					BUG();
		}
		if (priv->phy_data.data != 0xffff)
			break;
	}
	printk(KERN_INFO "%s: PHY detected, address = 0x%04x\n",
		dev->name, priv->phy_data.address);

	if ((ret = phy_process_wait (dev, PHY_OP_READ, MII_ADVERTISE,
		"PHY duplex detection anar")) < 0) {
			if (ret == -ERESTARTSYS)
				return ret;
			else
				BUG();
	}
	negotiated = priv->phy_data.data;
	if ((ret = phy_process_wait (dev, PHY_OP_READ, MII_LPA,
		"PHY duplex detection anlpar")) < 0) {
			if (ret == -ERESTARTSYS)
				return ret;
			else
				BUG();
	}
	negotiated &= priv->phy_data.data;
	duplex = mii_duplex (0, negotiated);
	u32_value = MPC8220_FEC_RECV_BUFFER_SIZE << 16;	/* max frame length */
	if (duplex) {
		printk(KERN_INFO "%s: FULL DUPLEX \n",dev->name);
		u32_value |= 0x00000024;
	} else {
		printk(KERN_INFO "%s: HALF DUPLEX \n",dev->name);
		u32_value |= 0x00000026;
	}
	out_be32(&fec->r_cntrl, u32_value);
	if (duplex)
		out_be32(&fec->x_cntrl, 0x00000006);
	else
		out_be32(&fec->x_cntrl, 0x00000000);
	out_be32(&fec->reset_cntrl, MPC8220_FEC_FIFO_RST_CTL);

	out_be32(&fec->op_pause, 0x00010020);
	/* mask out RXW */
	out_be32(&fec->rfifo_cntrl, 0x0e200000);
	out_be32(&fec->rfifo_alarm, 0x0000030c);
	out_be32(&fec->tfifo_cntrl, 0x0f000000);
	out_be32(&fec->tfifo_alarm, 0x00000180);
	out_be32(&fec->x_wmrk, 0x3);		/* xmit fifo watermark = 256 */
	out_be32(&fec->xmit_fsm, 0x03000000);	/* enable crc generation */
	out_be32(&fec->iaddr1, 0x00000000);	/* No individual filter */
	out_be32(&fec->iaddr2, 0x00000000);	/* No individual filter */

	mpc8220_fec_set_paddr(dev, dev->dev_addr);
	priv->link = 1;

	/* Enable the MIB counters */
	out_be32(&fec->mib_control, 0);

	/* Grab the Dma Channels for the device */
	if (dev->base_addr == MPC8220_FEC1) {
		xmit_channel_name = "FEC1 Xmit DMA";
		recv_channel_name = "FEC1 Recv DMA";
	} else {
		xmit_channel_name = "FEC2 Xmit DMA";
		recv_channel_name = "FEC2 Recv DMA";
	}
	if ((priv->r_tasknum = MSD_reserveChannel(-1, my_sdma_receive_interrupt,
						  recv_channel_name, dev)) < 0)
		goto eagain;

	if ((priv->t_tasknum =
	     MSD_reserveChannel(-1, my_sdma_transmit_interrupt,
				xmit_channel_name, dev)) < 0)
		goto eagain;

	/* Set the initiators needed for DMA */
	if (dev->base_addr == MPC8220_FEC1) {
		priv->xmit_initiator = DMA_REQ_M00_FEC1TX;
		priv->recv_initiator = DMA_REQ_M00_FEC1RX;
	} else {
		priv->xmit_initiator = DMA_REQ_M00_FEC2TX;
		priv->recv_initiator = DMA_REQ_M00_FEC2RX;
	}

	/* Flush edges in case adjacent memory is dirty. */
	/* Invalidate whole buffer so nothing gets flushed from our
	   middle of our data if the CPU needs our tag during the
	   DMA. */

	initRcvRing(dev, rcv_ring_size);

	ring = &priv->rcvRing;
	initXmtRing(dev, xmt_ring_size);
	xmtRing = &priv->xmtRing;

	/* Start Receive DMA */
	startRcvDma(priv, &ring->ring[ring->idx & ring->mask].pair->data);
	/* Start Xmit Dma on the *next* buffer since we preincrement
	   the index */
	startXmtDma(priv, xmtRing->ring[(xmtRing->head) & xmtRing->mask].b);

	netif_start_queue((void *) dev);

	/* Start up the FEC */
	u32_value = in_be32(&fec->ecntrl);
	u32_value |= MPC8220_FEC_ECNTRL_ENABLE;	/* enable FEC-Lite */
	out_be32(&fec->ecntrl, u32_value);

	/* Set up the PHY monitoring timer. */
	priv->phy_timer.expires = jiffies + 2 * HZ;
	priv->phy_timer.data = (unsigned long) dev;
	priv->phy_timer.function = &poll_mii;
	add_timer(&priv->phy_timer);
	clear_bit (0, &priv->phy_lock);
	return 0;

      eagain:
	if (dev->irq)
		free_irq(dev->irq, dev);
	return -EAGAIN;
}

/* This will only be invoked if your driver is _not_ in XOFF state.
 * What this means is that you need not check it, and that this
 * invariant will hold if you make sure that the netif_*_queue()
 * calls are done at the proper times.
 */
static int
mpc8220_fec_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	struct mpc8220_xmt_ring *ring = &priv->xmtRing;
	struct mpc8220_xmt_packet *r;
	unsigned long flags;
	int pad;
	short length;

	if (!priv->link)
		return 1;

	length = skb->len;
	pad = ETH_ZLEN - skb->len;
	if (pad > 0) {
		skb = skb_pad(skb, pad);
		length += pad;
	}

	spin_lock_irqsave(&priv->xmtLock, flags);

	if (ring->head - ring->tail == ring->size) {
		printk(KERN_INFO
		       "%s: failure - there isn't room in the ring,head=%d,tail=%d,size=%d\n",
		       __func__,ring->head,ring->tail,ring->size);
		BUG();
	}
	r = &ring->ring[ring->head++ & ring->mask];

	/* We can't stomp on a packet that's all ready to go out the door */
	if (r->b->statCtrl & MCD_FEC_BUF_READY) {
		printk(KERN_DEBUG
		       "%s: failure to stomp on a packet that's all ready to "
			"go out the door\n",
		       __func__);
		BUG();
	}
	/* The ready bit can not be set unless we loop around the ring */
	if (r->rdy) {
		printk(KERN_DEBUG
		       "%s: failure - the ready bit can not be set unless we "
			"loop around the ring\n",
		       __func__);
		BUG();
	}

	/* If the queue is full, stop the queue. The TFINT interrupt
	   will start it up again */
	if (ring->head - ring->tail >= (ring->size - 2)) {
		priv->tx_full = 1;
		netif_stop_queue(dev);
	}

	/* The coast is clear, let's send it. */
	r->skb = skb;
	r->len = length;

	/* Flush the packet out the door */
	flush_dcache_range((u32) skb->head, (u32) skb->end);

	/* Set the DMA source and length */
	r->b->dataPointer =
	    (u32) pci_map_single(NULL, skb->data, skb->len, PCI_DMA_TODEVICE);
	r->b->length = length;

	/* Mark the packet ready for DMA */
	if (!(ring->head & ring->mask))
		markXmtPacketReady(r->b, 1);
	else
		markXmtPacketReady(r->b, 0);

	/* Mark the rdy bit so the trasmit interrupt can now see that
	   the MCD_FEC_BUF_READY flag in r->b->statCtrl is valid */
	r->rdy = ring->head;

	/* Continue the DMA in case it stopped waiting for a buffer to
	   go ready */
	MCD_continDma(priv->t_tasknum);

	dev->trans_start = jiffies;

	spin_unlock_irqrestore(&priv->xmtLock, flags);
	return 0;
}

static void
my_sdma_transmit_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct net_device *dev = (struct net_device *) dev_id;
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;

	MSD_ack(priv->t_tasknum);
}

static void
my_sdma_receive_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct net_device *dev = (struct net_device *) dev_id;
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	struct mpc8220_rcv_ring *ring = &priv->rcvRing;
	struct sk_buff *skb, *new_skb;

	struct mpc8220_packet_pair *p;
	int numpackets = 0;
	u32 size;
	u32 pbuf;

	/* first ack the interrupt so if the DMA gets to the next buffer we'll notice */
	MSD_ack(priv->r_tasknum);

	/* look through the ring, stopping on the first buffer that is still
	   ready */
	for (p = &ring->ring[ring->idx & ring->mask];
	     !(p->pair->data.statCtrl & MCD_FEC_BUF_READY);
	     p = &ring->ring[++ring->idx & ring->mask], ++numpackets) {
		/* Grab a buffer.  If none available, recycle current and continue */
		if (p->skb_size < MTU_SIZE) {
			printk(KERN_DEBUG
			       "%s: failure - skb_size(%d) < MTU_SIZE(%d)\n",
			       __func__, p->skb_size, MTU_SIZE);
			BUG();
		}

		new_skb = dev_alloc_skb(p->skb_size);
		if (!new_skb) {
			/* Count it as a dropped packet and continue */
			priv->stats.rx_dropped++;

			/* Reset the buffers and continue */
			goto bottom;
		}

		/* Push old buffer up the stack */
		skb = p->skb;
		size = p->pair->data.length;
		if (size >= MTU_SIZE || skb->tail + size > skb->end) {
			printk(KERN_DEBUG
			       "%s: failure - wrong received packet size %d\n",
			       __func__, size);
			BUG();
		}
		skb_put(skb, size);	/* set the size in the buffer */
		skb->dev = dev;
		dev->last_rx = jiffies;
		skb->protocol = eth_type_trans(skb, dev);
		skb->ip_summed = CHECKSUM_HW;
		priv->stats.rx_packets++;
		priv->stats.rx_bytes += size;
		netif_rx(skb);

		/* Put the new buffer into the dma chain */
		p->skb = new_skb;

		p->pair->data.length = p->skb_size;
		p->pair->data.dataPointer =
		    (u32) pci_map_single(NULL, skb_put(new_skb, 0), p->skb_size,
					 PCI_DMA_FROMDEVICE);

	      bottom:
		pbuf = (u32) skb_put(p->skb, 0);
		invalidate_dcache_range(pbuf, pbuf + MTU_SIZE);

		/* mark the buffers as ready */
		if (!((ring->idx + 1) & ring->mask))
			markRcvPairReady(p->pair, 1);
		else
			markRcvPairReady(p->pair, 0);
	}

	/* Continue the DMA since the ring is now got some empty room in it */
	MCD_continDma(priv->r_tasknum);
}

static void
mpc8220_fec_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct net_device *dev = (struct net_device *) dev_id;
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	struct mpc8220_fec *fec = priv->fec;
	struct mpc8220_xmt_ring *xmtRing;
	struct mpc8220_xmt_packet *xpkt;
	int ievent, iievent;
	int npackets = 0;
	int tail_incr;

	iievent = ievent = in_be32(&fec->ievent);

	if (ievent & MPC8220_FEC_IEVENT_HBERR) {
		priv->stats.tx_heartbeat_errors++;
		out_be32(&fec->ievent, MPC8220_FEC_IEVENT_HBERR);
		iievent &= ~MPC8220_FEC_IEVENT_HBERR;
	}
	if (ievent & MPC8220_FEC_IEVENT_BABR) {
		priv->stats.rx_errors++;
		out_be32(&fec->ievent, MPC8220_FEC_IEVENT_BABR);
		iievent &= ~MPC8220_FEC_IEVENT_BABR;
	}
	if (ievent & MPC8220_FEC_IEVENT_BABT) {
		priv->stats.tx_errors++;
		out_be32(&fec->ievent, MPC8220_FEC_IEVENT_BABT);
		iievent &= ~MPC8220_FEC_IEVENT_BABT;
	}
	if (ievent & MPC8220_FEC_IEVENT_GRA) {
		/* Graceful Stop Complete...what do we do here? */
	}
	if (ievent & MPC8220_FEC_IEVENT_TFINT) {
		/* the packet is out the door, increment counters and free the buffer.
		 * This code should be in xmit_interrupt() success clause */
		/* Ack the interrupt so if another one comes along while
		 * we're processing this set of packets, we'll enter the
		 * interrupt again. */
		out_be32(&fec->ievent, MPC8220_FEC_IEVENT_TFINT);
		iievent &= ~MPC8220_FEC_IEVENT_TFINT;

		xmtRing = &priv->xmtRing;

		spin_lock_irq(&priv->xmtLock);

		/* Loop through looking for packets that have been sent.
		 * if tail<head, then there's a packet that's been queued
		 * up.  If its flag is *not* ready then it has been
		 * transmitted */
		for (;;) {
			/* If we caught up to the head, exit */
			if (xmtRing->tail == xmtRing->head)
				break;
				tail_incr = 1;
			xpkt =
			    &xmtRing->
				ring[(xmtRing->tail) & xmtRing->mask];

			/* If the packet hasn't been transmitted, exit */
			if (xpkt->rdy
			    && (xpkt->b->statCtrl & MCD_FEC_BUF_READY))
				break;

			npackets++;
			priv->stats.tx_packets++;
			priv->stats.tx_bytes += xpkt->len;

			if (!xpkt->skb) {
				printk(KERN_DEBUG "%s: failure - zero skb\n",
				       __func__);
				BUG();
			}

			dev_kfree_skb_irq(xpkt->skb);
			xpkt->skb = 0;

			xpkt->rdy = 0;	/* clear the ready bit since we're done
					   with the packet */
			/* increment the tail here since if we were full then
			   we'll wake the queue which can cause the transmitter to be
			   invoked *before* we're ready for it! */
			xmtRing->tail += tail_incr;

		}

		spin_unlock_irq(&priv->xmtLock);

		/* Restart the queue */
		if (npackets && priv->tx_full) {
			netif_wake_queue(dev);
			priv->tx_full = 0;
		}
	}
	if (ievent & MPC8220_FEC_IEVENT_MII) {
		int ret;

		/* MII completed data transfer request? */
		out_be32(&fec->ievent, MPC8220_FEC_IEVENT_MII);
		iievent &= ~MPC8220_FEC_IEVENT_MII;
		if ((ret=phy_action (priv, PHY_OP_CONTINUE,
			priv->phy_data.address)) < 0) {
				printk (KERN_INFO "phy_action ERROR %d\n",ret);
				BUG();
		}
		priv->phy_readCB(dev);
	}
	if (ievent & MPC8220_FEC_IEVENT_LATE_COL) {
		priv->stats.collisions++;
		/* anything else to do here? */
	}
	if (ievent & MPC8220_FEC_IEVENT_COL_RETRY_LIM) {
		priv->stats.tx_errors++;
		/* only get this in half-duplex */
	}
	if (ievent & MPC8220_FEC_IEVENT_XFIFO_UN) {
		/* xmit fifo underrun, try again, keep a count: TODO */
		priv->stats.tx_errors++;
		printk(KERN_INFO "fec xmit underrun error\n");
	}
	if (ievent & MPC8220_FEC_IEVENT_XFIFO_ERROR) {
		int xfifo_status, xfifo_cntrl;

		xfifo_status = in_be32(&fec->tfifo_status);
		xfifo_cntrl = in_be32(&fec->tfifo_cntrl);

		if ((xfifo_status & MPC8220_FEC_TSTATUS_IP)
		    && 0 == (xfifo_cntrl & MPC8220_FEC_TCNTL_IP_MASK)) {
			printk
			    (KERN_INFO
			     "fec xmit error illegal pointer written to fifo\n");
			out_be32(&fec->tfifo_status, MPC8220_FEC_TSTATUS_IP);
		}
		if ((xfifo_status & MPC8220_FEC_TSTATUS_FAE)
		    && 0 == (xfifo_cntrl & MPC8220_FEC_TCNTL_FAE_MASK)) {
			printk
			    (KERN_INFO
			     "fec xmit error, overwrite of data waiting to send\n");
			out_be32(&fec->tfifo_status, MPC8220_FEC_TSTATUS_FAE);
		}
		if ((xfifo_status & MPC8220_FEC_TSTATUS_RXW)
		    && 0 == (xfifo_cntrl & MPC8220_FEC_TCNTL_RXW_MASK)) {
			printk(KERN_INFO "fec xmit receive wait condition\n");
			out_be32(&fec->tfifo_status, MPC8220_FEC_TSTATUS_RXW);
		}
		if ((xfifo_status & MPC8220_FEC_TSTATUS_UF)
		    && 0 == (xfifo_cntrl & MPC8220_FEC_TCNTL_UF_MASK)) {
			printk(KERN_INFO "fec xmit underflow\n");
			out_be32(&fec->tfifo_status, MPC8220_FEC_TSTATUS_UF);
		}
		if ((xfifo_status & MPC8220_FEC_TSTATUS_OF)
		    && 0 == (xfifo_cntrl & MPC8220_FEC_TCNTL_OF_MASK)) {
			printk(KERN_INFO "fec xmit overflow\n");
			out_be32(&fec->tfifo_status, MPC8220_FEC_TSTATUS_OF);
		}
	}
	if (ievent & MPC8220_FEC_IEVENT_RFIFO_ERROR) {
		int rfifo_status, rfifo_cntrl;

		rfifo_status = in_be32(&fec->rfifo_status);
		rfifo_cntrl = in_be32(&fec->rfifo_cntrl);

		printk(KERN_INFO
		       "fec rfifo error, rfifo_status 0x%x rfifo_cntrl 0x%x\n",
		       rfifo_status, rfifo_cntrl);
	}

	/* If the event isn't cleared, state it and *then* clear it. */
	if (iievent) {
		printk(KERN_INFO "FEC %p: ievent %08x iievent %08x\n", fec,
		       ievent, iievent);
		out_be32(&fec->ievent, iievent);	/* reset bits we saw at top of function */
	}
}

static int
mpc8220_fec_close(struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	unsigned long timeout;

	priv->open_time = 0;

	netif_stop_queue(dev);

	/* Wait for Tx queue to drain */
	timeout = 2 * HZ;
	priv->link = 0;

	/* Disable the Rx and Tx queues. */
	MCD_killDma(priv->r_tasknum);
	MCD_killDma(priv->t_tasknum);

	freeRcvRing(&priv->rcvRing);
	freeXmtRing(&priv->xmtRing);

	MSD_releaseChannel(priv->r_tasknum);
	MSD_releaseChannel(priv->t_tasknum);

	free_irq(dev->irq, dev);

	mpc8220_fec_get_stats(dev);

	return 0;
}

/*
 * Get the current statistics.
 * This may be called with the card open or closed.
 */
static struct net_device_stats *
mpc8220_fec_get_stats(struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	struct net_device_stats *stats = &priv->stats;
#ifdef USE_MIB
	stats->rx_packets = in_be32(&fec->rmon_r_packets);
	stats->tx_packets = in_be32(&fec->rmon_t_packets);
	stats->rx_bytes = in_be32(&fec->rmon_r_octets);
	stats->tx_bytes = in_be32(&fec->rmon_t_octets);
	stats->rx_errors = stats->rx_packets - in_be32(&fec->ieee_r_frame_ok);
	stats->tx_errors = stats->tx_packets - in_be32(&fec->ieee_t_frame_ok);
	stats->multicast = in_be32(&fec->rmon_r_mc_pkt);
	stats->collisions = in_be32(&fec->rmon_t_col);

	/* detailed rx_errors: */
	stats->rx_length_errors = in_be32(&fec->rmon_r_undersize)
	    + in_be32(&fec->rmon_r_oversize)
	    + in_be32(&fec->rmon_r_frag)
	    + in_be32(&fec->rmon_r_jab);
	stats->rx_over_errors = in_be32(&fec->r_macerr);
	stats->rx_crc_errors = in_be32(&fec->ieee_r_crc);
	stats->rx_frame_errors = in_be32(&fec->ieee_r_align);
	stats->rx_fifo_errors = in_be32(&fec->rmon_r_drop);
	stats->rx_missed_errors = in_be32(&fec->rmon_r_drop);

	/* detailed tx_errors: */
	stats->tx_aborted_errors = in_be32(&fec->ieee_t_1col)
	    + in_be32(&fec->ieee_t_mcol);
	stats->tx_carrier_errors = in_be32(&fec->ieee_t_cserr);
	stats->tx_fifo_errors = in_be32(&fec->rmon_t_drop);
	stats->tx_heartbeat_errors = in_be32(&fec->ieee_t_sqe);
	stats->tx_window_errors = in_be32(&fec->ieee_t_lcol);
#endif
	return stats;
}

static void
mpc8220_5100_fec_update_stat(struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	struct net_device_stats *stats = &priv->stats;
	struct mpc8220_fec *fec = priv->fec;

	out_be32(&fec->mib_control, MPC8220_FEC_MIB_DISABLE);
	memset_io(&fec->rmon_t_drop, 0,
		  (u32) & fec->reserved10 - (u32) & fec->rmon_t_drop);
	out_be32(&fec->mib_control, 0);
	memset(stats, 0, sizeof *stats);
	mpc8220_fec_get_stats(dev);
}

/*
 * Set or clear the multicast filter for this adaptor.
 * num_addrs == -1	Promiscuous mode, receive all packets
 * num_addrs == 0	Normal mode, clear multicast list
 * num_addrs > 0	Multicast mode, receive normal and MC packets,
 *			and do best-effort filtering.
 */
static void
mpc8220_fec_set_multicast_list(struct net_device *dev)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *) dev->priv;
	struct mpc8220_fec *fec = priv->fec;
	u32 u32_value;

	if (dev->flags & IFF_PROMISC) {
		u32_value = in_be32(&fec->r_cntrl);
		u32_value |= MPC8220_FEC_RCNTL_PROM;
		out_be32(&fec->r_cntrl, u32_value);
	} else if (dev->flags & IFF_ALLMULTI) {
		u32_value = in_be32(&fec->r_cntrl);
		u32_value &= ~MPC8220_FEC_RCNTL_PROM;
		out_be32(&fec->r_cntrl, u32_value);
		out_be32(&fec->gaddr1, 0xffffffff);
		out_be32(&fec->gaddr2, 0xffffffff);
	} else {
		u32 crc;
		int i;
		struct dev_mc_list *dmi;
		u32 gaddr1 = 0x00000000;
		u32 gaddr2 = 0x00000000;

		dmi = dev->mc_list;
		for (i = 0; i < dev->mc_count; i++) {
			crc = ether_crc_le(6, dmi->dmi_addr) >> 26;
			if (crc >= 32)
				gaddr1 |= 1 << (crc - 32);
			else
				gaddr2 |= 1 << crc;
			dmi = dmi->next;
		}
		out_be32(&fec->gaddr1, gaddr1);
		out_be32(&fec->gaddr2, gaddr2);
	}
}

static int
mpc8220_ethtool_get_settings (struct net_device *dev, struct ethtool_cmd* ecmd)
{
	int ret;
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *)dev->priv;
	struct mpc8220_fec *fec = priv->fec;
	u16 mii_cmd;
	u16 mii_status;
	u16 mii_advControl;
	u32 x_reg;
	u32 r_reg;

	memset (ecmd, 0, sizeof(struct ethtool_cmd));
	if ((ret = phy_process_wait (dev, PHY_OP_READ, MII_BMCR,
			"ETHTOOL get settings MII_BMCR read")) < 0) {
		if (ret == -ERESTARTSYS)
			return ret;
		else
			BUG();
	}
	mii_cmd = priv->phy_data.data;
	if ((ret = phy_process_wait (dev, PHY_OP_READ, MII_BMSR,
			"ETHTOOL get settings MII_BMSR read")) < 0) {
		if (ret == -ERESTARTSYS)
			return ret;
		else
			BUG();
	}
	mii_status = priv->phy_data.data;
	if ((ret = phy_process_wait (dev, PHY_OP_READ, MII_ADVERTISE,
			"ETHTOOL get settings MII_ADVERTISE read")) < 0) {
		if (ret == -ERESTARTSYS)
			return ret;
		else
			BUG();
	}
	mii_advControl = priv->phy_data.data;

	r_reg = in_be32(&fec->r_cntrl);
	x_reg = in_be32(&fec->x_cntrl);
	if (((r_reg & 0x00000002) == 0x00000000) &&
		((x_reg & 0x00000006) == 0x00000006))
			priv->fec_duplex = 1;
	else if (((r_reg & 0x00000002) == 0x00000002) &&
		((x_reg & 0x00000006) == 0x00000000))
			priv->fec_duplex = 0;
	else {
		printk (KERN_INFO "%s: duplex ERROR: x_reg=%08x,r_reg=%08x\n",
			__func__, x_reg, r_reg);
		BUG ();
	}
	if (priv->fec_duplex)
		ecmd->duplex = DUPLEX_FULL;
	else
		ecmd->duplex = DUPLEX_HALF;
	if (mii_status & BMSR_100FULL)
		ecmd->supported |= SUPPORTED_100baseT_Full;
	if (mii_status & BMSR_100HALF)
		ecmd->supported |= SUPPORTED_100baseT_Half;
	if (mii_status & BMSR_10FULL)
		ecmd->supported |= SUPPORTED_10baseT_Full;
	if (mii_status & BMSR_10HALF)
		ecmd->supported |= SUPPORTED_10baseT_Half;
	ecmd->supported |= SUPPORTED_MII;
	if (mii_status & BMSR_ANEGCAPABLE)
		ecmd->supported |= SUPPORTED_Autoneg;
	if (mii_status & BMSR_ANEGCOMPLETE) {
		ecmd->autoneg = AUTONEG_ENABLE;
		ecmd->advertising |= ADVERTISED_Autoneg;
		if ((mii_advControl & ADVERTISE_100FULL) || (mii_advControl & ADVERTISE_100HALF))
			ecmd->speed = SPEED_100;
		else
			ecmd->speed = SPEED_10;
	} else {
		ecmd->autoneg = AUTONEG_DISABLE;
		if (mii_cmd & BMCR_SPEED100)
			ecmd->speed = SPEED_100;
		else
			ecmd->speed = SPEED_10;
	}
	if (mii_advControl & ADVERTISE_10FULL)
		ecmd->advertising |= ADVERTISED_10baseT_Full;
	if (mii_advControl & ADVERTISE_10HALF)
		ecmd->advertising |= ADVERTISED_10baseT_Half;
	if (mii_advControl & ADVERTISE_100FULL)
		ecmd->advertising |= ADVERTISED_100baseT_Full;
	if (mii_advControl & ADVERTISE_100HALF)
		ecmd->advertising |= ADVERTISED_100baseT_Half;
	ecmd->advertising |= ADVERTISED_MII;
	ecmd->port = PORT_MII;
	ecmd->phy_address = priv->phy_data.address;
	ecmd->transceiver = XCVR_INTERNAL;
	ecmd->maxtxpkt = 1;
	ecmd->maxrxpkt = 1;
	return 0;
}

static int
mpc8220_ethtool_ioctl(struct net_device *dev, struct ifreq *rq)
{
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *)dev->priv;
	struct ethtool_cmd ecmd;
	struct ethtool_drvinfo edrv;
	struct ethtool_ringparam erp;
	int ret = -EOPNOTSUPP;
	u16 mii_reg_sset;
	u16 mii_reg_autoneg;

	if (copy_from_user(&ecmd, rq->ifr_data, sizeof ecmd.cmd))
		return -EFAULT;

	switch (ecmd.cmd) {

	case ETHTOOL_GSET:
		ret = mpc8220_ethtool_get_settings(dev, &ecmd);
		if (ret >= 0) {
			if (copy_to_user(rq->ifr_data, &ecmd, sizeof (ecmd)))
				ret = -EFAULT;
		}
		break;
		/* Get driver info */
	case ETHTOOL_SSET:
		if (copy_from_user(&ecmd, rq->ifr_data, sizeof (struct ethtool_cmd)))
			return -EFAULT;
		mii_reg_sset = 0;
		if (ecmd.speed == SPEED_100)
			mii_reg_sset |= BMCR_SPEED100;
		if (ecmd.duplex == DUPLEX_FULL)
			mii_reg_sset |= BMCR_FULLDPLX;
		if (ecmd.autoneg == AUTONEG_ENABLE) {
			mii_reg_sset |= (BMCR_ANENABLE | BMCR_ANRESTART);
			priv->phy_data.data = mii_reg_sset;
			if ((ret = phy_process_wait (dev, PHY_OP_WRITE, MII_BMCR,
					"ETHTOOL bmcr write")) < 0) {
				if (ret == -ERESTARTSYS)
					break;
				else
					BUG();
			}
			if ((ret = phy_process_wait (dev, PHY_OP_READ, MII_ADVERTISE,
					"ETHTOOL mii_advertise read")) < 0) {
				if (ret == -ERESTARTSYS)
					break;
				else
					BUG();
			}
			mii_reg_sset = priv->phy_data.data;
			if (ecmd.speed == SPEED_100) {
				if (ecmd.duplex == DUPLEX_FULL) {
					mii_reg_sset |= (ADVERTISE_10FULL | ADVERTISE_100FULL |
						ADVERTISE_10HALF | ADVERTISE_100HALF);
				} else {
					mii_reg_sset |= (ADVERTISE_10HALF | ADVERTISE_100HALF);
					mii_reg_sset &= ~(ADVERTISE_10FULL | ADVERTISE_100FULL);
				}
			} else {
				if (ecmd.duplex == DUPLEX_FULL) {
					mii_reg_sset |= (ADVERTISE_10FULL | ADVERTISE_10HALF);
					mii_reg_sset &= ~(ADVERTISE_100FULL| ADVERTISE_100HALF);
				} else {
					mii_reg_sset |= (ADVERTISE_10HALF);
					mii_reg_sset &= ~(ADVERTISE_100FULL| ADVERTISE_100HALF | ADVERTISE_10FULL);
				}
			}
			priv->phy_data.data = mii_reg_sset;
			if ((ret = phy_process_wait (dev, PHY_OP_WRITE, MII_ADVERTISE,
					"ETHTOOL mii_advertise write")) < 0) {
				if (ret == -ERESTARTSYS)
					break;
				else
					BUG();
			}
		} else {
			mii_reg_sset &= ~(BMCR_ANENABLE | BMCR_ANRESTART);
			if (ecmd.duplex == DUPLEX_FULL) {
				mii_reg_sset |= BMCR_FULLDPLX;
			} else {
				mii_reg_sset &= ~BMCR_FULLDPLX;
			}
			if (ecmd.speed == SPEED_100) {
				mii_reg_sset |= BMCR_SPEED100;
			} else {
				mii_reg_sset &= ~BMCR_SPEED100;
			}
			priv->phy_data.data = mii_reg_sset;
			if ((ret = phy_process_wait (dev, PHY_OP_WRITE, MII_BMCR,
					"ETHTOOL mii_bmcr write")) < 0) {
				if (ret == -ERESTARTSYS)
					break;
				else
					BUG();
			}
		}
		ret = 0;
		break;
	case ETHTOOL_GDRVINFO:
		memset (&edrv, 0, sizeof(struct ethtool_drvinfo));
		strcpy (edrv.driver, drv_name);
		strcpy (edrv.version, version);
		if (copy_to_user(rq->ifr_data, &edrv, sizeof (struct ethtool_drvinfo)))
			return -EFAULT;
		ret = 0;
		break;
	case ETHTOOL_GRINGPARAM:
		memset (&erp, 0, sizeof(struct ethtool_ringparam));
		erp.rx_max_pending = priv->rcvRing.size;
		erp.tx_max_pending = priv->xmtRing.size;
		erp.rx_pending = priv->rcvRing.size;
		erp.tx_pending = priv->rcvRing.size;
		erp.cmd = edrv.cmd;
		if (copy_to_user(rq->ifr_data, &erp, sizeof (struct ethtool_ringparam)))
			ret = -EFAULT;
		ret = 0;
		break;
	case ETHTOOL_NWAY_RST:
		mii_reg_autoneg = 0;
		mii_reg_autoneg |= (BMCR_ANENABLE | BMCR_ANRESTART);
		priv->phy_data.data = mii_reg_autoneg;
		if ((ret = phy_process_wait (dev, PHY_OP_WRITE, MII_BMCR,
				"ETHTOOL nway MII_BMCR write")) < 0) {
			if (ret == -ERESTARTSYS)
				break;
			else
				BUG();
		}
		ret = 0;
		break;
	default:
		break;
	}

	return ret;
}

static int
mpc8220_fec_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	int retval;
	struct mpc8220_fec_priv *priv = (struct mpc8220_fec_priv *)dev->priv;
	struct mii_ioctl_data *data = (struct mii_ioctl_data *) &rq->ifr_data;

	switch (cmd) {
	case SIOCETHTOOL:
		wait_event_interruptible(priv->phy_queue,
			       (test_and_set_bit(0, &(priv->phy_lock)) == 0));
		priv->phy_readCB = phy_processCB;
		del_timer_sync(&priv->phy_timer);
		retval = mpc8220_ethtool_ioctl(dev, rq);
		clear_bit (0, &priv->phy_lock);
		priv->phy_readCB = cont_poll_mii;
		priv->phy_timer.expires = jiffies + 2 * HZ;
		add_timer(&priv->phy_timer);
		break;
	case SIOCGMIIPHY:	/* Get address of MII PHY in use. */
	case SIOCDEVPRIVATE:	/* for binary compat, remove in 2.5 */
		data->phy_id = priv->phy_data.address;
		/* Fall Through */

	case SIOCGMIIREG:	/* Read MII PHY register. */
	case SIOCDEVPRIVATE + 1:	/* for binary compat, remove in 2.5 */
		if (data->phy_id > 31 || data->reg_num > 31)
			return -ENXIO;

		wait_event_interruptible(priv->phy_queue,
			       (test_and_set_bit(0, &(priv->phy_lock)) == 0));
		priv->phy_readCB = phy_processCB;
		del_timer_sync(&priv->phy_timer);
		if ((retval = phy_process_wait (dev, PHY_OP_READ,data->reg_num,
				"PHY ioctl read register")) < 0) {
			if (retval == -ERESTARTSYS)
				return retval;
			else
				BUG();
		}
		data->val_out = priv->phy_data.data;
		clear_bit (0, &priv->phy_lock);
		priv->phy_readCB = cont_poll_mii;
		priv->phy_timer.expires = jiffies + 2 * HZ;
		add_timer(&priv->phy_timer);
		break;

	case SIOCSMIIREG:	/* Write MII PHY register. */
	case SIOCDEVPRIVATE + 2:	/* for binary compat, remove in 2.5 */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;

		if (data->phy_id > 31 || data->reg_num > 31)
			return -ENXIO;

		wait_event_interruptible(priv->phy_queue,
			       ((test_and_set_bit(0, &priv->phy_lock)) == 0));
		priv->phy_readCB = phy_processCB;
		del_timer_sync(&priv->phy_timer);
		priv->phy_data.data = data->val_in;
		if ((retval = phy_process_wait (dev, PHY_OP_WRITE,data->reg_num,
				"PHY ioctl read register")) < 0) {
			if (retval == -ERESTARTSYS)
				return retval;
			else
				BUG();
		}
		clear_bit (0, &priv->phy_lock);
		priv->phy_readCB = cont_poll_mii;
		priv->phy_timer.expires = jiffies + 2 * HZ;
		add_timer(&priv->phy_timer);
		break;

	default:
		retval = -EOPNOTSUPP;
		break;
	}
	return retval;
}

static void __init
mpc8220_fec_str2mac(char *str, unsigned char *mac)
{
	int i;
	u64 val64;

	val64 = simple_strtoull(str, NULL, 16);

	for (i = 0; i < 6; i++)
		mac[5 - i] = val64 >> (i * 8);
}

static int __init
mpc8220_fec_mac_setup(char *mac_address)
{
	mpc8220_fec_str2mac(mac_address, mpc8220_fec1_mac_addr);
	return 0;
}

static int
mpc8220_init_fec_dev(struct net_device *dev, int which)
{
	struct mpc8220_fec *fec;
	struct mpc8220_fec_priv *priv;
	u8 null_mac[6] = { 0, 0, 0, 0, 0, 0 };

	priv = (struct mpc8220_fec_priv *) dev->priv;

	priv->fec = fec =
	    (struct mpc8220_fec *) (which ? MPC8220_FEC2 : MPC8220_FEC1);

	dev->base_addr = (unsigned long) fec;

	priv->gpio = (struct mpc8220_gpio *) MPC8220_GPIO;

	spin_lock_init(&priv->lock);
	spin_lock_init(&priv->xmtLock);
	spin_lock_init(&priv->mii_irq_lock);

	dev->open = mpc8220_fec_open;
	dev->stop = mpc8220_fec_close;
	dev->hard_start_xmit = mpc8220_fec_hard_start_xmit;
	dev->do_ioctl = mpc8220_fec_ioctl;
	dev->get_stats = mpc8220_fec_get_stats;
	dev->set_mac_address = mpc8220_fec_set_mac_address;
	dev->set_multicast_list = mpc8220_fec_set_multicast_list;

	dev->tx_timeout = mpc8220_fec_tx_timeout;
	dev->watchdog_timeo = MPC8220_FEC_WATCHDOG_TIMEOUT;
	dev->flags &= ~IFF_RUNNING;

	if (memcmp(mpc8220_fec1_mac_addr, null_mac, 6) != 0) {
		memcpy(dev->dev_addr, mpc8220_fec1_mac_addr, 6);
	} else {
		*(u32 *) & dev->dev_addr[0] = in_be32(&fec->paddr1);
		*(u16 *) & dev->dev_addr[4] = in_be16((u16 *) & fec->paddr2);
	}

	/*
	 * Read MIB counters in order to reset them,
	 * then zero all the stats fields in memory
	 */
	mpc8220_5100_fec_update_stat(dev);
	return 0;
}

__setup("mpc8220_mac=", mpc8220_fec_mac_setup);

static int __init
mpc8220_fec_init(void)
{
	struct net_device *dev;
	struct mpc8220_fec_priv *priv;
	int ret;

	printk(KERN_INFO "MPC8220 FEC initialization\n");
	dev = init_etherdev(NULL, sizeof *priv);
	if (!dev)
		return -EIO;
	mpc8220_fec_dev[mpc8220_fec_dev_cnt++] = dev;
	ret = mpc8220_init_fec_dev(dev, 0);
	if (ret < 0)
		return ret;

	dev = init_etherdev(NULL, sizeof *priv);
	if (!dev)
		return -EIO;
	mpc8220_fec_dev[mpc8220_fec_dev_cnt++] = dev;
	ret = mpc8220_init_fec_dev(dev, 1);
	if (ret < 0)
		return ret;

	return 0;
}

#define NELEM(x) (sizeof(x)/sizeof((x)[0]))

static void __exit
mpc8220_fec_uninit(void)
{
	int i;
	struct net_device *dev;

	printk(KERN_INFO "MPC8220 FEC exit\n");
	for (i = 0; i < NELEM(mpc8220_fec_dev); ++i) {
		if ((dev = mpc8220_fec_dev[i]) != 0) {
			unregister_netdev(dev);
			free_irq(dev->irq, dev);
			kfree(dev->priv);
			dev->priv = NULL;
		}
	}
}

void __exit
mpc8220_fec_module_exit(void)
{
	printk(KERN_INFO "mpc8220_fec_module_exit\n");
	mpc8220_fec_uninit();
}

module_init(mpc8220_fec_init);
module_exit(mpc8220_fec_uninit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Ethernet driver for Freescale MPC8220 FEC");
