/*
 * hdlc_enet.c
 * Ethernet using HDLC on SCCx (proprietary hardware).
 * Copyright (c) 2002, 2004 Gary Jennejohn (gj@denx.de)
 * Modified to support the header used by IDS.
 *
 * Based on the original drivers:
 * UART driver for MPC860 CPM SCC or SMC
 * Copyright (c) 1997 Dan Malek (dmalek@jlc.net)
 *
 * Ethernet driver for Motorola MPC8xx.
 * Copyright (c) 1997 Dan Malek (dmalek@jlc.net)
 *
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/if_arp.h>

#include <asm/8xx_immap.h>
#include <asm/pgtable.h>
#include <asm/mpc8xx.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>

#include <asm/commproc.h>
/* my hdlc.h with board specific defines */
#include "hdlc.h"

/* stuff for the module */
MODULE_AUTHOR("Gary Jennejohn");
MODULE_DESCRIPTION("HDLC Ethernet driver");
MODULE_SUPPORTED_DEVICE("hdlc_enet");
/* there must be a better way to do this */
MODULE_PARM(eaddl, "i");
MODULE_PARM_DESC(eaddl, "The low byte of the ethernet address");
MODULE_PARM(devel_debug, "i");
MODULE_PARM_DESC(devel_debug, "Development DEBUG flag");
MODULE_LICENSE("GPL");

static int eaddl = 0;
static int devel_debug = 0;

/* Set of debugging defines */
#define DEBUG_INTR 1
#define DEBUG_INIT 2
#define DEBUG_READ 4
#define DEBUG_WRIT 8
#define DEBUG_IOCT 16
#define DEBUG_RX 32
#define DEBUG_TX   64

/*
 *				Theory of Operation
 *
 * The MPC8xx CPM performs the HDLC processing on SCCx.	 It can use
 * an aribtrary number of buffers on byte boundaries, but must have at
 * least two receive buffers to prevent constant overrun conditions.
 *
 * The buffer descriptors are allocated from the CPM dual port memory
 * with the data buffers allocated from host memory, just like all other
 * serial communication protocols.  The host memory buffers are allocated
 * from the free page pool, and then divided into smaller receive and
 * transmit buffers.  The size of the buffers should be a power of two,
 * since that nicely divides the page.	This creates a ring buffer
 * structure similar to the LANCE and other controllers.
 *
 * Like the LANCE driver:
 * The driver runs as two independent, single-threaded flows of control.  One
 * is the send-packet routine, which enforces single-threaded use by the
 * cep->tx_busy flag.  The other thread is the interrupt handler, which is
 * single threaded by the hardware and other software.
 *
 * The send packet thread has partial control over the Tx ring and the
 * 'cep->tx_busy' flag.	 It sets the tx_busy flag whenever it's queuing a Tx
 * packet. If the next queue slot is empty, it clears the tx_busy flag when
 * finished otherwise it sets the 'lp->tx_full' flag.
 *
 */

/* The transmitter timeout
 */
#define TX_TIMEOUT	(2*HZ)

/* The number of Tx and Rx buffers.  These are allocated from the page
 * pool.  The code may assume these are power of two, so it is best
 * to keep them that size.
 * We don't need to allocate pages for the transmitter.	 We just use
 * the skbuffer directly.
 */
#ifdef CONFIG_ENET_BIG_BUFFERS
#define CPM_ENET_RX_PAGES	32
#define CPM_ENET_RX_FRSIZE	2048
#define CPM_ENET_RX_FRPPG	(PAGE_SIZE / CPM_ENET_RX_FRSIZE)
#define RX_RING_SIZE		(CPM_ENET_RX_FRPPG * CPM_ENET_RX_PAGES)
#define TX_RING_SIZE		64	/* Must be power of two */
#define TX_RING_MOD_MASK	63	/*   for this to work */
#else
#define CPM_ENET_RX_PAGES	4
#define CPM_ENET_RX_FRSIZE	2048
#define CPM_ENET_RX_FRPPG	(PAGE_SIZE / CPM_ENET_RX_FRSIZE)
#define RX_RING_SIZE		(CPM_ENET_RX_FRPPG * CPM_ENET_RX_PAGES)
#define TX_RING_SIZE		8	/* Must be power of two */
#define TX_RING_MOD_MASK	7	/*   for this to work */
#endif

/* The CPM stores dest/src/type, data, and checksum for receive packets.
 */
#define PKT_MAXBUF_SIZE		1518
#define PKT_MINBUF_SIZE		64
#define PKT_MAXBLR_SIZE		1520

#ifdef MODULE
/* save the address of each page allocated in init_module so */
/* we can free it in cleanup_module */
static unsigned char *rx_pages[CPM_ENET_RX_PAGES];
#endif

/* The CPM buffer descriptors track the ring buffers.  The rx_bd_base and
 * tx_bd_base always point to the base of the buffer descriptors.  The
 * cur_rx and cur_tx point to the currently available buffer.
 * The dirty_tx tracks the current buffer that is being sent by the
 * controller.	The cur_tx and dirty_tx are equal under both completely
 * empty and completely full conditions.  The empty/ready indicator in
 * the buffer descriptor determines the actual condition.
 */
struct hdlc_enet_private {
	/* addresses in DPRAM so we can free them */
	uint rxdp_addr, txdp_addr;
	/* The saved address of a sent-in-place packet/buffer, for skfree(). */
	struct	sk_buff* tx_skbuff[TX_RING_SIZE];
	ushort	skb_cur;
	ushort	skb_dirty;

	/* CPM dual port RAM relative addresses.
	*/
	cbd_t	*rx_bd_base;		/* Address of Rx and Tx buffers. */
	cbd_t	*tx_bd_base;
	cbd_t	*cur_rx, *cur_tx;	/* The next free ring entry */
	cbd_t	*dirty_tx;		/* The ring entries to be free()ed. */
	scc_t	*sccp;
	/*
	 * Virtual addresses of the RX buffers because we can't do a __va()
	 * on them anymore.
	 */
	unsigned char *rx_vaddr[RX_RING_SIZE];
	struct	net_device_stats stats;
	uint	tx_full;
	spinlock_t lock;
	/* keep track of my HADDR */
	unsigned char myhaddr;
};

static int hdlc_enet_open(struct net_device *dev);
static int hdlc_enet_start_xmit(struct sk_buff *skb, struct net_device *dev);
static int hdlc_enet_rx(struct net_device *dev);
static void hdlc_enet_interrupt(void *dev_id, struct pt_regs *regs);
static int hdlc_enet_close(struct net_device *dev);
static struct net_device_stats *hdlc_enet_get_stats(struct net_device *dev);
static int hdlc_enet_set_mac_address(struct net_device *dev, void *addr);
#ifndef MODULE
int hdlc_enet_init(void);
#else
static int hdlc_enet_init(struct net_device *dev);
int hdlc_enet_probe(struct net_device *dev);
#endif

#ifdef CONFIG_TQM860L
/* TQM860L uses SCC2 */
#define CPM_CR_ENET CPM_CR_CH_SCC2
#define PROFF_ENET  PROFF_SCC2
#define SCC_ENET    1	    /* Index, not number! */
#define CPMVEC_ENET CPMVEC_SCC2
#else
/* IDS uses SCC4 */
#define CPM_CR_ENET CPM_CR_CH_SCC4
#define PROFF_ENET  PROFF_SCC4
#define SCC_ENET    3	    /* Index, not number! */
#define CPMVEC_ENET CPMVEC_SCC4
#endif

/* flags in the buffer descriptor not defined anywhere else */
#define BD_SC_CT 0x01
#define BD_SC_DE 0x80
#define BD_SC_FIRST 0x400
#define BD_SC_STATS (BD_SC_BR | BD_SC_FR | BD_SC_PR | BD_SC_NAK | BD_SC_CD | BD_SC_OV | BD_SC_DE)

#ifdef MODULE
int __init
hdlc_enet_probe( struct net_device *dev)
{
	SET_MODULE_OWNER(dev);

	if (hdlc_enet_init(dev) != 0)
		return -ENODEV;
	return 0;
}
#endif

static int
hdlc_enet_open(struct net_device *dev)
{

	/* I should reset the ring buffers here, but I don't yet know
	 * a simple way to do that.
	 */

	netif_start_queue(dev);
	return 0;					/* Always succeed */
}

static int
hdlc_enet_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct hdlc_enet_private *cep = (struct hdlc_enet_private *)dev->priv;
	volatile cbd_t	*bdp;
	struct ids_frame_hdr *ih;
	unsigned char thaddr;
	unsigned short proto;
	struct ethhdr *eh;
	struct ids_arp_hdr *iah;
	struct eth_arp_hdr *eah;

	/* check for an idle bus */
	if((cep->sccp->scc_sccs & 3) != 3)
		return 1;

	if (devel_debug & DEBUG_TX)
		printk("%s: \n", __FUNCTION__);

	if (devel_debug & DEBUG_TX) {
		int i, len;
		len = skb->len > 32 ? 32 : skb->len;
		printk("Raw data IN length %d dump length %d\n", skb->len, len);
		for (i = 0; i < len; i++)
			printk("%#x ", skb->data[i]);
		printk("\n");
	}
	eh = (struct ethhdr *)skb->data;
	/* see if the protocol is supported by IDS */
	proto = eh->h_proto;
	if (proto != ETH_P_IP && proto != ETH_P_ARP) {
		if (devel_debug & DEBUG_TX)
			printk("unsupported protocol type %#x!\n", proto);
		return 0; /* XXX */
	}
	/* Fill in a Tx ring entry */
	bdp = cep->cur_tx;

#ifndef final_version
	if (bdp->cbd_sc & BD_SC_READY) {
		/* Ooops.  All transmit buffers are full.  Bail out.
		 * This should not happen, since cep->tx_busy should be set.
		 */
		if (devel_debug & DEBUG_TX)
			printk("%s: tx queue full!\n", dev->name);
		return 1;
	}
#endif

	/* Clear all of the status flags.
	 */
	bdp->cbd_sc &= ~(BD_SC_CT|BD_SC_UN);

	/* IDS uses 0x00 for broadcast */
	if (eh->h_dest[(ETH_ALEN - 1)] == 0xff)
		thaddr = 0x00;
	else
		thaddr = eh->h_dest[(ETH_ALEN - 1)];

	/*
	 * put IDS' header over the ethernet frame header.
	 * this is done such that ih ends where eh ends.
	 */
	ih = (struct ids_frame_hdr *)&skb->data[IDSHDR_OFF];
	SETTARGADDR(ih, thaddr);
	SETSRCADDR(ih, cep->myhaddr);
	SETEND(ih);
	/* need to figure out whether this is IP or ARP */
	if (proto == ETH_P_IP) {
		if (devel_debug & DEBUG_TX)
			printk("IP\n");
		ih->proto_1 = IDS_IP_1;
		ih->proto_2 = IDS_IP_2;
		/* reuse proto for the length */
		proto = skb->len - sizeof(*eh) + sizeof(*ih);
	} else { /* must be ARP */
		if (devel_debug & DEBUG_TX)
			printk("ARP\n");
		ih->proto_1 = IDS_ARP_1;
		ih->proto_2 = IDS_ARP_2;
		/* convert arphdr to ids_arp_hdr */
		iah = (struct ids_arp_hdr *)((uint)ih + sizeof(*ih));
		eah = (struct eth_arp_hdr *)(skb->data + sizeof(*eh));
		iah->oah.ar_hrd = IDS_ARPTYPE;
		iah->oah.ar_pro = IDS_ARPPROTO;
		iah->oah.ar_hln = IDS_ARPHWALEN;
		iah->oah.ar_pln = eah->oah.ar_pln;
		iah->oah.ar_op = eah->oah.ar_op;
#ifdef IDSARP_USES_ETHADDR /* not normally defined */
		iah->ar_sha = eah->ar_sha[(ETH_ALEN - 1)];
		iah->ar_tha = eah->ar_tha[(ETH_ALEN - 1)];
#else
		iah->ar_sha = ih->source;
		iah->ar_tha = ih->target;
#endif
		memcpy(iah->ar_sip, eah->ar_sip, 4);
		memcpy(iah->ar_tip, eah->ar_tip, 4);
		/* reuse proto for the length */
		proto = sizeof(*ih) + sizeof(*iah);
	}


	/*
	 * the original driver handled padding short (< 60 bytes)
	 * packets here. It's easier to do it in the receive routine.
	 */

	/* Set buffer length and buffer pointer.
	*/
	bdp->cbd_datlen = proto;
	if (devel_debug & DEBUG_TX) {
		int i, len;
		len = proto > 32 ? 32 : proto;
		printk("Raw data OUT from cbd_bufaddr length %d dump length %d\n", proto, len);
		bdp->cbd_bufaddr = (uint)ih;
		printk("virt addr cbd_bufaddr %#x\n", bdp->cbd_bufaddr);
		for (i = 0; i < len; i++)
			printk("%#x ", ((unsigned char *)bdp->cbd_bufaddr)[i]);
		printk("\n");
	}
	bdp->cbd_bufaddr = __pa(ih);
	if (devel_debug & DEBUG_TX)
		printk("phys addr cbd_bufaddr %#x\n", bdp->cbd_bufaddr);

	/* Save skb pointer.
	*/
	cep->tx_skbuff[cep->skb_cur] = skb;

	cep->stats.tx_bytes += proto;
	cep->skb_cur = (cep->skb_cur+1) & TX_RING_MOD_MASK;

	/* Push the data cache so the CPM does not get stale memory
	 * data.
	 */
	flush_dcache_range((unsigned long)(skb->data),
					(unsigned long)(skb->data + skb->len));

	spin_lock_irq(&cep->lock);

	/* Send it on its way.	Tell CPM its ready, interrupt when done,
	 * its the last BD of the frame, and to put the CRC on the end.
	 */
	bdp->cbd_sc |= (BD_SC_READY | BD_SC_INTRPT | BD_SC_LAST | BD_SC_TC);

	dev->trans_start = jiffies;

	/* If this was the last BD in the ring, start at the beginning again.
	*/
	if (bdp->cbd_sc & BD_SC_WRAP)
		bdp = cep->tx_bd_base;
	else
		bdp++;

	if (bdp->cbd_sc & BD_SC_READY) {
		netif_stop_queue(dev);
		cep->tx_full = 1;
	}

	cep->cur_tx = (cbd_t *)bdp;

	spin_unlock_irq(&cep->lock);

	return 0;
}

static void
hdlc_enet_timeout(struct net_device *dev)
{
	struct hdlc_enet_private *cep = (struct hdlc_enet_private *)dev->priv;

	printk("%s: transmit timed out.\n", dev->name);
	cep->stats.tx_errors++;
#ifndef final_version
	{
		int	i;
		cbd_t	*bdp;
		printk(" Ring data dump: cur_tx %p%s cur_rx %p.\n",
		       cep->cur_tx, cep->tx_full ? " (full)" : "",
		       cep->cur_rx);
		bdp = cep->tx_bd_base;
		for (i = 0 ; i < TX_RING_SIZE; i++, bdp++)
			printk("%04x %04x %08x\n",
			       bdp->cbd_sc,
			       bdp->cbd_datlen,
			       bdp->cbd_bufaddr);
		bdp = cep->rx_bd_base;
		for (i = 0 ; i < RX_RING_SIZE; i++, bdp++)
			printk("%04x %04x %08x\n",
			       bdp->cbd_sc,
			       bdp->cbd_datlen,
			       bdp->cbd_bufaddr);
	}
#endif
	if (!cep->tx_full)
		netif_wake_queue(dev);
}

/* The interrupt handler.
 * This is called from the CPM handler, not the MPC core interrupt.
 */
#define SCCM_RF 0x08
static void hdlc_enet_interrupt (void *dev_id, struct pt_regs *regs)
{
	struct net_device *dev = dev_id;
	volatile struct hdlc_enet_private *cep;
	volatile cbd_t *bdp;
	ushort int_events;
	int must_restart;

	if (devel_debug & DEBUG_INTR)
		printk ("%s\n", __FUNCTION__);
	cep = (struct hdlc_enet_private *) dev->priv;

	/* Get the interrupt events that caused us to be here.
	 */
	int_events = cep->sccp->scc_scce;
	if (devel_debug & DEBUG_INTR)
		printk ("scce %#x\n", int_events);
	cep->sccp->scc_scce = int_events;
	must_restart = 0;

	/* Handle receive event in its own function.
	 */
	if (int_events & (SCCM_RF | SCCM_RX))
		hdlc_enet_rx (dev_id);

	/* Transmit OK, or non-fatal error.  Update the buffer descriptors.
	 */
	if (int_events & SCCM_TX) {
		spin_lock (&cep->lock);
		bdp = cep->dirty_tx;
		if (devel_debug & DEBUG_INTR)
			printk ("TX:cbd_sc %#x\n", bdp->cbd_sc);
		while ((bdp->cbd_sc & BD_SC_READY) == 0) {
			if ((bdp == cep->cur_tx) && (cep->tx_full == 0))
				break;

			if ((bdp->cbd_sc & (BD_SC_CT | BD_SC_UN)) != 0) {
				/* CTS lost - map to carrier lost */
				if ((bdp->cbd_sc & BD_SC_CT) != 0) {
					if (devel_debug & DEBUG_INTR)
						printk("TX CTS lost - must restart\n");
					cep->stats.tx_carrier_errors++;
				}
				/* underrun */
				if ((bdp->cbd_sc & BD_SC_UN) != 0) {
					if (devel_debug & DEBUG_INTR)
						printk("TX UNDERRUN - must restart\n");
					cep->stats.tx_fifo_errors++;
				}
				/* both errors cause the transmitter to stop */
				must_restart = 1;
				cep->stats.tx_errors++;
			}

			cep->stats.tx_packets++;

			/* Free the sk buffer associated with this last transmit.
			 */
			dev_kfree_skb_irq (cep->tx_skbuff[cep->skb_dirty]);
			cep->skb_dirty =
				(cep->skb_dirty + 1) & TX_RING_MOD_MASK;

			/* Update pointer to next buffer descriptor to be transmitted.
			 */
			if (bdp->cbd_sc & BD_SC_WRAP)
				bdp = cep->tx_bd_base;
			else
				bdp++;

			/* Since we have freed up a buffer, the ring is no longer
			 * full.
			 */
			if (cep->tx_full) {
				cep->tx_full = 0;
				if (netif_queue_stopped (dev))
					netif_wake_queue (dev);
			}

			cep->dirty_tx = (cbd_t *) bdp;
		}

		if (must_restart) {
			volatile cpm8xx_t *cp;

			/* Some transmit errors cause the transmitter to shut
			 * down.  We now issue a restart transmit.  Since the
			 * errors close the BD and update the pointers, the restart
			 * _should_ pick up without having to reset any of our
			 * pointers either.
			 */
			cp = cpmp;
			cp->cp_cpcr = mk_cr_cmd (CPM_CR_ENET, CPM_CR_RESTART_TX) |
					CPM_CR_FLG;
			while (cp->cp_cpcr & CPM_CR_FLG)
				;
		}
		spin_unlock (&cep->lock);
	}

	/* Check for receive busy, i.e. packets coming but no place to
	 * put them.  This "can't happen" because the receive interrupt
	 * is tossing previous frames.
	 */
	if (int_events & SCCM_BSY) {
		cep->stats.rx_dropped++;
		printk ("HDLC ENET: BSY can't happen.\n");
	}

	return;
}

/* During a receive, the cur_rx points to the current incoming buffer.
 * When we update through the ring, if the next incoming buffer has
 * not been given to the system, we just set the empty indicator,
 * effectively tossing the packet.
 */
static int hdlc_enet_rx (struct net_device *dev)
{
	struct hdlc_enet_private *cep;
	volatile cbd_t *bdp;
	struct sk_buff *skb;
	ushort pkt_len;

	unsigned char *cp;
	struct ids_frame_hdr *ih;
	struct ids_arp_hdr *iah;
	struct eth_arp_hdr *eah;
	struct ethhdr *eh;
	int i;

	if (devel_debug & DEBUG_RX)
		printk ("%s\n", __FUNCTION__);

	cep = (struct hdlc_enet_private *) dev->priv;

	/* First, grab all of the stats for the incoming packet.
	 * These get messed up if we get called due to a busy condition.
	 */
	bdp = cep->cur_rx;

	for (;;) {
		if (bdp->cbd_sc & BD_SC_EMPTY)
			break;

#ifndef final_version
		/* Since we have allocated space to hold a complete frame, both
		 * the first and last indicators should be set.
		 */
		if ((bdp->cbd_sc & (BD_SC_FIRST | BD_SC_LAST)) !=
		    (BD_SC_FIRST | BD_SC_LAST))
			printk ("HDLC ENET: rcv is not first+last\n");
#endif

		/*
		 * BD_SC_BR -> LG bit
		 * BD_SC_FR -> NO bit
		 * BD_SC_PR -> AB bit
		 * BD_SC_NAK -> CR bit
		 * 0x80 -> DE bit
		 */
		if (bdp->cbd_sc & BD_SC_STATS) {
			/*
			 * For statistics only
			 */
			/* frame length violation, shouldn't happen */
			if (bdp->cbd_sc & BD_SC_BR) {
				if (devel_debug & DEBUG_RX)
					printk ("rx_lg %d\n", bdp->cbd_datlen);
				cep->stats.rx_length_errors++;
			}
			/* nonoctet aligned */
			else if (bdp->cbd_sc & BD_SC_FR)
				cep->stats.rx_frame_errors++;
			/* rcv abort */
			else if (bdp->cbd_sc & BD_SC_PR)
				cep->stats.rx_missed_errors++;
			/* CRC error */
			if (bdp->cbd_sc & BD_SC_NAK)
				cep->stats.rx_crc_errors++;
			/* carrier detect lost - can't happen */
			if (bdp->cbd_sc & BD_SC_CD)
				cep->stats.rx_missed_errors++;
			/* overrun */
			if (bdp->cbd_sc & BD_SC_OV)
				cep->stats.rx_fifo_errors++;
			/* DPLL error */
			if (bdp->cbd_sc & BD_SC_DE)
				cep->stats.rx_missed_errors++;
		}

		cp = cep->rx_vaddr[(bdp - cep->rx_bd_base)];
		pkt_len = bdp->cbd_datlen;
		if (devel_debug & DEBUG_RX) {
			int len;
			len = pkt_len > 32 ? 32 : pkt_len;
			printk("raw packet dump IN length %d dump length %d\n", pkt_len, len);
			for (i = 0; i < len; i++)
				printk("%#x ", cp[i]);
			printk("\n");
		}
		/*
		 * ids_frame_hdr should start after scc_hdlc_hdr
		 */
		/*ih = (struct ids_frame_hdr *) &cp[(HDLC_HDR - 1)];*/
		ih = (struct ids_frame_hdr *) cp;
		/* It's hard to recognize collisions, but if HADDR is
		 * corrupted, then we can be fairly sure that it was due to
		 * a collision (NOTE 0x80 is the broadcast address used by IDS).
		 * The same goes for target and source in the IDS header.
		 */
		if (!CHKFORME (ih, cep->myhaddr)) {
			if (devel_debug & DEBUG_RX)
				printk ("not_for_me\n");
			cep->stats.rx_frame_errors++;
		} else {

			/* Process the incoming frame.
			 */
			/* if this is not a supported protocol then drop it */
			if ((ih->proto_1 != IDS_IP_1
			     && ih->proto_2 != IDS_IP_2)
			    && (ih->proto_1 != IDS_ARP_1
				&& ih->proto_2 != IDS_ARP_2)) {
				/* drop it */
				cep->stats.rx_dropped++;
				if (devel_debug & DEBUG_RX)
					printk ("bad protocol or not_for_me - packet dropped\n");
				goto badpkt;
			}
			cep->stats.rx_packets++;
			/* adjust pkt_len to remove the HDLC header, IDS header and CRC */
			pkt_len -= (sizeof (*ih) + HDLC_CRC);
			/* check for a short packet and pad it out */
			cep->stats.rx_bytes += pkt_len;

			/* This does 16 byte alignment, much more than we need.
			 */
			if (ih->proto_1 == IDS_IP_1
			    && ih->proto_2 == IDS_IP_2)
				i = pkt_len + sizeof (*eh);
			else	/* must be ARP */
				i = sizeof (*eh) + sizeof (*eah);
			skb = dev_alloc_skb (i);

			if (skb == NULL) {
				printk ("%s: Memory squeeze, dropping packet.\n", dev->name);
				cep->stats.rx_dropped++;
			} else {
				skb->dev = dev;
				skb_put (skb, i);	/* Make room */
				/* reconstruct the ether frame header */
				eh = (struct ethhdr *) skb->data;
				memset ((void *) eh, 0, sizeof (*eh));
				eh->h_dest[(ETH_ALEN - 1)] = cep->myhaddr;
				/* might have been a broadcast */
				if (ih->target == 0x80) {
					for (i = 0; i < ETH_ALEN; i++)
						eh->h_dest[i] = 0xff;
				}
				eh->h_source[(ETH_ALEN - 1)] =
					UNMKIDSADDR (ih->source);
				/* reconstruct the IP frame */
				if (ih->proto_1 == IDS_IP_1
				    && ih->proto_2 == IDS_IP_2) {
					eh->h_proto = ETH_P_IP;
					/*
					 * This was eth_copy_and_sum(), but we cannot use that here.
					 * Luckily, it merely copies and does not calculate a check
					 * sum.
					 */
					/* skip over our new ethhdr */
					memcpy (skb->data + sizeof (*eh), (void *) ih + sizeof (*ih),
						pkt_len);
				} else {
					/* must be ARP */
					eh->h_proto = ETH_P_ARP;
					/* construct an arphdr from the received packet */
					eah = (struct eth_arp_hdr *) (skb->data + sizeof(*eh));
					iah = (struct ids_arp_hdr *) ((void *)ih + sizeof(*ih));
					eah->oah.ar_hrd = ARPHRD_ETHER;
					eah->oah.ar_pro = ETH_P_IP;
					eah->oah.ar_hln = ETH_ALEN;
					eah->oah.ar_pln = iah->oah.ar_pln;
					eah->oah.ar_op = iah->oah.ar_op;
					memset (eah->ar_sha, 0, ETH_ALEN);
					memset (eah->ar_tha, 0, ETH_ALEN);
#ifdef IDSARP_USES_ETHADDR	/* not normally defined */
					eah->ar_sha[(ETH_ALEN - 1)] = iah->ar_sha;
#else
					eah->ar_sha[(ETH_ALEN - 1)] = UNMKIDSADDR(iah->ar_sha);
#endif
					/*
					 * if this is an ARPOP_REQUEST then ar_tha should be
					 * all zero, which the memset takes care of. Else...
					 */
					if (eah->oah.ar_op != ARPOP_REQUEST)
#ifdef IDSARP_USES_ETHADDR	/* not normally defined */
						eah->ar_tha[(ETH_ALEN - 1)] = iah->ar_tha;
#else
						eah->ar_tha[(ETH_ALEN - 1)] = UNMKIDSADDR (iah->ar_tha);
#endif
					/* copy the IP addresses */
					memcpy (eah->ar_sip, iah->ar_sip, 4);
					memcpy (eah->ar_tip, iah->ar_tip, 4);
				}
				skb->protocol = eth_type_trans (skb, dev);
				skb->pkt_type = PACKET_HOST;
				netif_rx (skb);
				if (devel_debug & DEBUG_RX) {
					int len;
					len = i > 32 ? 32 : i;
					printk("raw packet dump OUT length %d dump length %d\n", i, len);
					for (i = 0; i < len; i++)
						printk("%#x ", skb->data[i]);
					printk("\n");
				}
			}
		}

   badpkt:
		/* Clear the status flags for this buffer.
		 */
		bdp->cbd_sc &= ~BD_SC_STATS;

		/* Mark the buffer empty.
		 */
		bdp->cbd_sc |= BD_SC_EMPTY;

		/* Update BD pointer to next entry.
		 */
		if (bdp->cbd_sc & BD_SC_WRAP)
			bdp = cep->rx_bd_base;
		else
			bdp++;
	}
	cep->cur_rx = (cbd_t *) bdp;

	return 0;
}

static int
hdlc_enet_close(struct net_device *dev)
{
	/* Don't know what to do yet.
	*/
	netif_stop_queue(dev);

	return 0;
}

static struct net_device_stats *hdlc_enet_get_stats(struct net_device *dev)
{
	struct hdlc_enet_private *cep = (struct hdlc_enet_private *)dev->priv;

	return &cep->stats;
}

static int hdlc_enet_set_mac_address(struct net_device *dev, void *p)
{
	struct sockaddr *addr = p;
	struct hdlc_enet_private *cep = (struct hdlc_enet_private *)dev->priv;
	unsigned char nhaddr;
	volatile	scc_hdlc_t	*sup;


	if (netif_running(dev))
		return -EBUSY;


	nhaddr = addr->sa_data[(ETH_ALEN - 1)] & 0xff;
	/* check for valid range of addresses */
	if (nhaddr < MIN_HADDR || nhaddr > MAX_HADDR) {
		printk("%s: invalid HADDR %#x - myhaddr not changed\n", dev->name,
			nhaddr);
		return -EINVAL;
	}
	memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);
	eaddl = cep->myhaddr = nhaddr;
	if (devel_debug & DEBUG_IOCT)
		printk("set myhaddr to %d\n", nhaddr);

	/*
	 * Get pointer to SCC area in parameter RAM.
	 */
	sup = (scc_hdlc_t *)(&cpmp->cp_dparam[PROFF_ENET]);
	sup->scc_haddr2 = MKIDSADDR(cep->myhaddr);
	sup->scc_haddr3 = sup->scc_haddr2;
	sup->scc_haddr4 = sup->scc_haddr2;

	return 0;
}

#ifdef CONFIG_TQM860L
/* this should be done by m8xx_cpm_setbrg() but it doesn't seem
   to be exported ?? Stolen from commproc.c */
#define BRG_INT_CLK		(((bd_t *)__res)->bi_intfreq * 1000000)
#define BRG_UART_CLK		(BRG_INT_CLK/16)
#define BRG_UART_CLK_DIV16	(BRG_UART_CLK/16)

void
cpm_hdlc_setbaud(int baud)
{
	volatile uint *bp;

	/* use BRG4 */
	bp = (uint *)&cpmp->cp_brgc4;
	if (((BRG_UART_CLK / baud) - 1) < 4096)
		*bp = (((BRG_UART_CLK / baud) - 1) << 1) | CPM_BRG_EN;
	else
		*bp = (((BRG_UART_CLK_DIV16 / baud) - 1) << 1) |
				CPM_BRG_EN | CPM_BRG_DIV16;
}
#endif

/* Initialize the CPM Ethernet on SCC.	If EPPC-Bug loaded us, or performed
 * some other network I/O, a whole bunch of this has already been set up.
 * It is no big deal if we do it again, we just have to disable the
 * transmit and receive to make sure we don't catch the CPM with some
 * inconsistent control information.
 */
#ifdef MODULE
static struct net_device dev_hdlc_enet = {
	init: hdlc_enet_probe,
	base_addr: 0,
	irq: 0,
};

int
init_module(void)
{
	if (register_netdev(&dev_hdlc_enet) != 0)
		return -EIO;
	return 0;
}
#endif

#ifndef MODULE
int __init
hdlc_enet_init(void)
#else
static int
hdlc_enet_init(struct net_device *dev)
#endif
{
#ifndef MODULE
	struct net_device *dev;
#endif
	unsigned char *ba;
	dma_addr_t	mem_addr;
	struct hdlc_enet_private *cep;
	int i, j, k;
	bd_t		*bd;
	volatile	cbd_t		*bdp;
	volatile	cpm8xx_t	*cp;
	volatile	scc_t		*sccp;
	volatile	scc_hdlc_t	*sup;
	volatile	iop8xx_t	*iop;

	cp = cpmp;	/* Get pointer to Communication Processor */
	iop = (iop8xx_t *)&(((volatile immap_t *)IMAP_ADDR)->im_ioport);

	bd = (bd_t *)__res;

	/* Create an Ethernet device instance.
	*/
#ifndef MODULE
	dev = init_etherdev(0, 0);
#else
	dev = init_etherdev(dev, 0);
#endif

	/* Set Ethernet station address. */
	if (eaddl == 0) {
		/* always emit this warning */
		printk ("%s: Warning: ether addr not set! "
			"The driver will not work correctly.\n",
			__FUNCTION__);
	} else {
		/* set station_addr from eaddl */
		/* the first five are always 0 */
		for (i = 0; i < (ETH_ALEN - 1); i++)
			dev->dev_addr[i] = 0;
		dev->dev_addr[(ETH_ALEN - 1)] = eaddl & 0xff;
	}

	/* Allocate some private information.
	*/
	cep = (struct hdlc_enet_private *)kmalloc(sizeof(*cep), GFP_KERNEL);
	if (cep == NULL)
		return -ENOMEM;

	__clear_user(cep, sizeof(*cep));
	spin_lock_init(&cep->lock);

	/*
	 * eaddl not set at startup - assume that it will be set using
	 * ifconfig.
	 */
	if (eaddl == 0)
		goto no_eaddl;

	cep->myhaddr = eaddl & 0xff;
	/* check for valid range of addresses */
	if (cep->myhaddr < MIN_HADDR || cep->myhaddr > MAX_HADDR) {
		printk("%s: invalid HADDR %#x - interface inoperative\n", dev->name, cep->myhaddr);
		kfree((void *)cep);
		return -EINVAL;
	}

no_eaddl:
	/* Get pointer to SCC area in parameter RAM.
	*/
	sup = (scc_hdlc_t *)(&cp->cp_dparam[PROFF_ENET]);

	/* And another to the SCC register area.
	*/
	sccp = (volatile scc_t *)(&cp->cp_scc[SCC_ENET]);
	cep->sccp = (scc_t *)sccp;		/* Keep the pointer handy */

	/* Disable receive and transmit just in case.
	*/
	sccp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);


#ifdef CONFIG_TQM860L
	/* use pa12, pa13 pins for TXD2, RXD2 respectively */
	/* port A pin assignment reg */
	iop->iop_papar = iop->iop_papar | 0x000c;
#else
	/* use pa8, pa9 pins for TXD4, RXD4 respectively */
	/* port A pin assignment reg */
	/* also use pa0 as CLK8 - IDS only */
	iop->iop_papar = iop->iop_papar | 0x80c0;
#endif


	/* port A Data Direction reg */
#ifdef CONFIG_TQM860L
	iop->iop_padir = iop->iop_padir & ~0x000c;
#else
	iop->iop_padir = iop->iop_padir & ~0x80c0;
#endif

#ifdef CONFIG_TQM860L
	/*
	 * according to the 860UM CD2 defaults to being active (GND),
	 * use pb18=RTS2, use pc9=CD2, use pc8=CTS2
	 * port C pin assignment reg
	 * clear to select CD2 & CTS2
	 */
	iop->iop_pcpar = iop->iop_pcpar & ~0x00c0;
	/* clear to set CD2 & CTS2 as peripheral function */
	iop->iop_pcdir = iop->iop_pcdir & ~0x00c0;
	/* port C special options */
	/*  set pc7 as CTS2, leave pc6 as gpio */
	iop->iop_pcso  = (iop->iop_pcso  | 0x0040) & ~0x0080;
	/* set RTS2 */
	cp->cp_pbpar = cp->cp_pbpar | 0x2000;
	cp->cp_pbdir = cp->cp_pbdir | 0x2000;
#else
	/*
	 * according to the 866UM CD4 defaults to being active (GND),
	 * so no need to do anything here.
	 * use pc12=RTS4, use pc5=CTS4, use pc4=CD4
	 * port C pin assignment reg
	 * clear to select CTS4 & CD4 function
	 */
	iop->iop_pcpar = iop->iop_pcpar & ~0x0c00;
	/* set to select RTS4 function */
	iop->iop_pcpar = iop->iop_pcpar | 0x0008;
	/* clear to set CD4, RTS4 & CTS4 as peripheral functions */
	iop->iop_pcdir = iop->iop_pcdir & ~0x0c08;
	/* port C special options */
	/*  set pc5 as CTS4, leave pc4 as gpio  */
	iop->iop_pcso  = (iop->iop_pcso  | 0x0400) & ~0x0800;
#endif

#ifdef CONFIG_TQM860L
	/*
	 * SI clock routing
	 * Route BRG04 to produce baud clock for SCC3
	 * this also connects the SCC to NMSI
	 */
	cp->cp_sicr = (cp->cp_sicr & ~0x00ff0000) | 0x001b0000;
#else
	/*
	 * SI clock routing
	 * use CLK8
	 * this also connects SCC4 to NMSI
	 */
	cp->cp_sicr = (cp->cp_sicr & ~0xff000000) | 0x3f000000;
#endif

	/* Allocate space for the buffer descriptors in the DP ram.
	 * These are relative offsets in the DP ram address space.
	 * Initialize base addresses for the buffer descriptors.
	 */
	cep->rxdp_addr = i = m8xx_cpm_dpalloc(sizeof(cbd_t) * RX_RING_SIZE);
	sup->shd_genscc.scc_rbase = i;
	cep->rx_bd_base = (cbd_t *)&cp->cp_dpmem[i];

	cep->txdp_addr = i = m8xx_cpm_dpalloc(sizeof(cbd_t) * TX_RING_SIZE);
	sup->shd_genscc.scc_tbase = i;
	cep->tx_bd_base = (cbd_t *)&cp->cp_dpmem[i];

	cep->dirty_tx = cep->cur_tx = cep->tx_bd_base;
	cep->cur_rx = cep->rx_bd_base;

	/* Issue init Rx BD command for SCC.
	 * Manual says to perform an Init Rx parameters here.  We have
	 * to perform both Rx and Tx because the SCC may have been
	 * already running.
	 * In addition, we have to do it later because we don't yet have
	 * all of the BD control/status set properly.
	cp->cp_cpcr = mk_cr_cmd(CPM_CR_ENET, CPM_CR_INIT_RX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);
	 */

	/* Initialize function code registers for big-endian.
	*/
	sup->shd_genscc.scc_rfcr = SCC_EB;
	sup->shd_genscc.scc_tfcr = SCC_EB;

	/* Set maximum bytes per receive buffer.
	 * This appears to be an Ethernet frame size, not the buffer
	 * fragment size.  It must be a multiple of four.
	 */
	sup->shd_genscc.scc_mrblr = PKT_MAXBLR_SIZE;

	/* Setup CRC generator values for HDLC */
	sup->scc_c_mask = 0x0000F0B8;
	sup->scc_c_pres = 0x0000FFFF;

	/* Initialize all error counters to 0 */
	sup->scc_disfc = 0;
	sup->scc_crcec = 0;
	sup->scc_abtsc = 0;
	sup->scc_nmarc = 0;
	sup->scc_retrc = 0;

	/* Set maximum frame length size */
	sup->scc_mflr = PKT_MAXBLR_SIZE;

	/* set to 1 for per frame processing change later if needed */
	sup->scc_rfthr = 1;

	sup->scc_hmask	= 0xfd;

	sup->scc_haddr2 = MKIDSADDR(cep->myhaddr);
	sup->scc_haddr3 = sup->scc_haddr2;
	sup->scc_haddr4 = sup->scc_haddr2;
	/* broadcast */
	sup->scc_haddr1 = 0x80;

	/* Now allocate the host memory pages and initialize the
	 * buffer descriptors.
	 */
	bdp = cep->tx_bd_base;
	for (i = 0; i < TX_RING_SIZE; i++) {

		/* Initialize the BD for every fragment in the page.
		*/
		bdp->cbd_sc = 0;
		bdp->cbd_bufaddr = 0;
		bdp++;
	}

	/* Set the last buffer to wrap.
	*/
	bdp--;
	bdp->cbd_sc |= BD_SC_WRAP;

	bdp = cep->rx_bd_base;
	k = 0;
	for (i = 0; i < CPM_ENET_RX_PAGES; i++) {

		/* Allocate a page.
		*/
		ba = (unsigned char *)consistent_alloc(GFP_KERNEL, PAGE_SIZE, &mem_addr);
#ifdef MODULE
		rx_pages[i] = ba;
#endif


		/* Initialize the BD for every fragment in the page.
		*/
		for (j = 0; j < CPM_ENET_RX_FRPPG; j++) {
			bdp->cbd_sc = BD_SC_EMPTY | BD_SC_INTRPT;
			bdp->cbd_bufaddr = mem_addr;
			cep->rx_vaddr[k++] = ba;
			ba += CPM_ENET_RX_FRSIZE;
			mem_addr += CPM_ENET_RX_FRSIZE;
			bdp++;
		}
	}

	/* Set the last buffer to wrap.
	*/
	bdp--;
	bdp->cbd_sc |= BD_SC_WRAP;

	/* Let's re-initialize the channel now.	 We have to do it later
	 * than the manual describes because we have just now finished
	 * the BD initialization.
	 */
	cp->cp_cpcr = mk_cr_cmd(CPM_CR_ENET, CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	cep->skb_cur = cep->skb_dirty = 0;

#if 0
	sccp->scc_gsmrh = 0;
	sccp->scc_gsmrl = (SCC_GSMRL_MODE_HDLC|SCC_GSMRL_RENC_NRZI|
			/* SCC_GSMRL_TPL_16 -> more collisions */
			SCC_GSMRL_TENC_NRZI|SCC_GSMRL_TSNC_INF|SCC_GSMRL_TPL_8|
			/* doesn't work with [RT]DCR_{16,32} */
			/* SCC_GSMRL_TPP_ZEROS -> more collisions */
			SCC_GSMRL_TPP_10|SCC_GSMRL_TDCR_8|SCC_GSMRL_RDCR_8);
#ifdef HDLC_BUSMODE /* AFAICT BUSMODE can't work */
	if (devel_debug & DEBUG_INIT)
		printk("%s: using Busmode\n", __FUNCTION__);
	/* NOF=2, RTE=1, DRT=1, BUS=1 */
	sccp->scc_pmsr = 0x2260;
#else
	if (devel_debug & DEBUG_INIT)
		printk("%s: NOT using Busmode\n", __FUNCTION__);
	/* NOF=2, RTE=0, DRT=1, BUS=0 */
	sccp->scc_pmsr = 0x2040;
#endif	/* HDLC_BUSMODE */
#endif /* 0 */

#if 1
#ifdef HDLC_BUSMODE  /* AFAICT BUSMODE can't work with NC650 */
	sccp->scc_gsmrh = 0x00000080;
	sccp->scc_gsmrl = 0x00000000;
	if (devel_debug & DEBUG_INIT)
		printk("%s: using Busmode\n", __FUNCTION__);
	/* NOF=2, RTE=1, DRT=1, BUS=1 */
	sccp->scc_pmsr = 0x0220;
#else
	sccp->scc_gsmrh = 0x00000000;
	sccp->scc_gsmrl = 0x00000000;
	if (devel_debug & DEBUG_INIT)
		printk("%s: NOT using Busmode\n", __FUNCTION__);
	/* NOF=2, RTE=0, DRT=1, BUS=0 */
	sccp->scc_pmsr = 0x0040;
#endif	/* HDLC_BUSMODE */
#endif /* 1 */

/* loopback for local testing */
#ifdef GJTEST
	printk("LOOPBACK!\n");
	sccp->scc_gsmrl |= SCC_GSMRL_DIAG_LOOP;
#endif

	/* Disable all interrupts and clear all pending
	 * events.
	 */
	sccp->scc_sccm = 0;
	sccp->scc_scce = 0xffff;
	sccp->scc_dsr  = 0x7e7e;

#ifdef HDLC_BUSMODE /* AFAICT BUSMODE can't work */
	if (devel_debug & DEBUG_INIT)
		printk("%s: using Busmode\n", __FUNCTION__);
	/* NOF=2, RTE=1, DRT=1, BUS=1 */
	sccp->scc_pmsr = 0x0220;
#else
	if (devel_debug & DEBUG_INIT)
		printk("%s: NOT using Busmode\n", __FUNCTION__);
	/* NOF=2, RTE=0, DRT=1, BUS=0 */
	sccp->scc_pmsr = 0x0040;
#endif

	/* Install our interrupt handler.
	*/
	cpm_install_handler(CPMVEC_ENET, hdlc_enet_interrupt, dev);

#ifdef CONFIG_TQM860L
	/* Set up the baud rate generator	*/
	cpm_hdlc_setbaud(1024000);
#endif

	dev->base_addr = (unsigned long)sup;
	dev->priv = cep;

	/* The CPM Ethernet specific entries in the device structure. */
	dev->open = hdlc_enet_open;
	dev->hard_start_xmit = hdlc_enet_start_xmit;
	dev->tx_timeout = hdlc_enet_timeout;
	dev->watchdog_timeo = TX_TIMEOUT;
	dev->stop = hdlc_enet_close;
	dev->get_stats = hdlc_enet_get_stats;
	dev->set_mac_address = hdlc_enet_set_mac_address;

	/* enable interrupts */
	/* txbuf empty, rxbuf full, tx underrun, bsy */
	sccp->scc_sccm = 0x1f;

	/* And last, enable the transmit and receive processing.
	*/
	sccp->scc_gsmrl |= (SCC_GSMRL_ENR | SCC_GSMRL_ENT);

	printk("%s: HDLC ENET Version 0.3 on SCC%d, ", dev->name, SCC_ENET+1);
	for (i=0; i<5; i++)
		printk("%02x:", dev->dev_addr[i]);
	printk("%02x\n", dev->dev_addr[5]);

	return 0;
}

#ifdef MODULE
void
cleanup_module(void)
{
	volatile cpm8xx_t *cp;
	struct hdlc_enet_private *cep = NULL;
	int i;

	cep = (struct hdlc_enet_private *)dev_hdlc_enet.priv;
	unregister_netdev(&dev_hdlc_enet);
	dev_hdlc_enet.priv = NULL;
	/* Send the CPM a STOP TX command.
	*/
	cp = cpmp;	/* Get pointer to Communication Processor */
	cp->cp_cpcr = mk_cr_cmd(CPM_CR_ENET, CPM_CR_STOP_TX) | CPM_CR_FLG;
	/* could be dangerous */
	while (cp->cp_cpcr & CPM_CR_FLG);
	/* turn off the SCC */
	cep->sccp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);
	cpm_free_handler(CPMVEC_ENET);

	/* free pages allocated for the RX ring */
	for (i = 0; i < CPM_ENET_RX_PAGES; i++) {
		/* Free a page.
		*/
		consistent_free((void *)rx_pages[i]);
	}
	if (cep) {
		/* free memory in the parameter RAM */
		m8xx_cpm_dpfree(cep->rxdp_addr);
		m8xx_cpm_dpfree(cep->txdp_addr);

		/* free unsent TX buffers */
		if (cep->skb_cur != cep->skb_dirty)
			for (i = cep->skb_dirty; i < cep->skb_cur; i++)
				dev_kfree_skb_irq(cep->tx_skbuff[i]);

		kfree((void *)cep);
	}
}
#endif

#ifndef MODULE
/* set eaddl on the boot commandline */
static int __init
hdlc_enet_set_eaddl(char *str)
{
	eaddl = simple_strtol(str, NULL, 0);
	return 1;
}
__setup("eaddl=", hdlc_enet_set_eaddl);
/* set devel_debug on the boot commandline */
static int __init
hdlc_enet_set_devel_debug(char *str)
{
	devel_debug = simple_strtol(str, NULL, 0);
	return 1;
}
__setup("devel_debug=", hdlc_enet_set_devel_debug);
#endif
