/*
 * arch/ppc/kernel/mpc8220_pic.c
 *
 * Interupt Controller functions for the Motorola MPC8220 processor.
 *
 * Author: MontaVista Software, Inc.
 *         source@mvista.com
 *
 * Based on mgt5100_pic.c by Dale Farnsworth <dfarnsworth@mvista.com>
 * which was derived from code developed by Kent Borg
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */

#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/stddef.h>
#include <linux/delay.h>
#include <linux/irq.h>

#include <asm/io.h>
#include <asm/processor.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/mpc8220.h>

static void
mpc8220_ic_disable(unsigned int irq)
{
	struct mpc8220_intr *intr = (struct mpc8220_intr *)MPC8220_INTCTL;
	struct mpc8220_sdma *sdma = (struct mpc8220_sdma *)MPC8220_SDMA;
	u32 val;

	if (irq == MPC8220_IRQ0) {
		val = in_be32(&intr->ctrl);
		val &= ~(1 << 11);
		out_be32(&intr->ctrl, val);
	} else if (irq < MPC8220_IRQ1) {
		BUG();
	} else if (irq <= MPC8220_GPIO_IRQ) {
		val = in_be32(&intr->ctrl);
		val &= ~(1 << (10 - (irq - MPC8220_IRQ1)));
		out_be32(&intr->ctrl, val);
	} else if (irq < MPC8220_SDMA_IRQ_BASE) {
		val = in_be32(&intr->main_mask);
		val |= 1 << (16 - (irq - MPC8220_MAIN_IRQ_BASE));
		out_be32(&intr->main_mask, val);
	}
	else if (irq < MPC8220_PERP_IRQ_BASE) {
		val = in_be32(&sdma->IntMask);
		val |= 1 << (irq - MPC8220_SDMA_IRQ_BASE);
		out_be32(&sdma->IntMask, val);
	} else {
		val = in_be32(&intr->per_mask);
		val |= 1 << (31 - (irq - MPC8220_PERP_IRQ_BASE));
		out_be32(&intr->per_mask, val);
	}
}

static void
mpc8220_ic_enable(unsigned int irq)
{
	struct mpc8220_intr *intr = (struct mpc8220_intr *)MPC8220_INTCTL;
	struct mpc8220_sdma *sdma = (struct mpc8220_sdma *)MPC8220_SDMA;
	u32 val;

	if (irq == MPC8220_IRQ0) {
		val = in_be32(&intr->ctrl);
		val |= 1 << 11;
		out_be32(&intr->ctrl, val);
	} else if (irq < MPC8220_IRQ1) {
		BUG();
	} else if (irq <= MPC8220_GPIO_IRQ) {
		val = in_be32(&intr->ctrl);
		val |= 1 << (10 - (irq - MPC8220_IRQ1));
		out_be32(&intr->ctrl, val);
	} else if (irq < MPC8220_SDMA_IRQ_BASE) {
		val = in_be32(&intr->main_mask);
		val &= ~(1 << (16 - (irq - MPC8220_MAIN_IRQ_BASE)));
		out_be32(&intr->main_mask, val);
	}
	else if (irq < MPC8220_PERP_IRQ_BASE) {
		val = in_be32(&sdma->IntMask);
		val &= ~(1 << (irq - MPC8220_SDMA_IRQ_BASE));
		out_be32(&sdma->IntMask, val);
	} else {
		val = in_be32(&intr->per_mask);
		val &= ~(1 << (31 - (irq - MPC8220_PERP_IRQ_BASE)));
		out_be32(&intr->per_mask, val);
	}
}

static void
mpc8220_ic_ack(unsigned int irq)
{
	struct mpc8220_intr *intr = (struct mpc8220_intr *)MPC8220_INTCTL;
	struct mpc8220_sdma *sdma = (struct mpc8220_sdma *)MPC8220_SDMA;
	u32 val;

	/*
	 * Only external and SDMA irqs are reset here,
	 * others in interrupting hardware.
	 */

	switch (irq) {
	case MPC8220_IRQ0:
		val = in_be32(&intr->ctrl);
		val |= 0x08000000;
		out_be32(&intr->ctrl, val);
		break;
	case MPC8220_IRQ1:
		val = in_be32(&intr->ctrl);
		val |= 0x04000000;
		out_be32(&intr->ctrl, val);
		break;
	case MPC8220_GPIO_IRQ:
		val = in_be32(&intr->ctrl);
		val |= 0x02000000;
		out_be32(&intr->ctrl, val);
		break;
	default:
		if (irq >= MPC8220_SDMA_IRQ_BASE
		    && irq < (MPC8220_SDMA_IRQ_BASE + MPC8220_SDMA_IRQ_NUM)) {
			out_be32(&sdma->IntPend,
				 1 << (irq - MPC8220_SDMA_IRQ_BASE));
		}
		break;
	}
}

static void
mpc8220_ic_disable_and_ack(unsigned int irq)
{
	mpc8220_ic_disable(irq);
	mpc8220_ic_ack(irq);
}

static void
mpc8220_ic_end(unsigned int irq)
{
	mpc8220_ic_ack(irq);

	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS))) {
		mpc8220_ic_enable(irq);
	}
}

static struct hw_interrupt_type mpc8220_ic = {
	.typename	= "MPC8220",
	.startup	= NULL,
	.shutdown	= NULL,
	.enable		= mpc8220_ic_enable,
	.disable	= mpc8220_ic_disable,
	.ack		= mpc8220_ic_disable_and_ack,
	.end		= mpc8220_ic_end,
	.set_affinity	= 0
};

void __init
mpc8220_init_irq(void)
{
	struct mpc8220_intr *intr = (struct mpc8220_intr *)MPC8220_INTCTL;
	struct mpc8220_sdma *sdma = (struct mpc8220_sdma *)MPC8220_SDMA;
	int i;

	/* Disable all interrupt sources. */
	out_be32(&sdma->IntPend, 0xffffffff);	/* clear all pending irqs */
	out_be32(&sdma->IntMask, 0xffffffff);	/* disable all SDMA irqs */
	out_be32(&intr->per_mask, 0x7ffffc00);	/* disable all but SDMA */
	out_be32(&intr->ctrl, 0);
			/* - make sure IRQ0, IRQ1, GPIO_IRQ disabled */
	out_be32(&intr->main_mask, 0x00012fff);
			/* - disable all but IRQ1, GPIO_IRQ, and LO_int */
	out_be32(&intr->ctrl,
			0x0e000000 |	/* clear IRQ0, IRQ1, and GPIO_IRQ */
			0x00000000 |    /* all level sensitive, active high */
			0x00001000 |	/* MEE master external enable */
			0x00000000 |	/* disable IRQ0, IRQ1, and GPIO_IRQ */
			0x00000001);	/* CEb route critical normally */

	/* Zero a bunch of the priority settings.  */
	out_be32(&intr->per_pri1, 0);
	out_be32(&intr->per_pri2, 0);
	out_be32(&intr->per_pri3, 0);
	out_be32(&intr->main_pri1, 0);
	out_be32(&intr->main_pri2, 0);

	/* Initialize irq_desc[i].handler's with mpc8220_ic. */
	for (i = 0; i < NR_IRQS; i++) {
		irq_desc[i].handler = &mpc8220_ic;
	}
}

int
mpc8220_get_irq(struct pt_regs *regs)
{
	struct mpc8220_intr *intr = (struct mpc8220_intr *)MPC8220_INTCTL;
	struct mpc8220_sdma *sdma = (struct mpc8220_sdma *)MPC8220_SDMA;
	u32 status;
	int irq = -1;

	status = in_be32(&intr->enc_status);

	if (status & 0x00000400) {		/* critical */
		irq = (status >> 8) & 0x3;
		if (irq == 2)			/* high priority peripheral */
			goto peripheral;
		irq += MPC8220_CRIT_IRQ_BASE;
	}
	else if (status & 0x00200000) {		/* main */
		irq = (status >> 16) & 0x1f;
		if (irq == 4)			/* low priority peripheral */
			goto peripheral;
		irq += MPC8220_MAIN_IRQ_BASE;
	}
	else if (status & 0x20000000) {		/* peripheral */
peripheral:
		irq = (status >> 24) & 0x1f;
		if (irq == 0) {			/* smartcomm/smartdma */
			status = in_be32(&sdma->IntPend);
			irq = ffs(status) + MPC8220_SDMA_IRQ_BASE-1;
		}
		else
			irq += MPC8220_PERP_IRQ_BASE;
	}

	return irq;
}
