/*
 * arch/ppc/platforms/alaska_fpga_pic.c
 *
 * Functions for interrupt controller in Alaska FPGA.
 *
 * Author: MontaVista Software, Inc.
 *         source@mvista.com
 *
 * Derived from code developed by Kent Borg.
 * Copyright 2003, Motorola, Inc.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */

#include <linux/init.h>
#include <linux/irq.h>

#include <asm/io.h>
#include <platforms/alaska.h>

/* no spinlock is needed here - all this done in arch/ppc/kernel/irq.c */

static u8 fpga_irq_is_edge[] __initdata = {
	0, 0, 0, 0,	/* 4 unused */
	1,		/* ALASKA_FPGA_INT__VFIFO_EE */
	0,		/* ALASKA_FPGA_INT__VFIFO_AE */
	0,		/* ALASKA_FPGA_INT__VFIFO_AF */
	0,		/* ALASKA_FPGA_INT__VFIFO_FF */
	1, 1, 1, 1,	/* ALASKA_FPGA_INT__PCIx */
	1,		/* ALASKA_FPGA_INT__ENG_STS */
	1,		/* ALASKA_FPGA_INT__CONSOLE */
	0,		/* ALASKA_FPGA_INT__SERIAL1 */
	0,		/* ALASKA_FPGA_INT__SERIAL2 */
	0,		/* ALASKA_FPGA_INT__USB */
	0,		/* ALASKA_FPGA_INT__PHY1 */
	0,		/* ALASKA_FPGA_INT__PHY2 */
	1,		/* ALASKA_FPGA_INT__END_PAGE */
	1,		/* ALASKA_FPGA_INT__HSYNCH */
	1,		/* ALASKA_FPGA_INT__LED */
	0, 0, 0, 0, 0,
	0, 0, 0, 0, 0   /* 10 unused */
};

static void
alaska_fpga_ic_enable(unsigned int irq)
{
	unsigned long val;

	if ((irq <= ALASKA_FPGA_INT_HIGHEST)
	    && (irq >= ALASKA_FPGA_INT_LOWEST)) {
		val = in_be32((u32 *)FPGA_INT_ENABLE);
		val |= 1 << (irq - ALASKA_FPGA_INT_BASE);
		out_be32((u32 *)FPGA_INT_ENABLE, val);
	}
}

static void
alaska_fpga_ic_disable(unsigned int irq)
{
	unsigned long val;

	if ((irq <= ALASKA_FPGA_INT_HIGHEST)
	    && (irq >= ALASKA_FPGA_INT_LOWEST)) {
		val = in_be32((u32 *)FPGA_INT_ENABLE);
		val &= ~(1 << (irq - ALASKA_FPGA_INT_BASE));
		out_be32((u32 *)FPGA_INT_ENABLE, val);
	}
}

static void
alaska_fpga_ic_disable_and_ack(unsigned int irq)
{
	unsigned long val;
	unsigned long mask;

	mask = 1 << (irq - ALASKA_FPGA_INT_BASE);

	/* mask the interrupt */
	val = in_be32((u32 *)FPGA_INT_ENABLE);
	val &= ~mask;
	out_be32((u32 *)FPGA_INT_ENABLE, val);

	/* ack edge triggered interrupt */
	if (!(irq_desc[irq].status & IRQ_LEVEL)) {
		out_be32((u32 *)FPGA_INT_STATUS, mask);
	}
}

static void
alaska_fpga_ic_end(unsigned int irq)
{
	unsigned long val;
	unsigned long mask;

	mask = 1 << (irq - ALASKA_FPGA_INT_BASE);

	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS))) {
		/* ack level triggered interrupt */
		if (irq_desc[irq].status & IRQ_LEVEL) {
			out_be32((u32 *)FPGA_INT_STATUS, mask);
		}

		/* umnask the interrupt */
		val = in_be32((u32 *)FPGA_INT_ENABLE);
		val |= mask;
		out_be32((u32 *)FPGA_INT_ENABLE, val);
	}
}

static struct hw_interrupt_type alaska_fpga_ic = {
	.typename	= "Alaska FPGA PIC",
	.startup	= NULL,
	.shutdown	= NULL,
	.enable		= alaska_fpga_ic_enable,
	.disable	= alaska_fpga_ic_disable,
	.ack		= alaska_fpga_ic_disable_and_ack,
	.end		= alaska_fpga_ic_end,
	.set_affinity	= 0
};

int
alaska_get_irq(struct pt_regs *regs)
{
	int irq;

	irq = mpc8220_get_irq(regs);
	if (irq == ALASKA_FPGA_IRQ) {
		unsigned long enabled_irqs_asserted;

		enabled_irqs_asserted =
			in_be32((u32 *)FPGA_INT_ENABLE) &
			in_be32((u32 *)FPGA_INT_STATUS);

		if(enabled_irqs_asserted == 0)
			BUG();
		else
			return ffs(enabled_irqs_asserted) +
			       ALASKA_FPGA_INT_BASE-1;
			/* using ffs() ensures that if fpga_irq1 < fpga_irq2
			 * then irq1 is higher priority than irq2 */
	}

	return irq;
}

void __init
alaska_init_irq(void)
{
	struct mpc8220_intr *intr = (struct mpc8220_intr *)MPC8220_INTCTL;
	u32 val;
	int i;

	/* initialize MPC8220 interrupt controller */
	mpc8220_init_irq();

	/* reset and disable all FPGA interrupts */
	out_be32((u32 *)FPGA_INT_STATUS, 0xffffffff);
	out_be32((u32 *)FPGA_INT_ENABLE, 0);

	/* use alaska_fpga_ic for all possible interrupt sources */
	for (i = 0; i < 32; i++) {
		irq_desc[ALASKA_FPGA_INT_BASE + i].handler = &alaska_fpga_ic;
		if (fpga_irq_is_edge[i]) {
			irq_desc[ALASKA_FPGA_INT_BASE + i].status |= IRQ_LEVEL;
		} else {
			irq_desc[ALASKA_FPGA_INT_BASE + i].status &= ~IRQ_LEVEL;
		}
	}

	/* Make IRQ1 level sensitive, active-low */
	val = in_be32(&intr->ctrl);
	out_be32(&intr->ctrl, val | 0x00300000);

	/* enable ALASKA_FPGA_IRQ and make it visible in /proc */
	request_irq(ALASKA_FPGA_IRQ, no_action, 0,
			"Alaska FPGA PIC", 0);
}
