/***********************************************************************
 *
 * (C) Copyright 2005
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * This module provides support for the external
 * interrupts on CPC45 systems.
 ***********************************************************************/

/*
 * Standard in kernel modules
 */
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/rtc.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/time.h>
#include <asm/machdep.h>		/* ppc_md */
#include <platforms/cpc45.h>

#include "cpc45_intr.h"

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */


/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME		"cpc45_intr"
#define	CPC45_INTR_VERSION	"0.1-wd"

/***************************************************************************
 * Debug support
 ***************************************************************************
 */
#if 0
#define	DEBUG
#endif

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif

/***************************************************************************
 * Global stuff - export global symbols
 ***************************************************************************
 */
EXPORT_SYMBOL(request_cpc45_irq);
EXPORT_SYMBOL(free_cpc45_irq);

/***************************************************************************
 * Local stuff
 ***************************************************************************
 */
typedef	struct intr_hdlr_s {
	void		(*handler)(int, void *, struct pt_regs *);
	void		*arg;
} intr_hdlr_t;

typedef struct onbd_intr_hdlr_s {
	intr_hdlr_t	int_hand;
	char		irq;
	short		enable;
	char		*name;
} onbd_intr_hdlr_t;

/*
 * The following array is used as reference counter so we know when
 * to install or de-install an interrupt. Please note that only
 * some of the slots are used, the others remain uninitialized (0).
 */
static unsigned int intr_cnt[NUM_8259_INTERRUPTS];

/*
 * Prototypes
 */
static void obd_interrupt_handler (int, void *, struct pt_regs *);
static int onboard_interrupt_en_dis (unsigned int, unsigned int);

onbd_intr_hdlr_t onbd_intr_hdlr [] = {
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPC45_IRQ_LM_RTC,
		CPC45_ONBOARD_LM | \
		CPC45_ONBOARD_RTC | \
		CPC45_ONBOARD_SIO_1 | \
		CPC45_ONBOARD_SIO_2 | \
		CPC45_ONBOARD_DEC   | \
		CPC45_ONBOARD_KEY_2 | \
		CPC45_ONBOARD_KEY_1 | \
		CPC45_ONBOARD_ABORT_AC,
		"LM / RTC / KEY",
	},
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPC45_IRQ_PCMCIA,
		0,
		"PCMCIA",
	},
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPC45_IRQ_ETH,
		0,
		"Ethernet",
	},
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPC45_IRQ_PCI,
		CPC45_ONBOARD_CPCID | CPC45_ONBOARD_CPCIC |
		CPC45_ONBOARD_CPCIB | CPC45_ONBOARD_CPCIA,
		"CPCI-IRQ(D-A)",
	},
};

#define NR_ONBOARD_INTERRUPTS	(sizeof(onbd_intr_hdlr)/sizeof(onbd_intr_hdlr_t))

static volatile cpc45_bcsr_t	*bcsr;

/***************************************************************************
 *
 * Initialize the module
 */
static int __init cpc45_intr_init (void)
{

	printk (KERN_INFO
		"CPC45 Extended Interrupt Support version "
		CPC45_INTR_VERSION
		"\n"
	);


/* ToDo */


	debugk ("CPC45 Interrupt Support installed\n");
	return 0;
}

/***************************************************************************
 * Module Declarations
 ***************************************************************************
 */

module_init (cpc45_intr_init);

#ifdef MODULE
/*
 * Cleanup - unregister the appropriate file from /proc
 */
void cpc45_intr_cleanup (void)
{
	int irq;

	debugk ("Cleanup CPC45 Interrupt Support\n");

	/* make sure all interrupts are disabled */
	/* the Onboard Interrupts */
	for ( ; irq < NR_ONBOARD_INTERRUPTS; ++irq) {
		int index = irq;
		if (onbd_intr_hdlr[index].int_hand.handler) {
			debugk ("Cleanup Onboard Interrupt handler vector 0x%02x\n",
				irq);
			free_cpc45_irq (irq);
		}
	}

	debugk ("Cleanup completed\n");
}

module_exit (cpc45_intr_cleanup);

#endif	/* MODULE */


/***************************************************************************
 * Exported API functions
 ***************************************************************************
 */

/*
 * Register a On-Board interrupt handler
 */
int request_cpc45_irq (unsigned int irq,
		       void (*handler)(int, void *, struct pt_regs *),
	               void *argument)
{
	unsigned int	 real_irq;

	debugk ("Install CPC45 handler for IRQ %d\n", irq);

	real_irq = irq;

	if (real_irq < NR_ONBOARD_INTERRUPTS) {	/* Onboard interrupt */
		onbd_intr_hdlr_t *ih = &onbd_intr_hdlr[real_irq];
		int siu_irq = ih->irq;

		if (ih->int_hand.handler) {
			printk ("Nested %s interrupts not supported (yet)"
				" [IRQ: OnBD=%d, IRQ%d]\n",
				ih->name, real_irq, siu_irq);
			return (-EBUSY);
		}

		ih->int_hand.handler = handler;
		ih->int_hand.arg = argument;

		debugk ("Enabling CPC45 %s interrupt handler\n", ih->name);

		/*
		 * Enable interrupt when the first handler gets installed
		 */
		if (intr_cnt[siu_irq] == 0) {
		    int rc;

			rc = request_irq(siu_irq, obd_interrupt_handler,
					 0, ih->name, (void *)ih);
			if (rc != 0)
				panic ("Can't allocate CPC45_INTR IRQ");
		}
		++intr_cnt[siu_irq];	/* increment reference count */

		/* enable interrupt */
		onboard_interrupt_en_dis(ih->enable, 1);

		return 0;
	}


	/* real_irq -= NR_ONBOARD_INTERRUPTS; */
	/* ...more types could be added here */


	debugk ("Invalid CPC45 IRQ %d (0...%u only)\n",
		irq,
		NR_ONBOARD_INTERRUPTS
	);
	return (-EINVAL);
}

/*
 * De-register and free a On-Board interrupt handler
 */
int free_cpc45_irq (unsigned int irq)
{
	unsigned int	 real_irq;

	debugk ("Un-install handler for CPC45 IRQ %d\n", irq);

	real_irq = irq;

	if (real_irq < NR_ONBOARD_INTERRUPTS) {	/* Onboard interrupt */
		onbd_intr_hdlr_t *ih = &onbd_intr_hdlr[real_irq];
		int siu_irq = ih->irq;

		if (ih->int_hand.handler == NULL) {
			printk ("Can't uninstall NULL %s interrupt"
				" [IRQ: OnBD=%d, IRQ%d]\n",
				ih->name, real_irq, siu_irq);
			return (-EINVAL);
		}

		if (intr_cnt[siu_irq] == 0) {
			printk ("Error: dead CPC45 %s interrupt"
				" [IRQ: OnBD=%d, IRQ%d]\n",
				ih->name, real_irq, siu_irq);
			return (-EINVAL);
		}

		/*
		 * disable onboard IRQ
		 */
		onboard_interrupt_en_dis(ih->enable, 0);

		ih->int_hand.handler = NULL;
		ih->int_hand.arg = 0L;

		debugk ("CPC45 %s interrupt handler uninstalled\n", ih->name);

		--intr_cnt[siu_irq];	/* decrement reference count */

		/*
		 * Disable interrupt upon removal of last handler
		 */

		if (intr_cnt[siu_irq] == 0) {
			free_irq (siu_irq, (void *)ih);

			debugk ("CPC45 IRQ%d uninstalled\n", siu_irq);
		}

		return 0;
	}

	/* real_irq -= NR_ONBOARD_INTERRUPTS; */
	/* ...more types could be added here */

	debugk ("Invalid CPC45 IRQ %d (0...%u only)\n",
		irq,
		NR_ONBOARD_INTERRUPTS
	);
	return (-EINVAL);
}


/***************************************************************************
 * Utility functions
 ***************************************************************************
 */

/*
 * The handler for onboard interrupts needed to pass the "logical" IRQ
 */
static void obd_interrupt_handler (int irq, void *dev_id, struct pt_regs *regs)
{
	onbd_intr_hdlr_t *ih = (onbd_intr_hdlr_t *)dev_id;
	unsigned int ivect;

	ivect = (ih - &onbd_intr_hdlr[0]);

	/* run handler */
	if (ih->int_hand.handler)
		ih->int_hand.handler (ivect, ih->int_hand.arg, regs);
	else
		printk ("Bogus Onboard interrupt vector 0x%02x\n", ivect);
}


/*
 * Enable or disable an onboard interrupts
 */
static int onboard_interrupt_en_dis (unsigned int mask, unsigned int enable)
{
	unsigned char lie1_mask = mask & 0xFF;
	unsigned char lie2_mask = mask >> 8;

	if (enable) {
		bcsr->lie1 |=  lie1_mask;
		bcsr->lie2 |=  lie2_mask;
	} else {
		bcsr->lie1 &= ~lie1_mask;
		bcsr->lie2 &= ~lie2_mask;
	}

	return (0);
}
