/***********************************************************************
 *
 * (C) Copyright 2001-2004
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * This module provides support for the VMEBus (and other external)
 * interrupts on CPU86 VMEBus systems.
 *
 ***********************************************************************/

/*
 * Standard in kernel modules
 */
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/rtc.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <asm/immap_8260.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/time.h>
#include <asm/machdep.h>		/* ppc_md */
#include <asm/mpc8260.h>

#include "cpu86_mem.h"
#include "cpu86_intr.h"

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */


/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME			"cpu86_intr"
#define	CPU86_INTR_VERSION	"0.1-wd"

/***************************************************************************
 * Debug support
 ***************************************************************************
 */
#undef	DEBUG

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif

/***************************************************************************
 * Global stuff - export global symbols
 ***************************************************************************
 */
EXPORT_SYMBOL(request_cpu86_irq);
EXPORT_SYMBOL(free_cpu86_irq);
EXPORT_SYMBOL(vme_interrupt_enable);
EXPORT_SYMBOL(vme_interrupt_disable);

/***************************************************************************
 * Local stuff
 ***************************************************************************
 */
typedef	struct intr_hdlr_s {
	void		(*handler)(int, void *, struct pt_regs *);
	void		*arg;
} intr_hdlr_t;

intr_hdlr_t	vme_intr_hdlr [NR_VME_INTERRUPTS];

typedef struct onbd_intr_hdlr_s {
	intr_hdlr_t	int_hand;
	char		irq;
	char		enable;
	char		*name;
} onbd_intr_hdlr_t;

/*
 * Several of the SIU interrupts are shared by the VME Bus and
 * OnBoard interrupts; for instance, CPU86_IRQ_VME_INTR (SIU_INT_IRQ3)
 * is responsible for up to 8 VME Bus interrupts, etc.
 *
 * The following array is used as reference counter so we know when
 * to install or de-install a SIU interrupt. Please note that only
 * some of the slots are used, the others remain uninitialized (0).
 */
static unsigned int intr_cnt[NR_SIU_INTS];

/*
 * Prototypes
 */
static void obd_interrupt_handler (int, void *, struct pt_regs *);
static void vme_interrupt_handler (int, void *, struct pt_regs *);
static int  vme_interrupt_en_dis  (unsigned int, unsigned int);

onbd_intr_hdlr_t onbd_intr_hdlr [] = {
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPU86_IRQ_ABORT,
		CPU86_ONBOARD_ABORT,
		"Abort Key",
	},
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPU86_IRQ_SYSMON,
		CPU86_ONBOARD_SYSMON,
		"System Monitor",
	},
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPU86_IRQ_PMC,
		CPU86_ONBOARD_PMC,
		"PMC IRQ[A-D]",
	},
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPU86_IRQ_RTC,
		CPU86_ONBOARD_RTC,
		"Real Time Clock",
	},
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPU86_IRQ_FDC,
		CPU86_ONBOARD_FDC,
		"FDC37C78",
	},
	{	{ NULL, NULL, },	/* handler and argument uninitialized	*/
		CPU86_IRQ_VMEF_MBOX,
		CPU86_ONBOARD_MAILBOX,
		"VMEBus Mailbox / Sysfail",
	},
};

#define NR_ONBOARD_INTERRUPTS	(sizeof(onbd_intr_hdlr)/sizeof(onbd_intr_hdlr_t))

static volatile cpu86_bcsr_t	*bcsr;
static volatile unsigned char	*vme_intack;

/***************************************************************************
 *
 * Initialize the module
 */
static int __init cpu86_intr_init (void)
{
	vme_mem_t *p;

	printk (KERN_INFO
		"CPU86 Extended Interrupt Support version "
		CPU86_INTR_VERSION
		"\n"
	);

	if ((p = vme_mem_dev(VME_MEM_BCSR)) == NULL)
		return (-ENODEV);

	bcsr  = (cpu86_bcsr_t *)(p->addr);

	if ((p = vme_mem_dev(VME_MEM_SHORT)) == NULL)
		return (-ENODEV);

	vme_intack = (unsigned char *)(p->addr + CPU86_VME_INTACK_OFFSET);

	debugk ("CPU86 Interrupt Support installed\n");

	return 0;
}

/***************************************************************************
 * Module Declarations
 ***************************************************************************
 */

module_init (cpu86_intr_init);

#ifdef MODULE
/*
 * Cleanup - unregister the appropriate file from /proc
 */
void cpu86_intr_cleanup (void)
{
	int irq;

	debugk ("Cleanup CPU86 Interrupt Support\n");

	/* make sure all interrupts are disabled */

	/* VME Interrupts first */
	for (irq=0; irq < NR_VME_INTERRUPTS; ++irq) {
		if (vme_intr_hdlr[irq].handler) {
			debugk ("Cleanup VME Interrupt handler vector 0x%02x\n",
				irq);
			free_cpu86_irq (irq);
		}
	}

	/* the Onboard Interrupts */
	for ( ; irq < NR_VME_INTERRUPTS+NR_ONBOARD_INTERRUPTS; ++irq) {
		int index = irq - NR_VME_INTERRUPTS;
		if (onbd_intr_hdlr[index].int_hand.handler) {
			debugk ("Cleanup Onboard Interrupt handler vector 0x%02x\n",
				irq);
			free_cpu86_irq (irq);
		}
	}

	debugk ("Cleanup completed\n");
}

module_exit (cpu86_intr_cleanup);

#endif	/* MODULE */


/***************************************************************************
 * Exported API functions
 ***************************************************************************
 */

/*
 * Register a VMEBus or On-Board interrupt handler
 */
int request_cpu86_irq (unsigned int irq,
		       void (*handler)(int, void *, struct pt_regs *),
	               void *argument)
{
	unsigned int	 real_irq;
	volatile immap_t *immap = immr;

	debugk ("Install CPU86 handler for IRQ %d\n", irq);

	real_irq = irq;

	if (real_irq < NR_VME_INTERRUPTS) {	/* VME Bus interrupt */
		intr_hdlr_t	*ih = &vme_intr_hdlr[real_irq];

		if (ih->handler) {
			printk ("Nested VMEBus interrupts not supported (yet)"
				" [IRQ: VME=%d SIU=%d]\n",
				real_irq, CPU86_IRQ_VME_INTR);
			return (-EBUSY);
		}

		ih->handler = handler;
		ih->arg = argument;

		debugk ("VMEBus IRQ %d installed\n", real_irq);

		/*
		 * Enable VME Bus interrupt when the first handler gets installed
		 */
		if (intr_cnt[CPU86_IRQ_VME_INTR] == 0) {

			debugk ("Enabling CPU86 VME interrupt handler\n");

			if (request_8xxirq( CPU86_IRQ_VME_INTR,
					    vme_interrupt_handler,
					    0, "CPU86 VMEBus",
					    vme_intr_hdlr
					  ) != 0) {
				panic ("Can't allocate CPU86 VME IRQ");
			}
		}
		++intr_cnt[CPU86_IRQ_VME_INTR];	/* increment reference count */

		return (0);
	}

	real_irq -= NR_VME_INTERRUPTS;

	if (real_irq < NR_ONBOARD_INTERRUPTS) {	/* Onboard interrupt */
		onbd_intr_hdlr_t *ih = &onbd_intr_hdlr[real_irq];
		int siu_irq = ih->irq;

		if (ih->int_hand.handler) {
			printk ("Nested %s interrupts not supported (yet)"
				" [IRQ: OnBD=%d, SIU=%d]\n",
				ih->name, real_irq, siu_irq);
			return (-EBUSY);
		}

		ih->int_hand.handler = handler;
		ih->int_hand.arg = argument;

		debugk ("Enabling CPU86 %s interrupt handler\n", ih->name);

		/*
		 * Enable interrupt when the first handler gets installed
		 *
		 * For the onboard interrupts this is not really required
		 * since these don't share the IRQ, but it's easier
		 * to use identical code.
		 */
		if (intr_cnt[siu_irq] == 0) {
		    int rc;

			rc = request_8xxirq(siu_irq, obd_interrupt_handler,
					    0, ih->name, (void *)ih);
			if (rc != 0)
				panic ("Can't allocate CPU86_INTR IRQ");
				
			immap->im_intctl.ic_siexr |= (0x4000 >> (siu_irq - SIU_INT_IRQ1));
		}
		++intr_cnt[siu_irq];	/* increment reference count */

		/* enable interrupt */
		bcsr->onbd_imask |= ih->enable;

		return 0;
	}


	/* real_irq -= NR_ONBOARD_INTERRUPTS; */
	/* ...more types could be added here */


	debugk ("Invalid CPU86 IRQ %d (0...%u only)\n",
		irq,
		NR_VME_INTERRUPTS + NR_ONBOARD_INTERRUPTS
	);
	return (-EINVAL);
}


/*
 * De-register and free a VMEBus or On-Board interrupt handler
 */
int free_cpu86_irq (unsigned int irq)
{
	unsigned int	 real_irq;
	volatile immap_t *immap = immr;

	debugk ("Un-install handler for CPU86 IRQ %d\n", irq);

	real_irq = irq;

	if (real_irq < NR_VME_INTERRUPTS) {	/* VME Bus interrupt */
		intr_hdlr_t	*ih = &vme_intr_hdlr[real_irq];

		if (ih->handler == NULL) {
			printk ("Can't uninstall NULL VMEBus interrupt %d\n",
				real_irq);
			return (-EINVAL);
		}

		if (intr_cnt[CPU86_IRQ_VME_INTR] == 0) {
			printk ("Error: dead CPU86 VMEBus interrupt %d\n",
				real_irq);
			return (0);
		}

		--intr_cnt[CPU86_IRQ_VME_INTR];	/* decrement reference count */

		/*
		 * Disable VME Bus interrupt upon removal of last handler
		 */
		if (intr_cnt[CPU86_IRQ_VME_INTR] == 0) {

			debugk ("Disabling CPU86 VME interrupt handler\n");

			free_irq (CPU86_IRQ_VME_INTR, vme_intr_hdlr);
		}

		ih->handler = NULL;
		ih->arg = 0L;

		debugk ("VMEBus IRQ %d uninstalled\n", real_irq);

		return (0);
	}

	real_irq -= NR_VME_INTERRUPTS;

	if (real_irq < NR_ONBOARD_INTERRUPTS) {	/* Onboard interrupt */
		onbd_intr_hdlr_t *ih = &onbd_intr_hdlr[real_irq];
		int siu_irq = ih->irq;

		if (ih->int_hand.handler == NULL) {
			printk ("Can't uninstall NULL %s interrupt"
				" [IRQ: OnBD=%d, SIU=%d]\n",
				ih->name, real_irq, siu_irq);
			return (-EINVAL);
		}

		if (intr_cnt[siu_irq] == 0) {
			printk ("Error: dead CPU86 %s interrupt"
				" [IRQ: OnBD=%d, SIU=%d]\n",
				ih->name, real_irq, siu_irq);
			return (-EINVAL);
		}

		/*
		 * disable onboard IRQ
		 */
		bcsr->onbd_imask &= ~(ih->enable);

		ih->int_hand.handler = NULL;
		ih->int_hand.arg = 0L;

		debugk ("CPU86 %s interrupt handler uninstalled\n", ih->name);

		--intr_cnt[siu_irq];	/* decrement reference count */

		/*
		 * Disable interrupt upon removal of last handler
		 */
				
		if (intr_cnt[siu_irq] == 0) {
			immap->im_intctl.ic_siexr &= ~(0x4000 >> (siu_irq - SIU_INT_IRQ1));
			free_irq (siu_irq, (void *)ih);

			debugk ("CPU86 SIU IRQ %d uninstalled\n", siu_irq);
		}

		return 0;
	}


	/* real_irq -= NR_ONBOARD_INTERRUPTS; */
	/* ...more types could be added here */


	debugk ("Invalid CPU86 IRQ %d (0...%u only)\n",
		irq,
		NR_VME_INTERRUPTS + NR_ONBOARD_INTERRUPTS
	);
	return (-EINVAL);
}


int vme_interrupt_enable  (unsigned int level)
{
	return vme_interrupt_en_dis (level, 1);
}


int vme_interrupt_disable (unsigned int level)
{
	return vme_interrupt_en_dis (level, 0);
}

/***************************************************************************
 * Utility functions
 ***************************************************************************
 */

/*
 * The handler for onboard interrupts needed to pass the
 * "logical" IRQ instead of the SIU IRQ
 */
static void obd_interrupt_handler (int irq, void *dev_id, struct pt_regs *regs)
{
	onbd_intr_hdlr_t *ih = (onbd_intr_hdlr_t *)dev_id;
	unsigned int ivect;

	ivect = (ih - &onbd_intr_hdlr[0]) + NR_VME_INTERRUPTS;

	/* run handler */
	if (ih->int_hand.handler)
		ih->int_hand.handler (ivect, ih->int_hand.arg, regs);
	else
		printk ("Bogus Onboard interrupt vector 0x%02x\n", ivect);
}


/*
 * The VME Bus SIU interrupt handler
 */
static void vme_interrupt_handler (int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned char	imask, ipend, ivect, source_id;

	debugk ("## VMEBus interrupt %d called\n", irq);

	/*
	 * Check for pending VMEBus interrupts
	 * Ignore SysFail - SF interrupt causes an onboard interrupt
	 */
	ipend = bcsr->vme_ipend & 0xFE;

	if (ipend == 0) {
		debugk ("Spurious VMEBus interrupt\n");
		return;
	}

	/*
	 * Checking IRQ; start at VME-IRQ7 which has highest priority
	 */
	for (ivect = 7; ivect > 0; ivect--) {
		intr_hdlr_t	*ih;

		imask = 1 << ivect;

		if ((ipend & imask) == 0)	/* interrupt pending?	*/
			continue;

		/*
		 * VME Interrupt handling requires:
		 * mask interrupt, read vector, [run handler,] unmask interrupt
		 */
		bcsr->vme_imask &= ~imask;

		source_id = vme_intack[(ivect << 1) + 1];

		ih = &vme_intr_hdlr[source_id];

		/* run handler */
		if (ih->handler)
			ih->handler (source_id, ih->arg, regs);
		else
			printk ("Bogus VME interrupt vector 0x%02x\n", ivect);

		/* re-enable interrupt */
		bcsr->vme_imask |= imask;

		/* re-read the pending status */
		ipend = bcsr->vme_ipend & 0xFE;
		/* and start again with the highest priority level */
		ivect = 8;
	}
}


/*
 * Enable or disable a VME Bus interrupt level
 */
static int vme_interrupt_en_dis (unsigned int level, unsigned int enable)
{
	unsigned char mask;

	/* allow for VME interrupt levels 1 ... 7 */
	if ((level < 1) || (level > 7)) {
		return (-1);
	}

	mask = 1 << level;

	if (enable) {
		bcsr->vme_imask |=  mask;
	} else {
		bcsr->vme_imask &= ~mask;
	}

	return (0);
}
