/*
 * INKA4x0 Real Time Clock Driver
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/delay.h>

#include <asm/time.h>

#include <asm/mpc5xxx.h>

#define RTC_CE    0x01000000
#define RTC_WR    0x02000000
#define RTC_DATA  0x01
#define RTC_CLK   0x02

static int rtc_busy = 0;
static void inka_rtc_to_tm(unsigned long long rtc, struct rtc_time *t);
static void inka_rtc_mktime(int year, int month, int mday, int wday, int hour, int min, int sec, unsigned long long *time);
static void inka_rtc_read(unsigned long long *data);
static void inka_rtc_write(unsigned long long *data);

static int bit_weight[] = { 1, 2, 4, 8, 10, 20, 40, 80 };

static u8 bcd2number(char c, int bits)
{
	int i;
	u8 res = 0;

	for (i = 0; i < bits; i++) {
		if (c & (1 << i)) {
			res += bit_weight[i];
		}
	}
	return res;
}

static u8 number2bcd(char c)
{
	int i;
	char res = 0;

	for (i = 7; i >= 0; i--) {
		if (c >= bit_weight[i]) {
			res |= 1 << i;
			c -= bit_weight[i];
		}
	}

	return res;
}

static void inka_rtc_to_tm(unsigned long long rtc, struct rtc_time *t)
{

	memset(t, 0, sizeof(struct rtc_time));

	t->tm_sec  = bcd2number( rtc        & 0x7f, 7);
	t->tm_min  = bcd2number((rtc >>  8) & 0x7f, 7);
	t->tm_hour = bcd2number((rtc >> 16) & 0x3f, 6);
	t->tm_mday = bcd2number((rtc >> 28) & 0x3f, 6);
	t->tm_mon  = bcd2number((rtc >> 36) & 0x1f, 5);
	t->tm_year = bcd2number((rtc >> 44) & 0xff, 8);
	t->tm_wday = bcd2number((rtc >> 24) & 0x0f, 3);
}

static void inka_rtc_mktime(int year, int month, int mday, int wday,
		            int hour, int min, int sec, unsigned long long *time)
{
	*time = ((unsigned long long)number2bcd(year)  << 44) |
		((unsigned long long)number2bcd(month) << 36) |
		((unsigned long long)number2bcd(mday)  << 28) |
		(number2bcd(wday) << 24) |
		(number2bcd(hour) << 16) |
		(number2bcd(min)  <<  8) |
		number2bcd(sec);
}

static void inka_rtc_read(unsigned long long *data)
{
	struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	int i;

	*data = 0;

	/* Lower WR */
	*(volatile u_long *)MPC5xxx_WU_GPIO_DATAOUT &= ~RTC_WR;
	udelay(1);

	/* Rise CE */
	*(volatile u_long *)MPC5xxx_WU_GPIO_DATAOUT |= RTC_CE;
	udelay(1);

	/* PSC3_4 is input */
	gpio->sint_ddr &= ~RTC_DATA;

	for(i = 0; i < 52; i++) {
		gpio->sint_dvo |= RTC_CLK;
		udelay(10);
		*data |= ((unsigned long long)(gpio->sint_ival & RTC_DATA) << i);
		gpio->sint_dvo &= ~RTC_CLK;
		udelay(10);
	}

	/* Lower CE */
	*(volatile u_long *)MPC5xxx_WU_GPIO_DATAOUT &= ~RTC_CE;
}

static void inka_rtc_write(unsigned long long *data)
{
	struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	int i;

	/* PSC3_4 is output */
	gpio->sint_ddr |= RTC_DATA;

	/* Rise WR */
	*(volatile u_long *)MPC5xxx_WU_GPIO_DATAOUT |= RTC_WR;
	udelay(1);

	/* Rise CE */
	*(volatile u_long *)MPC5xxx_WU_GPIO_DATAOUT |= RTC_CE;
	udelay(1);

	for(i = 0; i < 52; i++) {
		gpio->sint_dvo &= ~RTC_DATA;
		if (*data & ((unsigned long long)1 << i))
			gpio->sint_dvo |= RTC_DATA;
		udelay(10);
		gpio->sint_dvo |= RTC_CLK;
		udelay(10);
		gpio->sint_dvo &= ~RTC_CLK;
		udelay(10);
	}

	/* Lower CE */
        *(volatile u_long *)MPC5xxx_WU_GPIO_DATAOUT &= ~RTC_CE;
}


/* Retrieve the current date and time from the real time clock. */
static void get_rtc_time(struct rtc_time *t)
{
	unsigned long long nowtime = 0;
	unsigned long flags;

	save_flags(flags);
	cli();

	inka_rtc_read(&nowtime);

	restore_flags(flags);

	inka_rtc_to_tm(nowtime, t);

	t->tm_mon -= 1;
	t->tm_wday -= 1;
}

/* Set the current date and time in the real time clock. */
static void set_rtc_time(struct rtc_time *t)
{
	unsigned long long nowtime;
	unsigned long flags;

	inka_rtc_mktime(t->tm_year, t->tm_mon + 1, t->tm_mday, t->tm_wday + 1,
			t->tm_hour, t->tm_min, t->tm_sec, &nowtime);

	save_flags(flags);
	cli();

	inka_rtc_write(&nowtime);
        inka_rtc_read (&nowtime);

	restore_flags(flags);
}

static loff_t rtc_lseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

static int rtc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg)
{
	struct rtc_time rtc_tm;

	switch (cmd) {
	case RTC_RD_TIME:
		get_rtc_time(&rtc_tm);

		if (copy_to_user((struct rtc_time*)arg, &rtc_tm, sizeof(struct rtc_time)))
			return -EFAULT;

		return 0;

	case RTC_SET_TIME:
		if (!capable(CAP_SYS_TIME))
			return -EPERM;

		if (copy_from_user(&rtc_tm, (struct rtc_time*)arg, sizeof(struct rtc_time)))
			return -EFAULT;

		set_rtc_time(&rtc_tm);

		return 0;

	default:
		return -EINVAL;
	}
}

static int rtc_open(struct inode *inode, struct file *file)
{
	if (rtc_busy)
		return -EBUSY;

	rtc_busy = 1;

	MOD_INC_USE_COUNT;

	return 0;
}

static int rtc_release(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	rtc_busy = 0;
	return 0;
}

static struct file_operations rtc_fops = {
	owner:		THIS_MODULE,
	llseek:		rtc_lseek,
	ioctl:		rtc_ioctl,
	open:		rtc_open,
	release:	rtc_release
};

static struct miscdevice rtc_dev = { RTC_MINOR, "rtc", &rtc_fops };

EXPORT_NO_SYMBOLS;

static int __init rtc_init(void)
{
	int error;
	struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *)MPC5xxx_GPIO;

	error = misc_register(&rtc_dev);
	if (error) {
		printk(KERN_ERR "rtc: unable to get misc minor %i\n", error);
		return error;
	}

	/* Configure Inputs and Outputs */
	/* PSC1_4 (CE) and PSC2_4 (MOSI) are output */
	*(volatile u_long *)MPC5xxx_WU_GPIO_ENABLE |= RTC_CE | RTC_WR;
	*(volatile u_long *)MPC5xxx_WU_GPIO_DIR |= RTC_CE | RTC_WR;

	/* PSC3_4 is input/output, PSC3_5 (CLK) is output */
	gpio->sint_gpioe |= RTC_DATA | RTC_CLK;
	gpio->sint_ddr |= RTC_CLK;
	gpio->sint_inten &= ~(RTC_DATA | RTC_CLK); /* no interrupts */

	/* Set '0' on CE and CLK */
	*(volatile u_long *)MPC5xxx_WU_GPIO_DATAOUT &= ~RTC_CE;
	gpio->sint_dvo &= ~RTC_CLK;

	return 0;
}

static void __exit rtc_exit(void)
{
	misc_deregister(&rtc_dev);
}

module_init(rtc_init);
module_exit(rtc_exit);
