/*
 * (C) Copyright 2004
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * Linux support for the M41ST85-series TIMEKEEPER Real-Time Clock.
 *
 * Based on code from the lm-sensors project which is available
 * at http://www.lm-sensors.nu/ and Russell King's PCF8583 Real-Time
 * Clock driver (linux/drivers/acorn/char/pcf8583.c).
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 */
#include <linux/delay.h>
#include <linux/miscdevice.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/smp_lock.h>
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <asm/types.h>
#include <asm/rtc.h>
#include <asm/semaphore.h>
#include <asm/machdep.h>
#include <asm/uaccess.h>

#define M41_BCD_TO_BIN(x) (((x) & 15) + ((x) >> 4) * 10)
#define M41_BIN_TO_BCD(x) ((((x) / 10) << 4) + (x) % 10)

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif


/* practically the same as struct rtc_time, but without tm_yday and tm_isdst */
struct m41st85_date {
	int tm_sec;
	int tm_min;
	int tm_hour;
	int tm_mday;
	int tm_mon;
	int tm_year;
	int tm_wday;
};

struct m41st85_memory {
	u8 offset;		/* 0 - 42 */
	u8 buf[43];		/* data */
	u8 length;		
};

/* size of the rtc non-volatile ram */
#define M41ST85_SIZE     43
#define M41ST85_OFFSET_TO_RAM 20
#define M41ST85_DATE_SIZE 8
#define M41ST85_SLAVE_ADDR (0xd0)


#define M41ST85_GET_DATE	_IOR ('d',0,struct m41st85_date *)
#define M41ST85_SET_DATE	_IOW ('d',1,struct m41st85_date *)
#define M41ST85_READ		_IOR ('d',6,struct m41st85_memory *)
#define M41ST85_WRITE	_IOW ('d',7,struct m41st85_memory *)
#define M41ST85_ENABLE	_IO  ('d',8)


#define  I2C_DRIVERID_M41ST85 0xf111

#define DEVICE_NAME "rtc"



extern struct machdep_calls ppc_md;




/*Pointers to get/set time via m41st85 chip to substitute undefined (yet)
 fields in struct ppc_md. For pp01 desk they are not defined, but if they
 would, we'll save current pointers concerning RTC from ppc_md and will
 restore them in cleanup_module
 */
static int             (*old_set_rtc_time)(unsigned long nowtime);
static unsigned long   (*old_get_rtc_time)(void);




static DECLARE_MUTEX (mutex);


/*prototypes*/

/*m41st85 rtc device related staff*/
static int m41st85_i2c_recv (struct i2c_client *client,char *buf,char addr,int count);
static int m41st85_i2c_send (struct i2c_client *client,const char *buf,int count);
static int m41st85_attach (struct i2c_adapter *adapter,int addr,unsigned short flags,int kind);
static int m41st85_attach_adapter (struct i2c_adapter *adapter);
static int m41st85_detach_client (struct i2c_client *client);
static int m41st85_getdate (struct i2c_client *client,void *arg);
static int m41st85_setdate (struct i2c_client *client,void *arg);
static int m41st85_enable (struct i2c_client *client,void *arg);
static int m41st85_read (struct i2c_client *client,void *arg);
static int m41st85_write (struct i2c_client *client,void *arg);


static void m41st85_inc_use (struct i2c_client *client);
static void m41st85_dec_use (struct i2c_client *client);

/*RTC related staff*/
static unsigned long m41st85_get_rtc_time(void);
static int m41st85_set_rtc_time(unsigned long);


/*I2c specific declarations*/


static struct i2c_driver m41st85;

static  unsigned short ignore[] = { I2C_CLIENT_END };
static  unsigned short addr[] = { 0x68, I2C_CLIENT_END };

static struct i2c_client_address_data m41st85_addr_data =
{
normal_i2c:			addr,
	normal_i2c_range:               ignore,
	probe:				ignore,
	probe_range:	                ignore,
	ignore:				ignore,
	ignore_range:	                ignore,
	force:				ignore
};

static struct i2c_client* m41st85_client;



/*==================================================================
 I2C staff
 ==================================================================*/

/*
 * The M41ST85 Real-Time Clock wants the address in a different
 * message, so we can't use the normal i2c_master_recv() routine
 * for receiving data.
 */
static int m41st85_i2c_recv (struct i2c_client *client,char *buf,char addr,int count)
{
	int result = 0;
	int res=0;

	struct i2c_msg msg[2] = {
		{ client->addr,  0, 1, &addr },
		{ client->addr,  I2C_M_RD, count, buf }
	};
	debugk("Communication with %x i2c ",client->addr);
	if (down_interruptible (&mutex))
		return (-ERESTARTSYS);

	if ((res=i2c_transfer (client->adapter,msg,2)) != 2)
	{
		result = -EIO;
		debugk("\ni2c_transfer=%d",res);
	}
	up (&mutex);

	return (result);
}

/*
 * Would've been nice to specify the address to this as well, but then we
 * would need to copy the buffer twice - not worth it...
 */
static int m41st85_i2c_send (struct i2c_client *client,const char *buf,int count)
{
	int result = 0;

	if (down_interruptible (&mutex))
		return (-ERESTARTSYS);

	if (i2c_master_send (client,(const char *) buf,count) != count)
		result = -EIO;

	up (&mutex);

	return (result);
}

static int m41st85_attach (struct i2c_adapter *adapter,int addr,unsigned short flags,int kind)
{
	/*  struct i2c_client *client;*/
	int result;
	debugk("Trying to create client...");
	if ((m41st85_client = (struct i2c_client *) kmalloc (sizeof (struct i2c_client),GFP_KERNEL)) == NULL)
		return (-ENOMEM);

	strcpy (m41st85_client->name,m41st85.name);
	m41st85_client->id = m41st85.id;
	m41st85_client->flags = I2C_CLIENT_ALLOW_USE | I2C_CLIENT_ALLOW_MULTIPLE_USE;
	m41st85_client->addr = addr;
	m41st85_client->adapter = adapter;
	m41st85_client->driver = &m41st85;
	m41st85_client->data = NULL;

	if ((result = i2c_attach_client (m41st85_client))) {
		kfree (m41st85_client);
		return (result);
	}

	return (0);
}

static int m41st85_attach_adapter (struct i2c_adapter *adapter)
{
	int rt;
	debugk("Trying to attach adapter...");
	rt=i2c_probe (adapter,&m41st85_addr_data,m41st85_attach);
	debugk("Probing result %d", rt);
/*	rt1=m41st85_attach(adapter,0x68,0,0);
	if(rt1)
	{
	    debugk("\n%s: Cannot attach slave!",__FUNCTION__);
	    return rt1;
	}*/
/*	debugk("\n%s: Trying to start occilator...",__FUNCTION__);
	if(m41st85_enable(m41st85_client,NULL))
	{
	    debugk("failed");
	    return -EIO;
	}
	debugk("OK");*/
	return (rt);
	
}

static int m41st85_detach_client (struct i2c_client *client)
{
	int result;

	if ((result = i2c_detach_client (client)))
		return (result);

	kfree (client);

	return (0);
}

/*
 Get time to device in RTC - suitable format
 */

static int m41st85_getdate (struct i2c_client *client,void *arg)
{
	struct m41st85_date *date = (struct m41st85_date *) arg;
	u8 buf[M41ST85_DATE_SIZE];

	/* this also enables the oscillator */
	memset (buf,0,M41ST85_DATE_SIZE);
	if(!client)
	{
	    debugk("%s: no client!",__FUNCTION__);
	    return -EIO;
	}

	if (m41st85_i2c_recv (client,(char *) buf,0, M41ST85_DATE_SIZE) < 0)
		return (-EIO);
	/*The very first is *0.01 seconds - no use */
	date->tm_sec = M41_BCD_TO_BIN (buf[1] & ~0x80);
	date->tm_min = M41_BCD_TO_BIN (buf[2]);
	date->tm_hour = M41_BCD_TO_BIN (buf[3] & 0x3f);
	date->tm_wday = M41_BCD_TO_BIN (buf[4]);
	date->tm_mday = M41_BCD_TO_BIN (buf[5]);
	date->tm_mon = M41_BCD_TO_BIN (buf[6]);
	date->tm_year = M41_BCD_TO_BIN (buf[7]) + 1970;
	return (0);
}
/*
 Set time to device in RTC - suitable format
 */
static int m41st85_setdate (struct i2c_client *client,void *arg)
{
	struct m41st85_date *date = (struct m41st85_date *) arg;
	u8 buf[M41ST85_DATE_SIZE+1];/*additional place for register addr*/

	/* select address 0 */
	buf[0] = 0;

	buf[2] = M41_BIN_TO_BCD (date->tm_sec);
	buf[3] = M41_BIN_TO_BCD (date->tm_min);
	buf[4] = M41_BIN_TO_BCD (date->tm_hour) | 0x40;
	buf[5] = M41_BIN_TO_BCD (date->tm_wday);
	buf[6] = M41_BIN_TO_BCD (date->tm_mday);
	buf[7] = M41_BIN_TO_BCD (date->tm_mon);
	buf[8] = M41_BIN_TO_BCD (date->tm_year - 1970);

	return (m41st85_i2c_send (client,(const char *) buf,M41ST85_DATE_SIZE+1));
}
/*
 Toggle internal occilator On/Off
 */
static int m41st85_enable (struct i2c_client *client,void *arg)
{
	static u8 buf[3];
	if(!client)
	    {
		debugk("%s : no client!",__FUNCTION__);
		return -EIO;
	    
	    }
	return (m41st85_i2c_send (client,(const char *) buf,3));
	
}
/*
 Read user memory of m41st85
 */
static int m41st85_read (struct i2c_client *client,void *arg)
{
	struct m41st85_memory *mem = (struct m41st85_memory *) arg;
	
	u8 buf[M41ST85_SIZE];

	if (mem->offset >= M41ST85_SIZE || mem->offset + mem->length > M41ST85_SIZE)
		return (-EINVAL);

	if (m41st85_i2c_recv (client,(char *) buf,mem->offset + M41ST85_OFFSET_TO_RAM,mem->length) < 0)
		return (-EIO);

	memcpy (mem->buf,buf,mem->length);

	return (0);
}
/*
 Write to user memory of m41st85
 */
static int m41st85_write (struct i2c_client *client,void *arg)
{
	struct m41st85_memory *mem = (struct m41st85_memory *) arg;
	u8 buf[M41ST85_SIZE + 1];

	if (mem->offset >= M41ST85_SIZE || mem->offset + mem->length > M41ST85_SIZE)
		return (-EINVAL);

	buf[0] = mem->offset +  M41ST85_OFFSET_TO_RAM;

	memcpy (buf + 1,mem->buf,mem->length);

	return (m41st85_i2c_send (client,(const char *) buf,mem->length + 1));
}

/*
 Remain existing unimplemented features of the m41st85 device
 such as Alarm registers and WatchDog timer
*/
static int m41st85_command (struct i2c_client *client,unsigned int cmd,void *arg)
{
	static const struct {
		int cmd;
		int (*function) (struct i2c_client *,void *arg);
	} ioctl[] = {
		{ M41ST85_ENABLE, m41st85_enable },
		{ M41ST85_GET_DATE, m41st85_getdate },
		{ M41ST85_SET_DATE, m41st85_setdate },
		{ M41ST85_READ, m41st85_read },
		{ M41ST85_WRITE, m41st85_write }
	};
	int i;

	for (i = 0; i < sizeof (ioctl) / sizeof (ioctl[0]); i++)
		if (ioctl[i].cmd == cmd)
		{	
			return (ioctl[i].function (client,arg));
		}
	debugk("Unrecognized command!");
	return (-EINVAL);
}

static void m41st85_inc_use (struct i2c_client *client)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif	/* #ifdef MODULE */
}

static void m41st85_dec_use (struct i2c_client *client)
{
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif	/* #ifdef MODULE */
}

static struct i2c_driver m41st85 = {
        name:                   "m41st85",
        id:                             I2C_DRIVERID_M41ST85,
        flags:                  I2C_DF_NOTIFY,
        attach_adapter: m41st85_attach_adapter,
        detach_client:  m41st85_detach_client,
        command:                m41st85_command,
        inc_use:                m41st85_inc_use,
        dec_use:                m41st85_dec_use
};

/*==========================================================================
                        HW staff
 =========================================================================*/

/*
 * Read the current RTC time and date
 */
static unsigned long m41st85_get_rtc_time(void)
{
	int ret;
	struct m41st85_date date;
	if(!m41st85_client)
	{
		debugk("No client!!");
		return -EIO;
	}
	ret=m41st85_command(m41st85_client,M41ST85_GET_DATE,&date);
	if(ret)
	{
            debugk("%s:Command returned %d",__FUNCTION__,ret);
		return ret;/*Something wrong */
	}
	debugk("\nCurent time %d:%d:%d %d %d %d\n",date.tm_hour,date.tm_min,date.tm_sec,date.tm_mday,date.tm_mon,date.tm_year);
	return mktime(date.tm_year,date.tm_mon,date.tm_mday,
		      date.tm_hour,date.tm_min,date.tm_sec);
}
/*
 * Set the current RTC time and date
 */
static int m41st85_set_rtc_time(unsigned long now)
{
	struct rtc_time tm;
        struct m41st85_date dt;
	to_tm(now,&tm);
        /*struct rtc_time have additional 2 fields -ignore them*/
	memcpy(&dt,&tm,sizeof(dt));
        return m41st85_command(m41st85_client,M41ST85_SET_DATE,&dt);
}


static int __init m41st85_init (void)
{
    int result;

    if ((result = i2c_add_driver (&m41st85)))
        return (result);
    printk ("M41ST85Y/W serial TIMEKEEPER Real-Time Clock driver (V0.01)\n");
    
    old_get_rtc_time=ppc_md.get_rtc_time;
    old_set_rtc_time=ppc_md.set_rtc_time;

    ppc_md.get_rtc_time=m41st85_get_rtc_time;
    ppc_md.set_rtc_time=m41st85_set_rtc_time;

    return (0);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
#undef __exit
#define __exit
#endif	/* #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)) */

static void __exit m41st85_exit (void)
{
	i2c_del_driver (&m41st85);
	/*restore old handlers */
	ppc_md.get_rtc_time=old_get_rtc_time;
	ppc_md.set_rtc_time=old_set_rtc_time;
	if(m41st85_client)
                kfree(m41st85_client);
	printk ("M41ST85 Real-Time Clock Driver de-registered\n");

}

module_init (m41st85_init);
module_exit (m41st85_exit);

MODULE_LICENSE ("GPL");
