/*
 * nc650_cpld.c - Driver for CPLD functionality on NC650
 *
 * Copyright (C) 2005 Detlev Zundel <dzu@denx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This driver publishes an interface to the CPLD on the CP650
 * hardware.  As most of these operations do not cleanly map into
 * the read/write abstraction with major/minor number device files
 * this driver hooks into the kernel as a misc device with only one
 * minor number and offers most services to userspace through
 * ioctl()s.
 *
 * Moreover this is the kernel abstraction for other drivers of this
 * hardware needing access to the CPLD functions so we export some
 * functions for other drivers (notably status_led and hdlc_enet).
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <asm/8xx_immap.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include <platforms/nc650.h>
#include <asm/nc650_cpld.h>


#define NC650_CPLD_MINOR 240	/* From the "Reserved for local use" range */

/*
 * PROCFS directory and file names
 */
#define NC650_CPLD_PROC_DIR "driver/nc650_cpld"
#define NC650_CPLD_PROC_SLOT_ADDR "driver/nc650_cpld/slot_addr"
#define NC650_CPLD_PROC_MODES "driver/nc650_cpld/modes"

#define DONTCARE 0xff                   /* Actual data is irrelevant for write
                                           access - the address is matched */

volatile static cpld_t *cpld = NULL;	/* Pointer to CPLD */
volatile static peri_t *peri = NULL; 	/* Pointer to Peri */
static int dev_open = 0;
static int errled_state;	/* Status of errled has to be buffered */
static int statusled_state;	/* The same goes for the other leds    */
static int slot_addr = -1;

static char *mode_string[]={
	[NC650_CPLD_MODE_INVALID] = "invalid",
	[NC650_CPLD_MODE_RS232]   = "rs232",
	[NC650_CPLD_MODE_RS485]   = "rs485",
	[NC650_CPLD_MODE_OC]      = "open collector"
};

static int
cpld_open(struct inode *inode, struct file *file)
{
	/* Implement single open policy */
	if (dev_open)
		return -EBUSY;

	dev_open = 1;

	return 0;
}

static int
cpld_release(struct inode *inode, struct file *file)
{
	dev_open = 0;

	return 0;
}

static ssize_t
cpld_read (struct file *filp, char *buffer, size_t length, loff_t *offset)
{
	unsigned char val;

	/* Can't seek on this device */
	if (offset != &filp->f_pos)
		return -ESPIPE;

	if (!length)
		return 0;

	/* Read status register */
	val = cpld->status_reg_read;
	if (copy_to_user((void *) buffer, &val, 1))
		return -EFAULT;

	return 1;
}

static int
cpld_ioctl(struct inode *inode, struct file *file,
	   unsigned int cmd, unsigned long arg)
{
        nc650_cpld_param_t param;
	unsigned mode;

	switch (cmd) {
	case NC650_CPLD_SLOTADDR_GET:
		if (copy_to_user((void *) arg, &slot_addr, sizeof(slot_addr)))
			return -EFAULT;
		break;

	case NC650_CPLD_COMM_MODE_GET:
		if (copy_from_user(&param, (void *)arg, sizeof (param)))
			return (-EFAULT);
		switch (param.port) {
		case 0:
			param.mode = (cpld->status_reg_read & CPLD_SR_SMC1_RS485)?
				NC650_CPLD_MODE_RS485 : NC650_CPLD_MODE_RS232;
			break;
		case 1:
			param.mode = (cpld->status_reg_read & CPLD_SR_SCC3_RS485)?
				NC650_CPLD_MODE_RS485 : NC650_CPLD_MODE_RS232;
			break;
		default:
			return -EINVAL;
		}
		if (copy_to_user((void *) arg, &param, sizeof(param)))
			return -EFAULT;
		break;

	case NC650_CPLD_COMM_MODE_SET:
		if (copy_from_user(&param, (void *)arg, sizeof(param)))
			return (-EFAULT);
		switch (param.port) {
		case 0:		/* ttyS0 */
			switch (param.mode) {
			case NC650_CPLD_MODE_RS485:
				cpld->smc1_rs485_mode = DONTCARE;
				break;
			case NC650_CPLD_MODE_RS232:
				cpld->smc1_rs232_mode = DONTCARE;
				break;
			default:
				return -EINVAL;
			}
		case 1:		/* ttyS1 */
			switch (param.mode) {
			case NC650_CPLD_MODE_RS485:
				cpld->scc3_rs485_mode = DONTCARE;
				break;
			case NC650_CPLD_MODE_RS232:
				cpld->scc3_rs232_mode = DONTCARE;
				break;
			default:
				return -EINVAL;
			}
		default:
			return -EINVAL;
		}
		if (copy_to_user((void *) arg, &param, sizeof(param)))
			return -EFAULT;
		break;

	case NC650_CPLD_HDLC_MODE_GET:
		mode = (cpld->status_reg_read & CPLD_SR_SCC4_OC)?
				NC650_CPLD_MODE_OC : NC650_CPLD_MODE_RS485;
		if (copy_to_user((void *) arg, &mode, sizeof(mode)))
			return -EFAULT;
		break;

	case NC650_CPLD_HDLC_MODE_SET:
		if (copy_from_user(&mode, (void *)arg, sizeof(mode)))
			return (-EFAULT);
		switch (param.mode) {
		case NC650_CPLD_MODE_RS485:
			cpld->scc4_rs485_mode = DONTCARE;
			break;
		case NC650_CPLD_MODE_OC:
			cpld->scc4_open_collector_mode = DONTCARE;
			break;
		default:
			return -EINVAL;
		}
		break;

	default:
		return -ENOTTY;
	}

	return 0;
}

#ifdef CONFIG_PROC_FS
static int cpld_slot_addr_proc_read(char *page, char **start, off_t off,
				    int count, int *eof, void *data)
{
	int    len = 0;

	len += sprintf(page, "%d\n", slot_addr);

	return len;
}

static int cpld_modes_proc_read(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int    len = 0;
	int    mode;


	mode = (cpld->status_reg_read & CPLD_SR_SMC1_RS485)?
		NC650_CPLD_MODE_RS485 : NC650_CPLD_MODE_RS232;
	len += sprintf(page+len, "ttyS0    : %s\n", mode_string[mode]);

	mode = (cpld->status_reg_read & CPLD_SR_SCC3_RS485)?
		NC650_CPLD_MODE_RS485 : NC650_CPLD_MODE_RS232;
	len += sprintf(page+len, "ttyS1    : %s\n", mode_string[mode]);

	mode = (cpld->status_reg_read & CPLD_SR_SCC4_OC)?
		NC650_CPLD_MODE_OC : NC650_CPLD_MODE_RS485;
	len += sprintf(page+len, "hdlc_enet: %s\n", mode_string[mode]);

	return len;
}
#endif

static struct file_operations cpld_fops = {
	.owner   = THIS_MODULE,
	.read    = cpld_read,
	.ioctl   = cpld_ioctl,
	.open    = cpld_open,
	.release = cpld_release,
};

static struct miscdevice cpld_miscdev = {
	.minor = NC650_CPLD_MINOR,
	.name = "nc650 cpld",
	.fops = &cpld_fops,
};

int __init
nc650_cpld_init(void)
{
	int ret = -ENODEV;
	volatile immap_t *immr = (immap_t *) IMAP_ADDR;	/* pointer to IMMR */
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry * proc;
#endif

	/* Configure memory controller (should really be moved into boot loader) */
#define BR_BA_MSK	0xffff8000	/* Base Address Mask			*/
#define BR_PS_8		0x00000400	/*  8 bit port size			*/
#define BR_PS_16	0x00000800	/*  16 bit port size			*/
#define BR_PS_32	0x00000000	/*  32 bit port size			*/
#define BR_V		0x00000001	/* Bank Valid				*/
	/* Configure CS4 through GPCM   */
	/* Timing for 20Mhz input clock */
	immr->im_memctl.memc_or4 = 0xFFFF8540;	/*(BR_BA_MSK | OR_SCY_5_CLK | OR_BI);*/
	immr->im_memctl.memc_br4 = ((CPLD_ADDR & BR_BA_MSK) | 0x00000400 | BR_V);
	/*((CPLD_ADDR & BR_BA_MSK) | BR_PS_8 | BR_V);*/

	/* to define card slot */
	immr->im_memctl.memc_or6 = 0xFFFF8140;
	immr->im_memctl.memc_br6 = ((PERI_ADDR & BR_BA_MSK) | 0x00000400 | BR_V);
	/* Get virtual pointers to hw */
	cpld = (cpld_t *)ioremap(CPLD_ADDR, sizeof(cpld_t));
	peri = (peri_t *)ioremap(PERI_ADDR, sizeof(peri_t));

	slot_addr = ~(peri->slot_addr_inv) & 0xF;

	/* Initialize LEDs to off */
	errled_state = 0;
	cpld->err_led_off = DONTCARE;
	statusled_state = 0;
	cpld->status_led = ~statusled_state;

	if ((ret = misc_register(&cpld_miscdev))) {
		printk(KERN_WARNING
		       "%s: could not register driver\n", __FILE__);
		return ret;
	}

#ifdef CONFIG_PROC_FS
	if (!proc_mkdir(NC650_CPLD_PROC_DIR, NULL)) {
		printk(KERN_ERR "failed to create /proc/" NC650_CPLD_PROC_DIR "\n");
		goto error;
	}

	if (!(proc = create_proc_entry(NC650_CPLD_PROC_SLOT_ADDR, S_IFREG|S_IRUGO, NULL))) {
		printk(KERN_ERR "failed to create /proc/" NC650_CPLD_PROC_SLOT_ADDR "\n");
		goto error;
	}

	proc->read_proc	 = cpld_slot_addr_proc_read;

	if (!(proc = create_proc_entry(NC650_CPLD_PROC_MODES, S_IFREG|S_IRUGO, NULL))) {
		printk(KERN_ERR "failed to create /proc/" NC650_CPLD_PROC_MODES "\n");
		goto error;
	}

	proc->read_proc	 = cpld_modes_proc_read;
#endif /* CONFIG_PROC_FS */

	printk (KERN_INFO "NC650 cpld driver initialized - running in slot %d\n", slot_addr);
	ret = 0;

 error:
	return ret;
}
EXPORT_SYMBOL(nc650_cpld_init);


int __init
cpld_init(void)
{
	/* Only register driver when no previous initialization was done */
	if (!cpld)
		cpld_init();

	return 0;
}

static void __exit
cpld_exit(void)
{
#ifdef CONFIG_PROC_FS
	remove_proc_entry(NC650_CPLD_PROC_DIR, NULL);
	remove_proc_entry(NC650_CPLD_PROC_SLOT_ADDR, NULL);
	remove_proc_entry(NC650_CPLD_PROC_MODES, NULL);
#endif
	misc_deregister(&cpld_miscdev);
}

#if 0
/* Not used right now - but might serve as an example later */
/* Parse commandline options */
static int __init
cpld_setup(char *options)
{
	char *sep;

	while (options && *options) {
		/* Parse mode for scc4 (hdlc) */
		if (strncmp (options, "scc4:", 4) == 0) {
			options += 4;
			if (!*options)
				return 0;

			if (strncmp (options, "oc", 2) == 0)
				scc4_busmode = NC650_CPLD_MODE_OC;
			else
				scc4_busmode = NC650_CPLD_MODE_RS485;

			if ((sep = strpbrk (options, ",")) != NULL)
				options = sep+1;
			else
				return 0;
		}
	}

	return 0;
}
__setup("cpld=", cpld_setup);
#endif

/* Functions also exported to other kernel modules (i.e. status_led) */
/* Read status register from CPLD */
unsigned char cpld_read_status(void)
{
	return cpld->status_reg_read;
}
EXPORT_SYMBOL(cpld_read_status);

/* Switch on a status led (incl. error led) */
void cpld_set_statusled(unsigned val)
{
	printk("cpld setting LED\n");
	if (val & CPLD_SL_ERR_LED) {
		if (!errled_state) {
			cpld->err_led_on  = DONTCARE;
			errled_state = 1;
		}
	} else {
		if (!(statusled_state & val)) {
			statusled_state |= val;
			cpld->status_led = ~statusled_state;
		}
	}
}
EXPORT_SYMBOL(cpld_set_statusled);

/* Switch off status led (incl. error led) */
void cpld_clear_statusled(unsigned val)
{
	if (val & CPLD_SL_ERR_LED) {
		if (errled_state) {
			cpld->err_led_off  = DONTCARE;
			errled_state = 0;
		}
	} else {
		if ((statusled_state & val)) {
			statusled_state &= ~val;
			cpld->status_led = ~statusled_state;
		}
	}
}
EXPORT_SYMBOL(cpld_clear_statusled);

/* Toggle status led (incl. error led) */
void cpld_toggle_statusled(unsigned val)
{
	if (val & CPLD_SL_ERR_LED) {
		if (errled_state) {
			cpld->err_led_off  = DONTCARE;
			errled_state = 0;
		} else {
			cpld->err_led_on   = DONTCARE;
			errled_state = 1;
		}
	} else {
		statusled_state ^= val;
		cpld->status_led = ~statusled_state;
	}
}
EXPORT_SYMBOL(cpld_toggle_statusled);

/* This publishes everything - only use it when wrappers would be
   trivial                                                        */
volatile cpld_t *cpld_get_ptr(void)
{
	return (cpld_t *)cpld;
}
EXPORT_SYMBOL(cpld_get_ptr);

module_init(cpld_init);
module_exit(cpld_exit);

MODULE_AUTHOR("Detlev Zundel <dzu@denx.de>");
MODULE_DESCRIPTION("NC650 CPLD Driver");
MODULE_LICENSE("GPL");
