/*
 * pp01_ad64.c - Driver for the AD64 display unit on the PP01 board 
 *
 * Copyright (C) 2004 Wolfgang Grandegger (wg@denx.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>

#include <asm/types.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/mpc5xxx.h>
#include <asm/pp01_ad64.h>


#define	AD64_VERSION	"0.9"
#define AD64_NAME	"ad64"
#define AD64_MSG	"ad64: "

#undef AD64_EXCLUSIVE_OPEN

#undef AD64_DEBUG
#ifdef AD64_DEBUG
# define debugk(fmt,args...)    printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Wolfgang Grandegger");
MODULE_DESCRIPTION("AD64 Display Driver");

#define AD64_MAJOR	248
static int major = AD64_MAJOR;
MODULE_PARM(major,"i");
MODULE_PARM_DESC(major, "Device major number (default=248)");

static struct mpc5xxx_gpt *gpt = (struct mpc5xxx_gpt *)MPC5xxx_GPT;

#define MPC5xxx_GPT_EMSR_OCT_TOGGLE	0x00300000
#define MPC5xxx_GPT_EMSR_ICT_RISING	0x00010000
#define MPC5xxx_GPT_EMSR_ICT_FALLING	0x00020000
#define MPC5xxx_GPT_EMSR_ICT_PULS	0x00030000
#define MPC5xxx_GPT_EMSR_CE		0x00001000
#define MPC5xxx_GPT_EMSR_CONTINOUS	0x00000400
#define MPC5xxx_GPT_EMSR_OPEN_DRAIN    	0x00000200
#define MPC5xxx_GPT_EMSR_INT_ENABLE    	0x00000100
#define MPC5xxx_GPT_EMSR_OUTPUT_1	0x00000030
#define MPC5xxx_GPT_EMSR_OUTPUT_0 	0x00000020
#define MPC5xxx_GPT_EMSR_INPUT    	0x00000000
#define MPC5xxx_GPT_EMSR_GPIO_MASK    	0x00000030
#define MPC5xxx_GPT_EMSR_INTERNAL	0x00000004
#define MPC5xxx_GPT_EMSR_PWM		0x00000003
#define MPC5xxx_GPT_EMSR_OUT_COMPARE	0x00000002
#define MPC5xxx_GPT_EMSR_INP_CAPTURE	0x00000001

#define MPC5xxx_GPT_SR_PIN		0x00000100
#define MPC5xxx_GPT_SR_TEXP		0x00000008
#define MPC5xxx_GPT_SR_PWMP		0x00000004
#define MPC5xxx_GPT_SR_COMP		0x00000002
#define MPC5xxx_GPT_SR_CAPT		0x00000001

/*
 * AD64 related definitions:
 *
 * The AD64 consists of a 4-Digit 5x7 Matrix LED Display MAX6952, 
 * one LED and three push buttoms. The third one is used as reset 
 * buttom. The MAX6952 is access via SPI on the GPIO timer ports 
 * 0..3. The LED and push buttons are connected to the GPIO timer
 * ports 4..7.
 */
#define AD64_CLK		0
#define AD64_DOUT		1
#define AD64_CS			2
#define AD64_DIN		3
#define AD64_LED		4
#define AD64_S1			5
#define AD64_S2			6

#ifdef AD64_DEBUG
static char *ad64_port_names[8] = {
    "CLK","DOUT","CS","DIN","LED","S1","S2","S3"
};
#endif

#define AD64_DELAY_US		10 /* in micro-seconds */

#define AD64_PORT_TO_IRQ(port)	(13 + port)
#define AD64_IRQ_TO_PORT(irq)	(irq - 13)

DECLARE_WAIT_QUEUE_HEAD(ad64_pushb_wait_queue);
static char ad64_pushb_state;

DECLARE_WAIT_QUEUE_HEAD(ad64_spi_wait_queue);
static int ad64_busy;

static int ad64_lock_spi(struct file *file)
{
    int err;
    if (ad64_busy) {
	if (file->f_flags & O_NONBLOCK) 
	    return -EAGAIN;
	if ((err = wait_event_interruptible(ad64_spi_wait_queue, 
					    !ad64_busy)))
	    return err;
    }
    return 0;
}

static inline void ad64_unlock_spi(void)
{
    ad64_busy = 0;
    wake_up(&ad64_spi_wait_queue);
}

#ifdef AD64_EXCLUSICE_OPEN
static char ad64_is_open = 0;
#endif

static void ad64_init_spi(void)
{
    /* SPI Data Output on GPIO Timer 3 */
    out_be32(&gpt[AD64_DOUT].emsr, 
	     MPC5xxx_GPT_EMSR_OUTPUT_0 |
	     MPC5xxx_GPT_EMSR_INTERNAL);

    /* SPI Chip Select on GPIO Timer 2 */
    out_be32(&gpt[AD64_CS].emsr, 
	     MPC5xxx_GPT_EMSR_OUTPUT_1 |
	     MPC5xxx_GPT_EMSR_INTERNAL);

    /* SPI Clock on GPIO Timer 0 */
    /* Simple GPIO should work together with watchdog */
    out_be32(&gpt[AD64_CLK].emsr, 
	     in_be32(&gpt[AD64_CLK].emsr) |
	     MPC5xxx_GPT_EMSR_OUTPUT_0 |
	     MPC5xxx_GPT_EMSR_INTERNAL);

    /* SPI Data Input on GPIO Timer 1 */
    out_be32(&gpt[AD64_DIN].emsr, 
	     MPC5xxx_GPT_EMSR_INPUT |
	     MPC5xxx_GPT_EMSR_INTERNAL);

    /* LED Output on GPIO Timer 4 */
    out_be32(&gpt[AD64_LED].emsr, 
	     MPC5xxx_GPT_EMSR_OUTPUT_0 |
	     MPC5xxx_GPT_EMSR_INTERNAL);

    /* Switch S1 on GPIO Timer 5 */
    out_be32(&gpt[AD64_S1].emsr, 
	     MPC5xxx_GPT_EMSR_INP_CAPTURE |
	     MPC5xxx_GPT_EMSR_ICT_RISING |
	     MPC5xxx_GPT_EMSR_INT_ENABLE);

    /* Switch S2 on GPIO Timer 6 */
    out_be32(&gpt[AD64_S2].emsr, 
	     MPC5xxx_GPT_EMSR_INP_CAPTURE |
	     MPC5xxx_GPT_EMSR_ICT_RISING |
	     MPC5xxx_GPT_EMSR_INT_ENABLE);
}

static void ad64_set_gpt(int num, int on)
{
    u32 emsr = in_be32(&gpt[num].emsr) & ~MPC5xxx_GPT_EMSR_GPIO_MASK;
    if (on) {
	out_be32(&gpt[num].emsr,
		 emsr | MPC5xxx_GPT_EMSR_OUTPUT_1);
    } else {
	out_be32(&gpt[num].emsr,
		 emsr | MPC5xxx_GPT_EMSR_OUTPUT_0);
    }
    udelay(AD64_DELAY_US);
    debugk("%s: %d => 0x%04x\n", ad64_port_names[num], on, gpt[num].emsr);
}

static void ad64_select_chip(int on)
{
    if (on) {
	ad64_set_gpt(AD64_CS, 0);
    } else {
	ad64_set_gpt(AD64_CS, 1);
    }
}

static int ad64_write_spi(u16 val)
{
    int i;

    ad64_select_chip(1);
    for (i = 15; i >= 0; i--) {
	ad64_set_gpt(AD64_DOUT, val & (1 << i));
	ad64_set_gpt(AD64_CLK, 1);
	if (i == 0)
	    ad64_select_chip(0);
	ad64_set_gpt(AD64_CLK, 0);
    }
    return 0;
}

static void ad64_reset_digits(void)
{
    ad64_write_spi(AD64_CONFIG | AD64_CONFIG_SHUTDOWN);
    ad64_write_spi(AD64_CONFIG | AD64_CONFIG_NORMAL);
    ad64_write_spi(AD64_DIGIT_P01(0) | ' ');
    ad64_write_spi(AD64_DIGIT_P01(1) | ' ');
    ad64_write_spi(AD64_DIGIT_P01(2) | ' ');
    ad64_write_spi(AD64_DIGIT_P01(3) | ' ');
}

static void ad64_set_digits(int op, char* digit)
{
	ad64_write_spi((op << 8) | digit[0]);
	op++;
	ad64_write_spi((op << 8) | digit[1]);
	op++;
	ad64_write_spi((op << 8) | digit[2]);
	op++;
	ad64_write_spi((op << 8) | digit[3]);
}

static void ad64_set_brightness(int val)
{
	if (val > 15) 
		val = 15;
	val += val << 4; 
	ad64_write_spi(AD64_INTENSITY10 | val);
	ad64_write_spi(AD64_INTENSITY32 | val);
}

static void ad64_set_led(int val)
{
    out_be32(&gpt[AD64_LED].emsr, 0); /* Reset first */
    switch (val) {
    case 0:			/* Off */
	out_be32(&gpt[AD64_LED].emsr,
		 MPC5xxx_GPT_EMSR_OUTPUT_0 |
		 MPC5xxx_GPT_EMSR_INTERNAL);	
	break;

    case 1:			/* On */
	out_be32(&gpt[AD64_LED].emsr,
		 MPC5xxx_GPT_EMSR_OUTPUT_1 |
		 MPC5xxx_GPT_EMSR_INTERNAL);	
	break;

    default:			/* Blink */
	out_be32(&gpt[AD64_LED].emsr, 0);
	/* Set blink rate in milli-seconds */
	out_be32(&gpt[AD64_LED].cir,
		 ((CONFIG_PPC_5xxx_IPBFREQ / 1000) << 16) | val);
	out_be32(&gpt[AD64_LED].emsr,
		 MPC5xxx_GPT_EMSR_OUT_COMPARE |
		 MPC5xxx_GPT_EMSR_CONTINOUS |
		 MPC5xxx_GPT_EMSR_OCT_TOGGLE); 	
    }
}

static void ad64_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{	
    int port = AD64_IRQ_TO_PORT(irq);
    if (port == AD64_S1 || port == AD64_S2) {
	/* Clear pending interrupt */
	out_be32(&gpt[port].sr,  MPC5xxx_GPT_SR_CAPT);
	ad64_pushb_state |= 1 << (port - AD64_S1);
	wake_up_interruptible(&ad64_pushb_wait_queue);
    } else {		
	printk(KERN_ERR AD64_MSG 
	       "Unexpected IRQ %d received", irq); 
    }
    debugk("irq=%d port=%d pushb=0x%x\n", irq, port, ad64_pushb_state); 
}

#ifdef CONFIG_PROC_FS
static int ad64_read_proc(char *page, char **start, off_t off,
                          int count, int *eof, void *data)
{
    int len = 0;

    len = sprintf(page, "%x\n", ad64_pushb_state);
    *eof = 1;
    return len;
}

static int ad64_write_proc(struct file *file, const char *buffer,
			   unsigned long count, void *data)
{    
    char buf[20], digit[4];
    int val, err;

    if (count > 20)
	return -EINVAL;
    if (copy_from_user(buf, buffer, count)) {
	return -EFAULT;
    }
    if (sscanf(buf, "brightness %i", &val) == 1) {
	if ((err = ad64_lock_spi(file)))
	    return err;
	ad64_set_brightness(val);
	ad64_unlock_spi();
    }
    else if (sscanf(buf, "digits %4s", digit) == 1) {
	if ((err = ad64_lock_spi(file)))
	    return err;
	ad64_set_digits(0x20, digit);
	ad64_unlock_spi();
    }
    else if (sscanf(buf, "led %i", &val) == 1) {
	ad64_set_led(val);
    }    
    else if (sscanf(buf, "%i", &val) == 1) {
	ad64_pushb_state &= ~val;
    }
    else {
	return -EINVAL;
    }

    return count;
}
#endif

/*
 * Device file operations
 */

static int ad64_open(struct inode *inode, struct file *file)
{
#ifdef AD64_EXCLUSICE_OPEN
    if (ad64_is_open != 0)
	return -EBUSY;
    ad64_is_open = 1;
#endif
    return 0;
}

static int ad64_close(struct inode *inode, struct file *file)
{
#ifdef AD64_EXCLUSICE_OPEN
    ad64_is_open = 1;
#endif
    return 0;
}

static ssize_t ad64_write(struct file *file, const char *buf, 
			  size_t count, loff_t *ppos)
{
    u8 digit[4] = {0x20, 0x20, 0x20, 0x20};
    int err;

    if (count <= 0)
	return 0;
    if (count > 4)
	count = 4;
    if (copy_from_user(digit, buf, count)) {
	return -EFAULT;
    }

    if ((err = ad64_lock_spi(file)))
	return err;
    ad64_write_spi(AD64_DIGIT_P0(0) | digit[0]);
    ad64_write_spi(AD64_DIGIT_P0(1) | digit[1]);
    ad64_write_spi(AD64_DIGIT_P0(2) | digit[2]);
    ad64_write_spi(AD64_DIGIT_P0(3) | digit[3]);
    ad64_unlock_spi();

    return count;
} 

static ssize_t ad64_read(struct file *file, char *buf, 
			 size_t count, loff_t *ppos)
{
    if (put_user(ad64_pushb_state, buf)) {
	return -EFAULT;
    }
    ad64_pushb_state = 0;
    return sizeof(ad64_pushb_state);
}

static u_int ad64_poll(struct file *file, 
		       struct poll_table_struct *ptbl)
{
    poll_wait(file, &ad64_pushb_wait_queue, ptbl);
    if (ad64_pushb_state)
	return  POLLIN | POLLRDNORM;
    else
	return 0;
}

static int ad64_ioctl(struct inode *inode, struct file *file, 
		      unsigned int cmd, unsigned long arg)
{
    int err = 0;

    /* 
     * Do as much checks as possible from the command definition.
     */
    if (_IOC_TYPE(cmd) != AD64_MAGIC) return -ENOTTY;

    if (_IOC_DIR(cmd) & _IOC_READ)
	err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
	err =  !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
    if (err)
	return -EFAULT;

    switch (cmd) {

    case AD64_RAW_WRITE:  {
	u16 val;
	if ((err = __get_user(val, (u16 *)arg)) != 0) {
	    return err;
	}
	if ((err = ad64_lock_spi(file)))
	    return err;
	ad64_write_spi(val);
	ad64_unlock_spi();
	break;
    }

    case AD64_CONFIGURE: {
	u8 val;
	if ((err = __get_user(val, (u8 *)arg)) != 0) {
	    return err;
	}
	if ((err = ad64_lock_spi(file)))
	    return err;
	ad64_write_spi(AD64_CONFIG | val);
	ad64_unlock_spi();
	break;
    }

    case AD64_BRIGHTNESS: {
	u8 val;
	if ((err = __get_user(val, (u8 *)arg)) != 0) {
	    return err;
	}
	if ((err = ad64_lock_spi(file)))
	    return err;
	ad64_set_brightness(val);
	ad64_unlock_spi();
	break;
    }

    case AD64_WRITE: {
	ad64_write_t ka, *ua = (ad64_write_t *)arg;
	int op;
	    
	if ((err = __copy_from_user(&ka, ua, sizeof(ka))) != 0) {
	    return err;
	}
	if (ka.plane > 2) {
	    err = -EINVAL;
	    return err;
	}
	op = (ka.plane + 1) * 0x20;
	if ((err = ad64_lock_spi(file)))
	    return err;
	ad64_set_digits(op, ka.digit);
	ad64_unlock_spi();
	break;
    }

    case AD64_RESET:  {
	if ((err = ad64_lock_spi(file)))
	    return err;
	ad64_reset_digits();
	ad64_unlock_spi();
	break;
    }

    case AD64_SET_LED:  {
	u32 val;
	if ((err = __get_user(val, (u32 *)arg)) != 0) {
	    return err;
	}
	ad64_set_led(val);
	break;
    }

    default:
	return -ENOTTY;
    }

    return 0;
}

static struct file_operations ad64_fops = {
    owner: THIS_MODULE,
    open:  ad64_open,
    release:  ad64_close,
    read: ad64_read,
    write: ad64_write,
    poll: ad64_poll,
    ioctl: ad64_ioctl,
};


static int __init ad64_init(void)
{
#ifdef CONFIG_PROC_FS
    struct proc_dir_entry * proc;
#endif
    int ret = -ENODEV;
	
    ret = register_chrdev(AD64_MAJOR, AD64_NAME, &ad64_fops);
    if (ret < 0) {
	printk(KERN_ERR AD64_MSG 
	       "Couldn't register " AD64_NAME " driver\n");
	goto abort;
    }
    if (major == 0)
	major = ret;   /* dynamic */

#ifdef CONFIG_PROC_FS
    proc = create_proc_entry(AD64_NAME, S_IFREG | S_IRUGO, NULL);
    if (proc == NULL) {
	printk(KERN_ERR AD64_MSG
	       "failed to create /proc/"AD64_NAME"\n");
	goto abort_unregister;
    }
    proc->read_proc  = ad64_read_proc;
    proc->write_proc = ad64_write_proc;
#endif
    /*
     * Initilaize display and set default mode
     */
    ad64_init_spi();
    ad64_reset_digits();
    ad64_write_spi(AD64_DIGIT_P0 (1) | 'H');
    ad64_write_spi(AD64_DIGIT_P0 (2) | 'i');

    /*
     * Register interrupt request(s)
     */
    if (request_irq(AD64_PORT_TO_IRQ(AD64_S1), 
		    ad64_interrupt, 0, "ad64 s1", NULL) ||	
	request_irq(AD64_PORT_TO_IRQ(AD64_S2), 
		    ad64_interrupt, 0, "ad64 s2", NULL)) {
	printk(KERN_ERR AD64_MSG
	       "couldn't register interrupts\n");
	goto abort_remove_proc;
    }	

    printk(KERN_INFO
	   "AD64 Display Driver v%s loaded\n", AD64_VERSION);
	
    return 0;
	
 abort_remove_proc:
    remove_proc_entry(AD64_NAME,  NULL);
 abort_unregister:
    unregister_chrdev(major, AD64_NAME);	
 abort:
    return ret;
}

static void __devexit ad64_cleanup(void)
{
    free_irq(AD64_PORT_TO_IRQ(AD64_S1), NULL);
    free_irq(AD64_PORT_TO_IRQ(AD64_S2), NULL);
    remove_proc_entry(AD64_NAME,  NULL);
    unregister_chrdev(major, AD64_NAME);
    printk(KERN_INFO "AD64 Display Driver v%s unloaded\n",
	   AD64_VERSION);
}

EXPORT_NO_SYMBOLS;

module_init(ad64_init)
module_exit(ad64_cleanup)
