/*
 *	sc26x198.c  -- Philips Semiconductors SC26x198 Octal UART driver.
 * 
 *	Copyright (C) 2000 Softwarecannery (kent@softwarecannery.com)
 *	Copyright (C) 2000 Stas Maximov (smaximov@yahoo.com).
 *
 *  This code is based on stallion multiport serial driver.
 *
 *	Copyright (C) 1996-1999  Stallion Technologies (support@stallion.oz.au).
 *	Copyright (C) 1994-1996  Greg Ungerer (gerg@stallion.oz.au).
 *
 *	This code is loosely based on the Linux serial driver, written by
 *	Linus Torvalds, Theodore T'so and others.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/comstats.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/ctype.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/uaccess.h>

/*****************************************************************************/

/*
 * Define important driver constants here.
 */
#define	SC26X_MAXCHIPS		1	/* Virtlab2 has only one SC28L198! */
#define	SC26X_PORTSPERCHIP	8
#define	SC26X_MAXDEVS		(SC26X_MAXCHIPS * SC26X_PORTSPERCHIP)

#define DRIVER_NAME "sc26x198"
#define DRIVER_DESCRIPTION "Philips Semiconductors sc26x198 Octal UART driver"

/*
 * Configuration structure for each chip. Each chip refered via its 
 * base address and irq. irqs numbered as they are passed to request_irq
 * function, thus on 8xx systems, these are internal numbers, like
 * SIU_IRQ5 = 10,  SIU_IRQ6 = 12, not irq line numbers of mpc8xx pins.
 *
 * Example:
 * static sc26x_conf_t	sc26x_chipconf[] = {
 *    {0x80010000, 10},
 *    {0x80010100, 12}
 * };
 */

typedef struct 
{
	int		ioaddr1;
	int		irq;
} sc26x_conf_t;

static sc26x_conf_t	sc26x_chipconf[] = {
#if defined(CONFIG_SC26X198_1_ADDR) && defined(CONFIG_SC26X198_1_IRQ)
    { CONFIG_SC26X198_1_ADDR, CONFIG_SC26X198_1_IRQ },
#else
    { 0x80000000, 0xA}, /* prevent Oopses if no options specified */
#endif
/*** only one SC28L198 on Virtlab2 
#if defined(CONFIG_SC26X198_2_ADDR) && defined(CONFIG_SC26X198_2_IRQ)
    { CONFIG_SC26X198_2_ADDR, CONFIG_SC26X198_2_IRQ },
#else
    { 0, 0},
#endif
    { 0, 0}
***/
	/*{ 0x2a0, 10 },*/
};

static int	sc26x_nrchips = sizeof(sc26x_chipconf) / sizeof(sc26x_conf_t);

/*
 * RTS signal can be configured in two ways - either on pin I/O1 or I/O2.
 * DTR and DCD can be configured on any remaning pins after CTS and RTS are
 * assigned. So we take sc26198.h as a basis and redefine pinouts according
 * to a specific board design.
 *
 * Below is only one of possible configurations. For detatils, see sc26x198
 * datasheet.
 */
#include <linux/sc26198.h>

#undef	IPR_CTS
#define	IPR_CTS         0x01
#undef	IPR_DTR
#define	IPR_DTR         0x08
#undef	IPR_RTS
#define	IPR_RTS         0x02
#undef	IPR_DCD
#define	IPR_DCD         0x04

#undef	IPR_CTSCHANGE
#define	IPR_CTSCHANGE   0x10
#undef	IPR_DTRCHANGE
#define	IPR_DTRCHANGE   0x80
#undef	IPR_RTSCHANGE
#define	IPR_RTSCHANGE   0x20
#undef	IPR_DCDCHANGE
#define	IPR_DCDCHANGE   0x40

#undef	IOPR_CTS
#define	IOPR_CTS        0x01
#undef	IOPR_DTR
#define	IOPR_DTR        0x08
#undef	IOPR_RTS
#define	IOPR_RTS        0x02
#undef	IOPR_DCD
#define	IOPR_DCD        0x04

#undef	IOPR_CTSCOS
#define	IOPR_CTSCOS     0x10
#undef	IOPR_DTRCOS
#define	IOPR_DTRCOS     0x80
#undef	IOPR_RTSCOS
#define	IOPR_RTSCOS     0x20
#undef	IOPR_DCDCOS
#define	IOPR_DCDCOS     0x40

#undef	IOPCR_SETCTS
#define	IOPCR_SETCTS    0x00
#undef	IOPCR_SETDTR
#define	IOPCR_SETDTR    0x04
#undef	IOPCR_SETRTS
#define	IOPCR_SETRTS    0x00
#undef	IOPCR_SETDCD
#define	IOPCR_SETDCD    0x40


/*****************************************************************************/

/*
 *	Define a ring queue structure for each port. This will hold the
 *	TX data waiting to be output. Characters are fed into this buffer
 *	from the line discipline (or even direct from user space!) and
 *	then fed into the UARTs during interrupts. Will use a classic ring
 *	queue here for this. The good thing about this type of ring queue
 *	is that the head and tail pointers can be updated without interrupt
 *	protection - since "write" code only needs to change the head, and
 *	interrupt code only needs to change the tail.
 */
typedef struct sc26x_rq
{
	char	*buf;
	char	*head;
	char	*tail;
} sc26x_rq_t;

/*
 *	Port, panel and board structures to hold status info about each.
 *	The board structure contains pointers to structures for each panel
 *	connected to it, and in turn each panel structure contains pointers
 *	for each port structure for each port on that panel. Note that
 *	the port structure also contains the board and panel number that it
 *	is associated with, this makes it (fairly) easy to get back to the
 *	board/panel info for a port.
 */
typedef struct sc26x_port 
{
	u32         magic;
	int			portnr;
	int			chipnr;
	u8 *        ioaddr;
	int			istate;
	int			flags;
	int			baud_base;
	int			custom_divisor;
	int			close_delay;
	int			closing_wait;
	int			refcount;
	int			openwaitcnt;
	int			brklen;
	long			session;
	long			pgrp;
	unsigned int		sigs;
	unsigned int		rxignoremsk;
	unsigned int		rxmarkmsk;
	unsigned int		imr;
	unsigned int		crenable;
	struct tty_struct	*tty;
	wait_queue_head_t	open_wait;
	wait_queue_head_t	close_wait;
	struct termios		normaltermios;
	struct termios		callouttermios;
	struct tq_struct	tqueue;
	comstats_t		stats;
	sc26x_rq_t		tx;
} sc26x_port_t;

typedef struct sc26x_chip 
{
	u32		magic;
	int		chipnr;
	int		state;
	int		irq;
	u8 *		ioaddr1_phys; /* unmapped */
	int		iosize1;
	u8 *		ioaddr1;
	sc26x_port_t	ports[SC26X_PORTSPERCHIP];
} sc26x_chip_t;

/*
 *	Define MAGIC numbers used for above structures.
 */
#define	SC26X_PORTMAGIC		0x5c261980
#define	SC26X_BOARDMAGIC	0x5c26198b

/*****************************************************************************/

/*
 *	Define some important driver characteristics. Device major numbers
 *	allocated as per Linux Device Registry.
 */
#ifndef	SC26X_SERIALMAJOR
#define	SC26X_SERIALMAJOR   24
#endif
#ifndef	SC26X_CALLOUTMAJOR
#define	SC26X_CALLOUTMAJOR  25
#endif

#define	SC26X_DRVTYPSERIAL	1
#define	SC26X_DRVTYPCALLOUT	2

/*
 *	Set the TX buffer size. Bigger is better, but we don't want
 *	to chew too much memory with buffers!
 */
#define	SC26X_TXBUFLOW      512
#define	SC26X_TXBUFSIZE     4096

/*****************************************************************************/

/*
 *	Define our local driver identity first. Set up stuff to deal with
 *	all the local structures required by a serial tty driver.
 */
static char	*sc26x_drvtitle = DRIVER_DESCRIPTION;
static char	*sc26x_drvname = DRIVER_NAME;
static char	*sc26x_drvversion = "0.1";
static char	*sc26x_serialname = "ttyE";
static char	*sc26x_calloutname = "cue";

static struct tty_driver	sc26x_serial;
static struct tty_driver	sc26x_callout;
static struct tty_struct	*sc26x_ttys[SC26X_MAXDEVS];
static struct termios		*sc26x_termios[SC26X_MAXDEVS];
static struct termios		*sc26x_termioslocked[SC26X_MAXDEVS];
static int			sc26x_refcount = 0;

/*
 *	We will need to allocate a temporary write buffer for chars that
 *	come direct from user space. The problem is that a copy from user
 *	space might cause a page fault (typically on a system that is
 *	swapping!). All ports will share one buffer - since if the system
 *	is already swapping a shared buffer won't make things any worse.
 */
static char *sc26x_tmpwritebuf;
static DECLARE_MUTEX(sc26x_tmpwritesem);

/*
 *	Define a local default termios struct. All ports will be created
 *	with this termios initially. Basically all it defines is a raw port
 *	at 9600, 8 data bits, 1 stop bit.
 */
static struct termios		sc26x_deftermios = {
#ifdef _PPC_TERMBITS_H
	0,
	0,
	(B9600 | CS8 | CREAD | HUPCL | CLOCAL),
	0,
	INIT_C_CC
#else
	0,
	0,
	(B9600 | CS8 | CREAD | HUPCL | CLOCAL),
	0,
	0,
	INIT_C_CC
#endif
};

/*
 *	Define global place to put buffer overflow characters.
 */
static char		sc26x_unwanted[SC26198_RXFIFOSIZE];

/*
 *	Keep track of what interrupts we have requested for us.
 *	We don't need to request an interrupt twice if it is being
 *	shared with another Stallion board.
 */
struct irq_tab
{
    int it_irq;
    int it_count;
};
static struct irq_tab sc26x_gotintrs[SC26X_MAXCHIPS];

/*****************************************************************************/

static sc26x_chip_t *sc26x_chips[SC26X_MAXCHIPS];

/*
 *	Per board state flags. Used with the state field of the board struct.
 *	Not really much here!
 */
#define	CHIP_FOUND	0x1

/*
 *	Define the port structure istate flags. These set of flags are
 *	modified at interrupt time - so setting and reseting them needs
 *	to be atomic. Use the bit clear/setting routines for this.
 */
#define	ASYI_TXBUSY	1
#define	ASYI_TXLOW	2
#define	ASYI_DCDCHANGE	3
#define	ASYI_TXFLOWED	4

/*****************************************************************************/

#ifdef MODULE
/*
 *	Define some string labels for arguments passed from the module
 *	load line. These allow for easy board definitions, and easy
 *	modification of the io, memory and irq resoucres.
 */

static char	*chip0[2];
static char	*chip1[2];
static char	*chip2[2];
static char	*chip3[2];

static char	**sc26x_chipsp[] = { chip0, chip1, chip2, chip3 };

/*
 *	Define the module agruments.
 */
MODULE_AUTHOR("Stas Maximov");
MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_LICENSE("GPL");

MODULE_PARM(chip0, "1-2s");
MODULE_PARM_DESC(chip0, "Chip 0 config -> ioaddr[,irq]");
/* //only one SC28L198 on Virtlab2 
MODULE_PARM(chip1, "1-2s");
MODULE_PARM_DESC(chip1, "Chip 1 config -> ioaddr[,irq]");
MODULE_PARM(chip2, "1-2s");
MODULE_PARM_DESC(chip2, "Chip 2 config -> ioaddr[,irq]");
MODULE_PARM(chip3, "1-2s");
MODULE_PARM_DESC(chip3, "Chip 3 config -> ioaddr[,irq]");
 */
#ifdef DEBUG
MODULE_PARM(debug_level, "i");
MODULE_PARM_DESC(debug_level, "Debug trace level (0-9)");
#endif /* DEBUG */

#endif /* MODULE */

#ifdef DEBUG

static int debug_level = 0;

#define debug_trace(lvl, fmt, arg...)                                   \
    do                                                                  \
    {                                                                   \
        if (debug_level >= lvl)                                         \
                printk(KERN_DEBUG DRIVER_NAME " "                       \
                __FILE__ ":" __MODULE_STRING(__LINE__) ": "             \
                fmt, ## arg);                                           \
    }                                                                   \
    while (0)
#else
#define debug_trace(lvl, fmt, arg...)
#endif /* DEBUG */

/*****************************************************************************/

#define	SC26X_MAXBAUD	    230400
#define	SC26X_BAUDBASE      115200
#define	SC26X_CLOSEDELAY    (5 * HZ / 10)

/*****************************************************************************/

/*
 *	Define macros to extract a chip/port number from a minor number.
 */
#define	MINOR2CHIP(min)		(((min) & 0xF8) >> 3)
#define	MINOR2PORT(min)		((min) & 0x7)

/*
 *	Define a baud rate table that converts termios baud rate selector
 *	into the actual baud rate value. All baud rate calculations are
 *	based on the actual baud rate required.
 */
static unsigned int	sc26x_baudrates[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
	9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600
};

/*
 *	Define some handy local macros...
 */
#undef	MIN
#define	MIN(a,b)	(((a) <= (b)) ? (a) : (b))

#undef	TOLOWER
#define	TOLOWER(x)	((((x) >= 'A') && ((x) <= 'Z')) ? ((x) + 0x20) : (x))

/*****************************************************************************/

/*
 *	Declare all those functions in this driver!
 */

#ifdef MODULE
int		init_module(void);
void		cleanup_module(void);
static void	sc26x_argchips(void);
static int	sc26x_parsechip(sc26x_conf_t *confp, char **argp);
static int	sc26x_unmapirq_all(void);
static unsigned long sc26x_atol(char *str);
#endif

int		sc26x_init(void);
static int	sc26x_initchips(void);
static int	sc26x_chipinit(sc26x_chip_t *chipp);
static int	sc26x_mapirq(int irq, char *name);
static int  sc26x_unmapirq(int irq);
static sc26x_chip_t *sc26x_allocchip(void);

static int	sc26x_open(struct tty_struct *tty, struct file *filp);
static void	sc26x_close(struct tty_struct *tty, struct file *filp);
static int	sc26x_write(struct tty_struct *tty, int from_user, 
    const unsigned char *buf, int count);
static void	sc26x_putchar(struct tty_struct *tty, unsigned char ch);
static void	sc26x_flushchars(struct tty_struct *tty);
static int	sc26x_writeroom(struct tty_struct *tty);
static int	sc26x_charsinbuffer(struct tty_struct *tty);
static int	sc26x_ioctl(struct tty_struct *tty, struct file *file, 
    unsigned int cmd, unsigned long arg);
static void	sc26x_settermios(struct tty_struct *tty, struct termios *old);
static void	sc26x_throttle(struct tty_struct *tty);
static void	sc26x_unthrottle(struct tty_struct *tty);
static void	sc26x_stop(struct tty_struct *tty);
static void	sc26x_start(struct tty_struct *tty);
static void	sc26x_flushbuffer(struct tty_struct *tty);
static void	sc26x_breakctl(struct tty_struct *tty, int state);
static void	sc26x_waituntilsent(struct tty_struct *tty, int timeout);
static void	sc26x_sendxchar(struct tty_struct *tty, char ch);
static void	sc26x_hangup(struct tty_struct *tty);
static int	sc26x_readproc(char *page, char **start, off_t off, int count, 
    int *eof, void *data);
static void	sc26x_getserial(sc26x_port_t *portp, struct serial_struct *sp);
static int	sc26x_setserial(sc26x_port_t *portp, struct serial_struct *sp);

static int	sc26x_waitcarrier(sc26x_port_t *portp, struct file *filp);
static void	sc26x_delay(int len);
static void	sc26x_offintr(void *private);
static void	sc26x_setport(sc26x_port_t *portp, struct termios *tiosp);
static int	sc26x_getsignals(sc26x_port_t *portp);
static void	sc26x_setsignals(sc26x_port_t *portp, int dtr, int rts);
static void	sc26x_enablerxtx(sc26x_port_t *portp, int rx, int tx);
static void	sc26x_startrxtx(sc26x_port_t *portp, int rx, int tx);
static void	sc26x_disableintrs(sc26x_port_t *portp);
static void	sc26x_sendbreak(sc26x_port_t *portp, int len);
static void	sc26x_flowctrl(sc26x_port_t *portp, int state);
static void	sc26x_sendflow(sc26x_port_t *portp, int state);
static void	sc26x_flush(sc26x_port_t *portp);
static int	sc26x_datastate(sc26x_port_t *portp);
static void	sc26x_wait(sc26x_port_t *portp);
static void	sc26x_txunflow(sc26x_port_t *portp, struct tty_struct *tty);

static void	sc26x_intr(int irq, void *dev_id, struct pt_regs *regs);
static void	sc26x_txisr(sc26x_port_t *port);
static void	sc26x_rxisr(sc26x_port_t *port, unsigned int cir);
static void	sc26x_rxbadch(sc26x_port_t *portp, unsigned char status, char ch);
static void	sc26x_rxbadchars(sc26x_port_t *portp);
static void	sc26x_otherisr(sc26x_port_t *port, unsigned int cir);

#if 0
static void dump_buf(char *b, int len);
#endif

/*****************************************************************************/

/*
 *	Define the sc26198 baud rate table. Offsets within the table
 *	represent the actual baud rate selector of sc26198 registers.
 */
static unsigned int	sc26198_baudtable[] = {
	50, 75, 150, 200, 300, 450, 600, 900, 1200, 1800, 2400, 3600,
	4800, 7200, 9600, 14400, 19200, 28800, 38400, 57600, 115200,
	230400, 460800, 921600
};

#define	SC26198_NRBAUDS		(sizeof(sc26198_baudtable) / sizeof(unsigned int))

/*****************************************************************************/

/*
 * Wrappers for accessing sc26x198 registers. May need modifications
 * when porting to a platform other than powerpc.
 */

static inline int
sc26x_getreg(sc26x_port_t *portp, int regnr)
{
    return in_8(portp->ioaddr + SC26198_PORTREG(portp->portnr, regnr));
}

static inline void 
sc26x_setreg(sc26x_port_t *portp, int regnr, int value)
{
    out_8(portp->ioaddr + SC26198_PORTREG(portp->portnr, regnr), value);
}

static inline int 
sc26x_updatereg(sc26x_port_t *portp, int regnr, int value)
{
    if (sc26x_getreg(portp, regnr) != value)
    {
        sc26x_setreg(portp, regnr, value);
        return 1;
    }
    return 0;
}

/*
 *	Functions to get and set the sc26198 global registers.
 */

static inline u8
sc26x_getglobreg(u8 *base_addr, int regnr)
{
    return in_8(base_addr + regnr);
}

static inline void 
sc26x_setglobreg(u8 *base_addr, int regnr, int value)
{
    out_8(base_addr + regnr, value);
}

static inline void 
sc26x_outsb(u8 *base_addr, int regnr, char *buf, int ns)
{
    base_addr += regnr;
    while (ns-- > 0)
        out_8(base_addr, *buf++);
}

static inline void 
sc26x_insb(u8 *base_addr, int regnr, char *buf, int ns)
{
    base_addr += regnr;
    while (ns-- > 0)
        *buf++ = in_8(base_addr);
}

/*****************************************************************************/

#ifdef MODULE

/*
 *	Loadable module initialization stuff.
 */

int init_module()
{
	unsigned long	flags;

	debug_trace(5, "init_module()\n");

	save_flags(flags);
	cli();
	sc26x_init();
	restore_flags(flags);

	return 0;
}

/*****************************************************************************/

void cleanup_module()
{
	unsigned long	flags;
	int		i, j;

	debug_trace(5, "cleanup_module()\n");

	printk(KERN_INFO "Unloading %s: version %s\n", sc26x_drvtitle,
		sc26x_drvversion);

	save_flags(flags);
	cli();

/*
 *	Free up all allocated resources used by the ports. This includes
 *	memory and interrupts. As part of this process we will also do
 *	a hangup on every open port - to try to flush out any processes
 *	hanging onto ports.
 */
	i = tty_unregister_driver(&sc26x_serial);
	j = tty_unregister_driver(&sc26x_callout);
	if (i || j) 
	{
		printk("sc26x198: failed to un-register tty driver, "
			"errno=%d,%d\n", -i, -j);
		restore_flags(flags);
		return;
	}

	if (sc26x_tmpwritebuf != (char *) NULL)
		kfree(sc26x_tmpwritebuf);

	for (i = 0; i < sc26x_nrchips; i++) 
	{
	    sc26x_chip_t    *chipp;

		chipp = sc26x_chips[i];
		if (!chipp)
			continue;

		for (j = 0; j < SC26X_PORTSPERCHIP; j++) 
		{
			sc26x_port_t    *portp;

			portp = chipp->ports + j;
			if (portp->tty)
				sc26x_hangup(portp->tty);
			if (portp->tx.buf)
				kfree(portp->tx.buf);
		}

		kfree(chipp);
		sc26x_chips[i] = (sc26x_chip_t *) NULL;
	}

	sc26x_unmapirq_all();

	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Check for any arguments passed in on the module load command line.
 */

static void 
sc26x_argchips()
{
	sc26x_conf_t	conf;
	sc26x_chip_t	*chipp;
	int		nrargs, i;

	debug_trace(5, "sc26x_argchips()\n");

	nrargs = sizeof(sc26x_chipsp) / sizeof(char **);

	for (i = sc26x_nrchips; i < nrargs; i++) 
	{
		memset(&conf, 0, sizeof(conf));
		if (sc26x_parsechip(&conf, sc26x_chipsp[i]) == 0)
			continue;
		chipp = sc26x_allocchip();
		if (!chipp)
			continue;
		sc26x_nrchips = i + 1;
		chipp->chipnr = i;
		chipp->ioaddr1_phys = (u8 *)conf.ioaddr1;
		chipp->irq = conf.irq;
		sc26x_chipinit(chipp);
	}
}

/*****************************************************************************/

/*
 *	Convert an ascii string number into an unsigned long.
 */

static unsigned long 
sc26x_atol(char *str)
{
	unsigned long	val;
	int		base, c;
	char		*sp;

	val = 0;
	sp = str;
	if (*sp == '0' && *(sp+1) == 'x') 
	{
		base = 16;
		sp += 2;
	} 
	else if (*sp == '0') 
	{
		base = 8;
		sp++;
	} 
	else
		base = 10;

	for (; *sp; sp++) 
	{
		c = *sp > '9' ? TOLOWER(*sp) - 'a' + 10 : *sp - '0';
		if (c < 0 || c >= base) 
		{
			printk("sc26x198: invalid argument %s\n", str);
			val = 0;
			break;
		}
		val = (val * base) + c;
	}
	return val;
}

/*****************************************************************************/

/*
 *	Parse the supplied argument string, into the board conf struct.
 */

static int 
sc26x_parsechip(sc26x_conf_t *confp, char **argp)
{
	int	i;

	debug_trace(5, "sc26x_parsechip(confp=%x,argp={%s,%s})\n", (int) confp, 
        argp[0] ? argp[0] : "<NULL>",
        argp[1] ? argp[1] : "<NULL>");

	if (!argp[0] || !*argp[0])
		return 0;

	i = 0;
	if (argp[i] && *argp[i])
		confp->ioaddr1 = sc26x_atol(argp[i]);
	i++;
	if (argp[i] && *argp[i])
		confp->irq = sc26x_atol(argp[i]);

	return 1;
}

#endif

/*****************************************************************************/

/*
 *	Allocate a new chip structure. Fill out the basic info in it.
 */

static sc26x_chip_t *
sc26x_allocchip()
{
	sc26x_chip_t *chipp;

	chipp = kmalloc(sizeof(sc26x_chip_t), GFP_KERNEL);
	if (!chipp)
	{
		printk("sc26x198: failed to allocate memory (size=%d)\n",
			sizeof(sc26x_chip_t));
		return 0;
	}

	memset(chipp, 0, sizeof(sc26x_chip_t));
	chipp->magic = SC26X_BOARDMAGIC;
	return chipp;
}

/*****************************************************************************/

static int 
sc26x_open(struct tty_struct *tty, struct file *filp)
{
	sc26x_port_t *portp;
	sc26x_chip_t *chipp;
	unsigned int minordev;
	int chipnr, portnr, rc;

	debug_trace(5, "sc26x_open(tty=%x,filp=%x): device=%x\n", (int) tty,
		(int) filp, tty->device);

	minordev = MINOR(tty->device);
	chipnr = MINOR2CHIP(minordev);
	if (chipnr >= sc26x_nrchips)
		return -ENODEV;
	chipp = sc26x_chips[chipnr];
	if (!chipp)
		return -ENODEV;
	portnr = MINOR2PORT(minordev);
	if (portnr < 0 || portnr >= SC26X_MAXDEVS)
		return -ENODEV;

	portp = chipp->ports + portnr;

	MOD_INC_USE_COUNT;

/*
 *	On the first open of the device setup the port hardware, and
 *	initialize the per port data structure.
 */
	portp->tty = tty;
	tty->driver_data = portp;
	portp->refcount++;

	if ((portp->flags & ASYNC_INITIALIZED) == 0) 
	{
		if (portp->tx.buf == 0) 
		{
			portp->tx.buf = kmalloc(SC26X_TXBUFSIZE, GFP_KERNEL);
			if (!portp->tx.buf)
				return -ENOMEM;
			portp->tx.head = portp->tx.buf;
			portp->tx.tail = portp->tx.buf;
		}
		sc26x_setport(portp, tty->termios);
		portp->sigs = sc26x_getsignals(portp);
		sc26x_setsignals(portp, 1, 1);
		sc26x_enablerxtx(portp, 1, 1);
		sc26x_startrxtx(portp, 1, 0);
		clear_bit(TTY_IO_ERROR, &tty->flags);
		portp->flags |= ASYNC_INITIALIZED;
	}

/*
 *	Check if this port is in the middle of closing. If so then wait
 *	until it is closed then return error status, based on flag settings.
 *	The sleep here does not need interrupt protection since the wakeup
 *	for it is done with the same context.
 */
	if (portp->flags & ASYNC_CLOSING)
	{
		interruptible_sleep_on(&portp->close_wait);
		if (portp->flags & ASYNC_HUP_NOTIFY)
			return -EAGAIN;
		return -ERESTARTSYS;
	}

/*
 *	Based on type of open being done check if it can overlap with any
 *	previous opens still in effect. If we are a normal serial device
 *	then also we might have to wait for carrier.
 */
	if (tty->driver.subtype == SC26X_DRVTYPCALLOUT) 
	{
		if (portp->flags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;

		if (portp->flags & ASYNC_CALLOUT_ACTIVE) 
        {
			if ((portp->flags & ASYNC_SESSION_LOCKOUT) &&
			    portp->session != current->session)
				return -EBUSY;
			if ((portp->flags & ASYNC_PGRP_LOCKOUT) &&
			    portp->pgrp != current->pgrp)
				return -EBUSY;
		}
		portp->flags |= ASYNC_CALLOUT_ACTIVE;
	} 
	else 
	{
		if (filp->f_flags & O_NONBLOCK) 
		{
			if (portp->flags & ASYNC_CALLOUT_ACTIVE)
				return -EBUSY;
		} 
		else 
		{
			if ((rc = sc26x_waitcarrier(portp, filp)) != 0)
				return rc;
		}

		portp->flags |= ASYNC_NORMAL_ACTIVE;
	}

	if (portp->refcount == 1 && (portp->flags & ASYNC_SPLIT_TERMIOS)) 
	{
		if (tty->driver.subtype == SC26X_DRVTYPSERIAL)
			*tty->termios = portp->normaltermios;
		else
			*tty->termios = portp->callouttermios;
		sc26x_setport(portp, tty->termios);
	}

	portp->session = current->session;
	portp->pgrp = current->pgrp;
	return 0;
}

/*****************************************************************************/

/*
 *	Possibly need to wait for carrier (DCD signal) to come high. Say
 *	maybe because if we are clocal then we don't need to wait...
 */

static int 
sc26x_waitcarrier(sc26x_port_t *portp, struct file *filp)
{
	unsigned long	flags;
	int		rc, doclocal;

	debug_trace(5, "sc26x_waitcarrier(portp=%x,filp=%x)\n", 
        (int) portp, (int) filp);

	rc = 0;
	doclocal = 0;

	if (portp->flags & ASYNC_CALLOUT_ACTIVE) 
	{
		if (portp->normaltermios.c_cflag & CLOCAL)
			doclocal++;
	} 
	else 
	{
		if (portp->tty->termios->c_cflag & CLOCAL)
			doclocal++;
	}

	save_flags(flags);
	cli();
	portp->openwaitcnt++;
	if (! tty_hung_up_p(filp))
		portp->refcount--;

	for (;;) 
	{
		if ((portp->flags & ASYNC_CALLOUT_ACTIVE) == 0)
			sc26x_setsignals(portp, 1, 1);

		if (tty_hung_up_p(filp) ||
		    ((portp->flags & ASYNC_INITIALIZED) == 0)) 
		{
			if (portp->flags & ASYNC_HUP_NOTIFY)
				rc = -EBUSY;
			else
				rc = -ERESTARTSYS;
			break;
		}

		if (((portp->flags & ASYNC_CALLOUT_ACTIVE) == 0) &&
		    ((portp->flags & ASYNC_CLOSING) == 0) &&
		    (doclocal || (portp->sigs & TIOCM_CD)))
			break;

		if (signal_pending(current)) 
		{
			rc = -ERESTARTSYS;
			break;
		}

		interruptible_sleep_on(&portp->open_wait);
	}

	if (!tty_hung_up_p(filp))
		portp->refcount++;
	portp->openwaitcnt--;
	restore_flags(flags);

	return(rc);
}

/*****************************************************************************/

static void sc26x_close(struct tty_struct *tty, struct file *filp)
{
	sc26x_port_t	*portp;
	unsigned long	flags;

	debug_trace(5, "sc26x_close(tty=%x,filp=%x)\n", (int) tty, (int) filp);

	portp = tty->driver_data;
	if (!portp)
		return;

	save_flags(flags);
	cli();
	if (tty_hung_up_p(filp)) 
	{
		MOD_DEC_USE_COUNT;
		restore_flags(flags);
		return;
	}
	if (tty->count == 1 && portp->refcount != 1)
		portp->refcount = 1;
	if (portp->refcount-- > 1) 
	{
		MOD_DEC_USE_COUNT;
		restore_flags(flags);
		return;
	}

	portp->refcount = 0;
	portp->flags |= ASYNC_CLOSING;

	if (portp->flags & ASYNC_NORMAL_ACTIVE)
		portp->normaltermios = *tty->termios;
	if (portp->flags & ASYNC_CALLOUT_ACTIVE)
		portp->callouttermios = *tty->termios;

/*
 *	May want to wait for any data to drain before closing. The BUSY
 *	flag keeps track of whether we are still sending or not - it is
 *	very accurate for the cd1400, not quite so for the sc26198.
 *	(The sc26198 has no "end-of-data" interrupt only empty FIFO)
 */
	tty->closing = 1;
	if (portp->closing_wait != ASYNC_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, portp->closing_wait);
	sc26x_waituntilsent(tty, HZ / 2);

	portp->flags &= ~ASYNC_INITIALIZED;
	sc26x_disableintrs(portp);
	if (tty->termios->c_cflag & HUPCL)
		sc26x_setsignals(portp, 0, 0);
	sc26x_enablerxtx(portp, 0, 0);
	sc26x_flushbuffer(tty);
	portp->istate = 0;
	if (portp->tx.buf) 
	{
		kfree(portp->tx.buf);
		portp->tx.buf = 0;
		portp->tx.head = 0;
		portp->tx.tail = 0;
	}
	set_bit(TTY_IO_ERROR, &tty->flags);
	if (tty->ldisc.flush_buffer)
		(tty->ldisc.flush_buffer)(tty);

	tty->closing = 0;
	portp->tty = 0;

	if (portp->openwaitcnt) 
	{
		if (portp->close_delay)
			sc26x_delay(portp->close_delay);
		wake_up_interruptible(&portp->open_wait);
	}

	portp->flags &= ~(ASYNC_CALLOUT_ACTIVE | ASYNC_NORMAL_ACTIVE |
		ASYNC_CLOSING);
	wake_up_interruptible(&portp->close_wait);
	MOD_DEC_USE_COUNT;
	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Wait for a specified delay period, this is not a busy-loop. It will
 *	give up the processor while waiting. Unfortunately this has some
 *	rather intimate knowledge of the process management stuff.
 */

static void 
sc26x_delay(int len)
{
	debug_trace(5, "sc26x_delay(len=%d)\n", len);

	if (len > 0) 
	{
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(len);
		set_current_state(TASK_RUNNING);
	}
}

/*****************************************************************************/

/*
 *	Write routine. Take data and stuff it in to the TX ring queue.
 *	If transmit interrupts are not running then start them.
 */

static int 
sc26x_write(struct tty_struct *tty, int from_user, const unsigned char *buf, 
    int count)
{
	sc26x_port_t	*portp;
	unsigned int	len, stlen;
	char		*head, *tail;

	debug_trace(5, "sc26x_write(tty=%x,from_user=%d,buf=%x,count=%d)\n",
		(int) tty, from_user, (int) buf, count);

	if (!tty || !sc26x_tmpwritebuf)
		return 0;
	portp = tty->driver_data;
	if (!portp || !portp->tx.buf)
		return 0;

/*
 *	If copying direct from user space we must cater for page faults,
 *	causing us to "sleep" here for a while. To handle this copy in all
 *	the data we need now, into a local buffer. Then when we got it all
 *	copy it into the TX buffer.
 */
	if (from_user) 
	{
		head = portp->tx.head;
		tail = portp->tx.tail;
		len = (head >= tail) ? (SC26X_TXBUFSIZE - (head - tail) - 1) :
			(tail - head - 1);
		count = MIN(len, count);

		down(&sc26x_tmpwritesem);
		copy_from_user(sc26x_tmpwritebuf, buf, count);
		buf = sc26x_tmpwritebuf;
	}

	head = portp->tx.head;
	tail = portp->tx.tail;
	if (head >= tail) {
		len = SC26X_TXBUFSIZE - (head - tail) - 1;
		stlen = SC26X_TXBUFSIZE - (head - portp->tx.buf);
	} else {
		len = tail - head - 1;
		stlen = len;
	}

	len = MIN(len, count);
	count = 0;
	while (len > 0) 
	{
		stlen = MIN(len, stlen);
		memcpy(head, buf, stlen);
		len -= stlen;
		buf += stlen;
		count += stlen;
		head += stlen;
		if (head >= portp->tx.buf + SC26X_TXBUFSIZE) 
		{
			head = portp->tx.buf;
			stlen = tail - head;
		}
	}
	portp->tx.head = head;

	clear_bit(ASYI_TXLOW, &portp->istate);
	sc26x_startrxtx(portp, -1, 1);

	if (from_user)
		up(&sc26x_tmpwritesem);

	return count;
}

/*****************************************************************************/

static void 
sc26x_putchar(struct tty_struct *tty, unsigned char ch)
{
	sc26x_port_t	*portp;
	unsigned int	len;
	char		*head, *tail;

	debug_trace(5, "sc26x_putchar(tty=%x,ch=%x)\n", (int) tty, (int) ch);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp || !portp->tx.buf)
		return;

	head = portp->tx.head;
	tail = portp->tx.tail;

	len = head >= tail ? SC26X_TXBUFSIZE - (head - tail) : tail - head;
	len--;

	if (len > 0) 
	{
		*head++ = ch;
		if (head >= portp->tx.buf + SC26X_TXBUFSIZE)
			head = portp->tx.buf;
	}	
	portp->tx.head = head;
}

/*****************************************************************************/

/*
 *	If there are any characters in the buffer then make sure that TX
 *	interrupts are on and get'em out. Normally used after the putchar
 *	routine has been called.
 */

static void 
sc26x_flushchars(struct tty_struct *tty)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_flushchars(tty=%x)\n", (int) tty);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp || !portp->tx.buf)
		return;

	if (!tty->stopped && !tty->hw_stopped && portp->tx.head != portp->tx.tail)
	    sc26x_startrxtx(portp, -1, 1);
}

/*****************************************************************************/

static int 
sc26x_writeroom(struct tty_struct *tty)
{
	sc26x_port_t	*portp;
	char    *head, *tail;

	debug_trace(5, "sc26x_writeroom(tty=%x)\n", (int) tty);

	if (!tty)
		return 0;
	portp = tty->driver_data;
	if (!portp || !portp->tx.buf)
		return 0;

	head = portp->tx.head;
	tail = portp->tx.tail;
	return head >= tail ? SC26X_TXBUFSIZE - (head - tail) - 1 : tail - head - 1;
}

/*****************************************************************************/

/*
 *	Return number of chars in the TX buffer. Normally we would just
 *	calculate the number of chars in the buffer and return that, but if
 *	the buffer is empty and TX interrupts are still on then we return
 *	that the buffer still has 1 char in it. This way whoever called us
 *	will not think that ALL chars have drained - since the UART still
 *	must have some chars in it (we are busy after all).
 */

static int 
sc26x_charsinbuffer(struct tty_struct *tty)
{
	sc26x_port_t	*portp;
	unsigned int	size;
	char		*head, *tail;

	debug_trace(5, "sc26x_charsinbuffer(tty=%x)\n", (int) tty);

	if (tty == (struct tty_struct *) NULL)
		return 0;
	portp = tty->driver_data;
	if (portp == (sc26x_port_t *) NULL)
		return 0;
	if (portp->tx.buf == (char *) NULL)
		return 0;

	head = portp->tx.head;
	tail = portp->tx.tail;
	size = (head >= tail) ? (head - tail) : (SC26X_TXBUFSIZE - (tail - head));
	if ((size == 0) && test_bit(ASYI_TXBUSY, &portp->istate))
		size = 1;
	return size;
}

/*****************************************************************************/

/*
 *	Generate the serial struct info.
 */

static void 
sc26x_getserial(sc26x_port_t *portp, struct serial_struct *sp)
{
	struct serial_struct	sio;
	sc26x_chip_t		*chipp;

	debug_trace(5, "sc26x_getserial(portp=%x,sp=%x)\n", (int) portp, (int) sp);

	memset(&sio, 0, sizeof(struct serial_struct));
	sio.line = portp->portnr;
	sio.port = (int)portp->ioaddr;
	sio.flags = portp->flags;
	sio.baud_base = portp->baud_base;
	sio.close_delay = portp->close_delay;
	sio.closing_wait = portp->closing_wait;
	sio.custom_divisor = portp->custom_divisor;
	sio.hub6 = 0;
	sio.type = PORT_UNKNOWN;
	sio.xmit_fifo_size = SC26198_TXFIFOSIZE;

	chipp = sc26x_chips[portp->chipnr];
	if (chipp != (sc26x_chip_t *) NULL)
		sio.irq = chipp->irq;

	copy_to_user(sp, &sio, sizeof(struct serial_struct));
}

/*****************************************************************************/

/*
 *	Set port according to the serial struct info.
 *	At this point we do not do any auto-configure stuff, so we will
 *	just quietly ignore any requests to change irq, etc.
 */

static int 
sc26x_setserial(sc26x_port_t *portp, struct serial_struct *sp)
{
	struct serial_struct	sio;

	debug_trace(5, "sc26x_setserial(portp=%x,sp=%x)\n", (int) portp, (int) sp);

	copy_from_user(&sio, sp, sizeof(struct serial_struct));
	if (!capable(CAP_SYS_ADMIN)) 
	{
		if ((sio.baud_base != portp->baud_base) ||
		    (sio.close_delay != portp->close_delay) ||
		    ((sio.flags & ~ASYNC_USR_MASK) !=
		    (portp->flags & ~ASYNC_USR_MASK)))
			return -EPERM;
	} 

	portp->flags = (portp->flags & ~ASYNC_USR_MASK) |
		(sio.flags & ASYNC_USR_MASK);
	portp->baud_base = sio.baud_base;
	portp->close_delay = sio.close_delay;
	portp->closing_wait = sio.closing_wait;
	portp->custom_divisor = sio.custom_divisor;
	sc26x_setport(portp, portp->tty->termios);
	return 0;
}

/*****************************************************************************/

static int 
sc26x_ioctl(struct tty_struct *tty, struct file *file, 
    unsigned int cmd, unsigned long arg)
{
	sc26x_port_t	*portp;
	unsigned int	ival;
	int		rc;

	debug_trace(5, "sc26x_ioctl(tty=%x,file=%x,cmd=%x,arg=%x)\n",
		(int) tty, (int) file, cmd, (int) arg);

	if (tty == (struct tty_struct *) NULL)
		return -ENODEV;
	portp = tty->driver_data;
	if (portp == (sc26x_port_t *) NULL)
		return -ENODEV;

	if (cmd != TIOCGSERIAL && cmd != TIOCSSERIAL &&
        (tty->flags & (1 << TTY_IO_ERROR)))
			return -EIO;

	rc = 0;

	switch (cmd) {
	case TIOCGSOFTCAR:
	    debug_trace(5, "sc26x_ioctl(TIOCGSOFTCAR)\n");
		rc = put_user(((tty->termios->c_cflag & CLOCAL) ? 1 : 0),
			(unsigned int *) arg);
		break;
	case TIOCSSOFTCAR:
	    debug_trace(5, "sc26x_ioctl(TIOCSSOFTCAR)\n");
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
		    sizeof(int))) == 0) {
			get_user(ival, (unsigned int *) arg);
			tty->termios->c_cflag =
				(tty->termios->c_cflag & ~CLOCAL) |
				(ival ? CLOCAL : 0);
		}
		break;
	case TIOCMGET:
	    debug_trace(5, "sc26x_ioctl(TIOCMGET)\n");
		if ((rc = verify_area(VERIFY_WRITE, (void *) arg,
		    sizeof(unsigned int))) == 0) {
			ival = sc26x_getsignals(portp);
			put_user(ival, (unsigned int *) arg);
		}
		break;
	case TIOCMBIS:
	    debug_trace(5, "sc26x_ioctl(TIOCMBIS)\n");
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
		    sizeof(unsigned int))) == 0) {
			get_user(ival, (unsigned int *) arg);
			sc26x_setsignals(portp, ((ival & TIOCM_DTR) ? 1 : -1),
				((ival & TIOCM_RTS) ? 1 : -1));
		}
		break;
	case TIOCMBIC:
	    debug_trace(5, "sc26x_ioctl(TIOCMBIC)\n");
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
		    sizeof(unsigned int))) == 0) {
			get_user(ival, (unsigned int *) arg);
			sc26x_setsignals(portp, ((ival & TIOCM_DTR) ? 0 : -1),
				((ival & TIOCM_RTS) ? 0 : -1));
		}
		break;
	case TIOCMSET:
	    debug_trace(5, "sc26x_ioctl(TIOCMSET)\n");
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
		    sizeof(unsigned int))) == 0) {
			get_user(ival, (unsigned int *) arg);
			sc26x_setsignals(portp, ((ival & TIOCM_DTR) ? 1 : 0),
				((ival & TIOCM_RTS) ? 1 : 0));
		}
		break;
	case TIOCGSERIAL:
	    debug_trace(5, "sc26x_ioctl(TIOCGSERIAL)\n");
		if ((rc = verify_area(VERIFY_WRITE, (void *) arg,
		    sizeof(struct serial_struct))) == 0)
			sc26x_getserial(portp, (struct serial_struct *) arg);
		break;
	case TIOCSSERIAL:
	    debug_trace(5, "sc26x_ioctl(TIOCSSERIAL)\n");
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
		    sizeof(struct serial_struct))) == 0)
			rc = sc26x_setserial(portp, (struct serial_struct *) arg);
		break;
	case TIOCSERCONFIG:
	case TIOCSERGWILD:
	case TIOCSERSWILD:
	case TIOCSERGETLSR:
	case TIOCSERGSTRUCT:
	case TIOCSERGETMULTI:
	case TIOCSERSETMULTI:
	default:
		debug_trace(5, "sc26x_ioctl(unknown TC)\n");
		rc = -ENOIOCTLCMD;
		break;
	}

	return(rc);
}

/*****************************************************************************/

static void 
sc26x_settermios(struct tty_struct *tty, struct termios *old)
{
	sc26x_port_t	*portp;
	struct termios	*tiosp;

	debug_trace(5, "sc26x_settermios(tty=%x,old=%x)\n", (int) tty, (int) old);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;

	tiosp = tty->termios;
	if ((tiosp->c_cflag == old->c_cflag) &&
	    (tiosp->c_iflag == old->c_iflag))
		return;

	sc26x_setport(portp, tiosp);
	sc26x_setsignals(portp, ((tiosp->c_cflag & (CBAUD & ~CBAUDEX)) ? 1 : 0),
		-1);
	if ((old->c_cflag & CRTSCTS) && ((tiosp->c_cflag & CRTSCTS) == 0)) {
		tty->hw_stopped = 0;
		sc26x_start(tty);
	}
	if (((old->c_cflag & CLOCAL) == 0) && (tiosp->c_cflag & CLOCAL))
		wake_up_interruptible(&portp->open_wait);
}

/*****************************************************************************/

/*
 *	Attempt to flow control who ever is sending us data. Based on termios
 *	settings use software or/and hardware flow control.
 */

static void 
sc26x_throttle(struct tty_struct *tty)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_throttle(tty=%x)\n", (int) tty);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;
	sc26x_flowctrl(portp, 0);
}

/*****************************************************************************/

/*
 *	Unflow control the device sending us data...
 */

static void 
sc26x_unthrottle(struct tty_struct *tty)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_unthrottle(tty=%x)\n", (int) tty);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;
	sc26x_flowctrl(portp, 1);
}

/*****************************************************************************/

/*
 *	Stop the transmitter. Basically to do this we will just turn TX
 *	interrupts off.
 */

static void 
sc26x_stop(struct tty_struct *tty)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_stop(tty=%x)\n", (int) tty);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;
	sc26x_startrxtx(portp, -1, 0);
}

/*****************************************************************************/

/*
 *	Start the transmitter again. Just turn TX interrupts back on.
 */

static void 
sc26x_start(struct tty_struct *tty)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_start(tty=%x)\n", (int) tty);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;
	sc26x_startrxtx(portp, -1, 1);
}

/*****************************************************************************/

/*
 *	Hangup this port. This is pretty much like closing the port, only
 *	a little more brutal. No waiting for data to drain. Shutdown the
 *	port and maybe drop signals.
 */

static void 
sc26x_hangup(struct tty_struct *tty)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_hangup(tty=%x)\n", (int) tty);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;

	portp->flags &= ~ASYNC_INITIALIZED;
	sc26x_disableintrs(portp);
	if (tty->termios->c_cflag & HUPCL)
		sc26x_setsignals(portp, 0, 0);
	sc26x_enablerxtx(portp, 0, 0);
	sc26x_flushbuffer(tty);
	portp->istate = 0;
	set_bit(TTY_IO_ERROR, &tty->flags);
	if (portp->tx.buf)
	{
		kfree(portp->tx.buf);
		portp->tx.buf = (char *) NULL;
		portp->tx.head = (char *) NULL;
		portp->tx.tail = (char *) NULL;
	}
	portp->tty = (struct tty_struct *) NULL;
	portp->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE);
	portp->refcount = 0;
	wake_up_interruptible(&portp->open_wait);
}

/*****************************************************************************/

static void 
sc26x_flushbuffer(struct tty_struct *tty)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_flushbuffer(tty=%x)\n", (int) tty);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;

	sc26x_flush(portp);
	wake_up_interruptible(&tty->write_wait);
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

/*****************************************************************************/

static void 
sc26x_breakctl(struct tty_struct *tty, int state)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_breakctl(tty=%x,state=%d)\n", (int) tty, state);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;

	sc26x_sendbreak(portp, state == -1 ? 1 : 2);
}

/*****************************************************************************/

static void 
sc26x_waituntilsent(struct tty_struct *tty, int timeout)
{
	sc26x_port_t	*portp;
	unsigned long	tend;

	debug_trace(5, "sc26x_waituntilsent(tty=%x,timeout=%d)\n", 
        (int) tty, timeout);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;

	if (timeout == 0)
		timeout = HZ;
	tend = jiffies + timeout;

	while (sc26x_datastate(portp)) 
	{
		if (signal_pending(current))
			break;
		sc26x_delay(2);
		if (time_after_eq(jiffies, tend))
			break;
	}
}

/*****************************************************************************/

static void 
sc26x_sendxchar(struct tty_struct *tty, char ch)
{
	sc26x_port_t	*portp;

	debug_trace(5, "sc26x_sendxchar(tty=%x,ch=%x)\n", (int) tty, ch);

	if (!tty)
		return;
	portp = tty->driver_data;
	if (!portp)
		return;

	if (ch == STOP_CHAR(tty))
		sc26x_sendflow(portp, 0);
	else if (ch == START_CHAR(tty))
		sc26x_sendflow(portp, 1);
	else
		sc26x_putchar(tty, ch);
}

/*****************************************************************************/

#if 0

#define	MAXLINE		80

/*
 *	Format info for a specified port. The line is deliberately limited
 *	to 80 characters. (If it is too long it will be truncated, if too
 *	short then padded with spaces).
 */

static int 
sc26x_portinfo(sc26x_port_t *portp, char *pos)
{
	char	*sp;
	int	sigs, cnt;

	debug_trace(5, "sc26x_portinfo(page=%x,start=%x,off=%x,count=%d,eof=%x,"
		"data=%x\n", (int) page, (int) start, (int) off, count,
		(int) eof, (int) data);

	sp = pos;
	sp += sprintf(sp, "%d: uart:%s #%d tx:%d rx:%d",
		portp->portnr, "SC26198", portp->chipnr,
		(int) portp->stats.txtotal, (int) portp->stats.rxtotal);

	if (portp->stats.rxframing)
		sp += sprintf(sp, " fe:%d", (int) portp->stats.rxframing);
	if (portp->stats.rxparity)
		sp += sprintf(sp, " pe:%d", (int) portp->stats.rxparity);
	if (portp->stats.rxbreaks)
		sp += sprintf(sp, " brk:%d", (int) portp->stats.rxbreaks);
	if (portp->stats.rxoverrun)
		sp += sprintf(sp, " oe:%d", (int) portp->stats.rxoverrun);

	sigs = sc26x_getsignals(portp);
	cnt = sprintf(sp, "%s%s%s%s%s ",
		(sigs & TIOCM_RTS) ? "|RTS" : "",
		(sigs & TIOCM_CTS) ? "|CTS" : "",
		(sigs & TIOCM_DTR) ? "|DTR" : "",
		(sigs & TIOCM_CD) ? "|DCD" : "",
		(sigs & TIOCM_DSR) ? "|DSR" : "");
	*sp = ' ';
	sp += cnt;

	for (cnt = sp - pos; cnt < MAXLINE - 1; cnt++)
		*sp++ = ' ';
    if (cnt >= MAXLINE)
        pos[MAXLINE - 2] = '+';
    pos[MAXLINE - 1] = '\n';

    return MAXLINE;
}

/*****************************************************************************/

/*
 * Port info, read from the /proc file system.
 *
 * page  - buffer where we play. First byte of our contents *always*
 *         begins here.
 * start -
 * off   - offset in 'page' of the region they want us to update.
 * count - length of the region in 'page' beginning from 'off' they
 *         want us to update.
 * eof   - set to 1 if the last byte written is EOF.
 * data  - ???
 */

static int 
sc26x_readproc(char *page, char **start, off_t off, int count, 
    int *eof, void *data)
{
	sc26x_chip_t *chipp;
	sc26x_port_t *portp;
	int chipnr, portnr;
	int curoff, maxoff;
	char *pos;

	debug_trace(5, "sc26x_readproc(page=%x,start=%x,off=%x,count=%d,eof=%x,"
		"data=%x\n", (int) page, (int) start, (int) off, count,
		(int) eof, (int) data);

	pos = page;
	curoff = 0;

	if (off == 0) 
    {
		pos += sprintf(pos, "%s: version %s", sc26x_drvtitle,
			sc26x_drvversion);
		while (pos < page + MAXLINE - 1)
			*pos++ = ' ';
		*pos++ = '\n';
	}
	curoff =  MAXLINE;

/*
 *	We scan through for each board and port. The offset is
 *	calculated on the fly, and irrelevant ports are skipped.
 */
	for (chipnr = 0; chipnr < sc26x_nrchips; chipnr++) 
    {
		chipp = sc26x_chips[chipnr];
		if (!chipp)
			continue;
		if (chipp->state == 0)
			continue;

		maxoff = curoff + (SC26X_PORTSPERCHIP * MAXLINE);
		if (off >= maxoff) 
        {
			curoff = maxoff;
			continue;
		}

		for (portnr = 0; portnr < SC26X_PORTSPERCHIP; portnr++) 
        {
			portp = chipp->ports + portnr;
			curoff += MAXLINE;
			if (off >= curoff)
				continue;
			if (pos - page + MAXLINE > count)
				goto sc26x_readdone;
			pos += sc26x_portinfo(portp, pos);
		}
	}

	*eof = 1;

sc26x_readdone:
	*start = page;
	return pos - page;
}

#endif 

static int 
sc26x_lineinfo(char *buf, sc26x_port_t *portp)
{
	char *sp;
	int	sigs;

	debug_trace(5, "sc26x_lineinfo(buf=0x%X, portnr=%d)\n", 
        (int)buf, portp->portnr);

	sp = buf;
	sp += sprintf(sp, "%d: uart:%s #%d tx:%d rx:%d",
		portp->portnr, "SC26x198", portp->chipnr,
		(int)portp->stats.txtotal, (int)portp->stats.rxtotal);

	if (portp->stats.rxframing)
		sp += sprintf(sp, " fe:%d", (int) portp->stats.rxframing);
	if (portp->stats.rxparity)
		sp += sprintf(sp, " pe:%d", (int) portp->stats.rxparity);
	if (portp->stats.rxbreaks)
		sp += sprintf(sp, " brk:%d", (int) portp->stats.rxbreaks);
	if (portp->stats.rxoverrun)
		sp += sprintf(sp, " oe:%d", (int) portp->stats.rxoverrun);

	sigs = sc26x_getsignals(portp);
	sp += sprintf(sp, "%s%s%s%s%s\n",
		(sigs & TIOCM_RTS) ? "|RTS" : "",
		(sigs & TIOCM_CTS) ? "|CTS" : "",
		(sigs & TIOCM_DTR) ? "|DTR" : "",
		(sigs & TIOCM_CD) ? "|DCD" : "",
		(sigs & TIOCM_DSR) ? "|DSR" : "");

	return sp - buf;
}


int sc26x_readproc(char *page, char **start, off_t off, int count,
    int *eof, void *data)
{  
    int chipnr, len;
    off_t begin = 0;

	debug_trace(5, "sc26x_readproc(page=%x,start=%x,off=%x,count=%d,eof=%x,"
		"data=%x\n", (int) page, (int) start, (int) off, count,
		(int) eof, (int) data);

	len = sprintf(page, "%s: version %s\n", sc26x_drvtitle, sc26x_drvversion);

	for (chipnr = 0; chipnr < sc26x_nrchips; chipnr++) 
	{
		int portnr;
		sc26x_chip_t *chipp;

		chipp = sc26x_chips[chipnr];
		if (!chipp)
			continue;
		if (chipp->state == 0)
			continue;

		for (portnr = 0; portnr < SC26X_PORTSPERCHIP; portnr++) 
		{
			sc26x_port_t *portp;

			portp = chipp->ports + portnr;

			len += sc26x_lineinfo(page + len, portp);
			if (len+begin > off+count)
				goto done;
			if (len+begin < off) 
			{
				begin += len;
				len = 0;
			}
		}
	}
	*eof = 1;
done:
	if (off >= len+begin)
		return 0;
	*start = page + (begin-off);
	return ((count < begin+len-off) ? count : begin+len-off);
}  


/*****************************************************************************/

/*
 *	All board interrupts are vectored through here first. This code then
 *	calls off to the approrpriate board interrupt handlers.
 */

static void 
sc26x_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	int		i;

	debug_trace(8, "sc26x_intr(irq=%d,regs=%x)\n", irq, (int) regs);

	for (i = 0; i < sc26x_nrchips; i++) 
	{
		sc26x_chip_t *chipp = sc26x_chips[i];
		if (chipp != 0 && chipp->state != 0 && chipp->irq == irq)
		{
			sc26x_port_t	*portp;
			unsigned int cir;

			/* get CIR updated */
			sc26x_setglobreg(chipp->ioaddr1, UCIR, 0);
			/* get fresh CIR value */
			cir = sc26x_getglobreg(chipp->ioaddr1, CIR);

			portp = chipp->ports + (cir & CIR_CHANMASK);

			debug_trace(8, "sc26x_isr(portp=%x,cir=%x)\n", (int) portp, cir);
        
			if (cir & CIR_RXDATA)
        			sc26x_rxisr(portp, cir);
        		else if (cir & CIR_TXDATA)
        			sc26x_txisr(portp);
        		else
        			sc26x_otherisr(portp, cir);
		}
	}
}

/*****************************************************************************/

/*
 *	Service an off-level request for some channel.
 */
static void 
sc26x_offintr(void *private)
{
	sc26x_port_t		*portp;
	struct tty_struct	*tty;
	unsigned int		oldsigs;

	portp = private;

	debug_trace(5, "sc26x_offintr(portp=%x)\n", (int)portp);

	if (!portp)
		return;

	tty = portp->tty;
	if (!tty)
		return;

	debug_trace(5, "sc26x_offintr: istate=0x%X flags=0x%lX sigs=0x%X\n",
        portp->istate, tty->flags, portp->sigs);

	lock_kernel();

	if (test_bit(ASYI_TXLOW, &portp->istate)) 
	{
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
	}

	if (test_bit(ASYI_DCDCHANGE, &portp->istate)) 
	{
		clear_bit(ASYI_DCDCHANGE, &portp->istate);
		oldsigs = portp->sigs;
		portp->sigs = sc26x_getsignals(portp);
		debug_trace(5, "sc26x_offintr: newsigs=0x%X\n", portp->sigs);

		if ((portp->sigs & TIOCM_CD) && (oldsigs & TIOCM_CD) == 0)
			wake_up_interruptible(&portp->open_wait);

		if ((oldsigs & TIOCM_CD) && (portp->sigs & TIOCM_CD) == 0) 
		{
			if (portp->flags & ASYNC_CHECK_CD) 
			{
				if (! ((portp->flags & ASYNC_CALLOUT_ACTIVE) &&
				    (portp->flags & ASYNC_CALLOUT_NOHUP)))
					tty_hangup(tty);
			}
		}
	}

	unlock_kernel();
}

/*****************************************************************************/

/*
 *	Map in interrupt vector to this driver. Check that we don't
 *	already have this vector mapped, we might be sharing this
 *	interrupt across multiple boards.
 */

static int __init sc26x_mapirq(int irq, char *name)
{
	int	rc, i, free_slot;

	debug_trace(5, "sc26x_mapirq(irq=%d,name=%s)\n", irq, name);

	rc = 0;
	free_slot = -1;
	for (i = 0; i < sizeof(sc26x_gotintrs)/sizeof(sc26x_gotintrs[0]); i++) 
		if (sc26x_gotintrs[i].it_count == 0)
			free_slot = i;
		else if (sc26x_gotintrs[i].it_irq == irq)
			break;

	if (i < sizeof(sc26x_gotintrs)/sizeof(sc26x_gotintrs[0])) 
		sc26x_gotintrs[i].it_count++;
	else if (free_slot != -1)
	{
		if (request_8xxirq(irq, sc26x_intr, 0, name, NULL) != 0) 
		{
			printk("sc26x198: failed to register interrupt "
				"routine for %s irq=%d\n", name, irq);
			rc = -ENODEV;
		} 
		else
		{
			sc26x_gotintrs[free_slot].it_irq = irq;
			sc26x_gotintrs[free_slot].it_count = 1;
		}
	}
	else
	{
		printk("sc26x198: no free slots in irq registration table; "
			"routine for %s irq=%d\n", name, irq);
		rc = -ENODEV;
	}

	return rc;
}

static int __init sc26x_unmapirq(int irq)
{
	int i;

	for (i = 0; i < sizeof(sc26x_gotintrs)/sizeof(sc26x_gotintrs[0]); i++) 
		if (sc26x_gotintrs[i].it_count > 0 && sc26x_gotintrs[i].it_irq == irq)
			break;

	if (i < sizeof(sc26x_gotintrs)/sizeof(sc26x_gotintrs[0])) 
	{
		sc26x_gotintrs[i].it_count--;
		if (sc26x_gotintrs[i].it_count <= 0)
			free_irq(irq, 0);
	}
	else
		printk("sc26x198: attempt to unmap unmapped irq=%d\n", irq);

	return 0;
}

#ifdef MODULE

static int __init sc26x_unmapirq_all(void)
{
	int i;

	for (i = 0; i < sizeof(sc26x_gotintrs)/sizeof(sc26x_gotintrs[0]); i++) 
		if (sc26x_gotintrs[i].it_count > 0)
		{
			sc26x_gotintrs[i].it_count = 0;
			free_irq(sc26x_gotintrs[i].it_irq, 0);
		}
	return 0;
}

#endif /* #ifdef MODULE */

/*****************************************************************************/

/*
 *	Initialize and configure the specified chip.
 */

static int __init sc26x_chipinit(sc26x_chip_t *chipp)
{
	int	i;
	sc26x_port_t	*portp;
	int rc = 0;
	char *name = "sc26x198 uart";

	debug_trace(5, "sc26x_chipinit(@0x%X,%d)\n", 
		(int)chipp->ioaddr1_phys, chipp->irq);

	sc26x_chips[chipp->chipnr] = chipp;
	chipp->iosize1 = 256;

	chipp->ioaddr1 = ioremap_nocache((int)chipp->ioaddr1_phys, chipp->iosize1);
	if (chipp->ioaddr1 == 0)
	{
		printk("sc26x198: ioremap_nocache(0x%X, %d) failed.\n",
			(int)chipp->ioaddr1_phys, chipp->iosize1);
		rc = -ENODEV;
		goto failure1;
	}

	debug_trace(5, "sc26x_chipinit(@0x%X,%d) ioremap() = 0x%X\n", 
	        (int)chipp->ioaddr1_phys, chipp->irq, (int)chipp->ioaddr1);

	if (sc26x_mapirq(chipp->irq, name) != 0)
	{
		rc = -ENODEV;
		goto failure2;
	}

	sc26x_setglobreg(chipp->ioaddr1, GCCR, GCCR_NOACK);
	
	/*
	 *	Check that each chip is present and started up OK.
	 */
	sc26x_setglobreg(chipp->ioaddr1, SCCR, CR_RESETALL); 
	udelay(10);

	/*
	printk ("status reg. is: 0x%x\n", sc26x_getglobreg(chipp->ioaddr1, SR));
	printk("chip addr.: %p, GCCR: %x, TSTR: %x, SCCR: %x \n",\
		chipp->ioaddr1, GCCR, TSTR, SCCR);
	*/
	if (sc26x_getglobreg(chipp->ioaddr1, TSTR) != 0) 
	{
		printk("sc26x198: chip #%d @0x%X,0x%X not responding\n", 
			chipp->chipnr, (int)chipp->ioaddr1_phys, chipp->irq);
		rc = -ENODEV;
		goto failure3;
	}

	chipp->state |= CHIP_FOUND;
	sc26x_setglobreg(chipp->ioaddr1, GCCR, GCCR_IVRTYPCHANACK);
	sc26x_setglobreg(chipp->ioaddr1, WDTRCR, 0xff);

	for (i = 0; (i < SC26X_PORTSPERCHIP); i++)
	{
		portp = chipp->ports + i;

		memset(portp, 0, sizeof(sc26x_port_t));
		portp->magic = SC26X_PORTMAGIC;
		portp->portnr = i;
		portp->chipnr = chipp->chipnr;
		portp->ioaddr = chipp->ioaddr1;
		portp->baud_base = SC26X_BAUDBASE;
		portp->close_delay = SC26X_CLOSEDELAY;
		portp->closing_wait = 30 * HZ;
		portp->normaltermios = sc26x_deftermios;
		portp->callouttermios = sc26x_deftermios;
		portp->tqueue.routine = sc26x_offintr;
		portp->tqueue.data = portp;
		init_waitqueue_head(&portp->open_wait);
		init_waitqueue_head(&portp->close_wait);

		sc26x_setreg(portp, IOPCR, IOPCR_SETSIGS);
	}

	debug_trace(5, "sc26x_chipinit: chip #%d @0x%X,0x%X initialized\n", 
			chipp->chipnr, (int)chipp->ioaddr1_phys, chipp->irq);

	return 0;

failure3:
	sc26x_unmapirq(chipp->irq);

failure2:
	iounmap((void *)chipp->ioaddr1);

failure1:
	return rc;
}

/*****************************************************************************/

/*
 *	Scan through all chips in the configuration and see what we can find.
 */

static int __init sc26x_initchips(void)
{
	sc26x_chip_t    *chipp;
	sc26x_conf_t    *confp;
	int		i;

	debug_trace(5, "sc26x_initchips()\n");

	if (sc26x_nrchips > SC26X_MAXCHIPS)
	{
		printk("sc26x198: too many boards in configuration table, "
			"truncating to %d\n", SC26X_MAXCHIPS);
		sc26x_nrchips = SC26X_MAXCHIPS;
	}

/*
 *	Firstly scan the list of static boards configured. Allocate
 *	resources and initialize the boards as found.
 */
	for (i = 0; i < sc26x_nrchips; i++) 
	{
		confp = &sc26x_chipconf[i];
#ifdef MODULE
		sc26x_parsechip(confp, sc26x_chipsp[i]);
#endif
		if ((chipp = sc26x_allocchip()) == 0)
			return -ENOMEM;
		chipp->chipnr = i;
		chipp->ioaddr1_phys = (u8 *)confp->ioaddr1;
		chipp->irq = confp->irq;
		sc26x_chipinit(chipp);
	}

/*
 *	Find any dynamically supported boards.
 */
#ifdef MODULE
	sc26x_argchips();
#endif

	return 0;
}

/*****************************************************************************/

int __init sc26x_init(void)
{
	memctl8xx_t     *memc = &((immap_t *)IMAP_ADDR)->im_memctl;
	memc->memc_or5 = 0xFFFF8D84;	/* SCY=8, ACS=2, also set burst inhibit! */
	memc->memc_br5 = 0x80000401;	/* PS=2, 8-bit access */
	debug_trace(5, "sc26x198: IMAP %x, MEMC %p, BR5 %x, OR5 %x\n",
		    IMAP_ADDR, memc, memc->memc_br5, memc->memc_or5);
	
	printk(KERN_INFO "%s: version %s\n", sc26x_drvtitle, sc26x_drvversion);

	sc26x_initchips();

/*
 *	Allocate a temporary write buffer.
 */
	sc26x_tmpwritebuf = kmalloc(SC26X_TXBUFSIZE, GFP_KERNEL);
	if (!sc26x_tmpwritebuf)
		printk("sc26x198: failed to allocate memory (size=%d)\n",
			SC26X_TXBUFSIZE);

/*
 *	Set up the tty driver structure and register us as a driver.
 *	Also setup the callout tty device.
 */
	memset(&sc26x_serial, 0, sizeof(struct tty_driver));
	sc26x_serial.magic = TTY_DRIVER_MAGIC;
	sc26x_serial.driver_name = sc26x_drvname;
	sc26x_serial.name = sc26x_serialname;
	sc26x_serial.major = SC26X_SERIALMAJOR;
	sc26x_serial.minor_start = 0;
	sc26x_serial.num = SC26X_MAXDEVS;
	sc26x_serial.type = TTY_DRIVER_TYPE_SERIAL;
	sc26x_serial.subtype = SC26X_DRVTYPSERIAL;
	sc26x_serial.init_termios = sc26x_deftermios;
	sc26x_serial.flags = TTY_DRIVER_REAL_RAW;
	sc26x_serial.refcount = &sc26x_refcount;
	sc26x_serial.table = sc26x_ttys;
	sc26x_serial.termios = sc26x_termios;
	sc26x_serial.termios_locked = sc26x_termioslocked;

	sc26x_serial.open = sc26x_open;
	sc26x_serial.close = sc26x_close;
	sc26x_serial.write = sc26x_write;
	sc26x_serial.put_char = sc26x_putchar;
	sc26x_serial.flush_chars = sc26x_flushchars;
	sc26x_serial.write_room = sc26x_writeroom;
	sc26x_serial.chars_in_buffer = sc26x_charsinbuffer;
	sc26x_serial.ioctl = sc26x_ioctl;
	sc26x_serial.set_termios = sc26x_settermios;
	sc26x_serial.throttle = sc26x_throttle;
	sc26x_serial.unthrottle = sc26x_unthrottle;
	sc26x_serial.stop = sc26x_stop;
	sc26x_serial.start = sc26x_start;
	sc26x_serial.hangup = sc26x_hangup;
	sc26x_serial.flush_buffer = sc26x_flushbuffer;
	sc26x_serial.break_ctl = sc26x_breakctl;
	sc26x_serial.wait_until_sent = sc26x_waituntilsent;
	sc26x_serial.send_xchar = sc26x_sendxchar;
	sc26x_serial.read_proc = sc26x_readproc;

	sc26x_callout = sc26x_serial;
	sc26x_callout.name = sc26x_calloutname;
	sc26x_callout.major = SC26X_CALLOUTMAJOR;
	sc26x_callout.subtype = SC26X_DRVTYPCALLOUT;
	sc26x_callout.read_proc = 0;

	if (tty_register_driver(&sc26x_serial))
		printk("sc26x198: failed to register serial driver\n");
	if (tty_register_driver(&sc26x_callout))
		printk("sc26x198: failed to register callout driver\n");

	return 0;
}

/*****************************************************************************/

/*
 *	Set up the sc26198 registers for a port based on the termios port
 *	settings.
 */

static void 
sc26x_setport(sc26x_port_t *portp, struct termios *tiosp)
{
	sc26x_chip_t	*chipp;
	unsigned long	flags;
	unsigned int	baudrate;
	unsigned char	mr0, mr1, mr2, clk;
	unsigned char	imron, imroff, iopr, ipr;

	mr0 = 0;
	mr1 = 0;
	mr2 = 0;
	clk = 0;
	iopr = 0;
	imron = 0;
	imroff = 0;

	chipp = sc26x_chips[portp->chipnr];
	if (!chipp)
		return;

/*
 *	Set up the RX char ignore mask with those RX error types we
 *	can ignore.
 */
	portp->rxignoremsk = 0;
	if (tiosp->c_iflag & IGNPAR)
		portp->rxignoremsk |= (SR_RXPARITY | SR_RXFRAMING |
			SR_RXOVERRUN);
	if (tiosp->c_iflag & IGNBRK)
		portp->rxignoremsk |= SR_RXBREAK;

	portp->rxmarkmsk = SR_RXOVERRUN;
	if (tiosp->c_iflag & (INPCK | PARMRK))
		portp->rxmarkmsk |= (SR_RXPARITY | SR_RXFRAMING);
	if (tiosp->c_iflag & BRKINT)
		portp->rxmarkmsk |= SR_RXBREAK;

/*
 *	Go through the char size, parity and stop bits and set all the
 *	option register appropriately.
 */
	switch (tiosp->c_cflag & CSIZE) 
	{
	case CS5:
		mr1 |= MR1_CS5;
		break;
	case CS6:
		mr1 |= MR1_CS6;
		break;
	case CS7:
		mr1 |= MR1_CS7;
		break;
	default:
		mr1 |= MR1_CS8;
		break;
	}

	if (tiosp->c_cflag & CSTOPB)
		mr2 |= MR2_STOP2;
	else
		mr2 |= MR2_STOP1;

	if (tiosp->c_cflag & PARENB) 
	{
		if (tiosp->c_cflag & PARODD)
			mr1 |= (MR1_PARENB | MR1_PARODD);
		else
			mr1 |= (MR1_PARENB | MR1_PAREVEN);
	} 
	else
		mr1 |= MR1_PARNONE;

	mr1 |= MR1_ERRBLOCK;

/*
 *	Set the RX FIFO threshold at 8 chars. This gives a bit of breathing
 *	space for hardware flow control and the like. This should be set to
 *	VMIN.
 */
	mr2 |= MR2_RXFIFOHALF;

/*
 *	Calculate the baud rate timers. For now we will just assume that
 *	the input and output baud are the same. The sc26198 has a fixed
 *	baud rate table, so only discrete baud rates possible.
 */
	baudrate = tiosp->c_cflag & CBAUD;
	if (baudrate & CBAUDEX) 
	{
		baudrate &= ~CBAUDEX;
		if ((baudrate < 1) || (baudrate > 4))
			tiosp->c_cflag &= ~CBAUDEX;
		else
			baudrate += 15;
	}

	baudrate = sc26x_baudrates[baudrate];
	if ((tiosp->c_cflag & CBAUD) == B38400) 
	{
		if ((portp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			baudrate = 57600;
		else if ((portp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			baudrate = 115200;
		else if ((portp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			baudrate = 230400;
		else if ((portp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			baudrate = 460800;
		else if ((portp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST)
			baudrate = (portp->baud_base / portp->custom_divisor);
	}

	if (baudrate > SC26X_MAXBAUD)
		baudrate = SC26X_MAXBAUD;

	if (baudrate > 0) 
	{
		for (clk = 0; (clk < SC26198_NRBAUDS); clk++)
			if (baudrate <= sc26198_baudtable[clk])
				break;
	}

/*
 *	Check what form of modem signaling is required and set it up.
 */
	if (tiosp->c_cflag & CLOCAL) 
	{
		portp->flags &= ~ASYNC_CHECK_CD;
	} 
	else 
	{
		iopr |= IOPR_DCDCOS;
		imron |= IR_IOPORT;
		portp->flags |= ASYNC_CHECK_CD;
	}

/*
 *	Setup sc26198 enhanced modes if we can. In particular we want to
 *	handle as much of the flow control as possible automatically. As
 *	well as saving a few CPU cycles it will also greatly improve flow
 *	control reliability.
 */
	if (tiosp->c_iflag & IXON) 
	{
		mr0 |= MR0_SWFTX | MR0_SWFT;
		imron |= IR_XONXOFF;
	} 
	else 
		imroff |= IR_XONXOFF;

	if (tiosp->c_iflag & IXOFF)
		mr0 |= MR0_SWFRX;

	if (tiosp->c_cflag & CRTSCTS) 
	{
		mr2 |= MR2_AUTOCTS;
		mr1 |= MR1_AUTORTS;
	}

/*
 *	All sc26198 register values calculated so go through and set
 *	them all up.
 */

	debug_trace(5, "sc26x_setport: chipnr=0x%X portnr=%d\n",
		(int)chipp->ioaddr1_phys, portp->portnr);
	debug_trace(5, "    mr0=0x%02X mr1=0x%02X mr2=0x%02X clk=0x%02X\n", 
        mr0, mr1, mr2, clk);
	debug_trace(5, "    iopr=0x%02X imron=0x%02X imroff=0x%02X\n", 
        iopr, imron, imroff);
	debug_trace(5, "    schr1=0x%02X schr2=0x%02X schr3=0x%02X schr4=0x%02X\n",
		tiosp->c_cc[VSTART], tiosp->c_cc[VSTOP],
		tiosp->c_cc[VSTART], tiosp->c_cc[VSTOP]);

	save_flags(flags);
	cli();
	sc26x_setreg(portp, IMR, 0);
	sc26x_updatereg(portp, MR0, mr0);
	sc26x_updatereg(portp, MR1, mr1);
	sc26x_setreg(portp, SCCR, CR_RXERRBLOCK);
	sc26x_updatereg(portp, MR2, mr2);
	sc26x_updatereg(portp, IOPIOR,
		((sc26x_getreg(portp, IOPIOR) & ~IPR_CHANGEMASK) | iopr));

	if (baudrate > 0) 
	{
		sc26x_setreg(portp, TXCSR, clk);
		sc26x_setreg(portp, RXCSR, clk);
	}

	sc26x_setreg(portp, XONCR, tiosp->c_cc[VSTART]);
	sc26x_setreg(portp, XOFFCR, tiosp->c_cc[VSTOP]);

	ipr = sc26x_getreg(portp, IPR);
	if (ipr & IPR_DCD)
		portp->sigs &= ~TIOCM_CD;
	else
		portp->sigs |= TIOCM_CD;

	portp->imr = (portp->imr & ~imroff) | imron;
	sc26x_setreg(portp, IMR, portp->imr);
	debug_trace(8, "port %x IMR: %x\n", portp->portnr, sc26x_getreg(portp, IMR)); 
	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Set the state of the DTR and RTS signals.
 */

static void 
sc26x_setsignals(sc26x_port_t *portp, int dtr, int rts)
{
	unsigned char	iopioron, iopioroff;
	unsigned long	flags;

	debug_trace(5, "sc26x_setsignals(portp=%x,dtr=%d,rts=%d)\n",
		(int) portp, dtr, rts);

	iopioron = 0;
	iopioroff = 0;
	if (dtr == 0)
		iopioroff |= IPR_DTR;
	else if (dtr > 0)
		iopioron |= IPR_DTR;
	if (rts == 0)
		iopioroff |= IPR_RTS;
	else if (rts > 0)
		iopioron |= IPR_RTS;

	save_flags(flags);
	cli();
	sc26x_setreg(portp, IOPIOR,
		((sc26x_getreg(portp, IOPIOR) & ~iopioroff) | iopioron));
	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Return the state of the signals.
 */

static int 
sc26x_getsignals(sc26x_port_t *portp)
{
	unsigned char	ipr;
	unsigned long	flags;
	int		sigs;

	debug_trace(5, "sc26x_getsignals(portp=%x)\n", (int) portp);

	save_flags(flags);
	cli();
	ipr = sc26x_getreg(portp, IPR);
	restore_flags(flags);

	sigs = 0;
	sigs |= (ipr & IPR_DCD) ? 0 : TIOCM_CD;
	sigs |= (ipr & IPR_CTS) ? 0 : TIOCM_CTS;
	sigs |= (ipr & IPR_DTR) ? 0: TIOCM_DTR;
	sigs |= (ipr & IPR_RTS) ? 0: TIOCM_RTS;
	sigs |= TIOCM_DSR;
	return sigs;
}

/*****************************************************************************/

/*
 *	Enable/Disable the Transmitter and/or Receiver.
 */

static void 
sc26x_enablerxtx(sc26x_port_t *portp, int rx, int tx)
{
	unsigned char	ccr;
	unsigned long	flags;

	debug_trace(5, "sc26x_enablerxtx(portp=%x,rx=%d,tx=%d)\n",
		(int) portp, rx, tx);

	ccr = portp->crenable;
	if (tx == 0)
		ccr &= ~CR_TXENABLE;
	else if (tx > 0)
		ccr |= CR_TXENABLE;
	if (rx == 0)
		ccr &= ~CR_RXENABLE;
	else if (rx > 0)
		ccr |= CR_RXENABLE;

	save_flags(flags);
	cli();
	sc26x_setreg(portp, SCCR, ccr);
	portp->crenable = ccr;
	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Start/stop the Transmitter and/or Receiver.
 */

static void 
sc26x_startrxtx(sc26x_port_t *portp, int rx, int tx)
{
	unsigned long	flags;

	debug_trace(5, "sc26x_startrxtx(portp=%x,rx=%d,tx=%d)\n",
		(int) portp, rx, tx);

	if (tx == 0)
		portp->imr &= ~IR_TXRDY;
	else if (tx == 1)
		portp->imr |= IR_TXRDY;
	if (rx == 0)
		portp->imr &= ~(IR_RXRDY | IR_RXBREAK | IR_RXWATCHDOG);
	else if (rx > 0)
		portp->imr |= IR_RXRDY | IR_RXBREAK | IR_RXWATCHDOG;

	save_flags(flags);
	cli();
	sc26x_setreg(portp, IMR, portp->imr);
	debug_trace(5, "IMR is 0x%x\n", sc26x_getreg(portp, IMR));
	if (tx > 0)
		set_bit(ASYI_TXBUSY, &portp->istate);
	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Disable all interrupts from this port.
 */

static void 
sc26x_disableintrs(sc26x_port_t *portp)
{
	unsigned long	flags;

	debug_trace(5, "sc26x_disableintrs(portp=%x)\n", (int) portp);

	save_flags(flags);
	cli();
	portp->imr = 0;
	sc26x_setreg(portp, IMR, 0);
	restore_flags(flags);
}

/*****************************************************************************/

static void 
sc26x_sendbreak(sc26x_port_t *portp, int len)
{
	unsigned long	flags;

	debug_trace(5, "sc26x_sendbreak(portp=%x,len=%d)\n", (int) portp, len);

	save_flags(flags);
	cli();

	if (len == 1) 
	{
		sc26x_setreg(portp, SCCR, CR_TXSTARTBREAK);
		portp->stats.txbreaks++;
	} 
	else
		sc26x_setreg(portp, SCCR, CR_TXSTOPBREAK);

	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Take flow control actions...
 */

static void 
sc26x_flowctrl(sc26x_port_t *portp, int state)
{
	struct tty_struct	*tty;
	unsigned long		flags;
	unsigned char		mr0;

	debug_trace(5, "sc26x_flowctrl(portp=%x,state=%x)\n", (int) portp, state);

	if (!portp)
		return;
	tty = portp->tty;
	if (!tty)
		return;

	save_flags(flags);
	cli();

	if (state) 
	{
		if (tty->termios->c_iflag & IXOFF) 
		{
			mr0 = sc26x_getreg(portp, MR0);
			sc26x_setreg(portp, MR0, (mr0 & ~MR0_SWFRXTX));
			sc26x_setreg(portp, SCCR, CR_TXSENDXON);
			mr0 |= MR0_SWFRX;
			portp->stats.rxxon++;
			sc26x_wait(portp);
			sc26x_setreg(portp, MR0, mr0);
		}

		if (tty->termios->c_cflag & CRTSCTS) 
		{
			sc26x_setreg(portp, MR1,
				(sc26x_getreg(portp, MR1) | MR1_AUTORTS));
			sc26x_setreg(portp, IOPIOR,
				(sc26x_getreg(portp, IOPIOR) | IOPR_RTS));
			portp->stats.rxrtson++;
		}
	} 
	else 
	{
		if (tty->termios->c_iflag & IXOFF) 
		{
			mr0 = sc26x_getreg(portp, MR0);
			sc26x_setreg(portp, MR0, (mr0 & ~MR0_SWFRXTX));
			sc26x_setreg(portp, SCCR, CR_TXSENDXOFF);
			mr0 &= ~MR0_SWFRX;
			portp->stats.rxxoff++;
			sc26x_wait(portp);
			sc26x_setreg(portp, MR0, mr0);
		}
		if (tty->termios->c_cflag & CRTSCTS) 
		{
			sc26x_setreg(portp, MR1,
				(sc26x_getreg(portp, MR1) & ~MR1_AUTORTS));
			sc26x_setreg(portp, IOPIOR,
				(sc26x_getreg(portp, IOPIOR) & ~IOPR_RTS));
			portp->stats.rxrtsoff++;
		}
	}

	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Send a flow control character.
 */

static void 
sc26x_sendflow(sc26x_port_t *portp, int state)
{
	struct tty_struct	*tty;
	unsigned long		flags;
	unsigned char		mr0;

	debug_trace(5, "sc26x_sendflow(portp=%x,state=%x)\n", (int) portp, state);

	if (portp == (sc26x_port_t *) NULL)
		return;
	tty = portp->tty;
	if (tty == (struct tty_struct *) NULL)
		return;

	save_flags(flags);
	cli();
	if (state) {
		mr0 = sc26x_getreg(portp, MR0);
		sc26x_setreg(portp, MR0, (mr0 & ~MR0_SWFRXTX));
		sc26x_setreg(portp, SCCR, CR_TXSENDXON);
		mr0 |= MR0_SWFRX;
		portp->stats.rxxon++;
		sc26x_wait(portp);
		sc26x_setreg(portp, MR0, mr0);
	} else {
		mr0 = sc26x_getreg(portp, MR0);
		sc26x_setreg(portp, MR0, (mr0 & ~MR0_SWFRXTX));
		sc26x_setreg(portp, SCCR, CR_TXSENDXOFF);
		mr0 &= ~MR0_SWFRX;
		portp->stats.rxxoff++;
		sc26x_wait(portp);
		sc26x_setreg(portp, MR0, mr0);
	}
	restore_flags(flags);
}

/*****************************************************************************/

static void 
sc26x_flush(sc26x_port_t *portp)
{
	unsigned long	flags;

	debug_trace(5, "sc26x_flush(portp=%x)\n", (int) portp);

	if (portp == (sc26x_port_t *) NULL)
		return;

	save_flags(flags);
	cli();
	sc26x_setreg(portp, SCCR, CR_TXRESET);
	sc26x_setreg(portp, SCCR, portp->crenable);
	portp->tx.tail = portp->tx.head;
	restore_flags(flags);
}

/*****************************************************************************/

/*
 *	Return the current state of data flow on this port. This is only
 *	really interresting when determining if data has fully completed
 *	transmission or not... The sc26198 interrupt scheme cannot
 *	determine when all data has actually drained, so we need to
 *	check the port statusy register to be sure.
 */

static int 
sc26x_datastate(sc26x_port_t *portp)
{
	unsigned long	flags;
	unsigned char	sr;

	debug_trace(5, "sc26x_datastate(portp=%x)\n", (int) portp);

	if (portp == (sc26x_port_t *) NULL)
		return(0);
	if (test_bit(ASYI_TXBUSY, &portp->istate))
		return(1);

	save_flags(flags);
	cli();
	sr = sc26x_getreg(portp, SR);
	restore_flags(flags);
	debug_trace (8,"ISR: 0x%x\n", sc26x_getreg(portp,ISR));
	return (sr & SR_TXEMPTY) ? 0 : 1;
}

/*****************************************************************************/

/*
 *	Delay for a small amount of time, to give the sc26198 a chance
 *	to process a command...
 */

static void 
sc26x_wait(sc26x_port_t *portp)
{
	int	i;

	debug_trace(5, "sc26x_wait(portp=%x)\n", (int) portp);

	if (portp == 0)
		return;

	for (i = 0; i < 20; i++)
		sc26x_getglobreg(portp->ioaddr, TSTR);
}

/*****************************************************************************/

/*
 *	If we are TX flow controlled and in IXANY mode then we may
 *	need to unflow control here. We gotta do this because of the
 *	automatic flow control modes of the sc26198.
 */

static void 
sc26x_txunflow(sc26x_port_t *portp, struct tty_struct *tty)
{
	unsigned char	mr0;

	mr0 = sc26x_getreg(portp, MR0);
	sc26x_setreg(portp, MR0, (mr0 & ~MR0_SWFRXTX));
	sc26x_setreg(portp, SCCR, CR_HOSTXON);
	sc26x_wait(portp);
	sc26x_setreg(portp, MR0, mr0);
	clear_bit(ASYI_TXFLOWED, &portp->istate);
}

/*****************************************************************************/

/*
 *	Transmit interrupt handler. This has gotta be fast!  Handling TX
 *	chars is pretty simple, stuff as many as possible from the TX buffer
 *	into the sc26198 FIFO.
 *	In practice it is possible that interrupts are enabled but that the
 *	port has been hung up. Need to handle not having any TX buffer here,
 *	this is done by using the side effect that head and tail will also
 *	be NULL if the buffer has been freed.
 */

static void 
sc26x_txisr(sc26x_port_t *portp)
{
	unsigned char	mr0;
	int		len, stlen;
	char		*head, *tail;

	debug_trace(8, "sc26x_txisr(portp=%x)\n", (int) portp);

	head = portp->tx.head;
	tail = portp->tx.tail;
	len = (head >= tail) ? (head - tail) : (SC26X_TXBUFSIZE - (tail - head));
	if ((len == 0) || ((len < SC26X_TXBUFLOW) &&
	    (test_bit(ASYI_TXLOW, &portp->istate) == 0))) {
		set_bit(ASYI_TXLOW, &portp->istate);
		schedule_task(&portp->tqueue);
	}

	if (len == 0) {
		mr0 = sc26x_getreg(portp, MR0);
		if ((mr0 & MR0_TXMASK) == MR0_TXEMPTY) {
			portp->imr &= ~IR_TXRDY;
			sc26x_setreg(portp, IMR, portp->imr);
			clear_bit(ASYI_TXBUSY, &portp->istate);
		} else {
			mr0 = ((mr0 & ~MR0_TXMASK) | MR0_TXEMPTY);
			sc26x_updatereg(portp, MR0, mr0);
		}
	} else {
		len = MIN(len, SC26198_TXFIFOSIZE);
		portp->stats.txtotal += len;
		stlen = MIN(len, ((portp->tx.buf + SC26X_TXBUFSIZE) - tail));
		sc26x_outsb(portp->ioaddr, GTXFIFO, tail, stlen);
		len -= stlen;
		tail += stlen;
		if (tail >= (portp->tx.buf + SC26X_TXBUFSIZE))
		{
			tail = portp->tx.buf;
		}
		if (len > 0) {
			sc26x_outsb(portp->ioaddr, GTXFIFO, tail, len);
			tail += len;
		}
		portp->tx.tail = tail;
	}
}

/*****************************************************************************/

/*
 *	Receive character interrupt handler. Determine if we have good chars
 *	or bad chars and then process appropriately. Good chars are easy
 *	just shove the lot into the RX buffer and set all status byte to 0.
 *	If a bad RX char then process as required. This routine needs to be
 *	fast!  In practice it is possible that we get an interrupt on a port
 *	that is closed. This can happen on hangups - since they completely
 *	shutdown a port not in user context. Need to handle this case.
 */

static void 
sc26x_rxisr(sc26x_port_t *portp, unsigned int cir)
{
	struct tty_struct	*tty;
	unsigned int		len, buflen;

	debug_trace(8, "sc26x_rxisr(portp=%x,cir=%x)\n", (int) portp, cir);

	tty = portp->tty;
	len = sc26x_getglobreg(portp->ioaddr, GIBCR) + 1;

	if ((cir & CIR_TYPEMASK) == CIR_RXDATA) 
	{
		if (!tty || !tty->flip.char_buf_ptr ||
		    (buflen = TTY_FLIPBUF_SIZE - tty->flip.count) <= 0) 
		{
			sc26x_insb(portp->ioaddr, GRXFIFO, &sc26x_unwanted[0], len);
			portp->stats.rxlost += len;
			portp->stats.rxtotal += len;
		} 
		else 
		{
			len = MIN(len, buflen);
			if (len > 0) 
			{
				sc26x_insb(portp->ioaddr, GRXFIFO, tty->flip.char_buf_ptr, len);

				memset(tty->flip.flag_buf_ptr, 0, len);
				tty->flip.flag_buf_ptr += len;
				tty->flip.char_buf_ptr += len;
				tty->flip.count += len;
				tty_schedule_flip(tty);
				portp->stats.rxtotal += len;
			}
		}
	} 
	else 
		sc26x_rxbadchars(portp);

/*
 *	If we are TX flow controlled and in IXANY mode then we may need
 *	to unflow control here. We gotta do this because of the automatic
 *	flow control modes of the sc26198.
 */
	if (test_bit(ASYI_TXFLOWED, &portp->istate)) 
	{
		if (tty && tty->termios && (tty->termios->c_iflag & IXANY))
			sc26x_txunflow(portp, tty);
	}
}

/*****************************************************************************/

/*
 *	Process an RX bad character.
 */

static void 
sc26x_rxbadch(sc26x_port_t *portp, unsigned char status, char ch)
{
	struct tty_struct	*tty;

	tty = portp->tty;

	if (status & SR_RXPARITY)
		portp->stats.rxparity++;
	if (status & SR_RXFRAMING)
		portp->stats.rxframing++;
	if (status & SR_RXOVERRUN)
		portp->stats.rxoverrun++;
	if (status & SR_RXBREAK)
		portp->stats.rxbreaks++;

	if (tty && (portp->rxignoremsk & status) == 0) 
	{
		if (portp->rxmarkmsk & status) 
		{
			if (status & SR_RXBREAK) 
			{
				status = TTY_BREAK;
				if (portp->flags & ASYNC_SAK) 
					do_SAK(tty);
			} 
			else if (status & SR_RXPARITY) 
				status = TTY_PARITY;
			else if (status & SR_RXFRAMING)
				status = TTY_FRAME;
			else if(status & SR_RXOVERRUN)
				status = TTY_OVERRUN;
			else
				status = 0;
		} 
		else
			status = 0;

		if (tty->flip.char_buf_ptr) 
		{
			if (tty->flip.count < TTY_FLIPBUF_SIZE) 
			{
				*tty->flip.flag_buf_ptr++ = status;
				*tty->flip.char_buf_ptr++ = ch;
				tty->flip.count++;
			}
			tty_schedule_flip(tty);
		}

		if (status == 0)
			portp->stats.rxtotal++;
	}
}

/*****************************************************************************/

/*
 *	Process all characters in the RX FIFO of the UART. Check all char
 *	status bytes as well, and process as required. We need to check
 *	all bytes in the FIFO, in case some more enter the FIFO while we
 *	are here. To get the exact character error type we need to switch
 *	into CHAR error mode (that is why we need to make sure we empty
 *	the FIFO).
 */

static void 
sc26x_rxbadchars(sc26x_port_t *portp)
{
	unsigned char	status, mr1;
	char		ch;

/*
 *	To get the precise error type for each character we must switch
 *	back into CHAR error mode.
 */
	mr1 = sc26x_getreg(portp, MR1);
	sc26x_setreg(portp, MR1, (mr1 & ~MR1_ERRBLOCK));

	while ((status = sc26x_getreg(portp, SR)) & SR_RXRDY) 
	{
		sc26x_setreg(portp, SCCR, CR_CLEARRXERR);
		ch = sc26x_getreg(portp, RXFIFO);
		sc26x_rxbadch(portp, status, ch);
	}

/*
 *	To get correct interrupt class we must switch back into BLOCK
 *	error mode.
 */
	sc26x_setreg(portp, MR1, mr1);
}

/*****************************************************************************/

/*
 *	Other interrupt handler. This includes modem signals, flow
 *	control actions, etc. Most stuff is left to off-level interrupt
 *	processing time.
 */

static void 
sc26x_otherisr(sc26x_port_t *portp, unsigned int cir)
{
	unsigned char	ipr, xisr;

	debug_trace(2, "sc26x_otherisr(portp=%x,cir=%x)\n", (int) portp, cir);

	switch (cir & CIR_SUBTYPEMASK) 
	{
	case CIR_SUBCOS:
		ipr = sc26x_getreg(portp, IPR);
		if (ipr & IPR_DCDCHANGE) 
		{
			set_bit(ASYI_DCDCHANGE, &portp->istate);
			schedule_task(&portp->tqueue);
			portp->stats.modem++;
		}
		break;
	case CIR_SUBXONXOFF:
		xisr = sc26x_getreg(portp, XISR);
		if (xisr & XISR_RXXONGOT) 
		{
			set_bit(ASYI_TXFLOWED, &portp->istate);
			portp->stats.txxoff++;
		}
		if (xisr & XISR_RXXOFFGOT) 
		{
			clear_bit(ASYI_TXFLOWED, &portp->istate);
			portp->stats.txxon++;
		}
		break;
	case CIR_SUBBREAK:
		sc26x_setreg(portp, SCCR, CR_BREAKRESET);
		sc26x_rxbadchars(portp);
		break;
	default:
		break;
	}
}

/*****************************************************************************/

#if 0

void
dump_buf(char *b, int len)
{   
    int width = 0x10;
    int start_addr, end_addr, addr;

    start_addr = (((int)b) / width) * width;
    end_addr = ((((int)b + len - 1) / width) + 1) * width;

    debug_trace(8, "b=0x%X len=%d start_addr=0x%X end_addr=0x%X\n", 
        (int)b, len, start_addr, end_addr);

    for (addr = start_addr; addr < end_addr; )
    {   
        static char hex_buf[128];
        static char ascii_buf[128];
        int j, hex_buf_o, ascii_buf_o;

        hex_buf_o = sprintf(hex_buf, "%08X ", addr);
        ascii_buf_o = 0;

        for (j = 0; j < width; j++, addr++)
            if (addr < (int)b || addr >= (int)b + len)
            {   
                hex_buf_o += sprintf(hex_buf + hex_buf_o, "   ");
                ascii_buf_o += sprintf(ascii_buf + ascii_buf_o, " ");
            }
            else
            {
                int v = *(unsigned char *)addr;
		hex_buf_o += sprintf(hex_buf + hex_buf_o, "%02X ", v);
		ascii_buf_o += sprintf(ascii_buf + ascii_buf_o, "%c",
					isalnum(v) ?  v : '.');
            }

        debug_trace(0, "%s%s\n", hex_buf, ascii_buf);
    }
}

#endif

