/*
 * Generic SRAM driver.
 * Copyright (C) 2004 DENX Software Engineering, Wolfgang Denk, wd@denx.de
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/errno.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#ifdef CONFIG_NC650
#include <platforms/nc650.h>
#endif
#ifdef CONFIG_CPC45
#include <platforms/cpc45.h>
#endif
#ifdef CONFIG_INKA4X0
#include <platforms/inka4x0.h>
#endif
#ifdef CONFIG_TQM5200
#include <platforms/tqm5200.h>
#endif

#define SRAM_MAJOR	32

static ulong sram_start_virt;

#define SRAM_PHYS	CFG_SRAM_PADDR
#define SRAM_START	sram_start_virt
#define SRAM_SIZE	CFG_SRAM_SIZE
#define SRAM_END	(SRAM_START + SRAM_SIZE)

static ssize_t sram_read(struct file * file, char * buf,
			size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;

	if (p >= SRAM_SIZE)
		return 0;

	if (count > SRAM_SIZE - p)
		count = SRAM_SIZE - p;

	p += SRAM_START;

	if (copy_to_user(buf, (void*) p, count))
		return -EFAULT;

	*ppos += count;

	return count;
}

static ssize_t sram_write(struct file * file, const char * buf,
			 size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;

	if (p >= SRAM_SIZE)
		return 0;

	if (count > SRAM_SIZE - p)
		count = SRAM_SIZE - p;

	p += SRAM_START;

	if (copy_from_user((void*) p, buf, count))
		return -EFAULT;

	*ppos += count;

	return count;
}

static int sram_mmap(struct file * file, struct vm_area_struct * vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long size = vma->vm_end - vma->vm_start;

	if (offset + size > SRAM_SIZE)
		return -EINVAL;

	offset += SRAM_PHYS;

	if (remap_page_range(vma->vm_start, offset, size, vma->vm_page_prot))
		return -EAGAIN;

	return 0;
}

static loff_t sram_lseek(struct file * file, loff_t offset, int orig)
{
	switch (orig) {
		case 0:
			file->f_pos = offset;
			return file->f_pos;
		case 1:
			file->f_pos += offset;
			return file->f_pos;
		case 2:
			file->f_pos = SRAM_SIZE - offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
}

static int sram_open(struct inode * inode, struct file * filp)
{
	return capable(CAP_SYS_RAWIO) ? 0 : -EPERM;
}

static struct file_operations sram_fops = {
	llseek:		sram_lseek,
	read:		sram_read,
	write:		sram_write,
	mmap:		sram_mmap,
	open:		sram_open,
};

int __init sram_drv_init(void)
{
	if ((sram_start_virt = (ulong)ioremap (SRAM_PHYS, SRAM_SIZE)) == 0)
	{
		printk (KERN_ERR "sram_drv: ioremap(%#x) failed\n", SRAM_PHYS);
		return -EIO;
	}
	if (register_chrdev (SRAM_MAJOR, "sram", & sram_fops))
	{
		printk (KERN_ERR "sram_drv: register_chrdev failed\n");
		return -EIO;
	}
  printk("SRAM_DRV: initialized\n");
	return 0;
}

/******************************
 **** Module Declarations *****
 **************************** */

module_init (sram_drv_init);

#ifdef MODULE

static void sram_drv_cleanup (void)
{
	iounmap (sram_start_virt);
	if (unregister_chrdev (SRAM_MAJOR, "sram"))
	{
		printk (KERN_ERR "sram_drv: failed to unregister character device\n");
	}
}

module_exit (sram_drv_cleanup);

#endif	/* MODULE */

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Wolfgang Denk <wd@denx.de>");
MODULE_DESCRIPTION("SRAM driver");
