/*
 * (C) Copyright 2005 DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * Device driver for ADC12048 Analog-Digital-Converter
 * on TQM823L-based VIRTLAB2 board
 *
 * 01/05:  Created, <agust7@web.de>
 *
 * $Id: vl_adc12048.c 1824 2005-05-10 07:30:07Z hg $
 */

#include <linux/kernel.h>	/* We're doing kernel work	*/
#include <linux/module.h>	/* Specifically, a module	*/
#include <asm/uaccess.h>	/* for put_user			*/
#include <linux/init.h>		/* for __initfunc		*/
#include <asm/io.h>		/* for ioremap_nocache		*/
#include <linux/delay.h>
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>

#if CONFIG_MODVERSIONS==1
# include <linux/modversions.h>
#endif
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/proc_fs.h>	/* proc definitions */
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */

#include <linux/adc12048.h>

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	((a)*65536+(b)*256+(c))
#endif

#if (defined(MODULE) && defined(DEBUG))
MODULE_PARM(dbg_level, "i");
MODULE_PARM_DESC(dbg_level, "Debug trace level (0-3)");
#endif

#ifdef	DEBUG

static int dbg_level;

#define dbg(dbg_lvl,fmt,args...)					\
	do {								\
		if (dbg_level >= dbg_lvl)				\
			printk(KERN_DEBUG "" 				\
			__FILE__ ": "__MODULE_STRING(__LINE__) ": "	\
			fmt ,##args);					\
	} while (0)
#else
#define dbg(dbg_lvl,fmt,args...)
#endif /* DEBUG */

/*
 * The name of our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME		"vl_adc"

#ifndef	VL2_ADC12048_MAJOR
#define VL2_ADC12048_MAJOR	240	/* free for local/experimental use */
#endif	/* VL_ADC12048_MAJOR */

/*
 * Hardware Settings
 */
#define	VL2_ADC12048_PHYSBASE	0xC0000000	/* physical base address of adc port */
#define	VL2_ADC12048_SIZE	0x1000		/* map 1 page of memory	*/
#define CS4_PHYSBASE		VL2_ADC12048_PHYSBASE

#define MEMC_OR4_VALUE		0xFFFF8924		/* see 16-10 ff	*/
#define MEMC_BR4_VALUE		(CS4_PHYSBASE | 0x401)	/* 8-bit bus, valid bit */

#define IO_PORT_OFF		0x00
#define IO_PORT_ON		0x01


typedef struct	vl_adc12048_port {
/*	u_char	offset[]; */
	u_char	adc0;			/* port0 - offset 0x000	*/
} adc12048_port_t;

static volatile adc12048_port_t *adc_port = NULL;

/*
 * prevent concurent access of the same device
 */
static int adc_used = 0;

static int adc_major;

static int channel;

/*
 * Prototypes
 */
int __init vl_adc_init (void);
static int vl_adc_open(struct inode *, struct file *);
static int vl_adc_release(struct inode *, struct file *);
static ssize_t vl_adc_read(struct file *, char *, size_t, loff_t *);
/* static ssize_t vl_adc_write(struct file *, const char *, size_t, loff_t *); */
static int vl_adc_ioctl(struct inode*, struct file*, unsigned int, unsigned long);

struct file_operations vl_adc_ops =
{
	owner:		THIS_MODULE,
	read:		vl_adc_read,
	open:		vl_adc_open,
	release:	vl_adc_release,
	ioctl:		vl_adc_ioctl,
};


/*
 * Toggle I/O-Pins as specified by flag arg.
 * Be careful here, don't touch neighbours!
 */
void toggle_io_port(int flag)
{
	cpm8xx_t        *cpm = &((immap_t *)IMAP_ADDR)->im_cpm;
	iop8xx_t        *iop = &((immap_t *)IMAP_ADDR)->im_ioport;
	volatile u32	*pb = NULL; //&cpm->cp_pbdir;
	volatile u16	*pc = NULL; //&iop->iop_pcdir;

	dbg (1, "PBDIR %04x, PBPAR %04x, PBODR %04x, PBDAT %04x\n",
		cpm->cp_pbdir, cpm->cp_pbpar, cpm->cp_pbodr, cpm->cp_pbdat);
	dbg (1, "PCDIR %04x, PCPAR %04x, PCSO %04x, PCDAT %04x, PCINT %04x\n",
		iop->iop_pcdir,	iop->iop_pcpar, iop->iop_pcso,
		iop->iop_pcdat, iop->iop_pcint);

	switch(flag) {
	case IO_PORT_ON:
		/* configure I/O port */
		pb = &cpm->cp_pbdir;
		pc = &iop->iop_pcdir;
		*pb |= 0x8330;	/* PB16, PB22, PB23, PB26, PB27 as output */	
		*pc |= 0x0C06;	/* PC4, PC5, PC13, PC14 as output */
	
		/* "warm up" temperature sensors */
		pb = &cpm->cp_pbdat;
		pc = &iop->iop_pcdat;
		*pb |= 0x8330;
		*pc |= 0x0C06;
		break;
	case IO_PORT_OFF:
		/* "shutdown" temperature sensors */
		pb = &cpm->cp_pbdat;
		pc = &iop->iop_pcdat;
		*pb &= 0x7CCF;
		*pc &= 0xF3F9;
	
		/* reconfigure I/O port  */
		pb = &cpm->cp_pbdir;
		pc = &iop->iop_pcdir;
		*pb &= 0x7CCF;	/* PB16, PB22, PB23, PB26, PB27 as input */	
		*pc &= 0xF3F9;	/* PC4, PC5, PC13, PC14 as input */
		break;
	default:
		break;
	}
	
	dbg (1, "PBDIR %04x, PBPAR %04x, PBODR %04x, PBDAT %04x\n",
		cpm->cp_pbdir, cpm->cp_pbpar, cpm->cp_pbodr, cpm->cp_pbdat);
	dbg (1, "PCDIR %04x, PCPAR %04x, PCSO %04x, PCDAT %04x, PCINT %04x\n",
		iop->iop_pcdir,	iop->iop_pcpar, iop->iop_pcso,
		iop->iop_pcdat, iop->iop_pcint);
}

/*
 * Issue an adc command.
 * Use the set ack flag (third arg) to wait for adc response.
 * Only works after successful ioremap!
 */
int adc_cmd(int upb, int lowb, int ack)
{
	volatile u16 *pcdat =  &((immap_t *)IMAP_ADDR)->im_ioport.iop_pcdat;
	int timeout = jiffies + HZ/20;

	dbg (3, "cmd: upb: %x, lowb: %x, ack: %d\n", upb, lowb, ack);

	adc_port->adc0 = (u8)lowb;

	if (upb >= 0 && upb < 0x80) {
		dbg (3, "setting upper byte to %x\n", upb);
		adc_port->adc0 = (u8)upb;
	}

	if (!ack)
		return 0;

	while ((*pcdat & 0x0001)) {
		dbg (3, "adc_cmd: PCDAT 0x%x\n", *pcdat);
		if (jiffies > timeout)
			return -EBUSY;
	}
	return 0;
}

/*
 * Select ADC channel and start conversion,
 * busy wait until conversion is done.
 * Return negative value if invalid channel
 * specified, or timeout condition.
 */
int adc_read_channel(int chan, short *result)
{
	volatile u16 *pcdat =  &((immap_t *)IMAP_ADDR)->im_ioport.iop_pcdat;
	volatile u32 *pbdat =  &((immap_t *)IMAP_ADDR)->im_cpm.cp_pbdat;
	int timeout = jiffies + HZ/20;
	int cnt = 10;

	dbg (3, "read from CH%d\n", chan);

	if (chan < 0 || chan > 7)
		return -ENXIO;

	if (chan != channel) {
		adc_cmd(-1, (ADC_MUX_CH0_COM + chan), 0);	/* select asked channel */
		channel = chan;
	}
	
	*pbdat |= 0x8000;	/* rising edge on PB16 starts conversion */
	while (cnt--);		/* SYNC time */	
	*pbdat &= 0x7fff;	/* drive PB16 low again */

	/* wait until conversion is done */
	while ((*pcdat & 0x0001)) {
		if (jiffies > timeout) {
			dbg (3, "timeout; PCDAT: %x\n", *pcdat);
			return -EBUSY;
		}
	}

	*result = adc_port->adc0;
	*result |= adc_port->adc0 << 8;
	return 0;
}

#ifdef CONFIG_PROC_FS
int adc_read_procmem(char *buf, char **start, off_t offset,
		     int count, int *eof, void *data)
{
	int i, len = 0, ret;
	short adc_data[8];

	dbg (2, "/proc: buf %p, *start %p, offset %i, count %i, *eof %d, data %p\n",
			buf, *start, (int)offset, count, *eof, data);
	
	if (offset == 49) {
		*eof = 1;
		return 0;	/* already got all you need */
	}

	for (i = 0; i < 8; i++) {
		ret = adc_read_channel(i, &adc_data[i]);
		if (ret < 0) {
			printk(DEVICE_NAME ": proc read CH%d failed: %d\n", i, ret);
			if (i) i--;
			break;
		}
		dbg (1, "adc CH%d value %d\n", i, adc_data[i]);
		len += sprintf(buf+len,"% 5d ", adc_data[i]);
	}

	len += sprintf(buf+len, "\n");

	*eof = 1;
	
	return len;
}
#endif /* CONFIG_PROC_FS */


/*
 * Initialize the driver, register us as a character device
 */
int __init vl_adc_init (void)
{
	memctl8xx_t     *memc = &((immap_t *)IMAP_ADDR)->im_memctl;
	int		ret = 0;
        
	/* //already done in tqm8xx_setup
	memc->memc_or4 = 0xFFFF8924;
	memc->memc_br4 = CS4_PHYSBASE | 0x401;
	 */

	dbg (3, "BR4 %x, OR4 %x\n", memc->memc_br4, memc->memc_or4);
	
	/*
	 * Remap I/O space into kernel memory
	 */
	adc_port = (adc12048_port_t *)ioremap_nocache (VL2_ADC12048_PHYSBASE,
							VL2_ADC12048_SIZE);
	if (adc_port == NULL) {
		printk (DEVICE_NAME ": unable to remap io memory\n");
		return -ENODEV;
	}
	
	adc_major = register_chrdev (VL2_ADC12048_MAJOR, DEVICE_NAME, &vl_adc_ops);
	
	if (adc_major < 0) {
		printk (DEVICE_NAME ": register_chrdev failed with %d\n", adc_major);
		iounmap ((void *)adc_port);
		return adc_major;
	}

	adc_major = VL2_ADC12048_MAJOR;

	/* Apply voltage to I/O-Port and prepare ADC */
	toggle_io_port(IO_PORT_ON);
	
	adc_cmd(ADC_CMD_RESET, ADC_HB, 0); //0xb8
	
	ret = adc_cmd(ADC_CMD_FULCAL, ADC_HB, 1);
	if (ret < 0)
		printk (DEVICE_NAME ": adc ful cal cmd timeout %d\n", ret);

	ret = adc_cmd(ADC_CMD_AUTOZERO, ADC_HB, 1);
	if (ret < 0)
		printk (DEVICE_NAME ": adc auto zero cmd timeout %d\n", ret);

	/* adc start (SYNC-IN mode) */
	ret = adc_cmd(ADC_CMD_START|ADC_SYNC_IN, ADC_HB|ADC_MUX_CH0_COM, 1);
	if (ret < 0)
		printk (DEVICE_NAME ": adc start cmd timeout %d\n", ret);

#ifdef CONFIG_PROC_FS
	create_proc_read_entry("driver/vl_adc", 0, NULL, adc_read_procmem, NULL);
#endif
	
	printk (DEVICE_NAME ": VIRTLAB2 A/D-Converter driver initialized, major: %d\n",
		adc_major);
	return 0;
}


/*
 * Called whenever a process attempts to open the device
 */
static int vl_adc_open (struct inode *inode, struct file *file)
{
	dbg (1, "vl_adc_open(%p,%p)\n", inode, file);

	/* exclusive open only */
	if (adc_used) {
		return -EBUSY;
	}

	/* only open for read is allowed */
	if((file->f_flags & O_ACCMODE) != O_RDONLY)
		return -EROFS;

	adc_used++;

	MOD_INC_USE_COUNT;
	
	return 0;
}


/*
 * Called when a process closes the device.
 */
static int vl_adc_release (struct inode *inode, struct file *file)
{

	dbg (1, "vl_adc_release(%p,%p)\n", inode, file);

	adc_used--;

	MOD_DEC_USE_COUNT;

	return 0;
}


/*
 * read entry point.
 * Read upto 8 channels (depends on length agrument) and
 * write array of short values to user.
 */
static ssize_t vl_adc_read (struct file *file,
			    char *buffer,	/* The buffer to fill with data	*/
			    size_t length,	/* The length of the buffer	*/
			    loff_t * offset)	/* Our offset in the file	*/
{
	short	adc_data[8];
	int	i, ret = 0;

	dbg (1, "read len=%d\n", length);

	if (length <= 0)
		return 0;
	
	if (length > 8)
		length = 8;

	for (i = 0; i < length; i++) {
		ret = adc_read_channel(i, &adc_data[i]);
		if (ret < 0) {
			printk(DEVICE_NAME ": CH%d read failed: %d\n", i, ret);
			if (i) i--;
			break;
		} else {
			dbg (1, "adc CH%d value %d\n", i, adc_data[i]);
		}
	}

	ret = copy_to_user (buffer, adc_data, i * sizeof(short));
	if (ret < 0) {
		printk(DEVICE_NAME ": copy_to_user failed %d\n", ret);
		return -EFAULT;
	}
	
	dbg (1, "actual read len=%d\n", i);

	return i;
}


/*
 * no write entry point for this driver
 *
static ssize_t vl_adc_write (struct file *file,
			     const char *buffer,
			     size_t length,
			     loff_t * offset)
{
	int ret;

	dbg (0, "write - len=%d\n", length);

	if (length < 1)
		return 0;

	return 0;
}
 */


/*
 * ioctl entry point.
 * Read only channels specified in channel bitmap given by user.
 * Write array of actually read values to user and update channel
 * bitmap appropriately signalling valid reads.
 */
int vl_adc_ioctl(struct inode* inode, struct file* file,
		  unsigned int cmd, unsigned long arg )
{
	int ret = 0;
	
	dbg (1, "act.cmd %#x, arg %#lx\n", cmd, arg);

	/* known command type? */
	if (_IOC_TYPE (cmd) != VL2_ADC_IOCTL_MGC)
		return -ENOTTY;

	if (_IOC_DIR (cmd) & _IOC_READ)
		ret = !access_ok (VERIFY_WRITE, (void *) arg, _IOC_SIZE (cmd));
	if (_IOC_DIR (cmd) & _IOC_WRITE)
		ret |= !access_ok (VERIFY_READ, (void *) arg, _IOC_SIZE (cmd));
	if (ret)
		return -EFAULT;

	switch (cmd) {
	case VL2_ADC_IOCTL_READ:
	{
		adc_read_rq_t req;
		short buf[8];
		int i;
		u8 mask = 0x01;

		if (copy_from_user(&req, (void *)arg, sizeof(req))) {
			return -EFAULT;
		}

		/* won't put kernel data to user! */
		memset(buf, 0, sizeof(buf));

		for (i = 0; i < 8; i++) {
			if (req.chan_bitmap & mask) {
				dbg (0, "ioctl read CH%d, mask 0x%x\n", i, mask);
				ret = adc_read_channel(i, &buf[i]);
				if (ret < 0) {
					req.chan_bitmap &= ~mask;
					buf[i] = 0;
				}
			}
			mask <<= 1;
		}
		/* update bitmap by user */
		if (put_user(req.chan_bitmap, (unsigned char*)arg))
			return -EFAULT;
		
		/* put adc data to user */
		if (copy_to_user(req.data_buf, buf, sizeof(buf))) {
			return -EFAULT;
		}
	}
	break;
	case VL2_ADC_IOCTL_ADCCMD:
		/*
		 * If you want to control ADC (reset, standby, calibration etc.)
		 * place appropriate code here.
		 */
		break;
	case VL2_ADC_IOCTL_IOPORT:
	/*
	 * Maybe you also want to be able to control I/O port states,
	 * so code the stuff here.
	 */
		break;
	default:
		return -ENOTTY;
		break;
	}
	
	return 0;
}

module_init(vl_adc_init);

#ifdef MODULE
void vl_adc_exit (void)
{
	int ret;

	toggle_io_port(IO_PORT_OFF);
	
	adc_cmd(ADC_CMD_STANDBY, ADC_HB, 0);

	iounmap ((void *)adc_port);

#ifdef CONFIG_PROC_FS
	remove_proc_entry("driver/vl_adc", NULL);
#endif
	ret = unregister_chrdev (adc_major, DEVICE_NAME);
	if (ret < 0) {
		printk ("unregister_chrdev: error %d\n", ret);
	}
}

module_exit(vl_adc_exit);

MODULE_AUTHOR("DENX Software Engineering");
MODULE_DESCRIPTION("ADC12048 A/D-Converter-Driver for TQM823L-based VirtLab2");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
#endif	/* MODULE */

