/***********************************************************************
 *
 * (C) Copyright 2000-2005
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * Device Driver for VIRTLAB-1 relay ports
 * - Modified for VIRTLAB-2 by dzu@denx.de  2004-08-26
 * - Some cleanups by agust7@web.de 2005-01-02
 *
 * Can be loaded as module
 *
 ***********************************************************************/

/*
 * Standard in kernel modules
 */
#include <linux/kernel.h>	/* We're doing kernel work	*/
#include <linux/module.h>	/* Specifically, a module	*/
#include <asm/uaccess.h>	/* for put_user			*/
#include <linux/init.h>		/* for __initfunc		*/
#include <asm/io.h>		/* for ioremap_nocache		*/
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>

#undef	DEBUG

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	((a)*65536+(b)*256+(c))
#endif

#define SUCCESS 0

#ifndef	VL_REL_MAJOR
#define	VL_REL_MAJOR	42	/* free for demo/sample use	*/
#endif	/* VL_REL_MAJOR */

/*
 * Hardware Settings
 */
#define	VL_REL_PHYSBASE	0xC0000000	/* physical base address of relay ports */
#define	VL_REL_SIZE	0x1000		/* map 1 page of memory			*/

//#define MEMC_BR4_VALUE (VL_REL_PHYSBASE | 0x401)	/* 8-bit PS, valid bit	*/
//#define MEMC_OR4_VALUE 0xFFFF8924			/* see 16-10 ff		*/

typedef struct	vl_rel_ports {
	u_char	reserved0[2048];
	u_char	rel0;			/* port0 - offset 0x800: rel 0...7	*/
} vl_rel_ports_t;

static volatile vl_rel_ports_t *ports = NULL;

/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME "vl_rel"

/*
 * prevent concurent access of the same device
 */
static int Device_Open = 0;

/*
 * Prototypes
 */
int __init vl_rel_init (void);
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);

struct file_operations vl_rel_ops =
{
	owner:		THIS_MODULE,
	read:		device_read,
	write:		device_write,
	open:		device_open,
	release:	device_release,
};

static int Major;

/*
 * Initialize the driver - Register the character device
 */
int __init vl_rel_init (void)
{
	/*
	 * Remap I/O space into kernel memory
	 */
	ports = (vl_rel_ports_t *)ioremap_nocache (VL_REL_PHYSBASE, VL_REL_SIZE);

	if (ports == NULL) {
		printk (DEVICE_NAME ": could not remap io memory\n");
		return -ENODEV;
	}

	/*
	 * Register the character device
	 */
	Major = register_chrdev (VL_REL_MAJOR, DEVICE_NAME, &vl_rel_ops);

	/* Negative values signify an error */
	if (Major < 0) {
		printk (DEVICE_NAME ": init_module failed with %d\n", Major);
		iounmap ((void *)ports);
		return Major;
	}

	Major = VL_REL_MAJOR;

	printk (DEVICE_NAME ": VIRTLAB2 Relays driver registered: major = %d\n", Major);
	return 0;
}

/*
 * called whenever a process attempts to open the device
 */
static int device_open (struct inode *inode, struct file *file)
{
	debugk (DEVICE_NAME " device_open(%p,%p)\n", inode, file);

	/*
	 * Get major / minor numbers when needed
	 */
	debugk (DEVICE_NAME " Device: %d.%d\n",
		inode->i_rdev >> 8, inode->i_rdev & 0xFF);

	/*
	 * exclusive open only
	 */
	if (Device_Open) {
		return -EBUSY;
	}

	/*
	 * Be careful here on SMP kernels!
	 */
	Device_Open++;

	/*
	 * Make sure that the module isn't removed while
	 * the file is open by incrementing the usage count
	 */
	MOD_INC_USE_COUNT;

	debugk (DEVICE_NAME ": open - port0=0x%x\n", ports->rel0);

	return SUCCESS;
}


/*
 * Called when a process closes the device.
 * Doesn't have a return value in version 2.0.x because it can't fail,
 * but in version 2.2.x it is allowed to fail
 */
static int device_release (struct inode *inode, struct file *file)
{
	debugk ("device_release(%p,%p)\n", inode, file);

	/* We're now ready for our next caller */
	Device_Open--;

	MOD_DEC_USE_COUNT;

	debugk (DEVICE_NAME ": release - port0=0x%x\n", ports->rel0);

	return 0;
}


/*
 * read entry point:
 */

static ssize_t device_read (struct file *file,
			    char *buffer,	/* The buffer to fill with data	*/
			    size_t length,	/* The length of the buffer	*/
			    loff_t * offset)	/* Our offset in the file	*/
{
	debugk (DEVICE_NAME ": read - len=%d, port0=0x%x\n",
		length, ports->rel0);

	if (length < 1)
		return (0);

	put_user (ports->rel0, buffer);

	return (1);		/* read() always returns normally 1 byte	*/
}


/*
 * write entry point:
 */
static ssize_t device_write (struct file *file,
			     const char *buffer,	/* buffer		*/
			     size_t length,		/* length of buffer	*/
			     loff_t * offset)		/* offset in the file	*/
{
	int error;
	unsigned char c;

	debugk (DEVICE_NAME ": write - len=%d, port0=0x%x\n",
		length, ports->rel0);

	if (length < 1)
		return (0);

	if ((error = get_user (c, buffer)) != 0) {
		return (error);
	}
	ports->rel0 = c;

	debugk (DEVICE_NAME ": write - port0=0x%x\n",
		ports->rel0);

	return (1);		/* write() uses (first) byte 			*/
}


/******************************
 **** Module Declarations *****
 **************************** */

module_init(vl_rel_init); /* also init if statically compiled into the kernel */

#ifdef MODULE

/*
 * Cleanup - unregister the appropriate device 
 */
void vl_rel_exit (void)
{
	int ret;

	/*
	 * Unmap device memory
	 */
	iounmap ((void *)ports);

	/*
	 * Unregister the device
	 */
	ret = unregister_chrdev (Major, DEVICE_NAME);

	/*
	 * If there's an error, report it
	 */
	if (ret < 0) {
		printk ("unregister_chrdev: error %d\n", ret);
	}
}

module_exit(vl_rel_exit);

MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
#endif	/* MODULE */
