/* drivers/mtd/ppcboot_env.c
 *
 * 2002-04-25: Initial Release
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/list.h>
#include <linux/errno.h>

#include <linux/mtd/mtd.h>

#include <asm/rwsem.h>
#include <asm/uaccess.h>

/* ** HG get it working with 2.4.4 kernel ...... */ 
#include <linux/compiler.h>
#if !defined(likely)
 /* see include/linux/compiler.h */
 #define likely(x)	__builtin_expect((x),1)
#endif

#if !defined(likely)
 /* see include/linux/compiler.h */
 #define unlikely(x)	__builtin_expect((x),0)
#endif

#if !defined(list_for_each_safe)
 /* see include/linux/list.h */
/**
 * list_for_each_safe   -   iterate over a list safe against removal of list entry
 * @pos:    the &struct list_head to use as a loop counter.
 * @n:      another &struct list_head to use as temporary storage
 * @head:   the head for your list.
 */
#define list_for_each_safe(pos, n, head) \
    for (pos = (head)->next, n = pos->next; pos != (head); \
        pos = n, n = pos->next)
#endif


/***** Begin stuff that belongs in a header *****/
/* This stuff belongs in a header, but doesn't really need
 * to be in one because it isn't used outside this file. The
 * exception to that is the two exported functions:
 * ppcboot_env_get/set_var, and I'm not really sure where
 * to put them.
 */

#define PPCBOOT_ENV_PROC_FILENAME "ppcboot_env"
/* ** HG .. */
#ifdef CONFIG_MVBLUE
	#ifdef CONFIG_MVBLUE_M
	#define CFG_ENV_SIZE 0x2000			/* 8 kB for mvBL-5M.. */
	#else
	#define CFG_ENV_SIZE 0x10000			/* 64 kB for mvBL-4.. */
	#endif
#else
	#error "please define your ppcboot environment size !"
#endif

#define _MIN(a,b)	((a) < (b) ? (a) : (b))
#define LOCKED_ETHADDR "ethaddr"
#define LOCKED_SERIAL_NUM "serial#"
#define START_ETHADDR1	"00:00:00:00:00:00"
#define START_ETHADDR2	"b6:b4:45:eb:fb:c0"
#define START_ETHADDR3	"\"b6:b4:45:eb:fb:c0\""
#define START_ETHADDR4a	'b'
#define START_ETHADDR4b	'6'
#define START_ETHADDR4c	'b'
#define START_ETHADDR4d	'4'
#define START_ETHADDR4e	'4'
#define START_ETHADDR4f	'5'
/* .. **HG */

typedef struct ppcboot_env_var_struct {
	char *name;
	char *value;
	struct list_head list;
} ppcboot_env_var_t;

typedef struct ppcboot_env_list_struct {
	struct list_head head;
	struct rw_semaphore lock;
} ppcboot_env_list_t;

static inline int ppcboot_env_try_init( void );
static int ppcboot_env_init( void );
static void ppcboot_env_exit( void );

static int ppcboot_env_read_proc( char *, char **, off_t, int, int *, void * );
static int ppcboot_env_write_proc( struct file *, const char *, unsigned long, void * );

static int ppcboot_env_parse( __u8 *, struct list_head *, int check_legal );
static int ppcboot_env_read( struct mtd_info *, ppcboot_env_list_t * );
static int ppcboot_env_write( struct mtd_info *, ppcboot_env_list_t * );

char *ppcboot_env_get_var( const char *name );
int ppcboot_env_set_var( const char *name, const char *value );
static void ppcboot_env_free_var( ppcboot_env_var_t * );

static void ppcboot_env_probe_mtd( struct mtd_info * );
static void do_nothing( struct mtd_info *mtd ) {}

static void make_crc_table( __u32 * );
static inline __u32 crc32( __u32, const __u8 *, __u32 );

/***** End stuff that belongs in a header *****/

static const char *ppcboot_env_proc_filename = PPCBOOT_ENV_PROC_FILENAME;

static int ppcboot_env_init_called = 0;

static ppcboot_env_list_t ppcboot_env_list = {
	head: LIST_HEAD_INIT( ppcboot_env_list.head ),
	lock: __RWSEM_INITIALIZER( ppcboot_env_list.lock )
};

static struct mtd_info *ppcboot_env_mtd = NULL;

static struct mtd_notifier mtd_user = {
	.add = ppcboot_env_probe_mtd,
	.remove = do_nothing,
/* ** HG	next: NULL */
};

static __u32 *crc_table = NULL;

/* I have no idea how the following two crc functions work.
 * They are borrowed and slightly modified for readability
 * from the PPCBoot code, which claims to have gotten them
 * from the zlib code.
 */

static void make_crc_table( __u32 *table )
{
	static const __u8 p[14] = {
		0, 1, 2, 4, 5, 7, 8, 10, 11, 12, 16, 22, 23, 26
	};
	__u32 c, poly;
	int n, k;

	for ( poly = 0, n = 0; n < 14; n++ ) {
		poly |= 1 << ( 31 - p[n] );
	}

	for ( n = 0; n < 256; n++ ) {
		c = ( __u32 )( n );

		for ( k = 0; k < 8; k++ ) {
			c = ( c & 1 ) ?
				( poly ^ ( c >> 1 ) ) :
				( c >> 1 );
		}

		table[n] = c;
	}
}

static inline __u32 crc32( __u32 crc, const __u8 *buf, __u32 len )
{
	crc = ~crc;

	while ( len-- ) {
		int index = ( ( int )( crc ) ^ ( *buf++ ) ) & 0x0ff;
		crc = crc_table[index] ^ ( crc >> 8 );
	}

	return ~crc;
}

static void ppcboot_env_probe_mtd( struct mtd_info *mtd )
{
	if ( !ppcboot_env_mtd ) {
		__u32 crc, newcrc;
		int n, off, ret;
                unsigned int sz;
		__u8 buf[4];
		int max_size = _MIN (CFG_ENV_SIZE, mtd->size );	

		/* read the location where the crc should be */
		ret = mtd->read( mtd, 0, 4, &n, ( __u8 * )( &crc ) );
		if ( ret < 0 ) {
                        printk( KERN_ERR "%s: Failed to read CRC\n", __FUNCTION__ );
			return;
		}
		/* read the data, crc'ing as we go */
		for ( off = 4, newcrc = 0, sz = max_size; off < sz; off += n ) {
			ret = mtd->read( mtd, off, 4, &n, buf );
			if ( ret < 0 ) {
                                printk( KERN_ERR "%s: Failed to read data\n", __FUNCTION__ );
				break;
			}

			newcrc = crc32( newcrc, buf, n );
		}

		/* if the crc matches, we'll hope this is the right device */
		if ( crc == newcrc ) {
                        printk( KERN_INFO "%s: CRC matches... using MTD %d: %s\n", __FUNCTION__, mtd->index, mtd->name );
			ppcboot_env_mtd = mtd;
		}
	}
}

static int ppcboot_env_read( struct mtd_info *mtd, ppcboot_env_list_t *env_list )
{
	__u8 *buf;
	int ret, n;
	int max_size = _MIN (CFG_ENV_SIZE, mtd->size );	

	/* I know this may be a lot of memory, but
	 * we only do it once at boot / load time,
	 * and we free it right away.
	 */

	buf = vmalloc( max_size );
	if ( !buf ) {
		ret = -ENOMEM;
		goto done;
	}

	/* start at offset 4 to skip the crc */
	ret = mtd->read( mtd, 4, max_size - 8, &n, buf );
	if ( ret < 0 ) {
		goto freemem;
	}

	/* place two zeros at the end so ppcboot_env_parse can cope */
	buf[n] = buf[n + 1] = 0;

	ret = ppcboot_env_parse( buf, &env_list->head, 0 );

 freemem:
	vfree( buf );
 done:
	return ret;
}

static int ppcboot_env_write( struct mtd_info *mtd, ppcboot_env_list_t *env_list )
{
	struct list_head *curr;
	struct erase_info instr;
	int ret = 0, pos = 4, n;
	__u32 tmp, crc = 0;
	int max_size = _MIN (CFG_ENV_SIZE, mtd->size );	

	instr.mtd = mtd;
	instr.addr = 0;
	instr.len = max_size;
	instr.callback = NULL;

	if ( mtd->erase ) {
		ret = mtd->erase( mtd, &instr );
	}
	if ( ret < 0 ) {
		goto done;
	}

	/* write out the data, crc'ing as we go */
	__down_read( &env_list->lock );
	list_for_each( curr, &env_list->head ) {
		ppcboot_env_var_t *env_var = list_entry( curr, ppcboot_env_var_t, list );
		char *buf;

		n = strlen( env_var->name ) + 1 + strlen( env_var->value ) + 1;

		buf = kmalloc( n, GFP_KERNEL );

		strcpy( buf, env_var->name );
		strcat( buf, "=" );
		strcat( buf, env_var->value );

		crc = crc32( crc, buf, n );

		ret = mtd->write( mtd, pos, n, &n, buf );

		kfree( buf );

		if ( ret < 0 ) {
			break;
		}

		pos += n;
	}
	__up_read( &env_list->lock );

	if ( ret < 0 ) {
		goto done;
	}

	ret = pos;

	/* write 0's over the rest of the space. it's what ppcboot does. */

	tmp = 0;
	while ( ( n = ( ( max_size ) - pos ) ) > 0 ) {
		if ( n > 4 ) {
			n = 4;
		}

		mtd->write( mtd, pos, n, &n, ( __u8 * )( &tmp ) );
		crc = crc32( crc, ( __u8 * )( &tmp ), n );
		pos += n;
	}

	mtd->write( mtd, 0, 4, &n, ( u_char * )( &crc ) );

 done:
	return ret;
}

static int ppcboot_env_parse( __u8 *buf, struct list_head *env_list, int check_legal )
{
	int i, ret = 0;

	for ( i = 0; buf[i] != 0; i++ ) {
		char *name, *value;
		int namelen, vallen;

		name = ( char * )( &buf[i] );
		namelen = ( int )( strchr( name, '=' ) - name );

		if ( namelen == -( int )( name ) ) {
			/* -EBADDATA would be good here. :) */
			return -ENODATA;
		}

		name[namelen] = '\0';

		value = name + namelen + 1;
		vallen = strlen( value );
/* ** HG ... */
		if (check_legal)
		{	/*
		 	 * Serial number may not be altered here at all!
			 * Ethernet address may only be altered if it is = START_ETHADDR
	 	 	 */
			char *tmp_value = NULL;

			if (strcmp (name, LOCKED_SERIAL_NUM) == 0)
	        {	/* trying to change the serial number! */
				if ((tmp_value = ppcboot_env_get_var (LOCKED_SERIAL_NUM)) != NULL)
				{	/* serial num. already set */
					printk( KERN_INFO "%s: Can't overwrite \"%s\"\n", __FUNCTION__, name );
					ret = 0; /* but continue... */
				}
				else
					/* not yet set, allow change */
					ret = ppcboot_env_set_var( name, value );
			}
			else if (strcmp (name, LOCKED_ETHADDR) == 0)
			{	/* trying to change the ethernet address */
				if ((tmp_value = ppcboot_env_get_var (LOCKED_ETHADDR)) != NULL)
				{	/* ethernet address already set */
					if ((strcmp (tmp_value, START_ETHADDR1) == 0)  ||
						(strcmp (tmp_value, START_ETHADDR2) == 0) ||
						(strcmp (tmp_value, START_ETHADDR3) == 0))
						/* already set but may be changed */
						ret = ppcboot_env_set_var( name, value );
					else if ((tmp_value[0] == START_ETHADDR4a) &&
							 (tmp_value[1] == START_ETHADDR4b) &&
							 (tmp_value[3] == START_ETHADDR4c) &&
							 (tmp_value[4] == START_ETHADDR4d) &&
							 (tmp_value[6] == START_ETHADDR4e) &&
							 (tmp_value[7] == START_ETHADDR4f) 
							)
						/* already set but may be changed */
						ret = ppcboot_env_set_var( name, value );
					else
					{	/* ethernet address != START_ETHADDR */
						printk( KERN_INFO "%s: Can't overwrite \"%s\" (%s <== \"%s\")\n",
							 __FUNCTION__, name, tmp_value, value );
						ret = 0; /* but continue... */
					}
					kfree (tmp_value);
				}
				else
					/* not yet set, allow change */
					ret = ppcboot_env_set_var( name, value );
			}
			else
				/* allow change for all other variables */
				ret = ppcboot_env_set_var( name, value );
		}
		else
/* ** ... HG */

		ret = ppcboot_env_set_var( name, value );
		if ( ret ) {
			break;
		}

		i += namelen + 1 + vallen;
	}

	return ret;
}

/* input is expected in the form: name=value
 * one per line. the 'write' command must appear
 * alone.
 */

static int ppcboot_env_write_proc( struct file *file, const char *buffer, unsigned long count, void *data )
{
	char *buf, *pos;
	int ret = 0;

	buf = kmalloc( count + 2, GFP_KERNEL );
	if ( !buf ) {
		ret = -ENOMEM;
		goto done;
	}

	memset( buf, 0, count + 2 );

	/* only deal with the data that was actually copied in */
	count -= copy_from_user( buf, buffer, count );
	if ( count <= 0 ) {
		ret = count;
		goto done;
	}

	/* allow the user to tell us to sync the environment we have
	 * in memory with that on the mtd.
	 */
	if ( strcmp( buf, "write\n" ) == 0 ) {
		ret = ppcboot_env_write( ppcboot_env_mtd, &ppcboot_env_list );
		if ( ret == 0 ) {
			goto success;
		} else {
			goto done;
		}
	}

	/* convert all newlines to 0's, so that ppcboot_env_parse can cope */
	for ( pos = buf; ( pos = strchr( pos, '\n' ) ); *pos = '\0', pos++ );

	ret = ppcboot_env_parse( buf, &ppcboot_env_list.head, 1 );
	if ( ret < 0 ) {
		goto done;
	}

 success:
	ret = count;

 done:
	if ( buf ) {
		kfree( buf );
	}

	return ret;
}

static int ppcboot_env_read_proc( char *page, char **start, off_t off, int count, int *eof, void *data )
{
	struct list_head *curr;
	int ret = 0;

	__down_read( &ppcboot_env_list.lock );
	list_for_each( curr, &ppcboot_env_list.head ) {
		ppcboot_env_var_t *env_var = list_entry( curr, ppcboot_env_var_t, list );
		int namelen, vallen;

		/* the + 1's are for the '=' and the
		 * '\n' that we write out.
		 */
		namelen = strlen( env_var->name ) + 1;
		vallen = strlen( env_var->value ) + 1;

		/* we have to find the offset before we
		 * start reading.
		 */
		if ( off > 0 ) {
			off -= namelen + vallen;
			continue;
		}

		/* we only write entire name/value
		 * pairs at a time.
		 */
		if ( count < namelen + vallen ) {
			break;
		}

		strcpy( &page[ret], env_var->name );
		ret += namelen;
		page[ret - 1] = '=';

		strcpy( &page[ret], env_var->value );
		ret += vallen;
		page[ret - 1] = '\n';

		count -= namelen + vallen;
	}
	__up_read( &ppcboot_env_list.lock );

	/* if we got all the way through the list, we're done */
	if ( curr == &ppcboot_env_list.head ) {
		*eof = 1;
	}

	/* if the offset is still positive, it must have been
	 * past the total length of the environment.
	 */
	if ( off > 0 ) {
		ret = -ESPIPE;
	}

	return ret;
}


/**
 * ppcboot_env_get_var - get a value from the ppcboot environment
 *
 * @name: a pointer to a null-terminated string containing the name of the environment variable
 *
 * return: a pointer to the value assocaited with the given name on success, NULL on failure
 *         the caller is responsible for kfree'ing the returned value
 *
 * we return kmalloc'd mem as opposed to the env var value itself because we don't want
 * the caller to directly be able to modify the value in our list.
 */

char *ppcboot_env_get_var( const char *name )
{
	struct list_head *curr;
	char *value;

	value = NULL;

	if ( ppcboot_env_try_init() ) {
		goto done;
	}

	__down_read( &ppcboot_env_list.lock );
	list_for_each( curr, &ppcboot_env_list.head ) {
		ppcboot_env_var_t *env_var = list_entry( curr, ppcboot_env_var_t, list );

		if ( strcmp( name, env_var->name ) == 0 ) {
			value = kmalloc( strlen( env_var->value ) + 1, GFP_KERNEL );
			if ( value ) {
				strcpy( value, env_var->value );
			}
			break;
		}
	}
	__up_read( &ppcboot_env_list.lock );

 done:
	return value;
}

/**
 * ppcboot_env_set_var - add/modify/delete a ppcboot environment variable
 *
 * @name:  a null terminated string containing the name of the variable
 * @value: a null terminated string containing the value of the variable
 *         zero length indicates deletion of the variable
 *
 * return: 0 on success, a negative errno on failure
 */

int ppcboot_env_set_var( const char *name, const char *value )
{
	ppcboot_env_var_t *env_var;
	struct list_head *curr;
	int ret = 0;

	ret = ppcboot_env_try_init();
	if ( ret ) {
		goto done;
	}

	if ( !( name && value ) ) {
		ret = -EINVAL;
		goto done;
	}

	__down_write( &ppcboot_env_list.lock );
	list_for_each( curr, &ppcboot_env_list.head ) {
		env_var = list_entry( curr, ppcboot_env_var_t, list );

		if ( strcmp( name, env_var->name ) == 0 ) {
			char *tmp_val = kmalloc( strlen( value ) + 1, GFP_KERNEL );
			if ( !tmp_val ) {
				ret = -ENOMEM;
				goto unlock;
			}

			/* change the value without checking length */
			kfree( env_var->value );
			env_var->value = tmp_val;

			strcpy( env_var->value, value );

			/* set value to NULL to indicate we used it */
			value = NULL;
		}

		/* if the value length is zero, delete the variable */
		if ( strlen( env_var->value ) == 0 ) {
			list_del( curr );
			ppcboot_env_free_var( env_var );
		}
	}

	/* the loop took care of it; the name existed */
	if ( !value ) {
		goto unlock;
	}

	/* new variable; insert it at the end */
	env_var = kmalloc( sizeof( ppcboot_env_var_t ), GFP_KERNEL );
	if ( !env_var ) {
		ret = -ENOMEM;
		goto unlock;
	}

	env_var->name = kmalloc( strlen( name ) + 1, GFP_KERNEL );
	env_var->value = kmalloc( strlen( value ) + 1, GFP_KERNEL );

	if ( !( env_var->name && env_var->value ) ) {
		ppcboot_env_free_var( env_var );
		ret = -ENOMEM;
		goto unlock;
	}

	strcpy( env_var->name, name );
	strcpy( env_var->value, value );
	list_add_tail( &env_var->list,  &ppcboot_env_list.head );

 unlock:
	__up_write( &ppcboot_env_list.lock );

 done:
	return ret;
}

static void ppcboot_env_free_var( ppcboot_env_var_t *var )
{
	if ( !var ) {
		return;
	}

	if ( var->name ) {
		kfree( var->name );
	}

	if ( var->value ) {
		kfree( var->value );
	}

	kfree( var );
}

static inline int ppcboot_env_try_init( void )
{
	/* Our exposed interfaces could be called before
	 * we've had a chance to initialize. Each time, they'll
	 * try to initialize us, but we can only let it happen
	 * the first time.
	 */

	if ( likely( ppcboot_env_init_called ) ) {
		return -( ppcboot_env_mtd == NULL );
	}

	return ppcboot_env_init();
}

static int __init ppcboot_env_init( void )
{
	struct proc_dir_entry *ppcboot_env_proc_entry = NULL;
	int err = 0;

	if ( ppcboot_env_init_called ) {
		return -( ppcboot_env_mtd == NULL );
	}

        printk( KERN_INFO "%s\n", __FUNCTION__ );

	ppcboot_env_init_called = 1;

	crc_table = kmalloc( sizeof( __u32 ) * 256, GFP_KERNEL );
	if ( !crc_table ) {
                printk( KERN_INFO "%s: Failed to allocate memory for CRC table\n", __FUNCTION__ );
		err = -ENOMEM;
		goto error;
	}

	make_crc_table( crc_table );

	/* This calls our probe function for
	 * every mtd device in the system.
	 */
	register_mtd_user( &mtd_user );

	if ( !ppcboot_env_mtd ) {
                printk( KERN_INFO "%s: Failed to find PPCBoot environment\n", __FUNCTION__ );
                err = -ENXIO;
                goto error;
	}

	/* ppcboot_env_mtd was set by our probe function which was
	 * called by our call to register_mtd_user
	 */
	get_mtd_device( ppcboot_env_mtd, -1 );

	err = ppcboot_env_read( ppcboot_env_mtd, &ppcboot_env_list );
	if ( err ) {
               printk( KERN_INFO "%s: Failed to load PPCBoot environment\n", __FUNCTION__ );
		goto error;
	}

	ppcboot_env_proc_entry = create_proc_entry( ppcboot_env_proc_filename, S_IRUGO | S_IWUSR, &proc_root );
	if ( !ppcboot_env_proc_entry ) {
                printk( KERN_INFO "%s: Failed to create proc entry\n", __FUNCTION__ );
		err = -ENOENT;
		goto error;
	}

	ppcboot_env_proc_entry->owner = THIS_MODULE;
	ppcboot_env_proc_entry->read_proc = ppcboot_env_read_proc;
	ppcboot_env_proc_entry->write_proc = ppcboot_env_write_proc;

	return 0;

 error:
	ppcboot_env_exit();
	return err;
}

static void __exit ppcboot_env_exit( void )
{
	struct list_head *curr, *next;

        printk( KERN_INFO "%s\n", __FUNCTION__ );

	remove_proc_entry( ppcboot_env_proc_filename, &proc_root );

	if ( ppcboot_env_mtd ) {
		ppcboot_env_write( ppcboot_env_mtd, &ppcboot_env_list );
		put_mtd_device( ppcboot_env_mtd );
		ppcboot_env_mtd = NULL;
	}

	unregister_mtd_user( &mtd_user );

	__down_write( &ppcboot_env_list.lock );
	list_for_each_safe( curr, next, &ppcboot_env_list.head ) {
		ppcboot_env_var_t *var = list_entry( curr, ppcboot_env_var_t, list );
		list_del( curr );
		ppcboot_env_free_var( var );
	}
	__up_write( &ppcboot_env_list.lock );

	if ( crc_table ) {
		kfree( crc_table );
	}
}

module_init( ppcboot_env_init );
module_exit( ppcboot_env_exit );

EXPORT_SYMBOL( ppcboot_env_set_var );
EXPORT_SYMBOL( ppcboot_env_get_var );

#if defined (MODULE_LICENSE)
  MODULE_LICENSE( "GPL" );
#endif
MODULE_AUTHOR( "John Tyner <jtyner@cs.ucr.edu>" );
MODULE_DESCRIPTION( "Read/Write PPCBoot environment from an MTD" );

