/*
 * ISP1362 USB controller core driver.
 * Copyright (C) 2004-2005 DENX Software Engineering, Wolfgang Denk, wd@denx.de
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <asm/io.h>
#ifdef CONFIG_NSCU
#include <asm/8xx_immap.h>
#endif

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/atomic.h>

#include <linux/usb_ch9.h>
#include <linux/usb_gadget.h>

#include "isp1362gapi.h"

/*#define DEBUG
  #define VERBOSE*/

#ifdef DEBUG
#undef KERN_DEBUG
#define KERN_DEBUG KERN_WARNING
#define	debug(fmt, params...)	printk(KERN_DEBUG "%s: " fmt, __FUNCTION__, ## params)
#define	debug0(fmt)	        printk(KERN_DEBUG "%s: " fmt, __FUNCTION__)
#else
#define	debug(fmt...)
#define	debug0(fmt)
#endif

#ifdef VERBOSE
#define	verbose(fmt, params...)	printk(KERN_DEBUG "%s: " fmt, __FUNCTION__, ## params)
#define	verbose0(fmt)	        printk(KERN_DEBUG "%s: " fmt, __FUNCTION__)
#else
#define	verbose(fmt...)
#define	verbose0(fmt...)
#endif

#define	DRIVER_VERSION	"4-Apr-2004"
#define	DRIVER_DESC	"isp1362 USB Device Controller driver"

#define	ISP1362_DC_DATA		(isp1362_io_base + 4)
#define	ISP1362_DC_COM		(isp1362_io_base + 6)

#ifdef CONFIG_NSCU
#define ISP1362_RESET 0x2000

static struct timer_list isp1362_timer;
#endif

static const char driver_name [] = "isp1362_udc";
static const char ep0name [] = "ep0";

static struct isp1362_udc *the_controller;
static void * isp1362_io_base = NULL;

static ulong isp1362_cnt_int = 0;
static ulong isp1362_cnt_reset = 0;
static ulong isp1362_cnt_resume = 0;
static ulong isp1362_cnt_suspend = 0;
static ulong isp1362_cnt_err_setup = 0;

static void isp1362_ep_int_hndlr(struct isp1362_ep* ep);

#if CONFIG_PROC_FS

#define DEBUG_DEVNAME        "debug"
#define DEBUG_USB_DIR        "usb"

typedef struct {
	int count;
	int size;
	wait_queue_head_t wq;
} debug_dev_t;

static debug_dev_t debug_dev;

static devfs_handle_t devfs_handle = NULL;

static struct proc_dir_entry *debug_proc_dir;
static struct proc_dir_entry *debug_proc_usb_dir;

struct file_operations debug_fops = {
};

#define	PROC_SUSPEND_NAME		"isp1362_suspend"

#define SAY( fmt, args... )  p += sprintf(p, fmt, ## args )
#define SAYV(  num )         p += sprintf(p, "%25.25s: %8.8lX\n", "Value", num )
#define SAYC( label, yn )    p += sprintf(p, "%25.25s: %s\n", label, yn )
#define SAYS( label, v )     p += sprintf(p, "%25.25s: %lu\n", label, v )

static int isp1362_read_proc(char *page, char **start, off_t off,
			    int count, int *eof, void *data)
{
	 char * p = page;
	 int len;

 	 SAY( "ISP1362-Gadget:\n" );

 	 SAYS( "Interrupt counter", isp1362_cnt_int );
 	 SAYS( "Reset counter", isp1362_cnt_reset );
 	 SAYS( "Resume counter", isp1362_cnt_resume );
 	 SAYS( "Suspend counter", isp1362_cnt_suspend );

 	 SAYS( "SETUP errors", isp1362_cnt_err_setup );
// 	 SAYS( "IN errors", isp1362_cnt_err_in );
// 	 SAYS( "OUT errors", isp1362_cnt_err_out );

	 len = ( p - page ) - off;
	 if ( len < 0 )
		  len = 0;
	 *eof = ( len <=count ) ? 1 : 0;
	 *start = page + off;
	 return len;
}

static void isp1362_reset_device(void);
static void reset_udc(struct isp1362_udc* dev);

static int isp1362_write_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	char mydata[0x100];
	volatile immap_t *immr;
	volatile cpm8xx_t *cp;
	struct isp1362_udc *dev = the_controller;

	immr = (immap_t*)IMAP_ADDR;
	cp = (cpm8xx_t*)&immr->im_cpm;

	/* TODO: Check for count > len */
	if (copy_from_user(mydata, buffer, count) ) {
		return -EFAULT;
	}

	if (strncmp(mydata, "reset", 5) == 0) {
		printk("ISP1362-Gadget: reset request!\n");
		cp->cp_pbdat &= ~ISP1362_RESET;
		mdelay(1);
		cp->cp_pbdat |= ISP1362_RESET;
	} else 	if (strncmp(mydata, "sreset", 6) == 0) {
		printk("ISP1362-Gadget: soft-reset request!\n");
		isp1362_reset_device();
		mdelay(500);
		reset_udc(dev);
	}


	return count;
}

#endif  /* CONFIG_PROC_FS */

#ifdef VERBOSE
static inline void isp1362_print_request( struct usb_ctrlrequest * pReq )
{
	 static char * tnames[] = { "dev", "intf", "ep", "oth" };
	 static char * rnames[] = { "std", "class", "vendor", "???" };
	 char * psz;
	 switch( pReq->bRequest ) {
	 case USB_REQ_GET_STATUS: psz = "get stat"; break;
	 case USB_REQ_CLEAR_FEATURE: psz = "clr feat"; break;
	 case USB_REQ_SET_FEATURE: psz = "set feat"; break;
	 case USB_REQ_SET_ADDRESS: psz = "set addr"; break;
	 case USB_REQ_GET_DESCRIPTOR: psz = "get desc"; break;
	 case USB_REQ_SET_DESCRIPTOR: psz = "set desc"; break;
	 case USB_REQ_GET_CONFIGURATION: psz = "get cfg"; break;
	 case USB_REQ_SET_CONFIGURATION: psz = "set cfg"; break;
	 case USB_REQ_GET_INTERFACE: psz = "get intf"; break;
	 case USB_REQ_SET_INTERFACE: psz = "set intf"; break;
	 default: psz = "unknown"; break;
	 }
	 debug( "- [%s: %s req to %s. dir=%s]\n", psz,
			 rnames[ (pReq->bRequestType >> 5) & 3 ],
			 tnames[ pReq->bRequestType & 3 ],
			 ( pReq->bRequestType & 0x80 ) ? "in" : "out" );
}
#endif /* VERBOSE */

static inline ulong isp1362_inw (volatile unsigned short * port)
{
	ulong res;
	mb();
	res = *port;
	mb();
	return res;
}

static inline void isp1362_outw (ulong data, volatile unsigned short * port)
{
	mb();
	*port = data;
	mb();
}

static ulong isp1362_read_reg32 (ulong reg)
{
	ulong res;

	isp1362_outw (reg, ISP1362_DC_COM);
	res = isp1362_inw (ISP1362_DC_DATA);
	res |= isp1362_inw (ISP1362_DC_DATA) << 16;

	return res;
}

static void isp1362_write_reg32 (ulong reg, ulong data)
{
	isp1362_outw (reg, ISP1362_DC_COM);
	isp1362_outw (data & 0xFFFF, ISP1362_DC_DATA);
	isp1362_outw (data >> 16, ISP1362_DC_DATA);
}

static ulong isp1362_read_reg16 (ulong reg)
{
	ulong res;

	isp1362_outw (reg, ISP1362_DC_COM);
	res = isp1362_inw (ISP1362_DC_DATA);

	return res;
}

static void isp1362_write_reg16 (ulong reg, ulong data)
{
	isp1362_outw (reg, ISP1362_DC_COM);
	isp1362_outw (data, ISP1362_DC_DATA);
}

static void isp1362_cmd_reg (ulong reg)
{
	isp1362_outw (reg, ISP1362_DC_COM);
}

  /* Set device address
   */
static void isp1362_set_address(ulong devaddr)
{
	isp1362_write_reg16(WR_DcAddress, 0x80 | devaddr);
}

  /* Check endpoint status
   */
ulong isp1362_check_status(ulong ep, ulong dir)
{
	ulong code = ep ? RD_DcEndpointStatusImage + ep : dir == EP_DIR_IN ?
		RD_DcEndpointStatusImage0_IN : RD_DcEndpointStatusImage0_OUT;

	return isp1362_read_reg16(code);
}

  /* Acknowledge Set-up.
   */
static void isp1362_acknowledge_setup(void)
{
	isp1362_cmd_reg(DO_AcknowledgeSetup);
}

  /* Validate endpoint buffer.
   */
static void isp1362_validate_buffer(ulong ep)
{
	ulong code = ep ? DO_ValidateEndpointBuffer + ep :
		DO_ValidateEndpointBuffer0;

	isp1362_cmd_reg(code);
}

  /* Clear endpoint buffer.
   */
void isp1362_clear_buffer(ulong ep)
{
	ulong code = ep ? DO_ClearEndpointBuffer + ep :
		DO_ClearEndpointBuffer0;

	isp1362_cmd_reg(code);
}

  /* Write to endpoint buffer memory
   */
void isp1362_write_buffer (ulong ep, void * buf, ulong size)
{
	ulong code = ep ? DO_WriteEndpointBuffer + ep : DO_WriteEndpointBuffer0;
	ushort data;
	int i;

#ifdef VERBOSE
	if (size)
	{
		unsigned char * pdb = (unsigned char *)buf;
		int i;

		verbose0("data:");
		for (i = 0; i < size; i++)
			printk(" %02X", (int)pdb[i]);
		printk("\n");
	}
#endif

	isp1362_cmd_reg (code);

	isp1362_outw (size, ISP1362_DC_DATA);

	for (i = 0; i < size / 2; i ++)
	{
		data = ((ushort *)buf)[i];
		cpu_to_le16s (&data);
		isp1362_outw (data, ISP1362_DC_DATA);
	}
	if (size & 1)
	{
		data = ((u_char *)buf)[size - 1];
		isp1362_outw (data, ISP1362_DC_DATA);
	}

	isp1362_validate_buffer (ep);
}

  /* Read from endpoint buffer memory
   */
ulong isp1362_read_buffer (ulong ep, void * buf, long size)
{
	ulong code = ep ? DO_ReadEndpointBuffer + ep : DO_ReadEndpointBuffer0;
	ulong res;
	ushort data;
	int i;

	isp1362_cmd_reg (code);

	res = isp1362_inw (ISP1362_DC_DATA);
	if (res > size)
	{
		printk ("isp1362: Packet is too long: %04X > %04X\n",
			(uint)res, (uint)size);
		res = size;
	}

	for (i = 0; i < res / 2; i ++)
	{
		data = isp1362_inw (ISP1362_DC_DATA);
		le16_to_cpus(&data);
		((ushort *)buf)[i] = data;
	}
	if (res & 1)
	{
		data = isp1362_inw (ISP1362_DC_DATA);
		((u_char *)buf)[res - 1] = data;
	}

	if (ep == 0 && (isp1362_check_status(0, EP_DIR_OUT) & EPSTAT_SETUPT))
		isp1362_acknowledge_setup();

	isp1362_clear_buffer (ep);

#ifdef VERBOSE
	if (res)
	{
		unsigned char * pdb = (unsigned char *)buf;
		int i;

		verbose("data %d bytes:", (int)res);
		for (i = 0; i < res; i++)
			printk(" %02X", (int)pdb[i]);
		printk("\n");
	}
#endif

	return res;
}

  /* Unlock device
   */
static void isp1362_unlock_device(void)
{
    isp1362_write_reg16(WR_UnlockDevice, 0xAA37);
}

  /* Chip identification code and hardware version number
   */
static ulong isp1362_chip_id(void)
{
    return isp1362_read_reg16(RD_DcChipID);
}

  /* Stall endpoint.
   */
static void isp1362_stall_endpoint(ulong ep, ulong dir)
{
    ulong code = ep ? DO_StallEndpoint + ep : dir == EP_DIR_IN ?
	DO_StallEndpoint0_IN : DO_StallEndpoint0_OUT;

    isp1362_cmd_reg(code);
}

  /* Unstall endpoint.
   */
static void isp1362_unstall_endpoint(ulong ep, ulong dir)
{
    ulong code = ep ? DO_UnstallEndpoint + ep : dir == EP_DIR_IN ?
	DO_UnstallEndpoint0_IN : DO_UnstallEndpoint0_OUT;

    isp1362_cmd_reg(code);
}

static void isp1362_send_empty_packet(ulong ep)
{
    isp1362_write_buffer(ep, NULL, 0);
}

static int isp1362_setup_data_send(void *in, ulong requested, ulong have)
{
	if (requested > have) {
		debug("isp1362: Wrong data size requested: %lu>%lu\n", requested,
		      have);
		requested = have;
		/* Note: must send an empty packet? */
		/*isp1362_ep0_empty_queued = 1;*/
	}

	isp1362_write_buffer(0, in, requested);
	return 0;
}

  /* Resets the DC in the same way as an external hardware reset
   */
static void isp1362_reset_device(void)
{
	isp1362_cmd_reg(DO_ResetDevice);
}

#ifdef VERBOSE
  /* Frame number of the last successfully received SOF
   */
static ulong isp1362_frame_number(void)
{
    return isp1362_read_reg16(RD_DcFrameNumber) & 0x07FF;
}
#endif

  /* Resume handler
   */
static void isp1362_resume(void)
{
	isp1362_unlock_device();
}

#ifdef CONFIG_NSCU
static int isp1362_read_susp_proc(char *page, char **start, off_t off,
				  int count, int *eof, void *data)
{
	 char * p = page;
	 int len;

 	 SAY( "ISP1362 Suspend Hangup detected!\n" );

	 len = ( p - page ) - off;
	 if ( len < 0 )
		  len = 0;
	 *eof = ( len <=count ) ? 1 : 0;
	 *start = page + off;
	 return len;
}

static void isp1362_timer_func(unsigned long data)
{
	debug("%s: ISP1362 Suspend Hangup detected!\n", __FUNCTION__);

	/*
	 * Create proc file to signal a hangup to user space
	 */
	create_proc_read_entry (PROC_SUSPEND_NAME, 0, NULL, isp1362_read_susp_proc, NULL);
}
#endif

  /* Suspend handler
   */
static void isp1362_suspend(void)
{
#ifdef DEBUG
	ushort mode = isp1362_read_reg16(RD_DcMode);
	debug("done, prev mode %#x\n", (uint)mode);
#endif

#ifdef CONFIG_NSCU
	isp1362_timer.expires = jiffies + 10 * HZ; /* 10 seconds after last suspend irq */
	isp1362_timer.data = 1;
	add_timer(&isp1362_timer);
#endif

/*	isp1362_write_reg16(WR_DcMode, mode | MODE_GOSUSP);
	isp1362_write_reg16(WR_DcMode, mode & ~MODE_GOSUSP);*/
}

  /* Read endpoint status
   */
ulong isp1362_read_status(ulong ep, ulong dir)
{
	ulong code = ep ? RD_DcEndpointStatus + ep : dir == EP_DIR_IN ?
		RD_DcEndpointStatus0_IN : RD_DcEndpointStatus0_OUT;

	return isp1362_read_reg16(code);
}

static void request_done(struct isp1362_ep* ep,
	struct isp1362_request* req, int flag)
{
	if (!list_empty(&req->queue))
		list_del_init(&req->queue);

	req->zlp_sent = 0;

	req->req.status = flag;
	req->req.complete(&ep->ep, &req->req);
}

static int isp1362_ep_enable (struct usb_ep *_ep,
		const struct usb_endpoint_descriptor *desc)
{
  	struct isp1362_ep   *ep;
	u8 ep_flags = EPCONF_FFOSZ_NONISO_64 | EPCONF_FIFOEN;
	u32 int_flags;

	ep = container_of (_ep, struct isp1362_ep, ep);
	if (!_ep || !desc || ep->desc || _ep->name == ep0name
			|| desc->bDescriptorType != USB_DT_ENDPOINT
			|| ep->bEndpointAddress != desc->bEndpointAddress
			|| ep->fifo_size < le16_to_cpu
						(desc->wMaxPacketSize)) {
		printk(KERN_WARNING "bad ep or descriptor\n");
		return -EINVAL;
	}

	/* xfer types must match, except that interrupt ~= bulk */
	if (ep->bmAttributes != desc->bmAttributes
			&& ep->bmAttributes != USB_ENDPOINT_XFER_BULK
			&& desc->bmAttributes != USB_ENDPOINT_XFER_INT) {
		printk(KERN_WARNING "%s type mismatch\n", _ep->name);
		return -EINVAL;
	}

	/* Note: should take in account iso eps and other fifo sizes? */
	/*ep_flags |= ep->bEndpointAddress & USB_DIR_IN ?
		EPCONF_EPDIR_IN : EPCONF_EPDIR_OUT;
	isp1362_write_reg16 (WR_DcEndpointConfiguration + ep_num(ep), ep_flags);*/
	ep_flags = isp1362_read_reg16(RD_DcEndpointConfiguration + ep_num(ep));

	/*isp1362_write_reg32 (WR_DcInterruptEnable,
			     (u32)isp1362_read_reg32 (RD_DcInterruptEnable) |
			     (INT_EP << ep_num(ep)));*/
	int_flags = isp1362_read_reg32 (RD_DcInterruptEnable);

	ep->desc = desc;

	debug("enabled endpoint %s flags %#x int_flags = %#x\n", _ep->name,
	      (int)ep_flags, int_flags);

	return 0;
}

static int isp1362_ep_disable (struct usb_ep *_ep)
{
  	struct isp1362_ep *ep;
	int ep_num;

	ep = container_of (_ep, struct isp1362_ep, ep);
	ep_num = ep->bEndpointAddress & ~USB_DIR_IN;
	ep->desc = 0;
	/* Note: changing of configuration of an ep needs revalidation
	 * of all enabled eps' buffers.
	 * See chapter 16.1.1 (remark) of isp1362 Product Data
	 */
	/*isp1362_write_reg16 (WR_DcEndpointConfiguration + ep_num, 0);
	isp1362_write_reg32 (WR_DcInterruptEnable,
	isp1362_read_reg32(WR_DcInterruptEnable) & ~(INT_EP << ep_num));*/
	return 0;
}

static struct usb_request *
isp1362_ep_alloc_request (struct usb_ep *_ep, int gfp_flags)
{
	struct isp1362_request *req;

	req = kmalloc (sizeof *req, gfp_flags);
	if (!req)
		return 0;

	memset (req, 0, sizeof *req);
	INIT_LIST_HEAD (&req->queue);
	return &req->req;
}

static void
isp1362_ep_free_request (struct usb_ep *_ep, struct usb_request *_req)
{
	struct isp1362_request *req;

	req = container_of (_req, struct isp1362_request, req);
	/*	WARN_ON (!list_empty (&req->queue));*/
	kfree(req);
}

static void *
isp1362_ep_alloc_buffer(struct usb_ep *_ep, unsigned bytes,
	dma_addr_t *dma, int gfp_flags)
{
	*dma = -1; /* dma is not supported so far */
	return kmalloc (bytes, gfp_flags);
}

static void
isp1362_ep_free_buffer(struct usb_ep *_ep, void *buf, dma_addr_t dma,
		unsigned bytes)
{
	kfree (buf);
}

  /* Returns nonzero if request is completed
   */
static int do_request(struct isp1362_ep* ep, struct isp1362_request* req)
{
	struct isp1362_udc *dev = the_controller;
	ulong dir_in = (!ep->bEndpointAddress && dev->ep0dir_in) ||
		(ep->bEndpointAddress & USB_DIR_IN);
	ulong status = isp1362_read_status(ep_num(ep), dir_in? EP_DIR_IN : EP_DIR_OUT);
	uint size;

	if (dir_in) {
		if (!(status & EPSTAT_EPFULL0)) {
			if (req->req.length == req->req.actual) {
				if ((!req->req.length ||
				     (req->req.zero &&
				      !(req->req.length % ep->ep.maxpacket))) &&
				    !req->zlp_sent) {
					isp1362_send_empty_packet(ep_num(ep));
					req->zlp_sent = 1;
					verbose("zlp sent to %s\n", ep->ep.name);
					return 0;
				}
				return 1;
			}

			size = min((uint)ep->ep.maxpacket,
				   req->req.length - req->req.actual);
			isp1362_write_buffer(ep_num(ep),
					     (char*)req->req.buf + req->req.actual,
					     size);
			req->req.actual += size;
		}
	} else {
		if ((status & EPSTAT_EPFULL0)) {
			size = isp1362_read_buffer(ep_num(ep),
				  (char*)req->req.buf + req->req.actual,
				  req->req.length - req->req.actual);
			req->req.actual += size;
			if (req->req.length == req->req.actual ||
			    size < ep->ep.maxpacket) {
				if (!ep->bEndpointAddress) {
					verbose0("sending OUT packet ACK\n");
					isp1362_send_empty_packet(0);
					dev->ep0dir_in = 1;
				}
				return 1;
			}
		}
	}
	return 0;
}

static int
isp1362_ep_queue(struct usb_ep *_ep, struct usb_request *_req, int gfp_flags)
{
	struct isp1362_request	*req;
	struct isp1362_ep	*ep;
	struct isp1362_udc	*dev = the_controller;
	unsigned long		flags;

	req = container_of(_req, struct isp1362_request, req);
	if (unlikely (!_req || !_req->complete || !_req->buf
			|| !list_empty(&req->queue))) {
		printk(KERN_WARNING "bad params _req = %#x complete %#x empty %d\n",
		      (int)_req, (int)_req->complete,
		      (int)list_empty(&req->queue));
		return -EINVAL;
	}

	ep = container_of(_ep, struct isp1362_ep, ep);
	if (unlikely (!_ep || (!ep->desc && ep->ep.name != ep0name))) {
		printk(KERN_WARNING "bad ep\n");
		return -EINVAL;
	}

	_req->status = -EINPROGRESS;
	_req->actual = 0;

	spin_lock_irqsave(&dev->lock, flags);

	/* kickstart this i/o queue? */
	if (list_empty (&ep->queue)) {
		if (do_request(ep, req)) {
			spin_unlock_irqrestore(&dev->lock, flags);
			verbose("req to %s done\n", _ep->name);
			request_done(ep, req, 0);
			return 0;
		}
	}

	list_add_tail(&req->queue, &ep->queue);
	verbose("req %#x queued to ep %s, size = %d fn = %d\n",
		(int)req, _ep->name,
		(int)_req->length, (int)isp1362_frame_number());

	spin_unlock_irqrestore(&dev->lock, flags);

	return 0;
}

static int isp1362_ep_dequeue(struct usb_ep *_ep, struct usb_request *_req)
{
	struct isp1362_request	*req;
	struct isp1362_ep	*ep;
	struct isp1362_udc	*dev = the_controller;
	unsigned long		flags;

	req = container_of(_req, struct isp1362_request, req);
	debug("req %#x\n", (int)req);
	if (unlikely (!_req || !_req->complete || !_req->buf
			|| !list_empty(&req->queue))) {
		debug0("bad params\n");
		return -EINVAL;
	}

	ep = container_of(_ep, struct isp1362_ep, ep);
	if (unlikely (!_ep || (!ep->desc && ep->ep.name != ep0name))) {
		debug0("bad ep\n");
		return -EINVAL;
	}

	spin_lock_irqsave(&dev->lock, flags);

	/* make sure it's still queued on this endpoint */
	list_for_each_entry (req, &ep->queue, queue) {
		if (&req->req == _req)
			break;
	}
	if (&req->req != _req) {
		spin_unlock_irqrestore (&dev->lock, flags);
		return -EINVAL;
	}

	request_done(ep, req, -ECONNRESET);

	spin_unlock_irqrestore(&dev->lock, flags);

	return 0;
}

static void set_halt(struct isp1362_ep* ep, int value)
{
	debug("%s %s halt\n", ep->ep.name, value ? "set" : "clear");
	if (value) {
		if (ep_num(ep) == 0)
			; /* Note: stall control ep */
				/*ep->dev->protocol_stall = 1;*/
		else
			isp1362_stall_endpoint(ep_num(ep), 0);
	} else
		isp1362_unstall_endpoint(ep_num(ep), 0);
}

static int isp1362_ep_set_halt(struct usb_ep *_ep, int value)
{
	struct isp1362_ep	*ep;
	struct isp1362_udc	*dev = the_controller;
	unsigned long		flags;
	int                     retval = 0;

	debug("ep = %s\n", _ep->name);

	ep = container_of(_ep, struct isp1362_ep, ep);

	spin_lock_irqsave(&dev->lock, flags);

	if (!list_empty(&ep->queue))
		retval = -EAGAIN;
	else
		set_halt(ep, value);

	spin_unlock_irqrestore(&dev->lock, flags);

	return retval;
}

static struct usb_ep_ops isp1362_ep_ops = {
	.enable		= isp1362_ep_enable,
	.disable	= isp1362_ep_disable,

	.alloc_request	= isp1362_ep_alloc_request,
	.free_request	= isp1362_ep_free_request,

	.alloc_buffer	= isp1362_ep_alloc_buffer,
	.free_buffer	= isp1362_ep_free_buffer,

	.queue		= isp1362_ep_queue,
	.dequeue	= isp1362_ep_dequeue,

	.set_halt	= isp1362_ep_set_halt,

	/* Unsupported */
	/*.fifo_status	= isp1362_ep_fifo_status,
	  .fifo_flush	= isp1362_ep_fifo_flush,*/
};

static int isp1362_udc_get_frame(struct usb_gadget *_gadget)
{
	return isp1362_read_reg16 (RD_DcFrameNumber) & 0x07FF;
}

static int isp1362_udc_wakeup(struct usb_gadget *_gadget)
{
	/* unimplemented */
	return -ENOSYS;
}

static const struct usb_gadget_ops isp1362_udc_ops = {
	.get_frame	 = isp1362_udc_get_frame,
	.wakeup		 = isp1362_udc_wakeup,
};

  /* The allocation of buffer memory
   * only takes place after all 16 endpoints
   * have been configured in sequence
   */
static void configure_endpoints (void)
{
	int i;
	struct isp1362_udc* dev = the_controller;

	isp1362_write_reg16 (
		WR_DcEndpointConfiguration0_OUT,
		EPCONF_FFOSZ_NONISO_64 | EPCONF_EPDIR_OUT | EPCONF_FIFOEN);

	isp1362_write_reg16 (
		WR_DcEndpointConfiguration0_IN,
		EPCONF_FFOSZ_NONISO_64 | EPCONF_EPDIR_IN | EPCONF_FIFOEN);

	isp1362_write_reg16 (
		WR_DcEndpointConfiguration + 1,
		EPCONF_FFOSZ_NONISO_64 | EPCONF_EPDIR_OUT | EPCONF_FIFOEN);

	isp1362_write_reg16 (
		WR_DcEndpointConfiguration + 2,
		EPCONF_FFOSZ_NONISO_64 | EPCONF_EPDIR_IN | EPCONF_FIFOEN);

	isp1362_write_reg16 (
		WR_DcEndpointConfiguration + 3,
		EPCONF_FFOSZ_NONISO_8 | EPCONF_EPDIR_IN | EPCONF_FIFOEN);

	for (i = 4; i <= 14; i++)
	{
		isp1362_write_reg16 (WR_DcEndpointConfiguration + i, 0);
	}

	/* device/ep0 records init */
	INIT_LIST_HEAD (&dev->gadget.ep_list);
	INIT_LIST_HEAD (&dev->gadget.ep0->ep_list);

	/* basic endpoint records init */
	for (i = 0; i < ISP1362_UDC_NUM_ENDPOINTS; i++) {
		struct isp1362_ep *ep = &dev->ep[i];

		if (!ep->ep.maxpacket)
			continue;

		if (i != 0)
			list_add_tail (&ep->ep.ep_list, &dev->gadget.ep_list);

		ep->desc = 0;
		INIT_LIST_HEAD (&ep->queue);
	}
}

static int start_udc( void )
{
	isp1362_write_reg16 (
		WR_DcHardwareConfiguration,
		HWCONF_CLKRUN | HWCONF_CKDIV_12MHZ);

	isp1362_write_reg32 (
		WR_DcInterruptEnable,
		INT_RST | INT_RESM | INT_SUSP | INT_EOT |
		/*INT_SOF |*/ /*INT_PSOF |*/ INT_SP_EOT |
		INT_EP0_OUT | INT_EP0_IN | (INT_EP << 1) | (INT_EP << 2) |
		(INT_EP << 3));

	verbose("interrupt status %02lx \n", isp1362_read_reg32 (RD_DcInterrupt));
	isp1362_set_address(0);

	isp1362_write_reg16 (WR_DcMode, MODE_SOFTCT | MODE_INTENA/* | MODE_DBGMOD*/);

	return 0;
}

static void reset_ep(struct isp1362_ep* ep)
{
	struct isp1362_request* req;
	while (!list_empty(&ep->queue)) {
		req = list_entry(ep->queue.next, struct isp1362_request, queue);
		request_done(ep, req, -ESHUTDOWN);
	}
}

  /* Reset handler
   */
static void reset_udc(struct isp1362_udc* dev)
{
	struct isp1362_ep* ep;
	unsigned long flags;

	spin_lock_irqsave(&dev->lock, flags);

	dev->gadget.speed = USB_SPEED_UNKNOWN;

	  /* Stop pending IO
	   */
	reset_ep(&dev->ep[0]);
	list_for_each_entry(ep, &dev->gadget.ep_list, ep.ep_list) {
		reset_ep(ep);
	}

	if (dev->driver)
		dev->driver->disconnect(&dev->gadget);

	spin_unlock_irqrestore (&dev->lock, flags);

	configure_endpoints();
	start_udc();
}

/* when a driver is successfully registered, it will receive
 * control requests including set_configuration(), which enables
 * non-control requests.  then usb traffic follows until a
 * disconnect is reported.  then a host may connect again, or
 * the driver might get unbound.
 */
int usb_gadget_register_driver(struct usb_gadget_driver *driver)
{
	struct isp1362_udc	*dev = the_controller;
	int			retval;

	if (!driver
	    || driver->speed != USB_SPEED_FULL
	    || !driver->bind
	    || !driver->unbind
	    || !driver->disconnect
	    || !driver->setup)
		return -EINVAL;
	if (!dev)
		return -ENODEV;
	if (dev->driver)
		return -EBUSY;

	/*
	 * We need to configure the endpoints _before_ we call the bind
	 * function. This is needed for the ether.c driver from 2.4.30er
	 * kernel.
	 */
	configure_endpoints();

	/* first hook up the driver ... */
	dev->driver = driver;

	retval = driver->bind(&dev->gadget);
	if (retval) {
		debug("bind to driver %s --> error %d\n",
				driver->driver.name, retval);
		dev->driver = 0;
		return retval;
	}

	/* ... then enable host detection and ep0; and we're ready
	 * for set_configuration as well as eventual disconnect.
	 * NOTE:  this shouldn't power up until later.
	 */
	start_udc();

	debug("registered gadget driver '%s'\n", driver->driver.name);

	return 0;
}
EXPORT_SYMBOL(usb_gadget_register_driver);

int usb_gadget_unregister_driver (struct usb_gadget_driver *driver)
{
	struct isp1362_udc *dev = the_controller;
	unsigned long	flags;

	if (!dev)
		return -ENODEV;
	if (!driver || driver != dev->driver)
		return -EINVAL;

	spin_lock_irqsave (&dev->lock, flags);
	/*stop_activity (dev, driver);*/
	isp1362_reset_device();
	spin_unlock_irqrestore (&dev->lock, flags);

	driver->unbind (&dev->gadget);
	dev->driver = 0;

	debug ("unregistered driver '%s'\n", driver->driver.name);
	return 0;
}
EXPORT_SYMBOL (usb_gadget_unregister_driver);

  /* Get Status request handler
   */
static void isp1362_get_status(struct usb_ctrlrequest * req)
{
	unsigned char status_buf[2] = { 0, 0 };
	int rc = 0;
	int n;

	/* return status bit flags */
	n = request_target(req->bRequestType);
	switch (n) {
	case USB_RECIP_DEVICE:
		if (0)			/* self_powered */
			status_buf[0] |= 1;
		break;
	case USB_RECIP_INTERFACE:
		break;
	case USB_RECIP_ENDPOINT:
		/* return stalled bit */
		n = windex_to_ep_num(__le16_to_cpu(req->wIndex));
		if (isp1362_check_status(n, 0) & EPSTAT_EPSTAL) {
			status_buf[0] |= 1;
		}
		break;
	default:
		printk("isp1362: Unknown target (%d) in GET_STATUS\n", n);
		break;
	}

	rc = isp1362_setup_data_send(status_buf,
				     __le16_to_cpu(req->wLength),
				     sizeof(status_buf));
	if (rc) {
		printk("isp1362: Get Status failed. Stall.\n");
		isp1362_cnt_err_setup++;
		isp1362_stall_endpoint(0, EP_DIR_IN);
	}
}

  /* Endpoint 0 OUT interrupt handler
   */
static void isp1362_ep0_out_int_hndlr (void)
{
	struct isp1362_udc	*dev = the_controller;
	u8 status = isp1362_read_status (0, EP_DIR_OUT);
	struct usb_ctrlrequest req;
	ulong res;

	verbose("EP0 OUT status: %02X\n", (uint)status);

	if (dev->gadget.speed == USB_SPEED_UNKNOWN)
	{
		dev->gadget.speed = USB_SPEED_FULL;
	}

	if (!(status & EPSTAT_SETUPT))
	{
#if 0
		dev->ep0dir_in = 0;
#endif
		isp1362_ep_int_hndlr(&dev->ep[0]);
		goto Done;
	}

#if 0
	isp1362_read_status (0, EP_DIR_OUT);
#endif

	res = isp1362_read_buffer (0, &req, sizeof(req));
	if (res != sizeof(req))
	{
		printk ("isp1362: Invalid SETUP size\n");
		isp1362_cnt_err_setup++;
		isp1362_stall_endpoint (0, EP_DIR_IN);
		goto Done;
	}

	le16_to_cpus (&req.wValue);
	le16_to_cpus (&req.wIndex);
	le16_to_cpus (&req.wLength);

#ifdef VERBOSE
	{
		unsigned char * pdb = (unsigned char *) &req;
		verbose ("query: %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
			pdb[0], pdb[1], pdb[2], pdb[3],
			pdb[4], pdb[5], pdb[6], pdb[7]);
		isp1362_print_request( &req );
	}
#endif

	dev->ep0dir_in = (req.bRequestType & USB_DIR_IN) || !req.wLength;

	if ((req.bRequestType & USB_TYPE_MASK) != USB_TYPE_STANDARD) {
		res = dev->driver->setup(&dev->gadget, &req);
		if (res < 0) {
		    debug("isp1362: gadget failed to process req, err %i\n", (int)res);
		}
		return;
	}

	/* Handle it */
	switch (req.bRequest)
	{

	  /* This first bunch have no data phase */

	case USB_REQ_SET_ADDRESS:
		isp1362_set_address(req.wValue & 0x7F);
		isp1362_send_empty_packet (0);
		break;

	case USB_REQ_CLEAR_FEATURE:
	case USB_REQ_SET_FEATURE:
		if ( req.wValue == 0 ) { /* setting ENDPOINT_HALT/STALL */
			int ep = windex_to_ep_num(req.wValue);
			set_halt(&dev->ep[ep], req.bRequest == USB_REQ_SET_FEATURE);
		}
		else {
			isp1362_cnt_err_setup++;
			debug( "isp1362: Unsupported feature selector "
				"(%d) in set feature\n", req.wValue);
		}
		isp1362_send_empty_packet (0);
		break;

	case USB_REQ_GET_STATUS:
		isp1362_get_status (&req);
		break;

	default :
		res = dev->driver->setup(&dev->gadget, &req);
		if (res < 0) {
		    debug("isp1362: Failed to change config, err %i\n", (int)res);
		}
		break;
	 }

Done:
	 return;
}

static void isp1362_ep_int_hndlr(struct isp1362_ep* ep)
{
	struct isp1362_request* req;
	unsigned long flags;
	struct isp1362_udc* dev = the_controller;

again:
	spin_lock_irqsave(&dev->lock, flags);

	if (list_empty(&ep->queue)) {
		/*verbose("queue empty, status %#x\n",
			(int)isp1362_read_status(ep_num(ep),
			dev->ep0dir_in ? EP_DIR_IN : EP_DIR_OUT));*/
		isp1362_read_status(ep_num(ep), dev->ep0dir_in ? EP_DIR_IN : EP_DIR_OUT);
		spin_unlock_irqrestore(&dev->lock, flags);
		return;
	}

	req = list_entry(ep->queue.next, struct isp1362_request, queue);

	if (do_request(ep, req)) {
		  /* Some upper drivers queue the same request
		   * in their completion routine, so we must
		   * clean up our portion of request data and
		   * unlock before signaling about completion.
		   */
		spin_unlock_irqrestore(&dev->lock, flags);
		verbose("req %#x to %s done dir %ud\n", (int)req, ep->ep.name, dev->ep0dir_in);
		request_done(ep, req, 0);
		if ((ep->bEndpointAddress & USB_DIR_IN) ||
		    (!ep->bEndpointAddress && dev->ep0dir_in))
			goto again;
		return;
	}

	/*verbose("req %#x to %s remains %d fn %d\n", (int)req, ep->ep.name,
	  req->req.length-req->req.actual,  (int)isp1362_frame_number());*/
	spin_unlock_irqrestore(&dev->lock, flags);
}

  /* ISP1362 DC interrupt handler
   */
static void isp1362_int_hndlr(int irq, void *dev_id, struct pt_regs *regs)
{
	ulong status = isp1362_read_reg32 (RD_DcInterrupt);
	struct isp1362_udc* dev = the_controller;

	if (!status)
		return;

#ifdef CONFIG_NSCU
	/* stop a running timer */
	if (isp1362_timer.data) {
		del_timer_sync(&isp1362_timer);
	}
#endif

	isp1362_cnt_int++;

	verbose ("[%lu] DC INTERRUPT: %08X\n",
		 isp1362_cnt_int, (uint)status);

	if (status & INT_RST) {
		isp1362_cnt_reset++;
		reset_udc(dev);
		return;
	}

	if (status & INT_RESM) {
		isp1362_cnt_resume++;
		isp1362_resume ();
	}

	if (status & INT_SUSP) {
		isp1362_cnt_suspend++;
		isp1362_suspend ();
	}

	if (status & INT_EP_ALL) {
		int cur_ep;
		for (cur_ep = 0; cur_ep < ISP1362_UDC_NUM_ENDPOINTS; cur_ep++)
			if (status & (INT_EP << cur_ep))
				isp1362_ep_int_hndlr(&dev->ep[cur_ep]);
	}

	if (status & INT_EP0_OUT) {
		isp1362_ep0_out_int_hndlr ();
	}
}

static struct isp1362_udc memory = {
	.gadget = {
		.ops		= &isp1362_udc_ops,
		.ep0		= &memory.ep[0].ep,
		.name		= driver_name,
		.dev = {
			.bus_id		= "gadget",
		},
	},

	.ep = {
	/* control endpoint */
	/*.ep[0] =*/ {
		.ep = {
			.name		= ep0name,
			.ops		= &isp1362_ep_ops,
			.maxpacket	= EP0_FIFO_SIZE,
		},
		.dev		= &memory,
	},

	/* first group of endpoints */
	/*.ep[1] =*/ {
		.ep = {
			.name		= "ep1out-bulk",
			.ops		= &isp1362_ep_ops,
			.maxpacket	= BULK_FIFO_SIZE,
		},
		.dev		= &memory,
		.fifo_size	= BULK_FIFO_SIZE,
		.bEndpointAddress = 1,
		.bmAttributes	= USB_ENDPOINT_XFER_BULK,
	},
	/*.ep[2] =*/ {
		.ep = {
			.name		= "ep2in-bulk",
			.ops		= &isp1362_ep_ops,
			.maxpacket	= BULK_FIFO_SIZE,
		},
		.dev		= &memory,
		.fifo_size	= BULK_FIFO_SIZE,
		.bEndpointAddress = USB_DIR_IN | 2,
		.bmAttributes	= USB_ENDPOINT_XFER_BULK,
	},
	/*.ep[3] =*/ {
		.ep = {
			.name		= "ep3in-int",
			.ops		= &isp1362_ep_ops,
			.maxpacket	= 8,
		},
		.dev		= &memory,
		.fifo_size	= 8,
		.bEndpointAddress = USB_DIR_IN | 3,
		.bmAttributes	= USB_ENDPOINT_XFER_INT,
	}

	}
};

int __init isp1362_init( void )
{
	int retval = 0;
	ushort chip_id;

#ifdef CONFIG_NSCU
	volatile immap_t *immr;
	volatile cpm8xx_t *cp;

	immr = (immap_t*)IMAP_ADDR;
	cp = (cpm8xx_t*)&immr->im_cpm;

	cp->cp_pbpar &= ~ISP1362_RESET;
	cp->cp_pbdir |= ISP1362_RESET;
	cp->cp_pbodr &= ~ISP1362_RESET;
	cp->cp_pbdat |= ISP1362_RESET;

	/*
	 * init the timer used for restart upon burst hangup
	 */
	init_timer(&isp1362_timer);
	isp1362_timer.data = 0;
	isp1362_timer.function = isp1362_timer_func;
#endif

#if CONFIG_PROC_FS
	devfs_handle = devfs_register(NULL, DEBUG_DEVNAME, DEVFS_FL_AUTO_DEVNUM,
				      0, 0,
				      S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP,
				      &debug_fops, &debug_dev);

	memset(&debug_dev, 0, sizeof(debug_dev_t));

	debug_proc_dir = proc_mkdir(DEBUG_DEVNAME, NULL);
	if (debug_proc_dir == NULL) {
		printk(KERN_ERR "debug: Unable to create /proc/%s entry\n", DEBUG_DEVNAME);
		return -ENOMEM;
	}

	debug_proc_usb_dir = create_proc_entry(DEBUG_USB_DIR, 0644, debug_proc_dir);
	if (debug_proc_usb_dir == NULL) {
		printk(KERN_ERR "DEBUG: Unable to create /proc/%s/%s entry\n", DEBUG_DEVNAME, DEBUG_USB_DIR);
		remove_proc_entry("debug", NULL);
		return -ENOMEM;
	}
	debug_proc_usb_dir->read_proc = isp1362_read_proc;
	debug_proc_usb_dir->write_proc = isp1362_write_proc;
#endif

	isp1362_io_base = ioremap(ISP1362_IO_PHYS, PAGE_SIZE);

	chip_id = isp1362_chip_id ();
	if (chip_id != 0x3630)
	{
		printk ("isp1362: Invalid chip identification, %#x\n", chip_id);
		return -1;
	}

	isp1362_reset_device();

	retval = request_8xxirq(ISP1362_IRQ_DC, isp1362_int_hndlr, 0,
				"ISP1362 DC", NULL);
	if (retval) {
		printk("isp1362: Couldn't request USB irq rc=%d\n", retval);
		iounmap((void *)isp1362_io_base);
		return retval;
	}

	the_controller = &memory;

	printk( "ISP1362 USB Controller Core Initialized\n");
	return 0;
}
module_init( isp1362_init );

void __exit isp1362_exit( void )
{
    printk("Unloading ISP1362 USB Controller\n");

#if CONFIG_PROC_FS
    devfs_unregister(devfs_handle);

    remove_proc_entry(DEBUG_USB_DIR, debug_proc_usb_dir);
    remove_proc_entry(DEBUG_DEVNAME, NULL);

    remove_proc_entry (PROC_SUSPEND_NAME, NULL);
#endif

    free_irq(ISP1362_IRQ_DC, NULL);

    iounmap((void *)isp1362_io_base);
}
module_exit( isp1362_exit );

MODULE_DESCRIPTION (DRIVER_DESC);
MODULE_AUTHOR ("Wolfgang Denk, DENX Software Engineering");
MODULE_LICENSE ("GPL");
