/*
 *  linux/drivers/video/fbcon-yuv16.c -- Low level frame buffer operations for
 *					 16 bpp YUV packed pixels
 *
 *	 Adapted from fbcon_cfb16.c 25 Nov 2002 by Wolfgang Grandegger
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of this archive for
 *  more details.
 *
 *  In this format each 4 bytes is two pixels. Each four bytes is two Y's,
 *  a Cb and a Cr. Each Y goes to one of the pixels, and the Cb and Cr belong
 *  to both pixels.
 */

#include <linux/module.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/string.h>
#include <linux/fb.h>
#include <asm/io.h>

#include <video/fbcon.h>
#include <video/fbcon-yuv16.h>

#undef DEBUG
#ifdef DEBUG
#define debugk(fmt,args...)	printk (fmt ,##args)
#else
#define debugk(fmt,args...)
#endif

static u32 tab_yuv16[] = {
#if defined(__BIG_ENDIAN)
    0x00000000, 0x0000ffff, 0xffff0000, 0xffffffff
#elif defined(__LITTLE_ENDIAN)
    0x00000000, 0xffff0000, 0x0000ffff, 0xffffffff
#else
#error FIXME: No endianness??
#endif
};

void fbcon_yuv16_setup(struct display *p)
{
    debugk(__FUNCTION__": disp=0x%p\n", p);
    p->next_line = p->line_length ? p->line_length : p->var.xres_virtual<<1;
    p->next_plane = 0;
}

void fbcon_yuv16_bmove(struct display *p, int sy, int sx, int dy, int dx,
		       int height, int width)
{
    int bytes = p->next_line, linesize = bytes * fontheight(p), rows;
    u8 *src, *dst;

    debugk(__FUNCTION__": disp=0x%p sy=%d sx=%d dy=%d dx=%d height=%d width=%d\n",
	   p, sy, sx, dy, dx, height, width);
    if (sx == 0 && dx == 0 && width * fontwidth(p) * 2 == bytes) {
	fb_memmove(p->screen_base + dy * linesize,
		  p->screen_base + sy * linesize,
		  height * linesize);
	return;
    }
    if (fontwidthlog(p)) {
	sx <<= fontwidthlog(p)+1;
	dx <<= fontwidthlog(p)+1;
	width <<= fontwidthlog(p)+1;
    } else {
	sx *= fontwidth(p)*2;
	dx *= fontwidth(p)*2;
	width *= fontwidth(p)*2;
    }
    if (dy < sy || (dy == sy && dx < sx)) {
	src = p->screen_base + sy * linesize + sx;
	dst = p->screen_base + dy * linesize + dx;
	for (rows = height * fontheight(p); rows--;) {
	    fb_memmove(dst, src, width);
	    src += bytes;
	    dst += bytes;
	}
    } else {
	src = p->screen_base + (sy+height) * linesize + sx - bytes;
	dst = p->screen_base + (dy+height) * linesize + dx - bytes;
	for (rows = height * fontheight(p); rows--;) {
	    fb_memmove(dst, src, width);
	    src -= bytes;
	    dst -= bytes;
	}
    }
}

static inline void rectfill(u8 *dest, int width, int height, u32 data,
			    int linesize)
{
    int i;
    u32 *p = (u32 *)dest;

    debugk(__FUNCTION__": dest=0x%p width=%d height=%d data=0x%x linesize=%d\n",
	   dest, width, height, data, linesize);

    for (i = 0; i < width/4; i++) {
	fb_writel(data, p++);
	fb_writel(data, p++);
    }
    if (width & 2)
	fb_writel(data, p++);
    if (width & 1)
	fb_writew(data, (u16*)p);
    dest += linesize;
}

void fbcon_yuv16_clear(struct vc_data *conp, struct display *p, int sy, int sx,
		       int height, int width)
{
    u8 *dest;
    int bytes = p->next_line, lines = height * fontheight(p);
    u32 bgx;

    debugk(__FUNCTION__": conp=0x%p disp=0x%p sy=%d sx=%d height=%d width=%d\n",
	   conp, p, sy, sx, height, width);

    dest = p->screen_base + sy * fontheight(p) * bytes + sx * fontwidth(p) * 2;

    bgx = ((u32 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];

    width *= fontwidth(p)/4;
    if (width * 8 == bytes)
	rectfill(dest, lines * width * 4, 1, bgx, bytes);
    else
	rectfill(dest, width * 4, lines, bgx, bytes);
}

void fbcon_yuv16_putc(struct vc_data *conp, struct display *p, int c, int yy,
		      int xx)
{
    u8 *dest, *cdat, bits;
    int bytes = p->next_line, rows;
    u32 eorx, fgx, bgx;

    debugk(__FUNCTION__": conp=0x%p disp=0x%p c=0x%x yy=%d xx=%d\n",
	   conp, p, c, yy, xx);

    dest = p->screen_base + yy * fontheight(p) * bytes + xx * fontwidth(p) * 2;

    fgx = ((u32 *)p->dispsw_data)[attr_fgcol(p, c)];
    bgx = ((u32 *)p->dispsw_data)[attr_bgcol(p, c)];

    eorx = fgx ^ bgx;

    switch (fontwidth(p)) {
    case 4:
    case 8:
	cdat = p->fontdata + (c & p->charmask) * fontheight(p);
	for (rows = fontheight(p); rows--; dest += bytes) {
	    bits = *cdat++;
	    fb_writel((tab_yuv16[bits >> 6] & eorx) ^ bgx, dest);
	    fb_writel((tab_yuv16[bits >> 4 & 3] & eorx) ^ bgx, dest+4);
	    if (fontwidth(p) == 8) {
		fb_writel((tab_yuv16[bits >> 2 & 3] & eorx) ^ bgx, dest+8);
		fb_writel((tab_yuv16[bits & 3] & eorx) ^ bgx, dest+12);
	    }
	}
	break;
    case 12:
    case 16:
	cdat = p->fontdata + ((c & p->charmask) * fontheight(p) << 1);
	for (rows = fontheight(p); rows--; dest += bytes) {
	    bits = *cdat++;
	    fb_writel((tab_yuv16[bits >> 6] & eorx) ^ bgx, dest);
	    fb_writel((tab_yuv16[bits >> 4 & 3] & eorx) ^ bgx, dest+4);
	    fb_writel((tab_yuv16[bits >> 2 & 3] & eorx) ^ bgx, dest+8);
	    fb_writel((tab_yuv16[bits & 3] & eorx) ^ bgx, dest+12);
	    bits = *cdat++;
	    fb_writel((tab_yuv16[bits >> 6] & eorx) ^ bgx, dest+16);
	    fb_writel((tab_yuv16[bits >> 4 & 3] & eorx) ^ bgx, dest+20);
	    if (fontwidth(p) == 16) {
		fb_writel((tab_yuv16[bits >> 2 & 3] & eorx) ^ bgx, dest+24);
		fb_writel((tab_yuv16[bits & 3] & eorx) ^ bgx, dest+28);
	    }
	}
	break;
    }
}

#define HISHORT(i)	((i >> 16)&0xffff)
#define LOSHORT(i)	(i & 0xffff)
#define HICHAR(s)	((i >> 8)&0xff)
#define LOCHAR(s)	(i & 0xff)
#define HI(c)		((c >> 4)&0xf)
#define LO(c)		(c & 0xf)
#define SWAPINT(i)	(HISHORT(i) | (LOSHORT(i) << 16))
#define SWAPSHORT(s)	(HICHAR(s) | (LOCHAR(s) << 8))
#define SWAPCHAR(c)	(HI(c) | (LO(c) << 4))

void fbcon_yuv16_putcs(struct vc_data *conp, struct display *p,
		       const unsigned short *s, int count, int yy, int xx)
{
    u8 *cdat, *dest, *dest0;
    u16 c;
    int rows, bytes = p->next_line;
    u32 eorx, fgx, bgx;

    debugk(__FUNCTION__": conp=0x%p disp=0x%p s=0x%p count=%d yy=%d xx=%d bytes=%d fw=%d\n",
	   conp, p, s, count, yy, xx, bytes, fontwidth(p));

    dest0 = p->screen_base + yy * fontheight(p)* bytes + xx * fontwidth(p) * 2;
#if 0
    printk(__FUNCTION__": xx=%d yy=%d dest0=0x%p\n", xx, yy, dest0);
    printk(__FUNCTION__": scr_readw(s)=%x attr_fgcol=%x attr_bgcol=%x\n",
	   scr_readw(s),
	   attr_fgcol(p, scr_readw(s)), attr_bgcol(p, scr_readw(s)));
#endif
    fgx = ((u32 *)p->dispsw_data)[attr_fgcol(p, scr_readw(s))];
    bgx = ((u32 *)p->dispsw_data)[attr_bgcol(p, scr_readw(s))];

#if 0
    if (xx & 1) {
	fgx = SWAPINT(fgx);
	bgx = SWAPINT(bgx);
    }
#endif
    eorx = fgx ^ bgx;

    switch (fontwidth(p)) {
    case 4:
    case 8:
	while (count--) {
	    c = scr_readw(s++) & p->charmask;
	    cdat = p->fontdata + c * fontheight(p);
	    for (rows = fontheight(p), dest = dest0; rows--; dest += bytes) {
		u8 bits = *cdat++;
	        fb_writel((tab_yuv16[bits >> 6] & eorx) ^ bgx, dest);
	        fb_writel((tab_yuv16[bits >> 4 & 3] & eorx) ^ bgx, dest+4);
		if (fontwidth(p) == 8) {
		    fb_writel((tab_yuv16[bits >> 2 & 3] & eorx) ^ bgx, dest+8);
		    fb_writel((tab_yuv16[bits & 3] & eorx) ^ bgx, dest+12);
		}
	    }
	    dest0 += fontwidth(p)*2;;
	}
	break;
    case 12:
    case 16:
	while (count--) {
	    c = scr_readw(s++) & p->charmask;
	    cdat = p->fontdata + (c * fontheight(p) << 1);
	    for (rows = fontheight(p), dest = dest0; rows--; dest += bytes) {
		u8 bits = *cdat++;
	        fb_writel((tab_yuv16[bits >> 6] & eorx) ^ bgx, dest);
	        fb_writel((tab_yuv16[bits >> 4 & 3] & eorx) ^ bgx, dest+4);
	        fb_writel((tab_yuv16[bits >> 2 & 3] & eorx) ^ bgx, dest+8);
	        fb_writel((tab_yuv16[bits & 3] & eorx) ^ bgx, dest+12);
		bits = *cdat++;
	        fb_writel((tab_yuv16[bits >> 6] & eorx) ^ bgx, dest+16);
	        fb_writel((tab_yuv16[bits >> 4 & 3] & eorx) ^ bgx, dest+20);
		if (fontwidth(p) == 16) {
		    fb_writel((tab_yuv16[bits >> 2 & 3] & eorx) ^ bgx, dest+24);
		    fb_writel((tab_yuv16[bits & 3] & eorx) ^ bgx, dest+28);
		}
	    }
	    dest0 += fontwidth(p)*2;
	}
	break;
    }
}

void fbcon_yuv16_revc(struct display *p, int xx, int yy)
{
    u8 *dest;
    int bytes = p->next_line, rows;

    debugk(__FUNCTION__": disp=0x%p yy=%d xx=%d\n", p, yy, xx);

    dest = p->screen_base + yy * fontheight(p) * bytes + xx * fontwidth(p)*2;
    for (rows = fontheight(p); rows--; dest += bytes) {
	switch (fontwidth(p)) {
	case 16:
	    fb_writel(fb_readl(dest+24) ^ 0xffffffff, dest+24);
	    fb_writel(fb_readl(dest+28) ^ 0xffffffff, dest+28);
	    /* FALL THROUGH */
	case 12:
	    fb_writel(fb_readl(dest+16) ^ 0xffffffff, dest+16);
	    fb_writel(fb_readl(dest+20) ^ 0xffffffff, dest+20);
	    /* FALL THROUGH */
	case 8:
	    fb_writel(fb_readl(dest+8) ^ 0xffffffff, dest+8);
	    fb_writel(fb_readl(dest+12) ^ 0xffffffff, dest+12);
	    /* FALL THROUGH */
	case 4:
	    fb_writel(fb_readl(dest+0) ^ 0xffffffff, dest+0);
	    fb_writel(fb_readl(dest+4) ^ 0xffffffff, dest+4);
	}
    }
}

void fbcon_yuv16_clear_margins(struct vc_data *conp, struct display *p,
			       int bottom_only)
{
    int bytes = p->next_line;
    u32 bgx;

    unsigned int right_start = conp->vc_cols*fontwidth(p);
    unsigned int bottom_start = conp->vc_rows*fontheight(p);
    unsigned int right_width, bottom_width;

    debugk(__FUNCTION__": conp=0x%p disp=0x%p bottom_only=%d\n",
	   conp, p, bottom_only);

    bgx = ((u32 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];

    if (!bottom_only && (right_width = p->var.xres-right_start))
	rectfill(p->screen_base+right_start*2, right_width,
		 p->var.yres_virtual, bgx, bytes);
    if ((bottom_width = p->var.yres-bottom_start))
	rectfill(p->screen_base+(p->var.yoffset+bottom_start)*bytes,
		 right_start, bottom_width, bgx, bytes);
}


    /*
     *  `switch' for the low level operations
     */

struct display_switch fbcon_yuv16 = {
    setup:		fbcon_yuv16_setup,
    bmove:		fbcon_yuv16_bmove,
    clear:		fbcon_yuv16_clear,
    putc:		fbcon_yuv16_putc,
    putcs:		fbcon_yuv16_putcs,
    revc:		fbcon_yuv16_revc,
    clear_margins:	fbcon_yuv16_clear_margins,
    fontwidthmask:	FONTWIDTH(4)|FONTWIDTH(8)|FONTWIDTH(12)|FONTWIDTH(16)
};


#ifdef MODULE
int init_module(void)
{
    return 0;
}

void cleanup_module(void)
{}
#endif /* MODULE */


    /*
     *  Visible symbols for modules
     */

EXPORT_SYMBOL(fbcon_yuv16);
EXPORT_SYMBOL(fbcon_yuv16_setup);
EXPORT_SYMBOL(fbcon_yuv16_bmove);
EXPORT_SYMBOL(fbcon_yuv16_clear);
EXPORT_SYMBOL(fbcon_yuv16_putc);
EXPORT_SYMBOL(fbcon_yuv16_putcs);
EXPORT_SYMBOL(fbcon_yuv16_revc);
EXPORT_SYMBOL(fbcon_yuv16_clear_margins);
