/*
 * drivers/video/mpc823lcd_fb.c
 *
 * LCD frame buffer driver for PowerPC MPC823 family.
 *
 * Copyright (c) 2000 Dan Malek (dmalek@jlc.net)
 *
 * Copyright (C) 2000, 2001 MontaVista Software, Inc.
 *
 * Copyright (C) 2002 DENX
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
*/

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/module.h>
#include <linux/bootmem.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>

#include <linux/console.h>

#include <linux/fb.h>
#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>

#ifdef CONFIG_PM
#include <linux/pm.h>
#endif

#if defined(CONFIG_R360MPI) && defined(CONFIG_8xx_PWM)
#include <linux/r360_pwm.h>
#endif

#include "mpc823_fb.h"

/* Comm processor */
extern cpm8xx_t *cpmp;

/*
 * Debug settings
 */
#ifdef CONFIG_FB_8xx_DEBUG
#define debugk(fmt,args...) printk("fb:" fmt ,##args)
#else
#define debugk(x...)
#endif /* CONFIG_FB_8xx_DEBUG */

/*
 * Information about displays we are using.  This is for configuring
 * the LCD controller and memory allocation.  Someone has to know what
 * is connected, as we can't autodetect anything.
 */
#define CFG_HIGH	0	/* Pins are active high */
#define CFG_LOW 	1	/* Pins are active low */

typedef struct vidinfo {
	ushort vl_col;		/* Number of columns (i.e. 640) */
	ushort vl_row;		/* Number of rows (i.e. 480) */
	ushort vl_width;	/* Width of display area in millimeters */
	ushort vl_height;	/* Height of display area in millimeters */

	/* LCD configuration register. */
	u_char vl_clkp;		/* Clock polarity */
	u_char vl_oep;		/* Output Enable polarity */
	u_char vl_hsp;		/* Horizontal Sync polarity */
	u_char vl_vsp;		/* Vertical Sync polarity */
	u_char vl_dp;		/* Data polarity */
	u_char vl_bpix;		/* Bits per pixel, 0 = 1, 1 = 2, 2 = 4, 3 = 8 */
	u_char vl_lbw;		/* LCD Bus width, 0 = 4, 1 = 8 */
#if 1				/* XXX FIXME ??? */
	u_char vl_splt;		/* DENX: Split display, 0 = single-scan, 1 = dual-scan */
#else
	u_char vl_splt;		/* MV:   Split display, 0 = dual, 1 = single */
#endif
	u_char vl_clor;		/* Color, 0 = mono, 1 = color */
	u_char vl_tft;		/* 0 = passive, 1 = TFT */

	/* Horizontal control register. Timing from data sheet. */
	ushort vl_wbl;		/* Wait between lines */

	/* Vertical control register. */
	u_char vl_vpw;		/* Vertical sync pulse width */
	u_char vl_lcdac;	/* LCD AC timing */
	u_char vl_wbf;		/* Wait between frames */
} vidinfo_t;

#define	LCD_MONOCHROME	0
#define LCD_COLOR2	1
#define LCD_COLOR4	2
#define LCD_COLOR8	3

/*=================================================================*/

/*-----------------------------------------------------------------*/
#ifdef CONFIG_EDT32F10
/*
* Emerging Display Technologies 320x240. Passive, monochrome, single scan.
*/
#define LCD_BPP	LCD_MONOCHROME
#define	LCD_DF	20

static vidinfo_t panel_info = {
	320, 240, 0, 0, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH,
	LCD_BPP, 0, 0, 0, 0, 0, 0, 0, 0
};
#endif

/*----------------------------------------------------------------------*/
#ifdef CONFIG_HITACHI_SP19X001_Z1A
/*
 *  Hitachi SP19X001-. Active, color, single scan.
 */
#define LCD_BPP	LCD_COLOR8

static vidinfo_t panel_info = {
    640, 480, 154, 116, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH,
    LCD_BPP, 1, 0, 1, 0, 0, 0, 0, 0
		/* wbl, vpw, lcdac, wbf */
};
#endif /* CONFIG_HITACHI_SP19X001_Z1A */

/*-----------------------------------------------------------------*/
#ifdef CONFIG_HITACHI_SX14Q001
/*
* Hitachi SX14Q001 320x240. Passive, color, single scan, 8 bit bus.
*/
#define LCD_BPP	LCD_COLOR8
#define	LCD_DF	1

static vidinfo_t panel_info = {
	320, 240, 0, 0, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH,
	LCD_BPP, 1, 0, 1, 0, 0, 0, 0, 4
};
#endif

/*-----------------------------------------------------------------*/
#ifdef CONFIG_HLD1045
/*
* HLD1045 display, 640x480. Active, color, single scan.
*/
#define LCD_BPP	LCD_COLOR8
#define	LCD_DF	1

static vidinfo_t panel_info = {
	640, 480, 0, 0, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_HIGH,
	LCD_BPP, 0, 0, 1, 1, 160, 3, 0, 48
/* wbl, vpw, lcdac, wbf */
};
#endif

/*-----------------------------------------------------------------*/
#ifdef CONFIG_KYOCERA_KCS057QV1AJ
/*
*  Kyocera KCS057QV1AJ-G23. Passive, color, single scan.
*/
#define LCD_BPP	LCD_COLOR4
#define LCD_DF	1

static vidinfo_t panel_info = {
	640, 480, 132, 99, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH,
	LCD_BPP, 1, 0, 1, 0, 5, 0, 0, 0
/* wbl, vpw, lcdac, wbf */
};
#endif				/* CONFIG_KYOCERA_KCS057QV1AJ */

/*-----------------------------------------------------------------*/
#ifdef CONFIG_NEC_NL6448AC33
/*
*  NEC NL6448AC33-18. Active, color, single scan.
*  Over time, we could either configure exactly one of these at build
*  time, or create an array of supported devices and find a way to
*  tell this driver what we have (EEPROM perhaps?).
*/
#define LCD_BPP	LCD_COLOR8
#define	LCD_DF	1

static vidinfo_t panel_info = {
	640, 480, 132, 99, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_HIGH,
	LCD_BPP, 0, 0, 1, 1, 144, 2, 0, 33
/* wbl, vpw, lcdac, wbf */
};
#endif

/*-----------------------------------------------------------------*/
#ifdef CONFIG_NEC_NL6448BC20
/*
*  NEC NL6448BC20-08. 6.5", 640x480. Active, color, single scan.
*/
#define LCD_BPP	LCD_COLOR8
#define	LCD_DF	1

static vidinfo_t panel_info = {
	640, 480, 132, 99, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_HIGH,
	LCD_BPP, 0, 0, 1, 1, 144, 2, 0, 33
/* wbl, vpw, lcdac, wbf */
};
#endif

/*-----------------------------------------------------------------*/
#ifdef CONFIG_OPTREX_BW
/*
* Hitachi SP14Q002 320x240. Passive, black & white, single scan(?).
*/
#define LCD_BPP	LCD_COLOR4
#define	LCD_DF	6

static vidinfo_t panel_info = {
	320, 240, 0, 0, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_LOW,
	LCD_BPP, 0, 0, 0, 0, 0, 0, 0, 4
};
#endif

/*----------------------------------------------------------------------*/
#ifdef CONFIG_PRIMEVIEW_V16C6448AC
/*
* Prime View V16C6448AC
*/
static vidinfo_t panel_info = {
	640, 480, 130, 98, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_HIGH,
	3, 0, 0, 1, 1, 144, 2, 0, 35
/* wbl, vpw, lcdac, wbf */
};
#endif				/* CONFIG_PRIMEVIEW_V16C6448AC */

/*-----------------------------------------------------------------*/
#ifdef CONFIG_SHARP_16x9
/*
* Sharp 320x240. Active, color, single scan.  It isn't 16x9, and I am
* not sure what it is.......
*/
#define LCD_BPP	LCD_COLOR8
#define	LCD_DF	1

static vidinfo_t panel_info = {
	320, 240, 0, 0, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH,
	LCD_BPP, 0, 0, 1, 1, 15, 4, 0, 3
};
#endif

/*-----------------------------------------------------------------*/
#ifdef CONFIG_SHARP_LQ104V7DS01
/*
*  SHARP LQ104V7DS01. 6.5", 640x480. Active, color, single scan.
*/
#define LCD_BPP	LCD_COLOR8
#define	LCD_DF	1

static vidinfo_t panel_info = {
	640, 480, 132, 99, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_LOW,
	LCD_BPP, 0, 0, 1, 1, 25, 1, 0, 33
/* wbl, vpw, lcdac, wbf */
};
#endif

/*-----------------------------------------------------------------*/
#ifdef CONFIG_SHARP_LQ057Q3DC02
/*
* Sharp LQ057Q3DC02 display, 320x240. Active, color, single scan.
*/
#define LCD_BPP	LCD_COLOR8
#define	LCD_DF	1

static vidinfo_t panel_info = {
	320, 240, 0, 0, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_HIGH,
	LCD_BPP, 0, 0, 1, 1, 15, 4, 0, 3
/* wbl, vpw, lcdac, wbf */
};
#endif

/*-----------------------------------------------------------------*/
#ifdef CONFIG_SHARP_LQ64D341
/*
* Sharp LQ64D341 display, 640x480. Active, color, single scan.
*/
#define LCD_BPP	LCD_COLOR8
#define	LCD_DF	1

static vidinfo_t panel_info = {
	640, 480, 0, 0, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_HIGH,
	LCD_BPP, 0, 0, 1, 1, 128, 16, 0, 32
/* wbl, vpw, lcdac, wbf */
};
#endif

/*-----------------------------------------------------------------*/
#ifdef CONFIG_SHARP_LQ084V1DG21
/*
* Sharp LQ084V1DG21 display, 640x480. Active, color, single scan.
*/
#define LCD_BPP LCD_COLOR8
#define LCD_DF  1

static vidinfo_t panel_info = {
	640, 480, 171, 129, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_LOW,
	LCD_BPP, 0, 0, 1, 1, 160, 3, 0, 48
/* wbl, vpw, lcdac, wbf */
};
#endif

/*=================================================================*/

#define NBITS(bit_code)		(1 << (bit_code))
#define NCOLORS(bit_code)	(1 << NBITS(bit_code))

static struct fb_info fb_info;
static struct display display;
static int currcon = 0;
static int inverse = 0;

#ifdef CONFIG_PM
/* Power management device */
static struct pm_dev *pm_dev;
#endif

/* Frame buffer memory information */
static unsigned long lcd_fb_base;
static unsigned long lcd_fb_line_length;

/* Desired mode */
static const char *mode_option __initdata = NULL;

/*
 * TODO: Add a modedb so the above actually works, instead of
 * relying on hard coded values in the panel_info. This way
 * we can add modes for our device (such as the Sharp and NEC
 * LCDs) and just specify which one to load via the unique name
 * identifier.
 */

static int
lcd823_getcolreg(unsigned regno, unsigned *red, unsigned *green, unsigned *blue,
		 unsigned *transp, struct fb_info *info)
{
	volatile cpm8xx_t *cp = cpmp;
	unsigned short colreg, *cmap_ptr;

	cmap_ptr = (unsigned short *) &cp->lcd_cmap[regno * 2];

	colreg = *cmap_ptr;
#if 0 /* was:  CONFIG_INVERSE_COLORS -- adjust vl_dp instead */
	colreg ^= 0x0FFF;
#endif

	*red = (colreg & 0x0F00) << 4;
	*green = (colreg & 0x00F0) << 8;
	*blue = (colreg & 0x000F) << 12;

	return 0;
}

static int
lcd823_setcolreg(unsigned regno, unsigned red, unsigned green, unsigned blue,
		 unsigned transp, struct fb_info *info)
{
	volatile cpm8xx_t *cp = cpmp;

#if !defined(CONFIG_OPTREX_BW) && !defined(CONFIG_EDT32F10)
	unsigned short colreg, *cmap_ptr;

	cmap_ptr = (unsigned short *) &cp->lcd_cmap[regno * 2];

	colreg = ((red & 0xF000) >> 4) |
	    ((green & 0xF000) >> 8) | ((blue & 0xF000) >> 12);
#if 0 /* was:  CONFIG_INVERSE_COLORS -- adjust vl_dp instead */
	colreg ^= 0x0FFF;
#endif
	*cmap_ptr = colreg;

#else /* is CONFIG_OPTREX_BW or CONFIG_EDT32F10 */

	if (regno == 0) {
		for (regno = 0; regno < 16; regno++) {
			cp->lcd_cmap[regno * 2] = regno & 0x0f;
			cp->lcd_cmap[(regno * 2) + 1] =
			    ((regno & 0x0f) << 4) | (regno & 0x0f);
		}
	} else {
		cp->lcd_cmap[regno * 2] = regno & 0x0f;
		cp->lcd_cmap[(regno * 2) + 1] =
		    ((regno & 0x0f) << 4) | (regno & 0x0f);
	}
#endif
	return 0;
}

static void
lcd823_install_cmap(int con, struct fb_info *info)
{
	struct fb_cmap *cmap;

	/* Make sure we're installing to the current console */
	if (con != currcon)
		return;

	if (fb_display[con].cmap.len != 0)
		cmap = &fb_display[con].cmap;
	else
		cmap = fb_default_cmap(NCOLORS(panel_info.vl_bpix));

	fb_set_cmap(cmap, 1, lcd823_setcolreg, info);
}

/*
 * To determine if these occur, just cat /proc/interrupts and look
 * at the interrupt count.
 */
static void
lcd823_interrupt(int irq, void *dev, struct pt_regs *regs)
{
	volatile lcd8xx_t *lcdp = &(((immap_t *) IMAP_ADDR)->im_lcd);
	lcdp->lcd_lcsr = 0xff;
}

/* ------------ Hardware Independent Functions ------------ */

static int
lcd823_open(struct fb_info *info, int user)
{
	MOD_INC_USE_COUNT;
	return 0;
}

static int
lcd823_release(struct fb_info *info, int user)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

unsigned long mpc8xx_fb_getsize(void);

static int
lcd823_get_fix(struct fb_fix_screeninfo *fix, int con, struct fb_info *info)
{
	debugk("%s(screeninfo: %p, con: %d, info: %p)\n",
		__FUNCTION__, fix, con, info);

	memset(fix, 0, sizeof (struct fb_fix_screeninfo));
	strcpy(fix->id, "MPC823");

	fix->smem_start  = virt_to_phys((void *) lcd_fb_base);
	fix->smem_len    = mpc8xx_fb_getsize();
	fix->line_length = lcd_fb_line_length;
	fix->type        = FB_TYPE_PACKED_PIXELS;

#if 1 /* XXX FIXME ??? */
	if (NBITS(panel_info.vl_bpix) == 1)	/* DENX */
#else
	if (info->var.bits_per_pixel == 1)	/* MV */
#endif
		fix->visual = FB_VISUAL_MONO01;
	else
		fix->visual = FB_VISUAL_PSEUDOCOLOR;

	return 0;
}

static int
lcd823_get_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	if (con == -1) {
		memset(var, 0, sizeof (struct fb_var_screeninfo));

		var->xres = var->xres_virtual = panel_info.vl_col;
		var->yres = var->yres_virtual = panel_info.vl_row;

		var->width = panel_info.vl_width;
		var->height = panel_info.vl_height;

		var->bits_per_pixel = NBITS(panel_info.vl_bpix);
		var->grayscale = (panel_info.vl_clor == 0);
		var->red.length = var->green.length = var->blue.length = 4;
	} else {
		*var = fb_display[con].var;
	}

	return 0;
}

static int
lcd823_set_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	if ((var->xres != panel_info.vl_col)
	    || (var->yres != panel_info.vl_row)
	    || (var->bits_per_pixel != NBITS(panel_info.vl_bpix))
	    || (var->grayscale && panel_info.vl_clor)
	    || ((var->grayscale == 0) && (panel_info.vl_clor == 0))) {
		return -EINVAL;
	}

	lcd823_get_var(var, con, info);

	return 0;
}

static int
lcd823_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
	struct fb_cmap *src_cmap;

	if (NBITS(panel_info.vl_bpix) == 1)
		return -EINVAL;

	if ((cmap->start + cmap->len) > NCOLORS(panel_info.vl_bpix))
		return -EINVAL;

	/* FIXME:  why can con be -1? */
	if ((con == currcon) || (con == -1))
		return fb_get_cmap(cmap, kspc, lcd823_getcolreg, info);

	if (fb_display[con].cmap.len)
		src_cmap = &fb_display[con].cmap;
	else
		src_cmap = fb_default_cmap(NCOLORS(panel_info.vl_bpix));
	fb_copy_cmap(src_cmap, cmap, kspc ? 0 : 2);

	return 0;
}

static int
lcd823_set_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
	int err;

	if (NBITS(panel_info.vl_bpix) == 1)
		return -EINVAL;

	if ((cmap->start + cmap->len) > NCOLORS(panel_info.vl_bpix))
		return -EINVAL;

	if (fb_display[con].cmap.len == 0) {
		int size = NCOLORS(panel_info.vl_bpix);
		if ((err = fb_alloc_cmap(&fb_display[con].cmap, size, 0)))
			return err;
	}

	/* FIXME:  why can con be -1? */
	if ((con == currcon) || (con == -1))
		return fb_set_cmap(cmap, kspc, lcd823_setcolreg, info);
	else
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);

	return 0;
}

static int
lcd823_pan_display(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	return -EINVAL;
}

static int
lcd823_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	     unsigned long arg, int con, struct fb_info *info)
{
	debugk("%s()\n", __FUNCTION__);
	return -EINVAL;
}

static int
lcd823_mmap(struct fb_info *info, struct file *fp, struct vm_area_struct *vma)
{
	unsigned long offset;

	debugk("%s(): \n", __FUNCTION__);

	offset = vma->vm_pgoff;

	if (offset != 0)
		return (-ENXIO);

	pgprot_val(vma->vm_page_prot) |=
	    (_PAGE_NO_CACHE | _PAGE_RW | _PAGE_HWWRITE | _PAGE_USER);

	vma->vm_pgoff += __pa(lcd_fb_base);

#ifdef CONFIG_FB_8xx_DEBUG
	printk("  virtAddr=%x\n", (int) lcd_fb_base);
	printk("  physAddr=%x\n", (int) __pa(lcd_fb_base));
	printk("  vma->vm_start= %08lX, vma->vm_offset= %08lX, \n"
	       "  vma->vm_end  = %08lX, vma->vm_page_prot= %08lX\n",
	       vma->vm_start, vma->vm_pgoff, vma->vm_end,
	       pgprot_val(vma->vm_page_prot));
#endif

	if (remap_page_range(vma->vm_start, vma->vm_pgoff,
			     vma->vm_end - vma->vm_start, vma->vm_page_prot))
		return -EAGAIN;
	vma->vm_file = fp;

	return 0;
}

static struct fb_ops lcd823_ops = {
      owner:		THIS_MODULE,
      fb_open:		lcd823_open,
      fb_release:	lcd823_release,
      fb_get_fix:	lcd823_get_fix,
      fb_get_var:	lcd823_get_var,
      fb_set_var:	lcd823_set_var,
      fb_get_cmap:	lcd823_get_cmap,
      fb_set_cmap:	lcd823_set_cmap,
      fb_pan_display:	lcd823_pan_display,
      fb_ioctl:		lcd823_ioctl,
      fb_mmap:		lcd823_mmap,	/* fb_mmap      */
      fb_rasterimg:	NULL		/* fb_rasterimg */
};

/* ------------------------------------------------------------------------- */

static int
lcd823_switch(int con, struct fb_info *info)
{
	debugk("%s(): con=%d -> %d\n", __FUNCTION__, currcon, con);

	/* Do we have to save the colormap ? */
	if (fb_display[currcon].cmap.len != 0)
		fb_get_cmap(&fb_display[currcon].cmap, 1, lcd823_getcolreg,
			    info);

	currcon = con;

	/* Install new colormap */
	lcd823_install_cmap(currcon, info);

	return 0;
}

static void
lcd823_blank(int blank, struct fb_info *info)
{
	u16 black[16];
	struct fb_cmap cmap;

	if (blank) {
		memset(black, 0, 16 * sizeof (u16));
		cmap.red = black;
		cmap.green = black;
		cmap.blue = black;
		cmap.transp = NULL;
		cmap.start = 0;
		cmap.len = 16;
		fb_set_cmap(&cmap, 1, lcd823_setcolreg, info);
	} else {
		debugk("%s(): con=%d\n", __FUNCTION__, currcon);
		lcd823_install_cmap(currcon, info);
	}
}

static int
lcd823_updatevar(int con, struct fb_info *info)
{
	return 0;
}

/* ------------------------------------------------------------------------- */

/*
 * This is called very early in the system initialization to
 * easily grab physically contiguous memory pages for the LCD
 * controller.
 */

unsigned long
mpc8xx_fb_getsize(void)
{
	unsigned long fb_size;

	lcd_fb_line_length =
	    (panel_info.vl_col * NBITS(panel_info.vl_bpix)) / 8;
	fb_size = lcd_fb_line_length * panel_info.vl_row;

	/* Round up to nearest full page */
	fb_size = (fb_size + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1);

	return (fb_size);
}

void __init
lcd_fb_setpage(unsigned long *memory_page_addr)
{
#ifdef CONFIG_FB_8xx_PRE_INIT_FB
	volatile lcd8xx_t *lcdp = &((immap_t *) IMAP_ADDR)->im_lcd;
	lcdp->lcd_lcfaa = __pa(*memory_page_addr);
	lcdp->lcd_lcfba = __pa(*memory_page_addr);
#endif

	/* Allocate pages for the frame buffer. */
	lcd_fb_base = *memory_page_addr;
//	*memory_page_addr = lcd_fb_base + mpc8xx_fb_getsize();
}

static void __init
lcd823_init_dev(void)
{
	volatile cpm8xx_t *cp = cpmp;
	volatile immap_t *immap = (immap_t *) IMAP_ADDR;
	volatile lcd8xx_t *lcdp = &immap->im_lcd;
	uint lccrtmp;
	uint lchcr_hpc_tmp;

	/*
	 * Initialize the LCD control register according to the LCD
	 * parameters defined.  We do everything here but enable
	 * the controller.
	 */

	lccrtmp = LCDBIT(LCCR_BNUM_BIT,
			 (((panel_info.vl_row * panel_info.vl_col) *
			   (1 << LCD_BPP)) / 128));

	lccrtmp |= LCDBIT(LCCR_CLKP_BIT, panel_info.vl_clkp) |
	    LCDBIT(LCCR_OEP_BIT, panel_info.vl_oep) |
	    LCDBIT(LCCR_HSP_BIT, panel_info.vl_hsp) |
	    LCDBIT(LCCR_VSP_BIT, panel_info.vl_vsp) |
	    LCDBIT(LCCR_DP_BIT, panel_info.vl_dp) |
	    LCDBIT(LCCR_BPIX_BIT, panel_info.vl_bpix) |
	    LCDBIT(LCCR_LBW_BIT, panel_info.vl_lbw) |
	    LCDBIT(LCCR_SPLT_BIT, panel_info.vl_splt) |
	    LCDBIT(LCCR_CLOR_BIT, panel_info.vl_clor) |
	    LCDBIT(LCCR_TFT_BIT, panel_info.vl_tft);

	lccrtmp |= ((SIU_LEVEL5 / 2) << 12);
	lccrtmp |= LCCR_EIEN;

	lcdp->lcd_lccr = lccrtmp;
	lcdp->lcd_lcsr = 0xff;	/* Clear pending interrupts */

	/* Initialize LCD controller bus priorities. */
#ifdef CONFIG_RBC823
	immap->im_siu_conf.sc_sdcr = (immap->im_siu_conf.sc_sdcr & ~0x0f) | 1;	/* RAID = 01, LAID = 00 */
#else
	immap->im_siu_conf.sc_sdcr &= ~0x0f;	/* RAID = LAID = 0 */

	/*
	 * set SHFT/CLOCK division factor 4
	 * This needs to be set based upon display type and processor
	 * speed.  The TFT displays run about 20 to 30 MHz.
	 * I was running 64 MHz processor speed.
	 * The value for this divider must be chosen so the result is
	 * an integer of the processor speed (i.e., divide by 3 with
	 * 64 MHz would be bad).
	 */
	immap->im_clkrst.car_sccr &= ~0x1f;
	immap->im_clkrst.car_sccr |= LCD_DF;
#endif /* CONFIG_RBC823 */

#if defined(CONFIG_RBC823)
	/* Enable LCD on port D.
	 */
	immap->im_ioport.iop_pddat &= 0x0300;
	immap->im_ioport.iop_pdpar |= 0x1CFF;
	immap->im_ioport.iop_pddir |= 0x1CFF;

	/* Configure LCD_ON, VEE_ON, CCFL_ON on port B.
	 */
	immap->im_cpm.cp_pbdat &= ~0x00005001;
	immap->im_cpm.cp_pbpar &= ~0x00005001;
	immap->im_cpm.cp_pbdir |=  0x00005001;
#elif !defined(CONFIG_EDT32F10)
	/* Enable LCD on port D. */
	immap->im_ioport.iop_pdpar |= 0x1fff;
	immap->im_ioport.iop_pddir |= 0x1fff;

	/* Enable LCD_A/B/C on port B. */
	cp->cp_pbpar |= 0x00005001;
	cp->cp_pbdir |= 0x00005001;
#else
	/* Enable LCD on port D. */
	immap->im_ioport.iop_pdpar |= 0x1DFF;
	immap->im_ioport.iop_pdpar &= ~0x0200;
	immap->im_ioport.iop_pddir |= 0x1FFF;
	immap->im_ioport.iop_pddat |= 0x0200;
#endif

	/*
	 * Load the physical address of the linear frame buffer
	 * into the LCD controller.
	 * BIG NOTE:  This has to be modified to load A and B depending
	 * upon the split mode of the LCD.
	 * Since we don't touch these pages, we don't care how the
	 * application assigns cache mode.
	 */
	lcdp->lcd_lcfaa = __pa(lcd_fb_base);
	lcdp->lcd_lcfba = __pa(lcd_fb_base);

	/*
	 * MORE HACKS...This must be updated according to 823 manual
	 * for different panels.
	 * Hack fix done by Udi Finkelstein - see below:
	 * Note: You better not try unsupported combinations such as
	 * 4-bit wide passive dual scan LCD at 4/8 Bit color.
	 */
	lchcr_hpc_tmp =
	    (panel_info.vl_col *
	     (panel_info.vl_tft ? 8 :
	      (((2 - panel_info.vl_lbw) << /* 4 bit=2, 8-bit = 1 */
	        /* use << to mult by: single scan = 1, dual scan = 2 */
	        panel_info.vl_splt) *
	       (panel_info.vl_bpix | 1)))) >> 3; /* 2/4 BPP = 1, 8/16 BPP = 3 */

	lcdp->lcd_lchcr = LCHCR_BO |
	    LCDBIT(LCHCR_AT_BIT, 4) |
	    LCDBIT(LCHCR_HPC_BIT, lchcr_hpc_tmp) |
	    panel_info.vl_wbl;

	lcdp->lcd_lcvcr = LCDBIT(LCVCR_VPW_BIT, panel_info.vl_vpw) |
	    LCDBIT(LCVCR_LCD_AC_BIT, panel_info.vl_lcdac) |
	    LCDBIT(LCVCR_VPC_BIT, panel_info.vl_row) | panel_info.vl_wbf;

#ifdef CONFIG_RPXLITE
	*((uint *) HIOX_CSR0_ADDR) &=
	    ~(HIOX_CSR0_ENVDOCLK | HIOX_CSR0_VDORST_HL);
#endif

	if (request_8xxirq(SIU_LEVEL5, lcd823_interrupt, 0, "lcd823", NULL) !=
	    0) {
		panic("fb: Couldn't allocate LCD IRQ!\n");
	}
}

static void __init
lcd823_enable(void)
{
	volatile immap_t *immap = (immap_t *) IMAP_ADDR;
	volatile lcd8xx_t *lcdp = &immap->im_lcd;

	/* See if the LCD has already been enabled */
	if (lcdp->lcd_lccr & LCCR_PON) {
		debugk("%s(): LCD already enabled\n", __FUNCTION__);
		return;
	}

	/* Enable the LCD panel */
#ifndef CONFIG_RBC823
	immap->im_siu_conf.sc_sdcr |= (1 << (31 - 25));	/* LAM = 1 */
#endif
	lcdp->lcd_lccr |= LCCR_PON;

#if defined(CONFIG_RBC823)
	/* Turn on display backlight */
	immap->im_cpm.cp_pbdat |= 0x00004000;
	mdelay(200); /* wait 200ms */
	/* Turn VEE_ON first */
	immap->im_cpm.cp_pbdat |= 0x00000001;
	mdelay(200); /* wait 200ms */
	/* Now turn on LCD_ON */
	immap->im_cpm.cp_pbdat |= 0x00001000;
#endif

#if defined(CONFIG_R360MPI) && defined(CONFIG_8xx_PWM)
	{
		extern int r360_pwm_write(unsigned char reg, unsigned char val);

		if (r360_pwm_write(PWM_SHDN, PWM_SHDN_DEF) < 0 ||
		    r360_pwm_write(PWM_LCDCT, PWM_LCDCT_DEF) < 0 ||
		    r360_pwm_write(PWM_LCDBL, PWM_LCDBL_DEF) < 0)
			printk("fb: %s(): Can't write PWM registers.\n", __FUNCTION__);
	}
#endif
}

#ifdef CONFIG_PM
static void
lcd823_disable(void)
{
	volatile immap_t *immap = (immap_t *) IMAP_ADDR;
	volatile lcd8xx_t *lcdp = &immap->im_lcd;

/* The following code is taken from U-Boot 0.4.8 - should this be enabled?
 */
#if defined(CONFIG_LWMON)
    {   uchar c = pic_read (0x60);
        c &= ~0x07;     /* Power off CCFL, Disable CCFL, Chip Disable LCD */
        pic_write (0x60, c);
    }
#elif defined(CONFIG_R360MPI)
    {
        extern void r360_i2c_lcd_write (uchar data0, uchar data1);

        r360_i2c_lcd_write(0x10, 0x00);
        r360_i2c_lcd_write(0x20, 0x00);
        r360_i2c_lcd_write(0x30, 0x00);
        r360_i2c_lcd_write(0x40, 0x00);
    }
#endif /* CONFIG_LWMON */
/* end of U-boot 0.4.8 code */

	/* Disable the LCD panel */
	lcdp->lcd_lccr &= ~LCCR_PON;
#ifdef CONFIG_RBC823
        /* Turn off display backlight, VEE and LCD_ON */
        immr->im_cpm.cp_pbdat &= ~0x00005001;
#else
	immap->im_siu_conf.sc_sdcr &= ~(1 << (31 - 25));	/* LAM = 0 */
#endif  /* CONFIG_RBC823 */
}

static int
lcd823_pm_request(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	debugk("%s(): rqst=%d\n", __FUNCTION__, rqst);

	switch (rqst) {
	case PM_SUSPEND:
		lcd823_disable();
		break;
	case PM_RESUME:
		lcd823_enable();
		break;
	}

	return 0;
}
#endif

int __init
lcd823_setup(char *options)
{
	char *this_opt;

	conswitchp = &fb_con;

	debugk("%s(): \n", __FUNCTION__);

	if (!options || !*options) {
		debugk("  no options\n");
		return 0;
	}

	fb_info.fontname[0] = '\0';

	while ((this_opt = strsep(&options, ","))) {
		if (!*this_opt)
			continue;
		if (!strncmp(this_opt, "inverse", 7)) {
			debugk("  inverse\n");
			inverse = 1;
			fb_invert_cmaps();
		} else if (!strncmp(this_opt, "font:", 5)) {
			strcpy(fb_info.fontname, this_opt + 5);
			debugk("  font=%s\n", fb_info.fontname);
		} else {
			mode_option = this_opt;
		}
	}

	return 0;
}

int __init
lcd823_init(void)
{
	strcpy(fb_info.modename, "fixed");
	fb_info.flags = FBINFO_FLAG_DEFAULT;
	fb_info.fbops = &lcd823_ops;
	fb_info.disp = &display;
	fb_info.switch_con = &lcd823_switch;
	fb_info.blank = &lcd823_blank;
	fb_info.updatevar = &lcd823_updatevar;
	fb_info.changevar = NULL;

	lcd823_get_var(&display.var, -1, &fb_info);
	display.screen_base = (char *) lcd_fb_base;
	display.line_length = lcd_fb_line_length;
	display.type = FB_TYPE_PACKED_PIXELS;
	display.inverse = inverse;

#if 1 /* XXX FIXME ??? */
	if (NBITS(panel_info.vl_bpix) == 1) {	/* DENX */
#else
	if (panel_info.vl_bpix == 1) {		/* MV */
#endif
		display.visual = FB_VISUAL_MONO01;
	} else {
		display.visual = FB_VISUAL_PSEUDOCOLOR;
		display.can_soft_blank = 1;
	}

	switch (display.var.bits_per_pixel) {
#ifdef FBCON_HAS_MFB
	case 1:
		display.dispsw = &fbcon_mfb;
		break;
#endif
#ifdef FBCON_HAS_CFB4
	case 4:
		display.dispsw = &fbcon_cfb4;
		break;
#endif
#ifdef FBCON_HAS_CFB8
	case 8:
		display.dispsw = &fbcon_cfb8;
		break;
#endif
	default:
		display.dispsw = &fbcon_dummy;
		break;
	}

	lcd823_init_dev();
	lcd823_install_cmap(0, &fb_info);
	lcd823_enable();

	if (register_framebuffer(&fb_info) < 0) {
		printk(KERN_ERR
		       "fb: Error registering MPC823 LCD frame buffer device\n");
		return -EINVAL;
	}

	printk(KERN_INFO "fb%d: MPC823 LCD frame buffer device\n",
	       GET_FB_IDX(fb_info.node));

#ifdef CONFIG_PM
	printk(KERN_INFO "fb%d: Enabling power management\n", GET_FB_IDX(fb_info.node));
	pm_dev = pm_register(PM_SYS_DEV, PM_SYS_UNKNOWN, lcd823_pm_request);
#endif

	return 0;
}

static void __exit
lcd823_exit(void)
{
#ifdef CONFIG_PM
	pm_unregister(pm_dev);
	lcd823_disable();
#endif
	unregister_framebuffer(&fb_info);
}

MODULE_DESCRIPTION("MPC823 LCD frame buffer device driver");
MODULE_LICENSE("GPL");

#ifdef MODULE
module_init(lcd823_init);
#endif
module_exit(lcd823_exit);
