/*
 *  linux/drivers/video/voyager.c -- Silicon Motion, Inc. Voyager GX frame buffer device
 *
 *      Copyright (C) 2003 Silicon Motion, Inc.
 *      Ge Wang, gewang@siliconmotion.com
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/pci.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb32.h>

#include "voyager.h"

#undef SMI_DEBUG

#ifdef	SMI_DEBUG
#define debug(fmt,args...)	printk (fmt ,##args)
#define debugX(level,fmt,args...) do { if (DEBUG>=level) printk(fmt ,##args); } while (0)
#else
#define debug(fmt,args...)
#define debugX(level,fmt,args...)
#endif	/* SMI_DEBUG */

static int smifb_pan_display (const struct fb_var_screeninfo *var,
			      struct fb_info *info);
static void set_screen_start (int base);

static char *VoyagerReg;	/* point to virtual Memory Map IO starting address	*/
static char *VoyagerFB;		/* point to virtual video memory starting address	*/
static struct par_info hw;	/* used to record hardware information			*/

/* default video mode */
static char * smi_mode = "640x480";	/* not used yet */
#if USE_CRT_ONLY
static int smi_displaytype = CRT;
static int smi_xres __initdata = CRT_DFLT_SCREEN_X_RES;
static int smi_yres __initdata = CRT_DFLT_SCREEN_Y_RES;
static int smi_bpp __initdata = CRT_DFLT_SCREEN_BPP;
#else
static int smi_displaytype = PANEL;
static int smi_xres __initdata = FP_DFLT_SCREEN_X_RES;
static int smi_yres __initdata = FP_DFLT_SCREEN_Y_RES;
static int smi_bpp __initdata = FP_DFLT_SCREEN_BPP;
#endif /* USE_CRT_ONLY */

static mode_table_t fp_modedb[] __initdata = {
	/*----------------------------------------------------------------------------------------
	 * H.   H.    H.     H.   H.        V.   V.    V.    V.   V.        Pixel     H.     V.
	 * tot. disp. sync   sync sync      tot. disp. sync  sync sinc      clock     freq.  freq.
	 *      end   start  wdth polarity       end   start hght polarity
	 *---------------------------------------------------------------------------------------*/
	/* 800 x 480 */
	{  939, 800,  813,   50,  NEGATIVE, 639, 480,  482,  2,   NEGATIVE, 36000000, 38340, 60 },
	{  888, 800,  808,   32,  NEGATIVE, 497, 480,  483,  3,   NEGATIVE, 28000000, 31469, 60 },//pixclock low, but seems to work
	{ 0, 0, 0, 0, NEGATIVE, 0, 0, 0, 0, NEGATIVE, 0, 0, 0 },
};

static mode_table_t crt_modedb[] __initdata = {
	/*----------------------------------------------------------------------------------------
	 * H.   H.    H.     H.   H.        V.   V.    V.    V.   V.        Pixel     H.     V.
	 * tot. disp. sync   sync sync      tot. disp. sync  sync sinc      clock     freq.  freq.
	 *      end   start  wdth polarity       end   start hght polarity
	 *---------------------------------------------------------------------------------------*/
	/* 640 x 480 */
	{  800, 640,  656,   96,  NEGATIVE, 525, 480,  490,  2,   NEGATIVE, 25175000, 31469, 60 },//vesa
	{  762, 640,  660,   61,  NEGATIVE, 525, 480,  494,  2,   NEGATIVE, 24000000, 31496, 60 },//sm501

	/* 800 x 600 */
	{ 1056, 800,  840,  128,  POSITIVE, 628, 600,  601,  4,   POSITIVE, 40000000, 37879, 60 },//vesa
	{ 1078, 800,  850,  128,  POSITIVE, 649, 600,  615,  4,   POSITIVE, 42000000, 38961, 60 },//sm501

	/* 1024 x 768*/
	{ 1377, 1024, 1061, 200,  NEGATIVE, 806, 768,  771,  6,   NEGATIVE, 72000000, 52291, 60 },//sm501 65 Hz
	{ 1344, 1024, 1048, 136,  NEGATIVE, 806, 768,  771,  6,   NEGATIVE, 65000000, 48363, 60 },//vesa

	/* End of table. */
	{ 0, 0, 0, 0, NEGATIVE, 0, 0, 0, 0, NEGATIVE, 0, 0, 0 },
};

/**********************************************************************
 *
 * regRead8
 *
 * Purpose
 *    Read the value of the 8-bit register specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of register to read
 *
 *    [out]
 *        None
 *
 * Returns
 *    Value of 8-bit register specified by nOffset
 *
 **********************************************************************/
UCHAR regRead8(ULONG nOffset)
{
    return *(volatile unsigned char *)(VoyagerReg+nOffset);
}


/**********************************************************************
 *
 * regRead16
 *
 * Purpose
 *    Read the value of the 16-bit register specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of register to read
 *
 *    [out]
 *        None
 *
 * Returns
 *    Value of 16-bit register specified by nOffset
 *
 **********************************************************************/
USHORT regRead16(ULONG nOffset)
{
    return *(volatile unsigned short *)(VoyagerReg+nOffset);
}


/**********************************************************************
 *
 * regRead32
 *
 * Purpose
 *    Read the value of the 32-bit register specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of register to read
 *
 *    [out]
 *        None
 *
 * Returns
 *    Value of 32-bit register specified by nOffset
 *
 **********************************************************************/
ULONG regRead32(ULONG nOffset)
{
#if SMI_BIG_ENDIAN
	/*
	 * swap bytes 1234 -> 4321
	 */
	ULONG nData1;
	ULONG nData2;
	char *pdata1;
	char *pdata2;

	nData1 = *(volatile unsigned long *)(VoyagerReg+nOffset);
	pdata1 = (char *) &nData1;
	pdata2 = (char *) &nData2;
	pdata2[0] = pdata1[3];
	pdata2[1] = pdata1[2];
	pdata2[2] = pdata1[1];
	pdata2[3] = pdata1[0];
	return nData2;
#else
	return *(volatile unsigned long *)(VoyagerReg+nOffset);
#endif /* SMI_BIG_ENDIAN */
}


/**********************************************************************
 *
 * regWrite8
 *
 * Purpose
 *    Write the 8-bit value, nData, to the 8-bit register specified by
 *    nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of register to write
 *        nData   - value to be written
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
void regWrite8(ULONG nOffset, UCHAR nData)
{
    *(volatile unsigned char *)(VoyagerReg+nOffset) = nData;
}


/**********************************************************************
 *
 * regWrite16
 *
 * Purpose
 *    Write the 16-bit value, nData, to the 16-bit register specified by
 *    nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of register to write
 *        nData   - value to be written
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
void regWrite16(ULONG nOffset, USHORT nData)
{
    *(volatile unsigned short *)(VoyagerReg+nOffset) = nData;
}


/**********************************************************************
 *
 * regWrite32
 *
 * Purpose
 *    Write the 32-bit value, nData, to the 32-bit register specified by
 *    nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of register to write
 *        nData   - value to be written
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
void regWrite32(ULONG nOffset, ULONG nData)
{
#if SMI_BIG_ENDIAN
	ULONG nData1;
	ULONG nData2;
	char *pdata1;
	char *pdata2;

	nData1 = nData;
	pdata1 = (char *) &nData1;
	pdata2 = (char *) &nData2;
	pdata2[0] = pdata1[3];
	pdata2[1] = pdata1[2];
	pdata2[2] = pdata1[1];
	pdata2[3] = pdata1[0];
	*(volatile unsigned long *)(VoyagerReg+nOffset) = nData2;
#else
	*(volatile unsigned long *)(VoyagerReg+nOffset) = nData;
#endif /* SMI_BIG_ENDIAN */
}


/**********************************************************************
 *
 * memRead8
 *
 * Purpose
 *    Read the value of the 8-bit memory location specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of memory location to read
 *
 *    [out]
 *        None
 *
 * Returns
 *    Value of 8-bit memory location specified by nOffset
 *
 **********************************************************************/
UCHAR memRead8(ULONG nOffset)
{
    return *(volatile unsigned char *)(VoyagerFB+nOffset);
}


/**********************************************************************
 *
 * memRead16
 *
 * Purpose
 *    Read the value of the 16-bit memory location specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of memory location to read
 *
 *    [out]
 *        None
 *
 * Returns
 *    Value of 16-bit memory location specified by nOffset
 *
 **********************************************************************/
USHORT memRead16(ULONG nOffset)
{
    return *(volatile unsigned short *)(VoyagerFB+nOffset);
}


/**********************************************************************
 *
 * memRead32
 *
 * Purpose
 *    Read the value of the 32-bit memory location specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of memory location to read
 *
 *    [out]
 *        None
 *
 * Returns
 *    Value of 32-bit memory location specified by nOffset
 *
 **********************************************************************/
ULONG memRead32(ULONG nOffset)
{
    return *(volatile unsigned long *)(VoyagerFB+nOffset);
}


/**********************************************************************
 *
 * memWrite8
 *
 * Purpose
 *    Write the 8-bit value, nData, to the 8-bit memory location
 *    specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of memory location to write
 *        nData   - value to be written
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
void memWrite8(ULONG nOffset, UCHAR nData)
{
    *(volatile unsigned char *)(VoyagerFB+nOffset) = nData;
}


/**********************************************************************
 *
 * memWrite16
 *
 * Purpose
 *    Write the 16-bit value, nData, to the 16-bit memory location
 *    specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of memory location to write
 *        nData   - value to be written
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
void memWrite16(ULONG nOffset, USHORT nData)
{
    *(volatile unsigned short *)(VoyagerFB+nOffset) = nData;
}


/**********************************************************************
 *
 * memWrite32
 *
 * Purpose
 *    Write the 32-bit value, nData, to the 32-bit memory location
 *    specified by nOffset
 *
 * Parameters
 *    [in]
 *        nOffset - zero-based offset of memory location to write
 *        nData   - value to be written
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
void memWrite32(ULONG nOffset, ULONG nData)
{
    *(volatile unsigned long *)(VoyagerFB+nOffset) = nData;
}

/**********************************************************************
 *
 * panelPlaneEnable
 *
 * Purpose
 *    Enable panel graphics plane
 *
 * Parameters
 *    [in]
 *        None
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
VOID panelPlaneEnable(VOID)
{
    ULONG panel_ctrl = 0;

    panel_ctrl = regRead32(PANEL_DISPLAY_CTRL);
    panel_ctrl = FIELD_SET(panel_ctrl, PANEL_DISPLAY_CTRL, PLANE, ENABLE);
    regWrite32(PANEL_DISPLAY_CTRL, panel_ctrl);
}


/**********************************************************************
 *
 * panelPlaneDisable
 *
 * Purpose
 *    Disable panel graphics plane
 *
 * Parameters
 *    [in]
 *        None
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
VOID panelPlaneDisable(VOID)
{
    ULONG panel_ctrl = 0;

    panel_ctrl = regRead32(PANEL_DISPLAY_CTRL);
    panel_ctrl = FIELD_SET(panel_ctrl, PANEL_DISPLAY_CTRL, PLANE, DISABLE);
    regWrite32(PANEL_DISPLAY_CTRL, panel_ctrl);
}


/**********************************************************************
 *
 * panelWaitVSync
 *
 * Purpose
 *    Wait for the specified number of panel Vsyncs
 *
 * Parameters
 *    [in]
 *        vsync_count - Number of Vsyncs to wait
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
static VOID panelWaitVSync(INT vsync_count)
{
	ULONG status;
	ULONG timeout;

	while (vsync_count-- > 0)
	{
		/* Wait for end of vsync */
		timeout = 0;
		do
		{
			status = FIELD_GET(regRead32(CMD_INTPR_STATUS),
							   CMD_INTPR_STATUS,
							   PANEL_SYNC);
			if (++timeout == VSYNCTIMEOUT)
				break;
		}
		while (status == CMD_INTPR_STATUS_PANEL_SYNC_ACTIVE);

		/* Wait for start of vsync */
		timeout = 0;
		do
		{
			status = FIELD_GET(regRead32(CMD_INTPR_STATUS),
							   CMD_INTPR_STATUS,
							   PANEL_SYNC);
			if (++timeout == VSYNCTIMEOUT)
				break;
		}
		while (status == CMD_INTPR_STATUS_PANEL_SYNC_INACTIVE);
	}
}


/**********************************************************************
 *
 * panelPowerSequence
 *
 * Purpose
 *    Turn the panel On/Off
 *
 * Parameters
 *    [in]
 *        on_off      - Turn panel On/Off. Can be:
 *                      PANEL_ON
 *                      PANEL_OFF
 *        vsync_delay - Number of Vsyncs to wait after each signal is
 *                      turned on/off
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
static VOID panelPowerSequence(panel_state_t on_off, INT vsync_delay)
{
	ULONG panelControl = regRead32(PANEL_DISPLAY_CTRL);

	if (on_off == PANEL_ON)
	{
		// Turn on FPVDDEN.
		panelControl = FIELD_SET(panelControl,
								 PANEL_DISPLAY_CTRL, FPVDDEN, HIGH);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn on FPDATA.
		panelControl = FIELD_SET(panelControl,
								 PANEL_DISPLAY_CTRL, DATA, ENABLE);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn on FPVBIAS.
		panelControl = FIELD_SET(panelControl,
								 PANEL_DISPLAY_CTRL, VBIASEN, HIGH);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn on FPEN.
		panelControl = FIELD_SET(panelControl,
								 PANEL_DISPLAY_CTRL, FPEN, HIGH);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
	}

	else
	{
		// Turn off FPEN.
		panelControl = FIELD_SET(panelControl,
								 PANEL_DISPLAY_CTRL, FPEN, LOW);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn off FPVBIASEN.
		panelControl = FIELD_SET(panelControl,
								 PANEL_DISPLAY_CTRL, VBIASEN, LOW);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn off FPDATA.
		panelControl = FIELD_SET(panelControl,
								 PANEL_DISPLAY_CTRL, DATA, DISABLE);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn off FPVDDEN.
		panelControl = FIELD_SET(panelControl,
								 PANEL_DISPLAY_CTRL, FPVDDEN, LOW);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
	}
}


/**********************************************************************
 *
 * panelOn
 *
 * Purpose
 *    Turn the panel ON
 *
 * Parameters
 *    [in]
 *        nVSyncDelay - Number of Vsyncs to wait after each signal is
 *                      turned on
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
VOID panelOn(ULONG nVSyncDelay)
{
    panelPowerSequence(PANEL_ON, nVSyncDelay);
}


/**********************************************************************
 *
 * panelOff
 *
 * Purpose
 *    Turn the panel OFF
 *
 * Parameters
 *    [in]
 *        nVSyncDelay - Number of Vsyncs to wait after each signal is
 *                      turned off
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
VOID panelOff(ULONG nVSyncDelay)
{
    panelPowerSequence(PANEL_OFF, nVSyncDelay);
}


/**********************************************************************
 *
 * panelUseCRT
 *
 * Purpose
 *    Enable/disable routing of panel output to CRT monitor
 *
 * Parameters
 *    [in]
 *        bEnable - TRUE enables routing of panel output to CRT monitor
 *                  FALSE disables routing of panel output to CRT monitor
 *
 *    [out]
 *        None
 *
 * Returns
 *    Nothing
 *
 **********************************************************************/
VOID panelUseCRT(BOOL bEnable)
{
    ULONG panel_ctrl = 0;
    ULONG crt_ctrl = 0;

    panel_ctrl = regRead32(PANEL_DISPLAY_CTRL);
    crt_ctrl = regRead32(CRT_DISPLAY_CTRL);

    if (bEnable)
	{
	    /* Enable panel graphics plane */
        panel_ctrl = FIELD_SET(panel_ctrl, PANEL_DISPLAY_CTRL, PLANE, ENABLE);

        /* Disable CRT graphics plane */
	    crt_ctrl = FIELD_SET(crt_ctrl, CRT_DISPLAY_CTRL, PLANE, DISABLE);

	    /* Route panel data to CRT monitor */
	    crt_ctrl = FIELD_SET(crt_ctrl, CRT_DISPLAY_CTRL, SELECT, PANEL);
	}
	else
	{
	    /* Disable panel graphics plane */
        panel_ctrl = FIELD_SET(panel_ctrl, PANEL_DISPLAY_CTRL, PLANE, DISABLE);

        /* Enable CRT graphics plane */
	    crt_ctrl = FIELD_SET(crt_ctrl, CRT_DISPLAY_CTRL, PLANE, ENABLE);

	    /* Do not route panel data to CRT monitor */
	    crt_ctrl = FIELD_SET(crt_ctrl, CRT_DISPLAY_CTRL, SELECT, CRT);
	}

    regWrite32(PANEL_DISPLAY_CTRL, panel_ctrl);
    regWrite32(CRT_DISPLAY_CTRL, crt_ctrl);
}

// Program new power mode.
void setPower(ULONG nGates, ULONG Clock)
{
	ULONG gate_reg, clock_reg;
	ULONG control_value;

	// Get current power mode.
	control_value = FIELD_GET(regRead32(POWER_MODE_CTRL),
							  POWER_MODE_CTRL,
							  MODE);

	switch (control_value)
	{
	case POWER_MODE_CTRL_MODE_MODE0:

		// Switch from mode 0 to mode 1.
		gate_reg = POWER_MODE1_GATE;
		clock_reg = POWER_MODE1_CLOCK;
		control_value = FIELD_SET(control_value,
					  POWER_MODE_CTRL, MODE, MODE1);
		break;

	case POWER_MODE_CTRL_MODE_MODE1:
	case POWER_MODE_CTRL_MODE_SLEEP:

		// Switch from mode 1 or sleep to mode 0.
		gate_reg = POWER_MODE0_GATE;
		clock_reg = POWER_MODE0_CLOCK;
		control_value = FIELD_SET(control_value, POWER_MODE_CTRL, MODE, MODE0);
		break;

	default:
		// Invalid mode
		return;
	}

	// Program new power mode.
	regWrite32(gate_reg, nGates);
	regWrite32(clock_reg, Clock);
	regWrite32(POWER_MODE_CTRL, control_value);

	// When returning from sleep, wait until finished.
	while (FIELD_GET(regRead32(POWER_MODE_CTRL),
					 POWER_MODE_CTRL,
					 SLEEP_STATUS) == POWER_MODE_CTRL_SLEEP_STATUS_ACTIVE) ;
}

// Set DPMS state.
void setDPMS(DPMS_t state)
{
	ULONG value;

	value = regRead32(SYSTEM_CTRL);
	switch (state)
	{
	case DPMS_ON:
		value = FIELD_SET(value, SYSTEM_CTRL, DPMS, VPHP);
		break;

	case DPMS_STANDBY:
		value = FIELD_SET(value, SYSTEM_CTRL, DPMS, VPHN);
		break;

	case DPMS_SUSPEND:
		value = FIELD_SET(value, SYSTEM_CTRL, DPMS, VNHP);
		break;

	case DPMS_OFF:
		value = FIELD_SET(value, SYSTEM_CTRL, DPMS, VNHN);
		break;
	}

	regWrite32(SYSTEM_CTRL, value);
}


// Perform a rounded division.
LONG roundDiv(LONG num, LONG denom)
{
	/* n / d + 1 / 2 = (2n + d) / 2d */
	return (2 * num + denom) / (2 * denom);
}

// Finds clock closest to the requested.
LONG findClock(LONG requested_clock, clock_select_t *clock, display_t display)
{
	LONG mclk;
	INT divider, shift;
	LONG best_diff = 999999999;

	// Try 288MHz and 336MHz clocks.
	for (mclk = 288000000; mclk <= 336000000; mclk += 48000000)
	{
		// For CRT, try dividers 1 and 3, for panel, try divider 5 as well.
		for (divider = 1; divider <= (display == PANEL ? 5 : 3);
			 divider += 2)
		{
			// Try all 8 shift values.
			for (shift = 0; shift < 8; shift++)
			{
				// Calculate difference with requested clock.
				LONG diff = roundDiv(mclk, divider << shift) - requested_clock;
				if (diff < 0)
				{
					diff = -diff;
				}

				// If the difference is less than the current, use it.
				if (diff < best_diff)
				{
					// Store best difference.
					best_diff = diff;

					// Store clock values.
					clock->mclk = mclk;
					clock->divider = divider;
					clock->shift = shift;
				}
			}
		}
	}

	// Return best clock.
	return clock->mclk / (clock->divider << clock->shift);
}

// Finds the requested mode in the mode table.
mode_table_t * __init findMode(mode_table_t *mode_table, INT width, INT height,
					   INT refresh_rate)
{
	// Walk the entire mode table.
	while (mode_table->pixel_clock != 0)
	{
		// If this mode matches the requested mode, return it!
		if ((mode_table->horizontal_display_end == width)
		&& (mode_table->vertical_display_end == height)
		&& (mode_table->vertical_frequency == refresh_rate))
		{
			return(mode_table);
		}

		// Next entry in the mode table.
		mode_table++;
	}

	// No mode found.
	return(NULL);
}

// Converts the VESA timing into Voyager timing.
void adjustMode(mode_table_t *vesaMode, mode_table_t *mode, display_t display)
{
	LONG blank_width, sync_start, sync_width;
	clock_select_t clock;

	// Calculate the VESA line and screen frequencies.
	vesaMode->horizontal_frequency = roundDiv(vesaMode->pixel_clock,
										   vesaMode->horizontal_total);
	vesaMode->vertical_frequency = roundDiv(vesaMode->horizontal_frequency,
										 vesaMode->vertical_total);

	// Calculate the sync percentages of the VESA mode.
	blank_width = vesaMode->horizontal_total
				- vesaMode->horizontal_display_end;
	sync_start = roundDiv((vesaMode->horizontal_sync_start -
					   vesaMode->horizontal_display_end) * 100, blank_width);
	sync_width = roundDiv(vesaMode->horizontal_sync_width * 100, blank_width);

	// Copy VESA mode into Voyager mode.
	*mode = *vesaMode;

	// Find the best pixel clock.
	mode->pixel_clock = findClock(vesaMode->pixel_clock * 2, &clock, display) / 2;

	// Calculate the horizontal total based on the pixel clock and VESA line
	// frequency.
	mode->horizontal_total = roundDiv(mode->pixel_clock,
								   vesaMode->horizontal_frequency);

	// Calculate the sync start and width based on the VESA percentages.
	blank_width = mode->horizontal_total - mode->horizontal_display_end;
	mode->horizontal_sync_start = mode->horizontal_display_end
								+ roundDiv(blank_width * sync_start, 100);
	mode->horizontal_sync_width = roundDiv(blank_width * sync_width, 100);

	// Calculate the line and screen frequencies.
	mode->horizontal_frequency = roundDiv(mode->pixel_clock,
									   mode->horizontal_total);
	mode->vertical_frequency = roundDiv(mode->horizontal_frequency,
									 mode->vertical_total);
}

// Fill the register structure.
void setModeRegisters(reg_table_t *register_table, mode_table_t *mode,
					  display_t display, INT bpp)
{
	clock_select_t clock;
	memset(&clock, 0, sizeof(clock));

	// Calculate the clock register values.
	findClock(mode->pixel_clock * 2, &clock, display);

	if (display == PANEL)
	{
		// Set clock value for panel.
		register_table->clock
			= (clock.mclk == 288000000
				? FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_SELECT, 288)
				: FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_SELECT, 336))
			| (clock.divider == 1
				? FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_DIVIDER, 1)
				: (clock.divider == 3
					? FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_DIVIDER, 3)
					: FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_DIVIDER, 5)))
			| FIELD_VALUE(0, CURRENT_POWER_CLOCK, P2XCLK_SHIFT, clock.shift);

		// Set control register value.
		register_table->control
			= (mode->vertical_sync_polarity == POSITIVE
				? FIELD_SET(0, PANEL_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_HIGH)
				: FIELD_SET(0, PANEL_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_LOW))
			| (mode->horizontal_sync_polarity == POSITIVE
				? FIELD_SET(0, PANEL_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_HIGH)
				: FIELD_SET(0, PANEL_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_LOW))
			| FIELD_SET(0, PANEL_DISPLAY_CTRL, TIMING, ENABLE)
			| FIELD_SET(0, PANEL_DISPLAY_CTRL, PLANE, ENABLE)
			| (bpp == 8
				? FIELD_SET(0, PANEL_DISPLAY_CTRL, FORMAT, 8)
				: (bpp == 16
					? FIELD_SET(0, PANEL_DISPLAY_CTRL, FORMAT, 16)
					: FIELD_SET(0, PANEL_DISPLAY_CTRL, FORMAT, 32)));

		// Set timing registers.
		register_table->horizontal_total
			= FIELD_VALUE(0, PANEL_HORIZONTAL_TOTAL, TOTAL,
						  mode->horizontal_total - 1)
			| FIELD_VALUE(0, PANEL_HORIZONTAL_TOTAL, DISPLAY_END,
						  mode->horizontal_display_end - 1);

		register_table->horizontal_sync
			= FIELD_VALUE(0, PANEL_HORIZONTAL_SYNC, WIDTH,
						  mode->horizontal_sync_width)
			| FIELD_VALUE(0, PANEL_HORIZONTAL_SYNC, START,
						  mode->horizontal_sync_start - 1);

		register_table->vertical_total
			= FIELD_VALUE(0, PANEL_VERTICAL_TOTAL, TOTAL,
						  mode->vertical_total - 1)
			| FIELD_VALUE(0, PANEL_VERTICAL_TOTAL, DISPLAY_END,
						  mode->vertical_display_end - 1);

		register_table->vertical_sync
			= FIELD_VALUE(0, PANEL_VERTICAL_SYNC, HEIGHT,
						  mode->vertical_sync_height)
			| FIELD_VALUE(0, PANEL_VERTICAL_SYNC, START,
						  mode->vertical_sync_start - 1);
	}
	else
	{
		// Set clock value for CRT.
		register_table->clock
			= (clock.mclk == 288000000
				? FIELD_SET(0, CURRENT_POWER_CLOCK, V2XCLK_SELECT, 288)
				: FIELD_SET(0, CURRENT_POWER_CLOCK, V2XCLK_SELECT, 336))
			| (clock.divider == 1
				? FIELD_SET(0, CURRENT_POWER_CLOCK, V2XCLK_DIVIDER, 1)
				: FIELD_SET(0, CURRENT_POWER_CLOCK, V2XCLK_DIVIDER, 3))
			| FIELD_VALUE(0, CURRENT_POWER_CLOCK, V2XCLK_SHIFT, clock.shift);

		// Set control register value.
		register_table->control
			= (mode->vertical_sync_polarity == POSITIVE
				? FIELD_SET(0, CRT_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_HIGH)
				: FIELD_SET(0, CRT_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_LOW))
			| (mode->horizontal_sync_polarity == POSITIVE
				? FIELD_SET(0, CRT_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_HIGH)
				: FIELD_SET(0, CRT_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_LOW))
			| FIELD_SET(0, CRT_DISPLAY_CTRL, SELECT, CRT)
			| FIELD_SET(0, CRT_DISPLAY_CTRL, TIMING, ENABLE)
			| FIELD_SET(0, CRT_DISPLAY_CTRL, PLANE, ENABLE)
			| (bpp == 8
				? FIELD_SET(0, CRT_DISPLAY_CTRL, FORMAT, 8)
				: (bpp == 16
					? FIELD_SET(0, CRT_DISPLAY_CTRL, FORMAT, 16)
					: FIELD_SET(0, CRT_DISPLAY_CTRL, FORMAT, 32)));

		// Set timing registers.
		register_table->horizontal_total
			= FIELD_VALUE(0, CRT_HORIZONTAL_TOTAL, TOTAL,
						  mode->horizontal_total - 1)
			| FIELD_VALUE(0, CRT_HORIZONTAL_TOTAL, DISPLAY_END,
						  mode->horizontal_display_end - 1);

		register_table->horizontal_sync
			= FIELD_VALUE(0, CRT_HORIZONTAL_SYNC, WIDTH,
						  mode->horizontal_sync_width)
			| FIELD_VALUE(0, CRT_HORIZONTAL_SYNC, START,
						  mode->horizontal_sync_start - 1);

		register_table->vertical_total
			= FIELD_VALUE(0, CRT_VERTICAL_TOTAL, TOTAL,
						  mode->vertical_total - 1)
			| FIELD_VALUE(0, CRT_VERTICAL_TOTAL, DISPLAY_END,
						  mode->vertical_display_end - 1);
		register_table->vertical_sync
			= FIELD_VALUE(0, CRT_VERTICAL_SYNC, HEIGHT,
						  mode->vertical_sync_height)
			| FIELD_VALUE(0, CRT_VERTICAL_SYNC, START,
						  mode->vertical_sync_start - 1);
	}

	// Calculate frame buffer width and height.
	register_table->fb_width = mode->horizontal_display_end * (bpp / 8);
	register_table->width = mode->horizontal_display_end;
	register_table->height = mode->vertical_display_end;

	// Save display type.
	register_table->display = display;
}

// Program the mode with the registers specified.
void programMode(reg_table_t *register_table)
{
	DWORD value, gate, clock;
	DWORD palette_ram;
	DWORD fb_size, offset;

	// Get current power configuration.
	gate = regRead32(CURRENT_POWER_GATE);
	clock = regRead32(CURRENT_POWER_CLOCK);

	// Program panel.
	if (register_table->display == PANEL)
	{
		// Program clock, enable display controller.
		gate = FIELD_SET(gate, CURRENT_POWER_GATE, DISPLAY, ENABLE);
		clock &= FIELD_CLEAR(CURRENT_POWER_CLOCK, P2XCLK_SELECT)
			  &  FIELD_CLEAR(CURRENT_POWER_CLOCK, P2XCLK_DIVIDER)
			  &  FIELD_CLEAR(CURRENT_POWER_CLOCK, P2XCLK_SHIFT);
		setPower(gate, clock | register_table->clock);

		// Calculate frame buffer address.
		value = 0;
		fb_size = register_table->fb_width * register_table->height;
		if (FIELD_GET(regRead32(CRT_DISPLAY_CTRL),
					  CRT_DISPLAY_CTRL,
					  PLANE) == CRT_DISPLAY_CTRL_PLANE_ENABLE)
		{
			value = FIELD_GET(regRead32(CRT_FB_ADDRESS),
							  CRT_FB_ADDRESS,
							  ADDRESS);
			if (fb_size < value)
			{
				value = 0;
			}
			else
			{
				value += FIELD_GET(regRead32(CRT_FB_WIDTH),
								   CRT_FB_WIDTH,
								   OFFSET)
					  *  (FIELD_GET(regRead32(CRT_VERTICAL_TOTAL),
								   CRT_VERTICAL_TOTAL,
								   DISPLAY_END) + 1);
			}
		}

		// Program panel registers.
		regWrite32 (PANEL_FB_ADDRESS,
			    FIELD_SET(0, PANEL_FB_ADDRESS, STATUS, PENDING)	|
			    FIELD_SET(0, PANEL_FB_ADDRESS, EXT, LOCAL)		|
			    FIELD_VALUE(0, PANEL_FB_ADDRESS, ADDRESS, value)	);

		regWrite32 (PANEL_FB_WIDTH,
			    FIELD_VALUE(0, PANEL_FB_WIDTH, WIDTH,
				    register_table->fb_width) 			|
			    FIELD_VALUE(0, PANEL_FB_WIDTH, OFFSET,
			    register_table->fb_width)				);

		regWrite32(PANEL_WINDOW_WIDTH,
			    FIELD_VALUE(0, PANEL_WINDOW_WIDTH, WIDTH,
			    register_table->width)				|
			    FIELD_VALUE(0, PANEL_WINDOW_WIDTH, X, 0)		);

		regWrite32(PANEL_WINDOW_HEIGHT,
			    FIELD_VALUE(0, PANEL_WINDOW_HEIGHT, HEIGHT,
				    register_table->height)			|
			    FIELD_VALUE(0, PANEL_WINDOW_HEIGHT, Y, 0)		);

		regWrite32(PANEL_PLANE_TL,
			    FIELD_VALUE(0, PANEL_PLANE_TL, TOP, 0)		|
			    FIELD_VALUE(0, PANEL_PLANE_TL, LEFT, 0)		);

		regWrite32(PANEL_PLANE_BR,
			    FIELD_VALUE(0, PANEL_PLANE_BR, BOTTOM,
					register_table->height)			|
			    FIELD_VALUE(0, PANEL_PLANE_BR, RIGHT,
					register_table->width)			);

		regWrite32(PANEL_HORIZONTAL_TOTAL, register_table->horizontal_total);
		regWrite32(PANEL_HORIZONTAL_SYNC,  register_table->horizontal_sync);
		regWrite32(PANEL_VERTICAL_TOTAL,   register_table->vertical_total);
		regWrite32(PANEL_VERTICAL_SYNC,    register_table->vertical_sync);

		// Program panel display control register.
		value = regRead32(PANEL_DISPLAY_CTRL)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, VSYNC_PHASE)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, HSYNC_PHASE)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, TIMING)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, PLANE)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, FORMAT);

		regWrite32(PANEL_DISPLAY_CTRL,
						  value | register_table->control);

		// Palette RAM.
		palette_ram = PANEL_PALETTE_RAM;

		// Turn on panel.
		panelPowerSequence(PANEL_ON, 4);
	}

	// Program CRT.
	else
	{
		// Program clock, enable display controller.
		gate = FIELD_SET(gate, CURRENT_POWER_GATE, DISPLAY, ENABLE);
		clock &= FIELD_CLEAR(CURRENT_POWER_CLOCK, V2XCLK_SELECT)
			  &  FIELD_CLEAR(CURRENT_POWER_CLOCK, V2XCLK_DIVIDER)
			  &  FIELD_CLEAR(CURRENT_POWER_CLOCK, V2XCLK_SHIFT);
		setPower(gate, clock | register_table->clock);

		// Turn on DAC.
		regWrite32(MISC_CTRL, FIELD_SET(regRead32(MISC_CTRL),
							   MISC_CTRL,
							   DAC_POWER,
							   ENABLE));

		// Calculate frame buffer address.
		value = 0;
		fb_size = register_table->fb_width * register_table->height;
		if (FIELD_GET(regRead32(PANEL_DISPLAY_CTRL),
					  PANEL_DISPLAY_CTRL,
					  PLANE) == PANEL_DISPLAY_CTRL_PLANE_ENABLE)
		{
			value = FIELD_GET(regRead32(PANEL_FB_ADDRESS),
							  PANEL_FB_ADDRESS,
							  ADDRESS);
			if (fb_size < value)
			{
				value = 0;
			}
			else
			{
				value += FIELD_GET(regRead32(PANEL_FB_WIDTH),
								   PANEL_FB_WIDTH,
								   OFFSET)
					  *  FIELD_GET(regRead32(PANEL_WINDOW_HEIGHT),
								   PANEL_WINDOW_HEIGHT,
								   HEIGHT);
			}
		}

		// Program CRT registers.
		regWrite32(CRT_FB_ADDRESS,
						  FIELD_SET(0, CRT_FB_ADDRESS, STATUS, PENDING)		|
						  FIELD_SET(0, CRT_FB_ADDRESS, EXT, LOCAL)			|
			              FIELD_VALUE(0, CRT_FB_ADDRESS, ADDRESS, value)	);

		regWrite32(CRT_FB_WIDTH,
						  FIELD_VALUE(0, CRT_FB_WIDTH, WIDTH,
									  register_table->fb_width)				|
						  FIELD_VALUE(0, CRT_FB_WIDTH, OFFSET,
									  register_table->fb_width)				);

		regWrite32(CRT_HORIZONTAL_TOTAL,
						  register_table->horizontal_total);
		regWrite32(CRT_HORIZONTAL_SYNC,
						  register_table->horizontal_sync);
		regWrite32(CRT_VERTICAL_TOTAL,
						  register_table->vertical_total);
		regWrite32(CRT_VERTICAL_SYNC,
						  register_table->vertical_sync);

		// Program CRT display control register.
		value = regRead32(CRT_DISPLAY_CTRL)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, VSYNC_PHASE)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, HSYNC_PHASE)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, SELECT)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, TIMING)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, PLANE)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, FORMAT);

		regWrite32(CRT_DISPLAY_CTRL,
						  value | register_table->control);

		// Palette RAM.
		palette_ram = CRT_PALETTE_RAM;

		// Turn on CRT.
		setDPMS(DPMS_ON);
	}

	// In case of 8-bpp, fill palette.
	if (FIELD_GET(register_table->control,
				  PANEL_DISPLAY_CTRL,
				  FORMAT) == PANEL_DISPLAY_CTRL_FORMAT_8)
	{
		// Start with RGB = 0,0,0.
		BYTE red = 0, green = 0, blue = 0;
		DWORD gray = 0;
		for (offset = 0; offset < 256 * 4; offset += 4)
		{
			// Store current RGB value.
			regWrite32(palette_ram + offset, gray
								? RGB((gray + 50) / 100,
									  (gray + 50) / 100,
									  (gray + 50) / 100)
								: RGB(red, green, blue));

			if (gray)
			{
				// Walk through grays (40 in total).
				gray += 654;
			}

			else
			{
				// Walk through colors (6 per base color).
				if (blue != 255)
				{
					blue += 51;
				}
				else if (green != 255)
				{
					blue = 0;
					green += 51;
				}
				else if (red != 255)
				{
					green = blue = 0;
					red += 51;
				}
				else
				{
					gray = 1;
				}
			}
		}
	}

	// For 16- and 32-bpp,  fill palette with gamma values.
	else
	{
		// Start with RGB = 0,0,0.
		value = 0x000000;
		for (offset = 0; offset < 256 * 4; offset += 4)
		{
			regWrite32(palette_ram + offset, value);

			// Advance RGB by 1,1,1.
			value += 0x010101;
		}
	}
}

VOID SetMode(display_t display, int bpp, struct fb_var_screeninfo *var,
	     struct smifb_info *sfb)
{
	mode_table_t mode;
	mode_table_t vesaMode;
	reg_table_t register_table;

	/* convert fb timings to more SM501 like timings */
	vesaMode.horizontal_total = var->left_margin + var->xres
		+ var->hsync_len + var->right_margin;
	vesaMode.horizontal_display_end = var->xres;
	vesaMode.horizontal_sync_start = var->xres + var->right_margin;
	vesaMode.horizontal_sync_width = var->hsync_len;
	vesaMode.horizontal_sync_polarity =
		(var->sync & FB_SYNC_HOR_HIGH_ACT) ? POSITIVE : NEGATIVE;

	vesaMode.vertical_total = var->upper_margin + var->yres
		+ var->lower_margin + var->vsync_len;
	vesaMode.vertical_display_end = var->yres;
	vesaMode.vertical_sync_start = var->yres + var->lower_margin;
	vesaMode.vertical_sync_height = var->vsync_len;
	vesaMode.vertical_sync_polarity =
		(var->sync & FB_SYNC_VERT_HIGH_ACT) ? POSITIVE : NEGATIVE;

	vesaMode.pixel_clock = (1000000000 / var->pixclock) * 1000;
	vesaMode.horizontal_frequency =
		roundDiv (vesaMode.pixel_clock, vesaMode.horizontal_total);
	vesaMode.vertical_frequency = roundDiv (vesaMode.horizontal_frequency,
						vesaMode.vertical_total);

	/*
	 * Convert VESA timing into Voyager timing:
	 *
	 * Find a SM501 pixel clock setting closest to the specified pixel
	 * clock. If there is no 100 % match, horizontal and vertical sync
	 * with and margins are adjusted to hold the spezified vertical scan
	 * rate.
	 */
	adjustMode(&vesaMode, &mode, display);

	/*
	 * write back adjusted values to fb_var_scrreninfo structure, so the
	 * user could see what values were adjusted.
	 */
	sfb->fb.var.hsync_len = mode.horizontal_sync_width;
	sfb->fb.var.right_margin = mode.horizontal_sync_start
		- mode.horizontal_display_end;
	sfb->fb.var.left_margin = mode.horizontal_total
		- mode.horizontal_display_end
		- mode.horizontal_sync_width
		- sfb->fb.var.right_margin;

	sfb->fb.var.vsync_len = mode.vertical_sync_height;
	sfb->fb.var.lower_margin = mode.vertical_sync_start
		- mode.vertical_display_end;
	sfb->fb.var.upper_margin = mode.vertical_total
		- mode.vertical_display_end
		- mode.vertical_sync_height
		- sfb->fb.var.lower_margin;

	sfb->fb.var.pixclock = 1000000000 / (mode.pixel_clock / 1000);

	// Fill the register structure
	setModeRegisters(&register_table, &mode, display, bpp);

	// Program the registers
	programMode(&register_table);
}

static void smi_set_timing(struct smifb_info *sfb,
			   struct fb_var_screeninfo *var)
{
	SetMode(smi_displaytype, sfb->fb.var.bits_per_pixel, var, sfb);
	/* Turn on CRT DAC. for simultanous display */
	regWrite32(MISC_CTRL, FIELD_SET(regRead32(MISC_CTRL),  MISC_CTRL,
					DAC_POWER, ENABLE));
}

/*
 * Get the VAR structure pointer for the specified console
 */
static inline struct fb_var_screeninfo *get_con_var(struct fb_info *info, int con)
{
	struct smifb_info *sfb = (struct smifb_info *)info;
	return (con == sfb->currcon) ? &sfb->fb.var : &fb_display[con].var;
}

/*
 * Get the DISPLAY structure pointer for the specified console
 */
static inline struct display *get_con_display(struct fb_info *info, int con)
{
	struct smifb_info *sfb = (struct smifb_info *)info;
	return (con < 0) ? sfb->fb.disp : &fb_display[con];
}

/*
 * Get the CMAP pointer for the specified console
 */
static inline struct fb_cmap *get_con_cmap(struct fb_info *info, int con)
{
	struct smifb_info *sfb = (struct smifb_info *)info;
	return (con == sfb->currcon) ? &sfb->fb.cmap : &fb_display[con].cmap;
}

/*
 * Set the color palette
 */
static int
smifb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
		   u_int trans, struct fb_info *info)
{
	struct smifb_info *sfb = (struct smifb_info *)info;
	struct display * display = (sfb->currcon < 0) ? sfb->fb.disp : &fb_display[sfb->currcon];
	void *pal;

	switch (display->visual) {
	case FB_VISUAL_TRUECOLOR:
		pal = sfb->fb.pseudo_palette;	// System has 16 default color
		if (regno >= 16)
			return 1;
		/* TODO: Fix tab_cfb16[] in fbcon-cfb16.c to get 16-bit console
			 but it is very likely, that it will interfere with mb86290/coralp driver.
			 probably we should disable endianess swapping on LPC CS1 */
		if (sfb->fb.var.bits_per_pixel == 16) {		// RGB 5:6:5 True Colour
			((u16 *)pal)[regno] = (red & 0xf800) | ((green & 0xfc00) >> 5) | ((blue & 0xf800) >> 11);
		} else {
			((u32 *)pal)[regno] = ((red & 0xff00) << 8) | ((green & 0xff00)) | ((blue & 0xff00) >> 8);
		}
		break;
	case FB_VISUAL_PSEUDOCOLOR:		/* 8-bit, update palette on chip*/
    		if (regno>255)
			 return 1;
		/* TODO: 1. Fix nibbletab_cfb8[] in fbcon-cfb8.c to get 8-bit console.
			 2. Filling palette here interfere with palette filling in programMode(),
			    fix it to get proper colors for 8-bit console */
		if (smi_displaytype == CRT) {
			regWrite32(CRT_PALETTE_RAM + regno*4, ((red & 0xff)<<16) | ((green & 0xff) << 8) | ((blue & 0xff)));
		} else {
			regWrite32(PANEL_PALETTE_RAM + regno*4, ((red & 0xff)<<16) | ((green & 0xff) << 8) | ((blue & 0xff)));
		}
		break;
	}
	return 0;
}

/*
 *  smifb_decode_var():
 *    Get the video params out of 'var'. If a value doesn't fit, round it up,
 *    if it's too big, return -EINVAL.
 *
 *    Suggestion: Round up in the following order: bits_per_pixel, xres,
 *    yres, xres_virtual, yres_virtual, xoffset, yoffset, grayscale,
 *    bitfields, horizontal timing, vertical timing.
 */
static int smifb_decode_var(struct fb_var_screeninfo *var, struct smifb_info *sfb, struct par_info *hw)
{
	if (var->xres_virtual > var->xres)
		return -EINVAL;

	var->xres_virtual =
	    var->xres_virtual < var->xres ? var->xres : var->xres_virtual;
	var->yres_virtual =
	    var->yres_virtual < var->yres ? var->yres : var->yres_virtual;

	if ((var->xres * var->yres * var->bits_per_pixel/8) > sfb->fb.fix.smem_len)
		return -EINVAL;
	if ((var->xres_virtual * var->yres_virtual * var->bits_per_pixel/8) > sfb->fb.fix.smem_len)
		return -EINVAL;

	hw->width = var->xres;
	hw->height = var->yres;
	hw->hz = 60;

	switch (var->bits_per_pixel) {
	case 8:
	case 16:		/* RGB 565 */
	case 32:
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static void
smifb_hw_set_var(struct fb_var_screeninfo *var, struct smifb_info *sfb)
{
	struct display *display = sfb->fb.disp;
	struct fb_cmap *cmap;

	sfb->palette_size = var->bits_per_pixel == 8 ? 256 : 16;

	if (display->cmap.len)
		cmap = &display->cmap;
	else
		cmap = fb_default_cmap(sfb->palette_size);

	fb_set_cmap(cmap, 1, smifb_setcolreg, &sfb->fb);
}

void
smifb_print_var(struct fb_var_screeninfo *var)
{
	debug ("xres           = %d\n", var->xres);
	debug ("yres           = %d\n", var->yres);
	debug ("xres_virtual   = %d\n", var->xres_virtual);
	debug ("yres_virtual   = %d\n", var->yres_virtual);
	debug ("xoffset        = %d\n", var->xoffset);
	debug ("yoffset        = %d\n", var->yoffset);
	debug ("bits_per_pixel = %d\n", var->bits_per_pixel);
	debug ("activate       = %d\n", var->activate);
	debug ("accel_flags    = %#x\n", var->accel_flags);
	debug (" red   = %d/%d/%d\n", var->red.length, var->red.offset, var->red.msb_right);
	debug (" green = %d/%d/%d\n", var->green.length, var->green.offset, var->green.msb_right);
	debug (" blue  = %d/%d/%d\n", var->blue.length, var->blue.offset, var->blue.msb_right);
	debug ("pixclock       = %d\n", var->pixclock);
	debug ("hsync_len      = %d\n", var->hsync_len);
	debug ("vsync_len      = %d\n", var->vsync_len);
}


static int
smifb_set_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	struct smifb_info *sfb = (struct smifb_info *)info;
	struct fb_var_screeninfo *dvar = get_con_var(&sfb->fb, con);
	struct display *display = get_con_display(&sfb->fb, con);
	int err, chgvar = 0;

	debug ("entering smifb_set_var\n");
	smifb_print_var(var);

	/*
	 * Decode var contents into a par structure, adjusting any
	 * out of range values.
	 */
	err = smifb_decode_var(var, sfb, &hw);
	if (err) {
		return err;
	}

	if (var->activate & FB_ACTIVATE_TEST)
		return 0;

	if ((var->activate & FB_ACTIVATE_MASK) != FB_ACTIVATE_NOW)
		return -EINVAL;

	if (dvar->xres != var->xres)
		chgvar = 1;
	if (dvar->yres != var->yres)
		chgvar = 1;
	if (dvar->xres_virtual != var->xres_virtual)
		chgvar = 1;
	if (dvar->yres_virtual != var->yres_virtual)
		chgvar = 1;
	if (dvar->bits_per_pixel != var->bits_per_pixel)
		chgvar = 1;
	if (con < 0)
		chgvar = 0;

	switch (var->bits_per_pixel) {
	case 8:
		var->red.offset		= 5;
		var->red.length		= 3;
		var->green.offset	= 2;
		var->green.length	= 3;
		var->blue.offset	= 0;
		var->blue.length	= 2;

		display->visual		= FB_VISUAL_PSEUDOCOLOR;
		display->line_length	= var->xres ;
		display->dispsw		= &fbcon_cfb8;
		display->dispsw_data	= sfb->fb.pseudo_palette;
		break;
	case 16:
		var->red.offset		= 11;
		var->red.length		= 5;
		var->green.offset	= 5;
		var->green.length	= 6;
		var->blue.offset	= 0;
		var->blue.length	= 5;

		display->visual		= FB_VISUAL_TRUECOLOR;
		display->line_length	= var->xres * 2;
		display->dispsw		= &fbcon_cfb16;
		display->dispsw_data	= sfb->fb.pseudo_palette;
		break;
	case 24:
	case 32:
		var->red.offset		= 16;
		var->red.length		= 8;
		var->green.offset	= 8;
		var->green.length	= 8;
		var->blue.offset	= 0;
		var->blue.length	= 8;

		display->visual		= FB_VISUAL_TRUECOLOR;
		display->line_length	= var->xres * 4;
		display->dispsw		= &fbcon_cfb32;
		display->dispsw_data	= sfb->fb.pseudo_palette;
		break;
	default:
		display->dispsw = &fbcon_dummy;
		break;
	}

	display->screen_base	= sfb->fb.screen_base;
	display->next_line	= display->line_length;
	display->type		= sfb->fb.fix.type;
	display->type_aux	= sfb->fb.fix.type_aux;
	display->ypanstep	= sfb->fb.fix.ypanstep;
	display->ywrapstep	= sfb->fb.fix.ywrapstep;
	display->can_soft_blank	= 1;
	display->inverse	= 0;

	display->scrollmode	= __SCROLL_YREDRAW |
				  __SCROLL_YPANREDRAW |
				  __SCROLL_YNOMOVE;

	*dvar			= *var;
	dvar->activate		&= ~FB_ACTIVATE_ALL;

	/*
	 * Update the old var.  The fbcon drivers still use this.
	 * Once they are using sfb->fb.var, this can be dropped.
	 */
	display->var = *dvar;

	/*
	 * If we are setting all the virtual consoles, also set the
	 * defaults used to create new consoles.
	 */
	if (var->activate & FB_ACTIVATE_ALL)
		sfb->fb.disp->var = *dvar;

	/*
	 * If the console has changed and the console has defined
	 * a changevar function, call that function.
	 */
	if (chgvar && info && sfb->fb.changevar)
		sfb->fb.changevar(con);

	/* If the current console is selected, activate the new var. */
	if (con != sfb->currcon)
		return 0;

	smifb_hw_set_var(dvar, sfb);
	smi_set_timing(sfb, var);

	return 0;
}

static int
__do_set_cmap(struct fb_cmap *cmap, int kspc, int con,
	      struct fb_info *info)
{
	struct smifb_info *sfb = (struct smifb_info *)info;
	struct fb_cmap *dcmap = get_con_cmap(info, con);
	int err = 0;

	/* no colormap allocated? */
	if (!dcmap->len)
		err = fb_alloc_cmap(dcmap, 256, 0);

	if (!err && con == sfb->currcon)
		err = fb_set_cmap(cmap, kspc, smifb_setcolreg, info);

	if (!err)
		fb_copy_cmap(cmap, dcmap, kspc ? 0 : 1);

	return err;
}

static int
smifb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
		  struct fb_info *info)
{
	struct display *disp = get_con_display(info, con);

	if (disp->visual == FB_VISUAL_TRUECOLOR)
		return -EINVAL;

	return __do_set_cmap(cmap, kspc, con, info);
}

static int
smifb_get_fix(struct fb_fix_screeninfo *fix, int con, struct fb_info *info)
{
	struct display *display = get_con_display(info, con);

	*fix = info->fix;

	fix->line_length = display->line_length;
	fix->visual	 = display->visual;

	return 0;
}

static int
smifb_get_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	debug ("entering smifb_get_var\n");
	*var = *get_con_var(info, con);
	smifb_print_var(var);

	return 0;
}

static int
smifb_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
	struct fb_cmap *dcmap = get_con_cmap(info, con);

	fb_copy_cmap(dcmap, cmap, kspc ? 0 : 2);
	return 0;
}

static struct fb_ops smifb_ops = {
	owner:			THIS_MODULE,
	fb_get_fix:		smifb_get_fix,
	fb_get_var:		smifb_get_var,
	fb_set_var:		smifb_set_var,
	fb_get_cmap:		smifb_get_cmap,
	fb_set_cmap:		smifb_set_cmap,
};

/*
 *  smifb_switch():
 *	Change to the specified console.  Palette and video mode
 *      are changed to the console's stored parameters.
 */
static int smifb_switch(int con, struct fb_info *info)
{
	struct smifb_info *sfb = (struct smifb_info *)info;
	struct display *disp;
	struct fb_cmap *cmap;

	if (con == sfb->currcon) {
		return 0;
	}

	if (sfb->currcon >= 0) {
		disp = fb_display + sfb->currcon;

		/*
		 * Save the old colormap and video mode.
		 */
		disp->var = sfb->fb.var;
		if (disp->cmap.len)
			fb_copy_cmap(&sfb->fb.cmap, &disp->cmap, 0);
	}

	sfb->currcon = con;
	disp = fb_display + con;

	if (disp->cmap.len)
		cmap = &disp->cmap;
	else
		cmap = fb_default_cmap(1 << disp->var.bits_per_pixel);

	fb_copy_cmap(cmap, &sfb->fb.cmap, 0);

	sfb->fb.var = disp->var;
	sfb->fb.var.activate = FB_ACTIVATE_NOW;

	smifb_set_var(&sfb->fb.var, con, info);

	return 0;
}

/*
 * Formal definition of the VESA spec:
 *  On
 *  	This refers to the state of the display when it is in full operation
 *  Stand-By
 *  	This defines an optional operating state of minimal power reduction with
 *  	the shortest recovery time
 *  Suspend
 *  	This refers to a level of power management in which substantial power
 *  	reduction is achieved by the display.  The display can have a longer
 *  	recovery time from this state than from the Stand-by state
 *  Off
 *  	This indicates that the display is consuming the lowest level of power
 *  	and is non-operational. Recovery from this state may optionally require
 *  	the user to manually power on the monitor
 *
 *  Now, the fbdev driver adds an additional state, (blank), where they
 *  turn off the video (maybe by colormap tricks), but don't mess with the
 *  video itself: think of it semantically between on and Stand-By.
 *
 *  So here's what we should do in our fbdev blank routine:
 *
 *  	VESA_NO_BLANKING (mode 0)	Video on,  front/back light on
 *  	VESA_VSYNC_SUSPEND (mode 1)  	Video on,  front/back light off
 *  	VESA_HSYNC_SUSPEND (mode 2)  	Video on,  front/back light off
 *  	VESA_POWERDOWN (mode 3)		Video off, front/back light off
 *
 *  This will match the matrox implementation.
 */
/*
 * smifb_blank():
 *	Blank the display by setting all palette values to zero.  Note, the
 * 	12 and 16 bpp modes don't really use the palette, so this will not
 *      blank the display in all modes.
 */
static void smifb_blank(int blank, struct fb_info *info)
{
//	struct smifb_info *sfb = (struct smifb_info *)info;

	switch (blank) {
	case VESA_POWERDOWN:
	case VESA_VSYNC_SUSPEND:
	case VESA_HSYNC_SUSPEND:
	case VESA_NO_BLANKING:
		break;
	}
}

static int smifb_updatevar(int con, struct fb_info *info)
{
	debug ("entering smifb_updatevar\n");
	smifb_print_var (&fb_display[con].var);
	return (smifb_pan_display(&fb_display[con].var, info));
}

/*
 * Alloc struct smifb_info and assign the default value
 */
static struct smifb_info * __devinit
smi_alloc_fb_info(struct pci_dev *dev, char *name)
{
    struct smifb_info *sfb;

    sfb = kmalloc(sizeof(struct smifb_info) + sizeof(struct display) +
                  sizeof(u32) * 16, GFP_KERNEL);

    if (!sfb)
        return NULL;

    memset(sfb, 0, sizeof(struct smifb_info) + sizeof(struct display));

    sfb->currcon        = -1;
    sfb->dev            = dev;

    strcpy(sfb->fb.fix.id, name);

    sfb->fb.fix.type		= FB_TYPE_PACKED_PIXELS;
    sfb->fb.fix.type_aux	= 0;
    sfb->fb.fix.xpanstep	= 0;
    sfb->fb.fix.ypanstep	= 1;
    sfb->fb.fix.ywrapstep	= 0;
    sfb->fb.fix.accel		= FB_ACCEL_NONE;

    sfb->fb.var.nonstd		= 0;
    sfb->fb.var.activate	= FB_ACTIVATE_NOW;
    sfb->fb.var.height		= -1;
    sfb->fb.var.width		= -1;
    sfb->fb.var.accel_flags	= 0;
    sfb->fb.var.vmode		= FB_VMODE_NONINTERLACED;

    strcpy(sfb->fb.modename, sfb->fb.fix.id);
    strcpy(sfb->fb.fontname, "VGA8x8");

    sfb->fb.fbops		= &smifb_ops;
    sfb->fb.changevar		= NULL;
    sfb->fb.switch_con		= smifb_switch;
    sfb->fb.updatevar		= smifb_updatevar;
    sfb->fb.blank		= smifb_blank;
    sfb->fb.flags		= FBINFO_FLAG_DEFAULT;
    sfb->fb.node		= -1;
    sfb->fb.disp		= (struct display *)(sfb + 1);
    sfb->fb.pseudo_palette	= (void *)(sfb->fb.disp + 1);

    return sfb;
}

/*
 * Unmap the memory mapped IO registers
 *
 */

static void __devinit
smi_unmap_mmio(struct smifb_info *sfb)
{
    if (sfb && VoyagerFB)
    {
        iounmap(VoyagerFB);
        VoyagerReg = NULL;
    }
}

/*
 * Map in the screen memory
 *
 */
static int __devinit
smi_map_smem(struct smifb_info *sfb, struct pci_dev *dev, u_long smem_len)
{
#if isPCI
    sfb->fb.fix.smem_start = pci_resource_start(dev, 0);
#else
    sfb->fb.fix.smem_start = FB_PHYSICAL_ADDR;
#endif
    sfb->fb.fix.smem_len  = smem_len;

    sfb->fb.screen_base = VoyagerFB;

    if (!sfb->fb.screen_base)
    {
        printk("%s: unable to map screen memory\n",sfb->fb.fix.id);
        return -ENOMEM;
    }

    return 0;
}


/*
 * Unmap the screen memory
 *
 */
static void __devinit
smi_unmap_smem(struct smifb_info *sfb)
{
    if (sfb && sfb->fb.screen_base)
    {
        iounmap(sfb->fb.screen_base);
        sfb->fb.screen_base = NULL;
    }
}

/*
 * We need to wake up the LynxEM+, and make sure its in linear memory mode.
 */
static inline void __devinit
smi_init_hw(void)
{
    outb(0x18, 0x3c4);
    outb(0x11, 0x3c5);
}

static void __devinit
smi_free_fb_info(struct smifb_info *sfb)
{
	if (sfb) {
		fb_alloc_cmap(&sfb->fb.cmap, 0, 0);
		kfree(sfb);
	}
}

/*
 * smi_set_m1xclk()
 *
 * Set the clocks MCLK and M1XCLK to working values for the TQM5200-Board:
 * MCLK:    PLL: 288 MHz, Divider: 12 (-> 0xA)
 * M1XCLK:  PLL: 288 MHz, Divider:  2 (-> 0x1)
 *
 * With this values the following resulutions are tested:
 * 640x480	<= @100 Hz
 * 800x600	<= @100 Hz
 * 1024x786	<= @85 Hz
 * 1152x864	<= @70 Hz
 * 1280x1024	<= @60 Hz
 *
 * At higher vertical scan rates (and thus higher pixel clock) the picture
 * could be disturbed. This is because of a to slow M1XCLK (SDRAM Clock)
 *
 * Hint: According to SM501 errata sheet (errata AA-5) the PLL source for
 * M1XCCLK and MCLK must be identical.
 */
void smi_set_m1xclk (void)
{
	ULONG clock;

	clock = regRead32(POWER_MODE0_CLOCK);
	clock &= 0xFFFF0000;
	clock |= (0 << 12) | (0xA << 8) | (0 << 4) | (0x1 << 0);
	regWrite32(POWER_MODE0_CLOCK, clock);

	clock = regRead32(POWER_MODE1_CLOCK);
	clock &= 0xFFFF0000;
	clock |= (0 << 12) | (0xA << 8) | (0 << 4) | (0x1 << 0);
	regWrite32(POWER_MODE1_CLOCK, clock);
}

/*
 * Pan the display (by changing the framebuffer start position)
 */
static int smifb_pan_display(const struct fb_var_screeninfo *var,
				   struct fb_info *info)
{
	unsigned int offset;
	struct smifb_info *sfb = (struct smifb_info *)info;

	offset = (var->xoffset + (var->yoffset * var->xres))
		* var->bits_per_pixel/8;
	sfb->fb.var.xoffset = var->xoffset;
	sfb->fb.var.yoffset = var->yoffset;
	set_screen_start(offset);

	return 0;
}

/*
 * Set start of framebuffer in grafic memory
 */
static void set_screen_start(int base)
{
	ulong tmp;
	if (smi_displaytype == CRT) {
		tmp = regRead32(CRT_FB_ADDRESS);
		tmp = FIELD_VALUE(tmp, CRT_FB_ADDRESS, ADDRESS, base);
		regWrite32(CRT_FB_ADDRESS, tmp);
	}
	else {
		tmp = regRead32(PANEL_FB_ADDRESS);
		tmp = FIELD_VALUE(tmp, PANEL_FB_ADDRESS, ADDRESS, base);
		regWrite32(PANEL_FB_ADDRESS, tmp);
	}
}

u16 Voyager_ChipIDs[numChipIDs] = {0x501};

int __init voyager_init(void)
{
    struct smifb_info *sfb;
    u_long smem_size= 0x00800000; //default 8MB
    char name[16];
    int err;
    char *m_pLAW;
    ulong pFramebufferPhysical;
    ulong pRegPhysical;
    struct pci_dev *pdev = NULL;
    mode_table_t * mode;
#if isPCI
    int i = 0;

    do {
		pdev = pci_find_device(0x126f,Voyager_ChipIDs[i], pdev);
		if (pdev == NULL)
			i++;
		else {
			hw.chipID = Voyager_ChipIDs[i];
			break;
		}
    } while (i< numChipIDs);

    err = pci_enable_device(pdev); // enable voyager chip
    if (err) {
        return err;
    }
#else
	hw.chipID = CHIP_ID;
#endif
    sprintf(name, "voyager");

    err = -ENOMEM;

    sfb = smi_alloc_fb_info(pdev, name);

    if (!sfb) {
        goto failed;
    }

	// Map address and memory detection
#if isPCI
	pFramebufferPhysical = pci_resource_start(pdev,0);
  	pRegPhysical = pci_resource_start(pdev,1);
#else
	pFramebufferPhysical = FB_PHYSICAL_ADDR;
  	pRegPhysical = REG_PHYSICAL_ADDR;
#endif

    switch (hw.chipID)
    {
    case 0x501:
        sfb->fb.fix.mmio_start = pRegPhysical;
        sfb->fb.fix.mmio_len   = 0x00200000;
        smem_size = 0x00800000; //8MB
	if (!request_mem_region(pFramebufferPhysical, smem_size, "voyager")) {
		printk("request_region pFramebufferPhysical failed!\n");
            return -1;
        };
        if (!request_mem_region(pRegPhysical, sfb->fb.fix.mmio_len, "voyager")) {
            printk("request_region pRegPhysical failed!\n");
            return -1;
        };
#if isACCELENT
        {
            volatile unsigned char *pCPLDS;

            MSC1 &= 0xffff0000;
            MSC1 |= 0x00009204; // 0x00009234

            MDREFR |= 0x01010000;	// Set free running clock and SDCLK[1] to 100MHz

            LCCR0 |= LCCR0_DIS;	// BIT:10 DISABLE lcd CONTROLLER

            GAFR2_L = (GAFR2_L & ~0x30000000) | 0x20000000;
            GAFR0_U = (GAFR0_U & ~0x30) | 0x10;
            GAFR1_U = (GAFR1_U & ~0xf) | 0xa;

            pCPLDS = ioremap_nocache(0x17C00000, 0x1000);

            *(volatile PULONG) (pCPLDS+0x38) = (*(volatile PULONG)(pCPLDS+0x38)) & ~0x10;

            iounmap(pCPLDS);
        }
#endif //isACCELENT
	hw.m_pLFB = VoyagerFB=m_pLAW = ioremap_nocache(pFramebufferPhysical, smem_size);
        hw.m_pMMIO = VoyagerReg = ioremap_nocache(pRegPhysical, sfb->fb.fix.mmio_len);
        break;
    }

#if isPC
	sfb->fb.var.xres = screen_info.lfb_width;
	sfb->fb.var.yres = screen_info.lfb_height;
	sfb->fb.var.bits_per_pixel = screen_info.lfb_depth;
#else
#if SMI_BIG_ENDIAN
	/* turn SM501 in big endian mode */
	regWrite32(ENDIAN_CTRL, 0xffffffff);
#endif /* SMI_BIG_ENDIAN */
#endif
	smi_set_m1xclk ();
	if (smi_displaytype == PANEL) {
		if (smi_xres != 800 || smi_yres != 480) {
			printk("Voyager: Currently only 800x480 panel supported,\n"\
			       " use CRT-interface for other resolutions\n");
		}
		sfb->fb.var.xres = FP_DFLT_SCREEN_X_RES;
		sfb->fb.var.yres = FP_DFLT_SCREEN_Y_RES;
		sfb->fb.var.bits_per_pixel = smi_bpp;
		sfb->fb.var.pixclock = FP_DFLT_PIXCLOCK;
		sfb->fb.var.left_margin = FP_DFLT_LEFT_MARGIN;
		sfb->fb.var.right_margin = FP_DFLT_RIGHT_MARGIN;
		sfb->fb.var.upper_margin = FP_DFLT_UPPER_MARGIN;
		sfb->fb.var.lower_margin = FP_DFLT_LOWER_MARGIN;
		sfb->fb.var.hsync_len = FP_DFLT_HSYNC_LEN;
		sfb->fb.var.vsync_len = FP_DFLT_VSYNC_LEN;
		sfb->fb.var.sync = SYNC;
	} else {
		mode = findMode (crt_modedb, smi_xres, smi_yres, 60);
		if (mode) {
			sfb->fb.var.xres = smi_xres;
			sfb->fb.var.yres = smi_yres;
			sfb->fb.var.bits_per_pixel = smi_bpp;
			sfb->fb.var.pixclock = 1000000000/(mode->pixel_clock/1000);
			sfb->fb.var.left_margin = mode->horizontal_total
						  - mode->horizontal_sync_start
						  - mode->horizontal_sync_width;
			sfb->fb.var.right_margin = mode->horizontal_sync_start
						   - mode->horizontal_display_end;
			sfb->fb.var.upper_margin = mode->vertical_total
						   - mode->vertical_sync_start
						   - mode->vertical_sync_height;
			sfb->fb.var.lower_margin = mode->vertical_sync_start
						   - mode->vertical_display_end;
			sfb->fb.var.hsync_len = mode->horizontal_sync_width;
			sfb->fb.var.vsync_len = mode->vertical_sync_height;
			sfb->fb.var.sync = 0;
			if (mode->horizontal_sync_polarity == POSITIVE)
				sfb->fb.var.sync |= FB_SYNC_HOR_HIGH_ACT;
			if (mode->vertical_sync_polarity == POSITIVE)
				sfb->fb.var.sync |= FB_SYNC_VERT_HIGH_ACT;
		} else {
			sfb->fb.var.xres = CRT_DFLT_SCREEN_X_RES;
			sfb->fb.var.yres = CRT_DFLT_SCREEN_Y_RES;
			sfb->fb.var.bits_per_pixel = smi_bpp;
			sfb->fb.var.pixclock = CRT_DFLT_PIXCLOCK;
			sfb->fb.var.left_margin = CRT_DFLT_LEFT_MARGIN;
			sfb->fb.var.right_margin = CRT_DFLT_RIGHT_MARGIN;
			sfb->fb.var.upper_margin = CRT_DFLT_UPPER_MARGIN;
			sfb->fb.var.lower_margin = CRT_DFLT_LOWER_MARGIN;
			sfb->fb.var.hsync_len = CRT_DFLT_HSYNC_LEN;
			sfb->fb.var.vsync_len = CRT_DFLT_VSYNC_LEN;
			sfb->fb.var.sync = SYNC;
		}
	}

	sfb->fb.var.accel_flags = 0;

	if (15 == sfb->fb.var.bits_per_pixel)
		sfb->fb.var.bits_per_pixel = 16;

   	sfb->fb.var.xres_virtual = sfb->fb.var.xres;

	sfb->fb.var.yres_virtual = sfb->fb.var.yres;

	err = smi_map_smem(sfb, pdev, smem_size);
	if (err) {
		goto failed;
	}
	smifb_set_var(&sfb->fb.var, -1, &sfb->fb);

	err = register_framebuffer(&sfb->fb);
	if (err < 0) {
		goto failed;
	}

	MOD_INC_USE_COUNT;
	printk("Silicon Motion, Inc. Voyager Init Complete.\n");

	return 0;

failed:
	smi_unmap_smem(sfb);
	smi_unmap_mmio(sfb);
	smi_free_fb_info(sfb);

	return err;
}

static int __init strtoi(const char *numstr)
{
	int val = 0;
	for (;; numstr++) {
		switch (*numstr) {
		case '0'...'9':
			val = 10*val + (*numstr - '0');
			break;
		default:
			return val;
		}
	}
}

static void __init parse_mode_opt(const char *mode)
{
	int i, yres_found = 0;
	int save_dflt_yres = smi_yres;

	for (i = strlen(mode) - 1; i >= 0; i--) {
		switch (mode[i]) {
		case '0'...'9':
			break;
		case 'x':
			if (!yres_found) {
				smi_yres = strtoi(&mode[i+1]);
				yres_found = 1;
				debug("Voyager: given yres: %d\n", smi_yres);
			} else { /* bad mode string, use predefined mode */
				smi_yres = save_dflt_yres;
				return;
			}
			break;
		default:
			if (yres_found) { /* bad mode string, use predefined mode */
				smi_yres = save_dflt_yres;
			}
			return;
		}
	}
	if (i < 0 && yres_found) {
		smi_xres = strtoi(mode);
		debug("Voyager: given xres: %d\n", smi_xres);
	}
}

/*
 * Parse user specified options (`video=voyager:')
 * example:
 * 	video=voyager:bpp=16,800x600,fp
 */
int __init voyager_setup(char *options)
{
	char * opt;
	if (!options || !*options)
		return 0;
	while((opt = strsep(&options,",")) != NULL ) {
		if (!*opt) continue;
		else if (!strncmp(opt,"fp",2))
			smi_displaytype = PANEL;
		else if (!strncmp(opt,"crt",3))
			smi_displaytype = CRT;
		else if (!strncmp(opt,"bpp=",4))
			smi_bpp = simple_strtoul(opt+4,NULL,0);
		else {
			smi_mode = opt;
			parse_mode_opt((const char *)smi_mode);
		}
	}
	return 0;
}
